/***************************************************************************
*
* PROJECT: Carnegie Mellon Planetary Rover Project
*          Task Control Architecture
*
* (c) Copyright 1991 Christopher Fedor and Reid Simmons.  All rights reserved.
* 
* MODULE: cells
*
* FILE: cells.h
*
* ABSTRACT:
*
* A "cell" is an entity that can take on different values under
* different contexts.  Each value of a cell is represented by a tms
* node.  Cells can have any number of values.  Regular cells can have
* many values (tms nodes) IN at once; "exclusive" cells can have only
* one value in at a time; "constrained" cells can have many values in at
* once, but only the "most-constrained" cell is active at any one time.
*
* "Rules" can be attached to a cell.  When a new value (tms node) is
* added to the cell the rule is run, its argument being the tms node for
* that value.  Typically, the rules are used to link up justifications
* between the tms_nodes of other cells. Rules are defined in the file
* "rules".
*
* For example, A+B can be implemented by two identical rules
* on the cells for both A and B.  Let's say B already has the value `B =
* 2'.  When a new value is added for A, such as `A = 1', the rule finds
* or creates a tms node for the sum `A+B = 3' and justifies the nodes
* with the tms nodes representing `A = 1' and `B = 2'.  There is a macro
* for automatically creating accessor and setting functions for the
* values of cells. 
*
* This file also contains mechanisms for maintaining a set of contexts.
* The context mechanism can be implemented using an ATMS or a regular
* TMS.  Currently it uses a modification of Williams' JTMS (in file
* "tms"), inning and outing nodes as contexts are switched, instead of
* keeping track of all contexts at once. 
*
***************************************************************************/

typedef struct { 
 char *name;
 LIST_PTR assumptions;
} CONTEXT_TYPE, *CONTEXT_PTR;

typedef enum {RegularCell, ExclusiveCell, ConstrainedCell} CELL_VALUE_TYPE;

typedef struct _CELL_TYPE { 
  char *datum;
  CELL_VALUE_TYPE type;
  LIST_PTR contextualValues;
  LIST_PTR rules;
  LIST_PTR outRules;
  LIST_PTR inRules;
  int (*sameValueFN)();
  void (*sayFN)();
  char *additional_field;
} CELL_TYPE, *CELL_PTR;

typedef struct {
  char *most_constrained_value;
  int (*most_constrained_fn)();
} CONSTRAINED_CELL_ADDL_TYPE, *CONSTRAINED_CELL_ADDL_PTR;

#define EXCLUSIVE_CELL_CURRENT_VALUE(cell) \
  ((cell)->additional_field)

#define CONSTRAINED_CELL_MOST_CONSTRAINED_VALUE(cell) \
  ((CONSTRAINED_CELL_ADDL_PTR) \
   (cell)->additional_field)->most_constrained_value

#define CONSTRAINED_CELL_MOST_CONSTRAINED_FN(cell) \
  ((CONSTRAINED_CELL_ADDL_PTR) \
   (cell)->additional_field)->most_constrained_fn

/* extern CELL_PTR cellCreate(); */
/* extern CELL_PTR cellCreateExclusive(); */
extern CELL_PTR cellCreateConstrained();
extern void cellFree();
/* extern void assume(); */
/* extern void unassume(); */
/* extern void assertPremise(); */
/* extern void unassertPremise(); */
extern void justify();
extern TMS_NODE_PTR addValueNode();
extern TMS_NODE_PTR getOrAddValueNode();
extern TMS_NODE_PTR getValueNode();
/* extern char *currentValue(); */
extern TMS_NODE_PTR currentValueSupporter(); 
/* extern LIST_PTR getCurrentValues(); */

extern void sayNodeValue();

/* extern CONTEXT_PTR THE_CURRENT_CONTEXT; */

