/*****************************************************************************
* PROJECT: Carnegie Mellon Planetary Rover Project
*          Task Control Architecture
*
* (c) Copyright 1991 Christopher Fedor and Reid Simmons.  All rights reserved.
*
* MODULE: central
*
* FILE: central.c
*
* ABSTRACT:
*
* This file contains the "main" of the central tca server.
*
* REVISION HISTORY:
*
* 25-Jul-91 Reid Simmons at School of Computer Science, CMU
* Removed the -b option (buffers are no longer managed).
* Added -r option to try resending messages after module crashes.
*
*  8-Jul-91 Reid Simmons at School of Computer Science, CMU
* Removed the -gc option.
* Added -b <n> option to pre-allocate a data message buffer.
*
*  7-Apr-90 Christopher Fedor at School of Computer Science, CMU
* Added -v option to simply display current server version info.
*
* 30-Mar-90 Christopher Fedor at School of Computer Science, CMU
* Revised to software standards.
*
*****************************************************************************/

#include "stdio.h"
#include "string.h"
#include "ctype.h"
#include "signal.h"
#include "time.h"
extern time_t time(); /* Is not in sys/time.h!! */

#include "tcaInternal.h"
#include "logging.h"

#include "idtable.h"

extern void listenLoop();
extern int serverInitialize();

extern int serverPortGlobal;

extern ID_TABLE_PTR dispatchTable;

int tcaDebugGlobal;
int qassertCheckGlobal;

extern int resendAfterCrashGlobal; /* defined in res.c */

/*****************************************************************************
*
* FUNCTION: void setOptions(option, logMessages, logStatus, logTime, logData, 
*                           logIgnore, logHandleSum)
*
* DESCRIPTION:
*
* INPUTS:
* char *option;
* int *logMessages, *logStatus, *logTime, *logData, *logIgnore, *logHandleSum;
*
* OUTPUTS: void
*
* HISTORY:
*  6-Oct-89 Christopher Fedor at School of Computer Science, CMU
*  eeeks -- this setting options is a mess! 
*
*****************************************************************************/

void setOptions(option, logMessages, logStatus, logTime, logData, 
		logIgnore, logHandleSum)
char *option;
int *logMessages, *logStatus, *logTime, *logData, *logIgnore, *logHandleSum;
{
  if (strlen(option) == 2) { 
    /* -l or -L -- use all the options, except i */
    *logMessages = *logStatus = *logTime = *logData = *logHandleSum = TRUE;
    *logIgnore = FALSE;  
  }
  else if (strpbrk(option, "xX") != NULL) {
    /* no options */
   *logMessages = *logStatus = *logTime = *logData = FALSE;
   *logHandleSum = FALSE;
   *logIgnore = FALSE;
  }
  else {
    *logMessages = (strpbrk(option, "mM") != NULL);
    *logStatus = (strpbrk(option, "sS") != NULL);
    *logTime = (strpbrk(option, "tT") != NULL);
    *logData = (strpbrk(option, "dD") != NULL);
    *logIgnore = (strpbrk(option, "iI") != NULL);
    *logHandleSum = (strpbrk(option, "hH") != NULL);
  }
}


/*****************************************************************************
*
* FUNCTION: void setLoggingOptions(argc, argv)
*
* DESCRIPTION:
*
* INPUTS:
* int argc;
* char **argv;
*
* OUTPUTS: void
*
*****************************************************************************/

void parseCommandLineOptions(argc, argv)
int argc;
char **argv;
{ 
  int i, foundFile = FALSE;

  Log_File = NULL;

  for (i=1; i<argc; i++)
    if (argv[i][0] == '-')
      switch (argv[i][1]) {
      case 'l':
	setOptions(argv[i], &Terminal_Log_Messages, &Terminal_Log_Status,
	    &Terminal_Log_Time, &Terminal_Log_Data, &Terminal_Log_Ignore,
		    &Terminal_Log_HandleSummary);
	break;
      case 'L': 
	setOptions(argv[i], &File_Log_Messages, &File_Log_Status,
		    &File_Log_Time, &File_Log_Data, &File_Log_Ignore,
		    &File_Log_HandleSummary);
	break;
      case 'f':
	if (strlen(argv[i]) > 2) {
	  foundFile = TRUE;
	  (void)strncpy(Log_File_Name, argv[i]+2, MAX_LOG_FILE_NAME_LENGTH);
	}
	break;
      case 'p':
	if (strlen(argv[i]) > 2) 
	  (void)sscanf(argv[i]+2, "%d", &serverPortGlobal);
	break;
      case 'd':
	tcaDebugGlobal = TRUE;
	break;
      case 'r':
	resendAfterCrashGlobal = TRUE;
	break;
      case 'q':
	qassertCheckGlobal = TRUE;
	break;
      default: 
	fprintf(stderr, "Ignoring unknown option '%s'\n", argv[i]);
      }
  
  if (!foundFile &&
    (File_Log_Messages || File_Log_Status || File_Log_Time || File_Log_Data)) {
    (void)sprintf(Log_File_Name, "tca.log.%d", time(((time_t *)NULL)));
  }
}


/*****************************************************************************
*
* FUNCTION: int parseExpectedMods(argc, argv, expectedMods)
*
* DESCRIPTION:
*
* INPUTS:
* int argc;
* char **argv;
* int *expectedMods;
*
* OUTPUTS: void
*
*****************************************************************************/

int parseExpectedMods(argc, argv, expectedMods)
int argc;
char **argv;
int *expectedMods;
{ 
  int i;

  for (i=1; i<argc; i++) {
    if (isdigit(argv[i][0])) {
      *expectedMods = atoi(argv[i]);
      return (TRUE);
    }
  }
  return (FALSE);
}

#if 0
/*****************************************************************************
*
* FUNCTION: void preallocateBuffers(argc, argv)
*
* DESCRIPTION:
*
* INPUTS:
* int argc;
* char **argv;
*
* OUTPUTS: void
*
*****************************************************************************/

void preallocateBuffers(argc, argv)
int argc;
char **argv;
{ 
  int i, size;

  for (i=1; i<argc; i++) {
    if (STREQ(argv[i], "-b")) {
      (void)sscanf(argv[++i], "%d", &size);
      if (size > 0) tcaAllocateDataBuffer(size);
    }
  }
}
#endif

/*****************************************************************************
*
* FUNCTION: void displayHelp(name)
*
* DESCRIPTION:
*
* INPUTS:
* char *name;
*
* OUTPUTS: void
*
*****************************************************************************/

void displayHelp(name)
char *name;
{
  printf("Usage: %s [expected-modules] -l[options] -L[options] -f[filename]\n",
	 name);
  printf("\n expected-modules: minimum number of modules being run\n");
  printf(" -l: logging onto terminal.\n");
  printf("     options are: m (message traffic)\n");
  printf("                  s (status of TCA)\n");
  printf("                  t (time messages are handled)\n");
  printf("                  d (data associated with message)\n");
  printf("                  i (ignore logging registration and deregistration messages)\n");
  printf("                  h (incoming message handle time summary)\n");
  printf("                  x (no logging)\n");
  printf("     -l is equivalent to -lmstdh; the default is -lmsi\n");
  printf(" -L: logging into file.\n");
  printf("     options are the same as above.\n");
  printf("     the default is -Lx\n");
  printf(" -f: filename to use for logging.\n");
  printf("     If not specified, name is automatically generated.\n");
  printf(" -v: display server version info.\n");
  printf(" -r: try resending non-completed messages when modules crash\n");
  printf(" -p: specify the server port - the compiled port is: %d\n", 
	 SERVER_PORT);
  printf(" -d: flag random debugging.\n");
  printf(" -q: test quantity lattice insertions for task tree nodes.\n");
}


/*****************************************************************************
*
* FUNCTION: void abortCentral(s, c, sp)
*
* DESCRIPTION:
* Trap fatal errors and attempt a core dump.
*
* INPUTS:
* int s, c;
* struct sigcontext *sp;
*
* OUTPUTS: void
*
*****************************************************************************/
extern void dataMsgDisplayStats();

/*
sun os 4.0.3 added a parameter
void abortCentral(s, c, sp, addr)
int s, c;
struct sigcontext *sp;
char *addr;
*/

void abortCentral(s, c, sp)
int s, c;
struct sigcontext *sp;
{
  idTablePrintInfo(dispatchTable);
  tcaModError("Central Abort");
}


/*****************************************************************************
*
* FUNCTION: void displayVersion()
*
* DESCRIPTION: displays the central server banner.
*
* INPUTS: none.
*
* OUTPUTS: void
*
*****************************************************************************/

void displayVersion()
{
  fprintf(stderr, "Task Control Server %d.%d (%s)\n",
	 TCA_VERSION_MAJOR, TCA_VERSION_MINOR, VERSION_DATE);
}


/*****************************************************************************
*
* FUNCTION: main(argc, argv)
*
* DESCRIPTION:
* start.
*
* INPUTS:
* int argc;
* char **argv;
*
* OUTPUTS: void
*
*****************************************************************************/

main(argc, argv)
int argc;
char **argv;
{ 
  int expectedMods = 0;

  serverPortGlobal = 0;

  tcaDebugGlobal = FALSE;

  qassertCheckGlobal = FALSE;

#ifdef SERVER_PORT
  serverPortGlobal = SERVER_PORT;
#endif

  (void)signal(SIGINT, abortCentral);
  (void)signal(SIGBUS, abortCentral);
  (void)signal(SIGSEGV, abortCentral);

  if (argc < 2 || STREQ(argv[1], "-h"))
    displayHelp(argv[0]);
  else if (STREQ(argv[1], "-v"))
    displayVersion();
  else {
    if (!parseExpectedMods(argc, argv, &expectedMods)) {
      displayHelp(argv[0]);
    }
    else {
      parseCommandLineOptions(argc, argv);

      /* 30-Mar-90: fedor: should change to allow start up with 0.
	 this was done once before - needs to be rechecked. */
      if (expectedMods < 1)
	expectedMods = 1;

      if (!serverInitialize(expectedMods)) {
	tcaError("ERROR: serverInitialize\n");
      }

      /* Need to do this here, after serverInitialize is called */
      /* preallocateBuffers(argc, argv);*/

      listenLoop();
    }
  }
  return 1;
}
