/******************************************************************************
*
* PROJECT: Carnegie Mellon Planetary Rover Project
*          Task Control Architecture 
* 
* (c) Copyright 1991 Christopher Fedor and Reid Simmons.  All rights reserved.
* 
* MODULE: communications
*
* FILE: comModule.c
*
* ABSTRACT:
* 
* Module Communications
*
* REVISION HISTORY
*
* 6-May-92 Christopher Fedor, School of Computer Science, CMU
* Added while loop to select if select dies on an EINTR then retry select.
*
* 17-Feb-92 Christopher Fedor, School of Computer Science, CMU
* Generalized call to a registered exit handler tcaExitHnd for all modules
* except lisp. VxWorks no longer has a special tcaVxExit routine.
*
* 28-Oct-91 Christopher Fedor, School of Computer Science, CMU
* Moved call for tcaVxExit so that it will be called when a close is detected
* by both a direct connection and central disconnect. Currently the VxWorks 
* part is not set up for direct connections and the exit call was only 
* happening for direct connections.
*
* 14-Aug-91 Christopher Fedor, School of Computer Science, CMU
* Changed VxWorks include files to vx5.0
*
* 25-Mar-91 Christopher Fedor, School of Computer Science, CMU
* Added lispExitGlobal as the VxWorks way of exiting lisp.
* Should eventually add a general abort handler and exit handler.
*
* 30-Jan-91 Christopher Fedor, School of Computer Science, CMU
* Added fflush(stdout) to printf for module code calls from lisp
*
*  6-Dec-90 Christopher Fedor, School of Computer Science, CMU
* Removed signal.h because the signal to call tcaClose on various signals
* was commeted out and signal.h for sunos 4.1 eventually includes
* sys/stdtypes which defines size_t as an int - this conflicts with the
* vxworks includes which eventually include types.h which defines
* size_t as a long.
* Move sys/time.h to non VxWorks and added vxworks utime.h because the
* timeval struct is needed and the size_t conflict exits.
*
*  7-Apr-90 Christopher Fedor, School of Computer Science, CMU
* Revised to Software Standards.
*
*  6-Mar-89 Christopher Fedor, School of Computer Science, CMU
* Created.  
* fedor: 6-Mar-89 problem with many defined globals from com_server and lots
* of repetitive code needs to be removed!  perhaps a com_com file for
* common communications between the two - actually i want to avoid 
* differences - module and server should be alike as much as possible.
*
******************************************************************************/

#ifdef VXWORKS
#include "stdioLib.h"
#else
#include "stdio.h"
#endif

#ifdef VXWORKS

#include "systime.h"
#include "vxWorks.h"
#include "socket.h"
#include "in.h"

#else

#include "errno.h"
extern int errno;

#include "sys/time.h"
#include "sys/types.h"
#include "sys/socket.h"
#include "netinet/in.h"
#include "netdb.h"

#endif

#include "ctype.h"

#include "tcaInternal.h"
#include "formatters.h"
#include "key.h"

#include "datamsg.h"
#include "idtable.h"
#include "hash.h"
#include "list.h"

#include "dispatch.h"

#define TCP 6
#define BACKLOG 5
#define PROTOCOL_NAME "tcp"

extern char LispFlagGlobal;
extern int (*lispDecodeMsgGlobal)();
extern int (*lispReplyMsgGlobal)();
extern int (*lispExitGlobal)();

extern void msgInfoMsgInitialize();

extern void (*dPrintSTR_FN)();
extern void (*dPrintFORMAT_FN)();
extern void (*dPrintMAP_FN)();
extern void (*dPrintTCA_FN)();
extern void (*dPrintCHAR_FN)();
extern void (*dPrintSHORT_FN)();
extern void (*dPrintLONG_FN)();
extern void (*dPrintINT_FN)();
extern void (*dPrintFLOAT_FN)();
extern void (*dPrintDOUBLE_FN)();
extern void (*dPrintBYTE_FN)();
extern void (*dPrintTWOBYTE_FN)();

extern void classModInitialize(); /* defined in sendMsg.c */
extern void formatInitialize(); /* defined in formatters.c */
extern void execHnd(); /* defined in behaviors.c */
extern void classModInitialize(); /* defined in sendMsg.c */

#ifndef VXWORKS
extern char *getenv();
#endif

int tcaServerGlobal;

char *modNameGlobal, *servHostGlobal;

void (*tcaRegInitGlobal)();

int tcaListenGlobal;
int directFlagGlobal;
LIST_PTR moduleConnectionList;

fd_set tcaConnectionListGlobal;

TCA_VERSION_TYPE versionGlobal;

ID_TABLE_PTR msgIdTable;
ID_TABLE_PTR hndIdTable;
extern ID_TABLE_PTR dispatchTable;

HASH_TABLE_PTR messageTable;
HASH_TABLE_PTR handlerTable;
HASH_TABLE_PTR classFormatTable;

/* Indicates the top of the task tree */
TCA_REF_PTR tcaRootNodeGlobal;

/* List of reply dispatches waiting to be received. */
LIST_PTR tcaPendingRepliesGlobal = NULL;

void (*tcaExitHnd)();

/******************************************************************************
*
* FUNCTION: void printDataModuleInitialize();
*
* DESCRIPTION: 
* Sets print data routines for modules.
*
* INPUTS: none.
*
* OUTPUTS: void.
*
******************************************************************************/

void sampleModulePrintData(stream, item, next)
FILE *stream;
char *item;
int next;
{
  fprintf(stderr, "WARNING: DPrint called!\n");
#ifdef LISP
  fflush(stderr);
#endif
}

void sampleModuleMapPrintData(stream, map, format, next, mapElements)
FILE *stream;
char *map;
FORMAT_PTR format;
int next;
char *mapElements;
{
  fprintf(stderr, "WARNING: DPrint on maps called!\n");
#ifdef LISP
  fflush(stderr);
#endif
}

void printDataModuleInitialize()
{
  dPrintSTR_FN = sampleModulePrintData;
  dPrintFORMAT_FN = sampleModulePrintData;
  dPrintMAP_FN = sampleModuleMapPrintData;
  dPrintTCA_FN = sampleModulePrintData;
  dPrintCHAR_FN = sampleModulePrintData;
  dPrintSHORT_FN = sampleModulePrintData;
  dPrintLONG_FN = sampleModulePrintData;
  dPrintINT_FN = sampleModulePrintData;
  dPrintFLOAT_FN = sampleModulePrintData;
  dPrintDOUBLE_FN = sampleModulePrintData;
  dPrintBYTE_FN = sampleModulePrintData;
  dPrintTWOBYTE_FN = sampleModulePrintData;
}


/******************************************************************************
*
* FUNCTION: void tcaModuleInitialize()
*
* DESCRIPTION: 
* Sets globals and calls main initialize routines for initializing a module.
*
* INPUTS: none.
*
* OUTPUTS: void.
*
******************************************************************************/

void tcaModuleInitialize()
{
  tcaExitHnd = NULL;
  
  directFlagGlobal = FALSE;
  moduleConnectionList = listCreate();

  versionGlobal.tcaMajor = TCA_VERSION_MAJOR;
  versionGlobal.tcaMinor = TCA_VERSION_MINOR;

  dataMsgInitialize();

  tcaRefInitialize();
  formatInitialize();

  msgIdTable = idTableCreate("Message Table", 10);
  hndIdTable = idTableCreate("Handler Table", 10);

  messageTable = hashTableCreate(100, strHashFunc, strKeyEqFunc);
  handlerTable = hashTableCreate(100, hndHashFunc, hndKeyEqFunc);

  classFormatTable = hashTableCreate(10, classHashFunc, classEqFunc);

  printDataModuleInitialize();
}


/******************************************************************************
*
* FUNCTION: char *tcaServerMachine()
*
* DESCRIPTION: Simply calls getenv to read the server machine from CENTRALHOST
*
* INPUTS: none.
*
* OUTPUTS: char *
*
******************************************************************************/

#ifndef VXWORKS
char *tcaServerMachine()
{
  char *servHost;

  servHost = getenv("CENTRALHOST");
  if (!servHost) {
    tcaModError("TC ERROR: CENTRALHOST environment variable not set.\n");
  };

  return servHost;
}
#endif

/***************************************************************************/

int connectAtPort(machine, port, sd)
char *machine;
int port, *sd;
{
#ifndef VXWORKS
  struct hostent *hp;
#endif

  struct sockaddr_in serverAddr;

  bzero((char *)&serverAddr, sizeof(struct sockaddr_in));

  serverAddr.sin_port = port;
  serverAddr.sin_family = AF_INET;

#ifndef VXWORKS
  if ((hp = gethostbyname(machine)) == NULL)
    return FALSE;
  bcopy((char *)hp->h_addr, (char *)&serverAddr.sin_addr, hp->h_length);
#else
  if ((serverAddr.sin_addr.s_addr = hostGetByName(machine)) == ERROR)
    return FALSE;
#endif

  if ((*sd = socket(AF_INET, SOCK_STREAM, TCP)) < 0)
    return FALSE;

  /******* 17-aug-91: fedor: causes NeXT machine to fail connecting **
  if ((setsockopt(*sd, SOL_SOCKET, SO_REUSEADDR, (char *)0, 0)) < 0)
    return FALSE;
    *****/

  /* 24-Aug-90: fedor: VxWorks sockaddr_in is defined differently
     this should be checked to make a VxWorks version. */
  if (connect(*sd, (struct sockaddr *)&serverAddr, sizeof(serverAddr)) < 0)
    return FALSE;

  return TRUE;
}

/***************************************************************************/

/* 7-Jun-91: fedor: may need some changes for vxworks */

int listenAtPort(port, sd)
int port, *sd;
{
  int reuse;
  struct sockaddr_in server;

  bzero((char *)&server, sizeof(struct sockaddr_in));

  server.sin_port = port;

  server.sin_family = AF_INET;
  server.sin_addr.s_addr = INADDR_ANY;

  if ((*sd = socket(AF_INET, SOCK_STREAM, TCP)) < 0)
    return FALSE;

  reuse = TRUE;

  if ((setsockopt(*sd, SOL_SOCKET, SO_REUSEADDR, &reuse, sizeof(int))) < 0)
    return FALSE;

  reuse = TRUE;

  if ((bind(*sd, (struct sockaddr *)&server, sizeof(struct sockaddr_in))) < 0)
    return FALSE;

  if (listen(*sd, BACKLOG) < 0)
    return FALSE;

  return TRUE;
}

/***************************************************************************/



/******************************************************************************
*
* FUNCTION int connectSocket(servHost, sd)
*
* DESCRIPTION: 
* Returns TRUE if a connection was successfully opened to servHost, 
* FALSE otherwise. sd is set to the value of the opened connection.
*
* INPUTS: 
* char *servHost;
* int *sd;
*
* OUTPUTS: int (TRUE or FALSE)
*
******************************************************************************/

int connectSocket(servHost, sd)
char *servHost;
int  *sd;
{ 
  char *tmp;
  int i, serverPort;

  i = 0;
  serverPort = SERVER_PORT;
  while(servHost[i] != '\0' && servHost[i] != ':') 
    i++;

  if (servHost[i] != '\0') {
    tmp = servHost+i+1;
    servHost[i] = '\0';
    /* 24-Oct-90: fedor: could use atoi here but it is missing in VxWorks */
    (void)sscanf(tmp, "%d", &serverPort);
  }

  return connectAtPort(servHost, serverPort, sd);
}


/******************************************************************************
*
* FUNCTION: void tcaClose()
*
* DESCRIPTION: 
*
* INPUTS: none.
*
* OUTPUTS: void.
*
******************************************************************************/

void tcaClose()
{
  (void)tcaInform("tcaCloseMsg", (void *)NULL);
}

#ifdef LISP

TCA_REF_PTR returnRootGlobal()
{
  return tcaRootNodeGlobal;
}

#endif


/******************************************************************************
*
* FUNCTION: void tcaConnectModule(modName, serverHost)
*
* DESCRIPTION: 
*
* INPUTS: char *modName, *serverHost;
*
* OUTPUTS: void.
*
******************************************************************************/

void tcaConnectModule(modName, serverHost)
char *modName, *serverHost;
{
  MOD_DATA_TYPE modData;
  TCA_VERSION_TYPE vData;
  DIRECT_INFO_PTR directInfo;
  TCA_RETURN_VALUE_TYPE statusV;

  if (!serverHost) {
    tcaModError("TC ERROR: ServerHost parameter is NULL.\n");
  }

  /* 14-Dec-91: fedor: this is cheating this should be all bcopied */
  modNameGlobal = modName;
  servHostGlobal = serverHost;
  tcaRegInitGlobal = NULL;

  FD_ZERO(&tcaConnectionListGlobal);

  tcaModuleInitialize();

  if (!connectSocket(serverHost, &tcaServerGlobal))
    tcaModError("ERROR: Can not establish a connection to the server.");

  FD_SET(tcaServerGlobal, &tcaConnectionListGlobal);

  directInfo = NEW(DIRECT_INFO_TYPE);
  directInfo->sd = tcaServerGlobal;

  listInsertItem((char *)directInfo, moduleConnectionList);

  msgInfoMsgInitialize();
  classModInitialize();

  tcaRootNodeGlobal = CREATE_NULL_REF();

  tcaPendingRepliesGlobal = listCreate();

  modData.modName = modName;
  modData.hostName = (char *)tcaMalloc(sizeof(char)*32);
  (void)gethostname(modData.hostName, 32);  

  statusV = tcaQuery("NewModuleConnectMsg", (void *)&modData, (void *)&vData);
  if (statusV != Success) {
    tcaModError("ERROR: ConnectModule: Query: ");
  }

  if (vData.tcaMajor != TCA_VERSION_MAJOR) {
      fprintf(stderr, "*** ERROR ***\n");
      fprintf(stderr, 
	      "Major Version Number is different from Task Control Server.\n");
      fprintf(stderr, "Module Version     : %d.%d\n", 
	      versionGlobal.tcaMajor, versionGlobal.tcaMinor);
      fprintf(stderr, 
	      "Task Control Server: %d.%d\n", vData.tcaMajor, vData.tcaMinor);
      tcaModError((char *)NULL);
    } else if (vData.tcaMinor != TCA_VERSION_MINOR) {
      fprintf(stderr, "*** WARNING ***\n");
      fprintf(stderr,
	      "Minor Version Number is different from Task Control Server.\n");
      fprintf(stderr, "Module Version     : %d.%d\n", 
       versionGlobal.tcaMajor, versionGlobal.tcaMinor);
      fprintf(stderr,
	      "Task Control Server: %d.%d\n", vData.tcaMajor, vData.tcaMinor);
#ifdef LISP
      fflush(stderr);
#endif
    } else {
      fprintf(stderr, "Task Control Connected %d.%d\n", vData.tcaMajor, vData.tcaMinor);
#ifdef LISP
      fflush(stderr);
#endif
    }
}


/******************************************************************************
*
* FUNCTION: void tcaWaitUntilReady()
*
* DESCRIPTION: 
*
* INPUTS: none.
*
* OUTPUTS: void.
*
******************************************************************************/

void tcaWaitUntilReady()
{
  int go;
  TCA_RETURN_VALUE_TYPE statusV;

  fprintf(stderr, "Waiting ...\n");
#ifdef LISP
  fflush(stderr);
#endif

  statusV = tcaQuery("WaitMsg", (void *)NULL, (void *)&go);
  if (statusV != Success) {
    tcaModError("ERROR: tcaWaitUntilReady: Query: ");
  }

  fprintf(stderr, "Done.\n");
#ifdef LISP
  fflush(stderr);
#endif
}


/******************************************************************************
*
* FUNCTION: int tcaSingleMsgHnd(time)
*
* DESCRIPTION: 
*
* INPUTS: struct timeval *time;
*
* OUTPUTS: int
*
* NOTES:
* Assumes that this connection is only listening to the central server
* not very general here but it fits our current tca model of control.
*
* tcaConnectionListGlobal was inited from tcaServerGlobal
*
******************************************************************************/

#ifndef VXWORKS

void tcaRegisterInitProc(proc)
     void (*proc)();
{
  tcaRegInitGlobal = proc;
}

tryReconnect()
{
  char *modName, *serverHost;

  MOD_DATA_TYPE modData;
  TCA_VERSION_TYPE vData;
  DIRECT_INFO_PTR directInfo;
  TCA_RETURN_VALUE_TYPE statusV;

  fprintf(stderr, "Trying to reconnect to server.\n");

  if (!modNameGlobal || !servHostGlobal) {
    tcaModError("Sorry: module name or server host unknown.\n");
  }

  modName = modNameGlobal;
  serverHost = servHostGlobal;

  /* this will make current direct connections invalid */
  FD_ZERO(&tcaConnectionListGlobal); 

  tcaModuleInitialize();

  sleep(5);

  if (!connectSocket(serverHost, &tcaServerGlobal)) {
    fprintf(stderr, "Looking for central on %s ", serverHost);
    fflush(stderr);
  
    while (!connectSocket(serverHost, &tcaServerGlobal)) {
      fprintf(stderr, ".");
      fflush(stderr);
      sleep(2);
    }
    fprintf(stderr, "found.\n");
  }

  FD_SET(tcaServerGlobal, &tcaConnectionListGlobal);

  directInfo = NEW(DIRECT_INFO_TYPE);
  directInfo->sd = tcaServerGlobal;

  listInsertItem((char *)directInfo, moduleConnectionList);

  msgInfoMsgInitialize();
  classModInitialize();

  tcaRootNodeGlobal = CREATE_NULL_REF();

  tcaPendingRepliesGlobal = listCreate();

  modData.modName = modName;
  modData.hostName = (char *)tcaMalloc(sizeof(char)*32);
#ifndef VXWORKS
  (void)gethostname(modData.hostName, 32);  
#else
  modData.hostName[0]='V';
  modData.hostName[1]='X';
  modData.hostName[2]='\0';
#endif

  statusV = tcaQuery("NewModuleConnectMsg", (void *)&modData, (void *)&vData);
  if (statusV != Success) {
    tcaModError("ERROR: ConnectModule: Query: ");
  }

  if (vData.tcaMajor != TCA_VERSION_MAJOR) {
      fprintf(stderr, "*** ERROR ***\n");
      fprintf(stderr,
	      "Major Version Number is different from Task Control Server.\n");
      fprintf(stderr, "Module Version     : %d.%d\n", 
       versionGlobal.tcaMajor, versionGlobal.tcaMinor);
      fprintf(stderr, "Task Control Server: %d.%d\n", 
	      vData.tcaMajor, vData.tcaMinor);
      tcaModError((char *)NULL);
    } else if (vData.tcaMinor != TCA_VERSION_MINOR) {
      fprintf(stderr, "*** WARNING ***\n");
      fprintf(stderr,
	      "Minor Version Number is different from Task Control Server.\n");
      fprintf(stderr, "Module Version     : %d.%d\n", 
       versionGlobal.tcaMajor, versionGlobal.tcaMinor);
      fprintf(stderr, "Task Control Server: %d.%d\n", 
	      vData.tcaMajor, vData.tcaMinor);
#ifdef LISP
      fflush(stderr);
#endif
    } else {
      fprintf(stderr, "Task Control Connected %d.%d\n", 
	      vData.tcaMajor, vData.tcaMinor);
#ifdef LISP
      fflush(stderr);
#endif
    }

  if (tcaRegInitGlobal) {
    (*tcaRegInitGlobal)();
  }

  fprintf(stderr, "done.");
}

#endif

int iterateModuleConnections(readMask, directInfo)
fd_set *readMask;
DIRECT_INFO_PTR directInfo;
{
  DATA_MSG_PTR dataMsg;

  if (FD_ISSET(directInfo->sd, readMask)) {
    switch(dataMsgRecv(directInfo->sd, &dataMsg)){
    case StatOK:
      execHnd(directInfo, dataMsg);
      return 1; /* 21-Apr-91: fedor: give a value to tcaHandleMessage */
    case StatError:
      tcaModError("ERROR: tcaSingleMsgHnd: Reading Data Message.");
    case StatEOF:
      FD_CLR(directInfo->sd, &tcaConnectionListGlobal);
#ifndef VXWORKS
      (void)shutdown(directInfo->sd, 2);
      (void)close(directInfo->sd);
      if (directInfo->sd == tcaServerGlobal) {
	fprintf(stderr,"Closed Connection Detected from: Central Server: sd: %d:\n",
		directInfo->sd);

#if 0
#ifndef VXWORKS
	if (tcaRegInitGlobal) {
	  if (tcaExitHnd) {
	    (*tcaExitHnd)();
	  }
	  tryReconnect();
	  tcaWaitUntilReady();
	} else {
#ifdef LISP
	  fflush(stdout);
#endif /* LISP */
	  listDeleteItem((char *)directInfo, moduleConnectionList);
	  tcaFree((char *)directInfo);
#ifdef LISP
	  (*lispExitGlobal)();
#else  
	  if (!tcaExitHnd) {
	    exit(-1);
	  } else {
	    (*tcaExitHnd)();
	  }
#endif /* LISP */
	}
#endif /* VXWORKS */

#endif /* 0 */


#ifdef LISP
	fflush(stdout);
#endif
	listDeleteItem((char *)directInfo, moduleConnectionList);
	tcaFree((char *)directInfo);
#ifdef LISP
	(*lispExitGlobal)();
#else
	if (!tcaExitHnd) {
	  exit(-1);
	} else {
	  (*tcaExitHnd)();
	}
#endif

      } else {
	fprintf(stderr, "Closed Connection Detected from: Module: sd: %d:\n", 
		directInfo->sd);
#ifdef LISP
	fflush(stderr);
#endif
	listDeleteItem((char *)directInfo, moduleConnectionList);
	tcaFree((char *)directInfo);
      }      
#endif
#ifdef VXWORKS
      (void)close(directInfo->sd);
      if (directInfo->sd == tcaServerGlobal) {
	fprintf(stderr,"Closed Connection Detected from: Central Server: sd: %d:\n",
		directInfo->sd);
	listDeleteItem((char *)directInfo, moduleConnectionList);
	tcaFree((char *)directInfo);
      } else {
	fprintf(stderr, "Closed Connection Detected from: Module: sd: %d:\n", 
		directInfo->sd);
	listDeleteItem((char *)directInfo, moduleConnectionList);
	tcaFree((char *)directInfo);
      }

      if (!tcaExitHnd) {
	exit(-1);
      } else {
	(*tcaExitHnd)();
      }

#endif
      break;
    default:
      tcaModError("ERROR: tcaSingleMsgHnd: UNKNOWN INTENT\n");
    }
  }
  return TRUE;
}

/****************************************************************************/

int tcaSingleMsgHnd(time)
struct timeval *time;
{
  int ret, sd;
  fd_set readMask; 
  DIRECT_INFO_PTR directInfo;

#ifndef VXWORKS
  readMask = tcaConnectionListGlobal;
  do {
    ret = select(FD_SETSIZE, &readMask, (fd_set *)NULL, (fd_set *)NULL, time);
  } while (ret < 0 && errno == EINTR);

  if (ret < 1) 
    return ret;
#endif

  if (directFlagGlobal && FD_ISSET(tcaListenGlobal, &readMask)) {
    sd = accept(tcaListenGlobal, (struct sockaddr *)NULL, (int *)NULL);

    directInfo = NEW(DIRECT_INFO_TYPE);
    directInfo->sd = sd;

    fprintf(stderr, "New Connection Established for a Module at sd: %d\n",
	    directInfo->sd);
#ifdef LISP
    fflush(stderr);
#endif

    listInsertItem((char *)directInfo, moduleConnectionList);
    FD_SET(sd, &tcaConnectionListGlobal);
    /* 25-Jun-91: fedor: set sd in readMask as well? */
  }

#ifdef VXWORKS
  /* 26-Jun-91: fedor: VxWorks sits in a read point-to-point work
     will have to wait until this is rethought */
  FD_ZERO(&readMask);
  FD_SET(tcaServerGlobal, &readMask);
#endif
  (void)listIterate(iterateModuleConnections, (char *)&readMask, 
		    moduleConnectionList);
  return 1; /* Success */
}


/******************************************************************************
*
* FUNCTION: TCA_RETURN_VALUE_TYPE processQueryReply(dataMsg, msg, reply)
*
* DESCRIPTION: 
* Decodes dataMsg using msg format information and storing it in reply. 
* Reclaims dataMsg storage.
*
* INPUTS:
* DATA_MSG_PTR dataMsg;
* MSG_PTR msg;
* char *reply;
*
* OUTPUTS: TCA_RETURN_VALUE_TYPE
*
******************************************************************************/

TCA_RETURN_VALUE_TYPE processQueryReply(dataMsg, msg, reply, sel)
DATA_MSG_PTR dataMsg;
MSG_PTR msg;
char *reply;
int sel;
{
  char *r2;
  int amount;
  BUFFER_TYPE buffer;

  if (!msg) {
    tcaModError("ERROR: processQueryReply: NULL msg.\n");
  }

  if (dataMsg->msgTotal == 0) {
    dataMsgFree(dataMsg);
    return NullReply;
  }

  if (reply == &LispFlagGlobal) {
    buffer.bstart = 0;
    buffer.buffer = NULL;

    if (msg->msgData->resForm || (sel && msg->msgData->msgForm)) {
      buffer.buffer = dataMsg->msgData;
      if (sel)
	r2 = (char *)(*lispDecodeMsgGlobal)(msg->msgData->msgForm, &buffer);
      else
	r2 = (char *)(*lispDecodeMsgGlobal)(msg->msgData->resForm, &buffer);
    }
    else
      r2 = (char *)(*lispDecodeMsgGlobal)(0, NULL);      

    (*lispReplyMsgGlobal)(r2);
  }
  else {
    if (sel) {
      r2 = (char *)dataMsgDecodeMsg(msg->msgData->msgForm, dataMsg);
      amount = dataStructureSize(msg->msgData->msgForm);
    }
    else {
      r2 = (char *)dataMsgDecodeMsg(msg->msgData->resForm, dataMsg);
      amount = dataStructureSize(msg->msgData->resForm);
    }
    bcopy(r2, reply, amount);
    tcaFree((char *)r2);
  }

  dataMsgFree(dataMsg);
  return Success;
}


/******************************************************************************
*
* FUNCTION: TCA_RETURN_VALUE_TYPE waitForReply(ref, reply, timeout)
*
* DESCRIPTION: Loops to handle messages while waiting for a query reply.
*
* INPUTS:
* TCA_REF_PTR ref;
* char *reply;
* long timeout;
*
* OUTPUTS: TCA_RETURN_VALUE_TYPE
*
* NOTES: Modified from tcaSingleMsgHnd
*
******************************************************************************/

int testMsgRef(refId, dataMsg)
int *refId;
DATA_MSG_PTR dataMsg;
{
  if (dataMsg->msgRef == *refId)
    return TRUE;

  return FALSE;
}

TCA_RETURN_VALUE_TYPE waitForReply2(ref, reply, sel)
TCA_REF_PTR ref;
char *reply;
int sel;
{
  int ret;
  fd_set readMask; 
  DATA_MSG_PTR dataMsg;

  if (!ref) {
    tcaModError("ERROR: waitForReply: NULL ref.\n");
  }

  for (;;) {
    dataMsg = (DATA_MSG_PTR)listMemReturnItem(testMsgRef, (char *)&ref->refId, 
					      tcaPendingRepliesGlobal);
    if (dataMsg) {
      listDeleteItem((char *)dataMsg, tcaPendingRepliesGlobal);
      return processQueryReply(dataMsg, ref->msg, reply, sel);
    }

#ifndef VXWORKS
    readMask = tcaConnectionListGlobal;
    do {
      ret = select(FD_SETSIZE, &readMask, (fd_set *)NULL, (fd_set *)NULL, 
		   (struct timeval *)NULL);
    } while (ret < 0 && errno == EINTR);
    if (ret < 1) {
      tcaModError("ERROR: WaitForReply: select: ");
    }
#endif
    switch(dataMsgRecv(tcaServerGlobal, &dataMsg)){
    case StatOK:
      if (dataMsg->intent == QUERY_REPLY_INTENT) {
	if (dataMsg->msgRef == ref->refId) 
	  return processQueryReply(dataMsg, ref->msg, reply, sel);
	listInsertItem((char *)dataMsg, tcaPendingRepliesGlobal);
      }
      else 
	execHnd((DIRECT_INFO_PTR)NULL, dataMsg);
      break;
    case StatError:
      tcaModError("ERROR: waitForReply: Reading Data Message.");
    case StatEOF:
      fprintf(stderr, "Closed Connection Detected from Central Server: sd: %d:\n", 
	      tcaServerGlobal);
#ifdef LISP
      fflush(stderr);
#endif
      (void)shutdown(tcaServerGlobal, 2);
      (void)close(tcaServerGlobal);
#ifdef LISP
      (*lispExitGlobal)();
#else
      if (!tcaExitHnd) {
	exit(-1);
      } else {
	(*tcaExitHnd)();
      }

#endif
      break;
    default:
      tcaModError("ERROR: waitForReply: tcaServerGlobal: UNKNOWN INTENT\n");
    }
  }
}

int iterateModuleReply(wfr, directInfo)
WAIT_FOR_REPLY_PTR wfr;
DIRECT_INFO_PTR directInfo;
{
  int sel;
  char *reply;
  TCA_REF_PTR ref;
  fd_set *readMask;
  DATA_MSG_PTR dataMsg;

  sel = wfr->sel;
  reply = wfr->replyData;
  ref = wfr->ref;
  readMask = wfr->readMask;
  
  if (FD_ISSET(directInfo->sd, readMask)) {
    FD_CLR(directInfo->sd, readMask);
    switch(dataMsgRecv(directInfo->sd, &dataMsg)){
    case StatOK:
      if (dataMsg->intent == QUERY_REPLY_INTENT) {
	if (dataMsg->msgRef == ref->refId) {
	  wfr->status = processQueryReply(dataMsg, ref->msg, reply, sel);
	  return 0;
	}
	listInsertItem((char *)dataMsg, tcaPendingRepliesGlobal);
      }
      else 
	execHnd(directInfo, dataMsg);
      return 1;
    case StatError:
      tcaModError("ERROR: waitForReply3: Reading Data Message.");
    case StatEOF:
      FD_CLR(directInfo->sd, &tcaConnectionListGlobal);
#ifndef VXWORKS
      (void)shutdown(directInfo->sd, 2);
      (void)close(directInfo->sd);
      if (directInfo->sd == tcaServerGlobal) {
	fprintf(stderr,"Closed Connection Detected from: Central Server: sd: %d:\n",
		directInfo->sd);
#ifdef LISP
	fflush(stderr);
#endif
	listDeleteItem((char *)directInfo, moduleConnectionList);
	tcaFree((char *)directInfo);
#ifdef LISP
	(*lispExitGlobal)();
#else
	if (!tcaExitHnd) {
	  exit(-1);
	} else {
	  (*tcaExitHnd)();
	}

#endif
      } else {
	fprintf(stderr, "Closed Connection Detected from: Module: sd: %d:\n", 
		directInfo->sd);
#ifdef LISP
	fflush(stderr);
#endif
	listDeleteItem((char *)directInfo, moduleConnectionList);
	tcaFree((char *)directInfo);
      }      
#endif
#ifdef VXWORKS
      (void)close(directInfo->sd);
      if (directInfo->sd == tcaServerGlobal) {
	fprintf(stderr,"Closed Connection Detected from: Central Server: sd: %d:\n",
		directInfo->sd);
	listDeleteItem((char *)directInfo, moduleConnectionList);
	tcaFree((char *)directInfo);
      } else {
	fprintf(stderr, "Closed Connection Detected from: Module: sd: %d:\n", 
		directInfo->sd);
	listDeleteItem((char *)directInfo, moduleConnectionList);
	tcaFree((char *)directInfo);

	if (!tcaExitHnd) {
	  exit(-1);
	} else {
	  (*tcaExitHnd)();
	}
	
      }
#endif
      break;
    default:
      tcaModError("ERROR: waitForReply3: UNKNOWN INTENT\n");
    }
  }
  return TRUE;
}

fd_set readMaskGlobal;

/*****************
 * Returns Timeout if "timeout" seconds elapse without receiving a message.
 *   WAITFOREVER always waits until reply received (a very long time, at least).
 ****************/
TCA_RETURN_VALUE_TYPE waitForReply3(ref, reply, sel, timeout)
TCA_REF_PTR ref;
char *reply;
int sel;
long timeout;
{
  int ret, sd;
  DATA_MSG_PTR dataMsg;
  WAIT_FOR_REPLY_TYPE wfr;
  DIRECT_INFO_PTR directInfo;
  struct timeval time;
  
  if (timeout != WAITFOREVER) {
    time.tv_usec = 0;
    time.tv_sec = timeout;
  }

  if (!ref) {
    tcaModError("ERROR: waitForReply: NULL ref.\n");
  }

  for(;;) {

    dataMsg = (DATA_MSG_PTR)listMemReturnItem(testMsgRef, (char *)&ref->refId, 
					      tcaPendingRepliesGlobal);
    if (dataMsg) {
      listDeleteItem((char *)dataMsg, tcaPendingRepliesGlobal);
      return processQueryReply(dataMsg, ref->msg, reply, sel);
    }

#ifndef VXWORKS
    readMaskGlobal = tcaConnectionListGlobal;
    do {
      ret = select(FD_SETSIZE, &readMaskGlobal, (fd_set *)NULL, (fd_set *)NULL,
		   ((timeout == WAITFOREVER) ? NULL : &time));
    } while (ret < 0 && errno == EINTR);
    if (ret == 0) {
      return TimeOut;
    } else if (ret < 1) {
      tcaModError("ERROR: WaitForReply: select: ");
    }
#endif

    if (directFlagGlobal && FD_ISSET(tcaListenGlobal, &readMaskGlobal)) {
      sd = accept(tcaListenGlobal, (struct sockaddr *)NULL, (int *)NULL);

      directInfo = NEW(DIRECT_INFO_TYPE);
      directInfo->sd = sd;

      fprintf(stderr, "New Connection Established for a Module at sd: %d\n",
	      directInfo->sd);
#ifdef LISP
      fflush(stderr);
#endif
      listInsertItem((char *)directInfo, moduleConnectionList);
      FD_SET(sd, &tcaConnectionListGlobal);
      /* 25-Jun-91: fedor: set sd in readMask as well? */
    }

#ifdef VXWORKS
    /* 26-Jun-91: fedor: VxWorks sits in a read point-to-point work
       will have to wait until this is rethought */
    FD_ZERO(&readMaskGlobal);
    FD_SET(tcaServerGlobal, &readMaskGlobal);
#endif

    wfr.sel = sel;
    wfr.readMask = &readMaskGlobal;
    wfr.ref = ref;
    wfr.replyData = reply;

    if (!listIterate(iterateModuleReply, (char *)&wfr, moduleConnectionList)) {
      return wfr.status;
    }
  }
}


TCA_RETURN_VALUE_TYPE waitForReply(ref, reply)
TCA_REF_PTR ref;
char *reply;
{
  return waitForReply3(ref, reply, 0, WAITFOREVER);
}


/******************************************************************************
*
* FUNCTION: TCA_RETURN_VALUE_TYPE tcaHandleMessage(sec)
*
* DESCRIPTION: 
*
* INPUTS: long sec;
*
* OUTPUTS: TCA_RETURN_VALUE_TYPE
*
******************************************************************************/

TCA_RETURN_VALUE_TYPE tcaHandleMessage(sec)
long sec;
{
  int ret;
  struct timeval time;

  if (sec == WAITFOREVER)
      ret = tcaSingleMsgHnd((struct timeval *)NULL);
  else {
    time.tv_sec = sec;
    time.tv_usec = 0;
    ret = tcaSingleMsgHnd(&time);
  }

  switch(ret) {
  case 0:
    return TimeOut;
  case -1:
    return Failure;
  default:
    return Success;
  }
}


/******************************************************************************
*
* FUNCTION: void tcaModuleListen()
*
* DESCRIPTION: 
*
* INPUTS: none.
* 
* OUTPUTS: void.
*
******************************************************************************/

void tcaModuleListen()
{
  for(;;)
    (void)tcaHandleMessage(WAITFOREVER);
}


/******************************************************************************
*
* FUNCTION: void tcaRegisterExitHandler(proc)
*
* DESCRIPTION: 
*
* INPUTS:
* int (*proc)();
* 
* OUTPUTS: void.
*
******************************************************************************/

void tcaRegisterExitProc(proc)
void (*proc)();
{
  tcaExitHnd = proc;
}

/****************************************************************************/

void tcaDirectResource(name)
char *name;
{
  int port;
  DIRECT_TYPE direct;

  port = LISTEN_PORT;
  while ((port < LISTEN_PORT_MAX) && !listenAtPort(port, &tcaListenGlobal))
    port++;

  if (port == LISTEN_PORT_MAX) {
    tcaModError("ERROR: Exceeded available ports for direct connection.\n");
  }

  directFlagGlobal = TRUE;
  FD_SET(tcaListenGlobal, &tcaConnectionListGlobal);

  direct.port = port;
  direct.name = name;
  
  (void)tcaInform("directResMsg", (void *)&direct);
}

/****************************************************************************/

int testInfo(directInfo, infoItem)
DIRECT_INFO_PTR directInfo, infoItem;
{
  if (directInfo->port == infoItem->port &&
      strKeyEqFunc(directInfo->host, infoItem->host))
    return TRUE;
  else
    return FALSE;
}

void establishDirect(msg)
MSG_PTR msg;
{
  int sd;
  DIRECT_INFO_PTR directInfo, oldInfo;

  directInfo = NEW(DIRECT_INFO_TYPE);

  (void)tcaQuery("directInfoMsg", (void *)&msg->msgData->name, (void *)directInfo);

  msg->direct = TRUE;
  msg->intent = directInfo->intent;

  oldInfo = (DIRECT_INFO_PTR)listMemReturnItem(testInfo, (char *)directInfo,
					       moduleConnectionList);
  if (oldInfo) {
    msg->sd = oldInfo->sd;
    tcaFreeReply("directInfoMsg", (void *)directInfo);
    fprintf(stderr, "Using existing direct connection for %s to %s.\n",
	    msg->msgData->name, oldInfo->module);
#ifdef LISP
    fflush(stderr);
#endif
    return;
  } 

  if (!connectAtPort(directInfo->host, directInfo->port, &sd)) {
    TCA_MOD_ERROR2("ERROR: Can not establish direct connection to %s at %d.\n",
		   directInfo->host, directInfo->port);
  } else {
    directFlagGlobal = TRUE;
    fprintf(stderr, "Direct Connection Established for %s to %s.\n", 
	    msg->msgData->name, directInfo->module);
#ifdef LISP
    fflush(stderr);
#endif
  }

  directInfo->sd = sd;

  /* 26-Jun-91: fedor
     attempt to make the server the first item
     otherwise things get overrun on the server socket
     in waitForReply - going to have to redesign this nonsense! */
  listInsertItemLast((char *)directInfo, moduleConnectionList);
  FD_SET(sd, &tcaConnectionListGlobal);  

  msg->sd = sd;
}
