/******************************************************************************
*
* PROJECT: Carnegie Mellon Planetary Rover Project
*          Task Control Architecture 
* 
* (c) Copyright 1991 Christopher Fedor and Reid Simmons.  All rights reserved.
* 
* MODULE: dispatch
*
* FILE: dispatch.c
*
* ABSTRACT:
* 
* A dispatch represents a single message and the message's progress.
*
* REVISION HISTORY
*
* 26-Jul-91 Reid Simmons, School of Computer Science, CMU
* Added "refCount" field -- the problem is that a dispatch could be freed
* when another module still has a refence to it.  For example, one module
* can kill a task tree node that another wants to add temporal constraints to.
* This can cause nasty things to happen.  As a result, free a dispatch (and
* task tree node) only if the refCount is zero.
*
* 14-May-91 Christopher Fedor, School of Computer Science, CMU
* Moved dispatch update and display code to here from main recv message loop.
*
* 14-Mar-91  Reid Simmons, School of Computer Science, CMU
* Fixed constraint messages from printing out a blank line when they completed.
*
* 16-Sep-90 Christopher Fedor, School of Computer Science, CMU
* Revised to Software Standards.
*
*  6-Jul-89 Reid Simmons, School of Computer Science, CMU
* Split out server (central) related code into "dispatch_server.c"
*
* 24-Apr-89 Christopher Fedor, School of Computer Science, CMU
* Changed dispatch to cache selected Msg and Hnd.
* Removed arguments from AllocDispatch.
*
*  5-Apr-89 Christopher Fedor, School of Computer Science, CMU
* Changed to use IdTable.
*
* 20-Mar-89 Christopher Fedor, School of Computer Science, CMU
* created
*
******************************************************************************/

#include "stdio.h"

#include "basics.h"
#include "tcaInternal.h"
#include "logging.h"
#include "datamsg.h"
#include "idtable.h"
#include "hash.h"
#include "formatters.h"

#include "dispatch.h"
#include "res.h"
#include "msgTap.h"

#define DISPATCH_FREE_INC 10

extern ID_TABLE_PTR hndIdTable;
extern MODULE_PTR tcaServerModGlobal;

extern HASH_TABLE_PTR classFormatTable;

ID_TABLE_PTR dispatchTable;
DISPATCH_PTR dispatchFreeListGlobal = NULL;

extern void freeTapInfo(); /* defined in msgTap.c */
extern char *messageClassName(); /* defined in behaviors.c */
extern void recvMessage(); /* defined in recvMsg.c */
extern void HandleKillAfterAttendingNodes(); /* defined in taskTree.c */

/* 15-Oct-90: fedor: old logging code - this needs to be redone!! */

extern int Ignore_Logging_Message();

extern void classDataFree(); /* defined in formatters.c */

/******************************************************************************
*
* FUNCTION: void dispatchIncrementFreeList()
*
* DESCRIPTION:
* Add DISPATCH_FREE_INC number of dispatches to the free list.
* Insert each newly created dispatch into the dispatchTable.
*
* INPUTS: none.
*
* OUTPUTS: none.
*
******************************************************************************/

void dispatchIncrementFreeList()
{
  int i;
  DISPATCH_PTR new, current;

  current = NULL;
  for(i=0;i<DISPATCH_FREE_INC;i++) {
    new = NEW(DISPATCH_TYPE);

    new->refId = NO_REF;
    new->orgId = 0;
    new->desId = 0;
    new->pRef = NO_REF; /* 9-Jul-90: fedor: change to ROOT_NODE */
    new->org = NULL;
    new->des = NULL;

    new->msg = NULL;
    new->hnd = NULL;
    new->resource = NULL;

    new->msgData = NULL;
    new->resData = NULL;

    new->treeNode = NULL;
    new->blockCom = NULL;
    new->status = UnallocatedDispatch;

    new->classData = NULL;
    new->class = UNKNOWN;

    new->respProc = NULL;
    new->respData = NULL;

    new->tapInfo = NULL;

    new->refCount = 0;

    new->locId = idTableInsert((char *)new, dispatchTable);

    new->next = current;
    current = new;
  }
  dispatchFreeListGlobal = current;
}


/******************************************************************************
*
* FUNCTION: DISPATCH_PTR dispatchAllocate()
*
* DESCRIPTION:
* Return a newly initialized dispatch from the free list.
*
* INPUTS: none.
*
* OUTPUTS: DISPATCH_PTR
*
******************************************************************************/

DISPATCH_PTR dispatchAllocate()
{
  DISPATCH_PTR tmp;

  if (!dispatchFreeListGlobal)
    dispatchIncrementFreeList();

  tmp = dispatchFreeListGlobal;

  dispatchFreeListGlobal = dispatchFreeListGlobal->next;

  tmp->next = NULL;

  return tmp;
}


/******************************************************************************
*
* FUNCTION: void dispatchFree(dispatch)
*
* DESCRIPTION:
* Initialize the dispatch and return it to the free list.
* Call dataMsgFree on the msgData and resData slots.
*
* INPUTS: DISPATCH_PTR dispatch
*
* OUTPUTS: none.
*
* NOTES: 
* 17-Oct-91: fedor: added check that dispatch has not already been freed.
* The next field should be NULL. If it gets freed more than once the 
* dispatchFreeListGlobal will become a circular list of one dispatch.
*
* 26-Jul-91: Reid: Do not free if refCount > 0.
*
******************************************************************************/

void dispatchFree(dispatch)
DISPATCH_PTR dispatch;
{
  if (dispatch->refCount <= 0 && !dispatch->next) {
    dispatch->refId = NO_REF;
    dispatch->orgId = 0;
    dispatch->desId = 0;
    dispatch->pRef = NO_REF;	/* 9-Jul-90: fedor: change to ROOT_NODE */
    dispatch->org = NULL;
    dispatch->des = NULL;

    dispatch->msg = NULL;
    dispatch->hnd = NULL;

    dispatch->resource = NULL;

    dispatch->treeNode = NULL;
    dispatch->blockCom = NULL;
    dispatch->status = UnallocatedDispatch;

    dataMsgFree(dispatch->msgData);
    dataMsgFree(dispatch->resData);

    dispatch->msgData = NULL;
    dispatch->resData = NULL;

    /* Command class data is stored in "blockCom"; It has already been freed */
    if (dispatch->class != CommandClass) 
      classDataFree(dispatch->class, dispatch->classData);
    dispatch->classData = NULL;
    dispatch->class = UNKNOWN;

    dispatch->respProc = NULL;
    dispatch->respData = NULL;

    freeTapInfo(dispatch->tapInfo);
    dispatch->tapInfo = NULL;

    dispatch->refCount = 0;

    dispatch->next = dispatchFreeListGlobal;
    dispatchFreeListGlobal = dispatch;
  }
}

/******************************************************************************
*
* FUNCTION: void dispatchSetMsgData(msgData, dispatch)
*
* DESCRIPTION:
* Set the dispatch msgData to the new msgData calling dataMsgFree on 
* any existing msgData. Increment the new msgData reference count.
*
* INPUTS: 
* DATA_MSG_PTR msgData;
* DISPATCH_PTR dispatch;
*
* OUTPUTS: none.
*
******************************************************************************/

void dispatchSetMsgData(msgData, dispatch)
DATA_MSG_PTR msgData;
DISPATCH_PTR dispatch;
{
  dataMsgFree(dispatch->msgData);
  dispatch->msgData = msgData;
  if (msgData)
    msgData->refCount++;
}


/******************************************************************************
*
* FUNCTION: void dispatchSetResData(resData, dispatch)
*
* DESCRIPTION:
* Set the dispatch resData to the new resData calling dataMsgFree on 
* any existing resData. Increment the new resData reference count.
*
* INPUTS: 
* DATA_MSG_PTR resData;
* DISPATCH_PTR dispatch;
*
* OUTPUTS: none.
*
******************************************************************************/

void dispatchSetResData(resData, dispatch)
DATA_MSG_PTR resData;
DISPATCH_PTR dispatch;
{
  dataMsgFree(dispatch->resData);
  dispatch->resData = resData;
  if (resData)
    resData->refCount++;
}


/******************************************************************************
*
* FUNCTION: DISPATCH_PTR dispatchCopy(dispatch)
*
* DESCRIPTION:
* Create a copy of the dispatch, including all its slots (except for locId)
*
* INPUTS: DISPATCH_PTR dispatch
*
* OUTPUTS: DISPATCH_PTR: Returns the duplicated dispatch.
*
* NOTES: 
******************************************************************************/

DISPATCH_PTR dispatchCopy(dispatch)
DISPATCH_PTR dispatch;
{
  DISPATCH_PTR newDispatch;
  int locId;

  newDispatch = dispatchAllocate();
  locId = newDispatch->locId;

  bcopy((char *)dispatch, (char *)newDispatch, sizeof(DISPATCH_TYPE));

  DISPATCH_MSG_DATA(newDispatch) = NULL;
  dispatchSetMsgData(DISPATCH_MSG_DATA(dispatch), newDispatch);
  newDispatch->locId = locId;

  return newDispatch;
}


/******************************************************************************
*
* FUNCTION: char *dispatchStatusName(status)
*
* DESCRIPTION: Return a printable string version of the dispatch status.
*
* INPUTS: DISPATCH_STATUS_TYPE status;
*
* OUTPUTS: char *
*
******************************************************************************/

char *dispatchStatusName(status)
DISPATCH_STATUS_TYPE status;
{
  switch(status) {
  case UnallocatedDispatch:
    return("Unallocated");
  case AllocatedDispatch:
    return("Allocated");
  case InactiveDispatch:
    return("Inactive");
  case PendingDispatch:
    return("Pending");
  case AttendingDispatch:
    return("Attending");
  case HandledDispatch:
    return("Handled");
  case KilledDispatch:
    return("Killed");
  case ReplyDispatch:
    return("Reply");
  case SuccessDispatch:
    return("Success");
  case FailureDispatch:
    return("Failure");
  default:
    return("UNKNOWN");
  }
}


/******************************************************************************
*
* FUNCTION: void dispatchUpdateAndDisplay1(newStatus, dispatch)
*
* DESCRIPTION:
*
* INPUTS: 
*
* OUTPUTS:
*
******************************************************************************/

void dispatchUpdateAndDisplay1(newStatus, dispatch)
DISPATCH_STATUS_TYPE newStatus;
DISPATCH_PTR dispatch;
{
  int error = FALSE;
  int printed = TRUE;

  switch(dispatch->status) {
  case AllocatedDispatch:
    switch(newStatus) {
    case AttendingDispatch:
      Log_Message1("%-9s", messageClassName(dispatch->class));
      Log_Message1("%15s: ", dispatch->msg->msgData->name);
      Log_Message1("%10s", dispatch->org->modData->modName);
      Log_Message1(" --> %-15s (Sent)", dispatch->des->modData->modName);
      break;
    case PendingDispatch:
      Log_Message1("%-9s", messageClassName(dispatch->class));
      Log_Message1("%15s: ", dispatch->msg->msgData->name);
      Log_Message1("%10s", dispatch->org->modData->modName);

      Log_Message1(" --> Resource %-6s (Pending)", 
		   dispatch->hnd->resource->name); 
      break;
    case InactiveDispatch:
      Log_Message1("%-9s", messageClassName(dispatch->class));
      Log_Message1("%15s: ", dispatch->msg->msgData->name);
      Log_Message1("%10s", dispatch->org->modData->modName);

      Log_Message(" --> ON HOLD         (Inactive)");
      break;
    case HandledDispatch:
      if (dispatch->class == FailureClass) {
	Log_Message2("Failure  %15s: %s",
		     "*Top Level*",
		     *(char **)dispatch->classData);
      }
      break;
    default:
      error = TRUE;
      break;
    }
    break;
  case PendingDispatch:
    switch(newStatus) {
    case AttendingDispatch:
      Log_Message1("%-9s", messageClassName(dispatch->class));
      Log_Message1("%15s: ", dispatch->msg->msgData->name);
      Log_Message1("Resource %s", dispatch->hnd->resource->name); 
      Log_Message1(" --> %-15s (Sent)", dispatch->des->modData->modName);
      break;
    default:
      error = TRUE;
      break;
    }
    break;
  case CentralPendingDispatch:
    switch(newStatus) {
    case AttendingDispatch:
      Log_Message1("%-9s", messageClassName(dispatch->class));
      Log_Message1("%15s: ", dispatch->msg->msgData->name);
      Log_Message1("%10s", dispatch->org->modData->modName);
      Log_Message1(" --> %-15s (Sent)", dispatch->des->modData->modName);
      break;
    default:
      error = TRUE;
      break;
    }
    break;
  case AttendingDispatch:
    switch(newStatus) {
    case HandledDispatch:
      if (TWO_WAY_MSG(dispatch->msg)) {
	Log_Message("  Reply  "); 
	Log_Message1("%15s: ", dispatch->msg->msgData->name);
	Log_Message2("%10s --> %-15s", 
		     dispatch->des->modData->modName, 
		     dispatch->org->modData->modName);
	if (!dispatch->org->alive) {
	  Log_Message1("\nWARNING: Reply Ignored for dead module: %s",
		      dispatch->org->modData->modName);
	}
      }
      else if (monitorClass(dispatch->class)) {
	Log_Message1("%-9s", messageClassName(dispatch->class));
	Log_Message1("%15s: ", dispatch->msg->msgData->name);
	Log_Message("                               (Completed)");
      }
      else if (dispatch->class == SuccessClass) {
	if (dispatch->msg->msgData->class == InformClass ||
	    dispatch->msg->msgData->class == BroadcastClass)
	  Log_Message("     Done");
	else
	  Log_Message("  Success");
	Log_Message1("%15s: ", dispatch->msg->msgData->name);
      }
      else if (dispatch->class == FailureClass) {
	Log_Message2("  Failure%15s:   %s",
		     dispatch->msg->msgData->name, 
		     *(char **)dispatch->classData);
      }
      else if (dispatch->class == ReplyClass) {
	Log_Message("  Reply  "); 
	Log_Message1("%15s: ", dispatch->msg->msgData->name);
	Log_Message2("%10s --> %-15s", 
		     tcaServerModGlobal->modData->modName,
		     dispatch->org->modData->modName);
      } else {
	printed = FALSE;
      }
      break;
    case CentralPendingDispatch:
      Log_Message1("%-9s", messageClassName(dispatch->class));
      Log_Message1("%15s: ", dispatch->msg->msgData->name);
      Log_Message1("%10s", dispatch->org->modData->modName);
      Log_Message1(" --> %-15s (Pending)", dispatch->des->modData->modName);
      break;
    default:
      error = TRUE;
      break;
    }
    break;
  case InactiveDispatch:
    Log_Message1("%-9s", messageClassName(dispatch->class));
    Log_Message1("%15s: ", dispatch->msg->msgData->name);
    Log_Message("   ON HOLD");
    switch(newStatus) {
    case AttendingDispatch:
      Log_Message1(" --> %-15s (Sent)", dispatch->des->modData->modName);
      break;
    case PendingDispatch:
      Log_Message1(" --> Resource %-6s (Pending)", dispatch->hnd->resource->name); 
      break;
    default:
      error = TRUE;
      break;
    }
    break;
  default:
    error = TRUE;
    break;
  }
  
  if (error) {
    Log_Message2("WARNING: dispatchUpdateAndDisplay: %s --> %s",
		 dispatchStatusName(dispatch->status),
		 dispatchStatusName(newStatus));
  }

  /* 17-Nov-90: fedor: quick fix to display time. */
  if (printed) {
    Log_Time(1); 

    Log_Message("\n");
  }

  DISPATCH_SET_STATUS(newStatus, dispatch);
}


/******************************************************************************
*
* FUNCTION: void dispatchUpdateAndDisplay(newStatus, dispatch)
*
* DESCRIPTION: 
*
* INPUTS:
*
* OUTPUTS:
*
******************************************************************************/

void dispatchUpdateAndDisplay(newStatus, dispatch)
DISPATCH_STATUS_TYPE newStatus;
DISPATCH_PTR dispatch;
{
  char *Data;
  FORMAT_PTR Format;
  int ignored;

  if (Terminal_Log_Messages || File_Log_Messages) {

    ignored = Ignore_Logging_Message(dispatch->msg);
    if (ignored) Start_Ignore_Logging();

    if (Terminal_Log_Messages || File_Log_Messages) { 

      dispatchUpdateAndDisplay1(newStatus, dispatch);

      if ((Terminal_Log_Data || File_Log_Data) &&
	  ((newStatus == AttendingDispatch) ||
	   (dispatch->class == ReplyClass))) {

	if (dispatch->msg) {
	  Data = NULL;
	  if (newStatus == HandledDispatch) {
	    Format = dispatch->msg->msgData->resForm;
	    Data = dataMsgDecodeMsg(Format, dispatch->resData);
	  }
	  else {
	    Format = dispatch->msg->msgData->msgForm;
	    Data = dataMsgDecodeMsg(Format, dispatch->msgData);
	  }

	  Log_Data(Format, Data, 4);
	  if (Data) { /* free decoded data, including all its subelements */
	    freeDataStructure(Format, Data);
	  }
	}
      }
    }
    if (ignored) End_Ignore_Logging();
  }
  else
    DISPATCH_SET_STATUS(newStatus, dispatch);
}



/******************************************************************************
*
* FUNCTION: void deliverDispatch(dispatch)
*
* DESCRIPTION: 
* Delivers a dispatch to the correct module to select a handler.
* Triggers a msgFind on the module side if the format has changed for this
* handler request. This allows multiple formats (multiple messages with 
* different formats) for a single handler.  The handler code can simply cast 
* based on a tag value - or request the message name as a tag from the ref.
*
* INPUTS: DISPATCH_PTR dispatch;
*
* OUTPUTS: void.
*
* NOTES:
* 
*
******************************************************************************/

int monitorClass(class)
TCA_MSG_CLASS_TYPE class;
{
  return(class == PointMonitorClass || class == PollingMonitorClass ||
	 class == DemonMonitorClass);
}

void deliverDispatch(dispatch)
DISPATCH_PTR dispatch;
{
  char *data;
  DATA_MSG_PTR newDataMsg;
  TCA_MSG_CLASS_TYPE class;
  CLASS_FORM_PTR classForm;

  data = NULL;

  dispatchUpdateAndDisplay(AttendingDispatch, dispatch);

  if (dispatch->des == tcaServerModGlobal) {
    if (monitorClass(dispatch->msg->msgData->class))
      data = dispatch->classData;
    else
      data = dataMsgDecodeMsg(dispatch->msg->msgData->msgForm, 
			      DISPATCH_MSG_DATA(dispatch));
    (*dispatch->hnd->hndProc)(dispatch, data);
  }
  else {
    if (dispatch->hnd->msg != dispatch->msg) {
      /* handler to recieve new message */
      class = ExecHndClass;
      classForm = (CLASS_FORM_PTR)hashTableFind((char *)&class, 
						classFormatTable);
      if (!classForm)
	tcaError("ERROR: deliverDispatch: missing ExecHndClass class format.");

      newDataMsg = 
	dataMsgReplaceClassData(classForm->format,
				(char *)&dispatch->msg->msgData->name,
				dispatch->msgData);
      
      /* 31-Oct-90: fedor: the dataMsgReplaceClassData routine calls
	 dataMsgFree - this avoids calling it twice and gets the 
	 refs correct. blah! */
      dispatch->msgData = NULL;

      dispatchSetMsgData(newDataMsg, dispatch);
    }
    else {
      /* The datamags may already have old class data.
	 Since class data follows message data, this
	 will prevent old class data from being sent. */
      dispatch->msgData->classTotal = 0;
    }

    dispatch->msgData->intent = dispatch->hnd->hndData->refId;
    dispatch->msgData->msgRef = dispatch->locId;

    (void)dataMsgSend(dispatch->desId, dispatch->msgData);
  }
}


/******************************************************************************
*
* FUNCTION: void deliverResponse(dispatch)
*
* DESCRIPTION: 
* Delivers a response dispatch to an internal procedure or to a query reply.
*
* INPUTS: 
* DISPATCH_PTR dispatch;
*
* OUTPUTS: void.
*
******************************************************************************/

void deliverResponse(dispatch)
DISPATCH_PTR dispatch;
{
  if (dispatch->respProc) {
    (*dispatch->respProc)(dispatch, dispatch->respData);
    return;
  }

  pendingReplyRemove(dispatch);

  if (dispatch->org->alive) {
    dispatch->resData->intent = QUERY_REPLY_INTENT;
    dispatch->resData->msgRef = dispatch->refId;
  
    (void)dataMsgSend(dispatch->orgId, DISPATCH_RES_DATA(dispatch));
  } else if (!dispatch->org->repliesPending) {
    moduleFree(dispatch->org);
  }
}


/******************************************************************************
*
* FUNCTION: void centralReply(dispatch, data)
*
* DESCRIPTION: allows central to reply like a module.
*
* INPUTS: 
* DISPATCH_PTR dispatch;
* char *data;
*
* OUTPUTS: void
*
******************************************************************************/

void centralReply(dispatch, data)
DISPATCH_PTR dispatch;
char *data;
{
  dispatch->resData = dataMsgCreate(0, 0, 0, 0, 0,
				    dispatch->msg->msgData->resForm, data,
				    (FORMAT_PTR)NULL, (char *)NULL);

  /* 20-Nov-90: fedor: to be consistent with recvMesgBuild */
  dispatch->class = ReplyClass;
  recvMessage(dispatch, ReplyClass, dispatch->classData);
}


/******************************************************************************
*
* FUNCTION: char *dispatchDecodeResponse(dispatch)
*
* DESCRIPTION: Make use of the msg resForm to decode the res datamsg.
*
* INPUTS: DISPATCH_PTR dispatch;
*
* OUTPUTS: char *
*
******************************************************************************/

char *dispatchDecodeResponse(dispatch)
DISPATCH_PTR dispatch;
{
  return(dataMsgDecodeMsg(dispatch->msg->msgData->resForm,
			  DISPATCH_RES_DATA(dispatch)));
}

/******************************************************************************
*
* FUNCTION: void reserveDispatch(dispatch)
*           void releaseDispatch(dispatch)
*
* DESCRIPTION: Some module has an outstanding reference to this dispatch.
*              "reserveDispatch" records that fact, "releaseDispatch" removes
*              that assertion.
*
* INPUTS: DISPATCH_PTR dispatch;
*
* OUTPUTS: none
*
* NOTES: With "releaseDispatch", need to check if the associated task tree
*        node is waiting to be killed.
*
******************************************************************************/
void reserveDispatch(dispatch)
     DISPATCH_PTR dispatch;
{
  if (dispatch) {
    dispatch->refCount++;
  }
}

void releaseDispatch(dispatch)
     DISPATCH_PTR dispatch;
{
  if (dispatch) {

    if (dispatch->refCount == 0) {
      Log2("WARNING: Releasing an unreserved dispatch %s (%d)\n",
	   dispatch->msg->msgData->name, (char *)dispatch->locId);
    } else {
      dispatch->refCount--;
    }

    if (dispatch->refCount <= 0) {
      HandleKillAfterAttendingNodes(dispatch);
      if ((dispatch->status == HandledDispatch) && !dispatch->treeNode) {
	dispatchFree(dispatch);
      }
    }
  }
}

/******************************************************************************
*
* FUNCTION: void dispatchStats
*
* DESCRIPTION: Print stats about the utilization of dispatches in central
*
* INPUTS: 
*
* OUTPUTS: 
*
******************************************************************************/

void dispatchStats ()
{
  int i, num=0;
  DISPATCH_PTR dispatch;

  for (i=0; i<dispatchTable->currentSize; i++) {
    dispatch = DISPATCH_FROM_ID(i);
    if (dispatch->status != UnallocatedDispatch) {
      Log2("%d. %s\n", (char *)(++num), dispatch->msg->msgData->name);
    }
  }
}
