/******************************************************************************
*
* PROJECT: Carnegie Mellon Planetary Rover Project
*          Task Control Architecture 
* 
* (c) Copyright 1991 Christopher Fedor and Reid Simmons.  All rights reserved.
* 
* MODULE: exception handling
*
* FILE: exception.c
*
* ABSTRACT:
* 
* exception code.
*
* REVISION HISTORY
*
* 17-Feb-92 Reid Simmons, School of Computer Science, CMU
* Fixed choosing exception handlers so that task tree is killed if no
*  exception is found (fix of 31-Oct-91 just worked for byPass)
*
* 31-Dec-91 Reid Simmons, School of Computer Science, CMU
* Added facility to add exception handlers to class of messages, 
*  rather than just a single task tree node.
*
* 31-Oct-91 Christopher Fedor, School of Computer Science, CMU
* If exception handler could not be found by byPass display warning
* and kill SubTaskTree of tcaTaskTreeRootGlobal.
*
* 29-May-91 Christopher Fedor, School of Computer Science, CMU
* Cut the temporal constraints between the node to be retried
* and its children to prevent a confusion when moving "Now" backwards.
*
* 12-May-91 Christopher Fedor, School of Computer Science, CMU
* Revised to verion tca 5.x
*
* 25-Nov-89 Long-Ji Lin, School of Computer Science, CMU
* created.
*
******************************************************************************/

#include "stdio.h"

#include "tcaInternal.h"
#include "list.h"
#include "logging.h"
#include "formatters.h"
#include "hash.h"
#include "idtable.h"

#include "datamsg.h"
#include "dispatch.h"

#include "tms.h"
#include "cells.h"
#include "rules.h"
#include "qlattice.h"

#include "res.h"
#include "tplConstr.h"
#include "taskTree.h"
#include "msgTap.h"

#include "exception.h"

extern void killSubTaskTree();

extern MSG_PTR findOrRegisterMessage();

extern int removeParentChildRelations(); /* defined in tplConstr.c */

extern QUANTITY_PTR Now;

extern HASH_TABLE_PTR handlerTable;

extern TASK_TREE_NODE_PTR tcaTaskTreeRootGlobal;

extern int findHnd(); /* forward reference: defined later below */


/******************************************************************************
*
* FUNCTION: void killTopLevelTaskTreeHandler(ref, dummy)
*
* DESCRIPTION:
* Kills the top-level task tree.
*
* INPUTS:  DISPATCH_PTR dispatch;
*          char *dummy;
*
* OUTPUTS: void.
*
******************************************************************************/
void killTopLevelTaskTreeHandler (dispatch, dummy)
     DISPATCH_PTR dispatch;
     char *dummy;
{
  Log_Message1("WARNING: Unable to find a handler for the exception %s.\n",
	       dispatch->msg->msgData->name);
  Log_Message("WARNING: Killing task tree under root node.\n");

  killSubTaskTree(tcaTaskTreeRootGlobal);
  processHandledMessage(dispatch);
}    

/******************************************************************************
*
* FUNCTION: HND_PTR chooseExceptionHandler(dispatch)
*
* DESCRIPTION:
* Chooses the execption handler for the first call and for by passing handlers.
* chooseExceptionHandler will also create or update the context information.
*
* INPUTS:  DISPATCH_PTR dispatch;
*
* OUTPUTS: void.
*
******************************************************************************/

static char *findName;
static HND_PTR foundHnd;
static HND_PTR byPassHnd;

int findMatchingHnd(hndList, hnd)
LIST_PTR hndList;
HND_PTR hnd;
{
  /* byPassHnd forces the exception handler selection to pass over
     previously executed handlers in the node exception handler list. */

  if (!byPassHnd) {
    return(listMemberItem((char *)hnd, hndList));
  }
  else if (byPassHnd == hnd) {
    byPassHnd = NULL;
  }

  return 0;
}

HND_PTR chooseExceptionHandler(dispatch)
DISPATCH_PTR dispatch;
{
  HND_PTR hnd;
  int size, nodeSize;
  LIST_PTR hndList, nodeHndList;
  TASK_TREE_NODE_PTR ENode, node;

  hndList = dispatch->msg->hndList;
  size = listLength(hndList);

  if (!size)
    return NULL;

  ENode = node = dispatch->treeNode->parent;

  /* test for byPass Exception to continue from the context node */
  if (dispatch->treeNode->ECNode) {
    node = dispatch->treeNode->ECNode->CNode;
  }

  while (node) {
    nodeHndList = node->excHndList;
    nodeSize = listLength(nodeHndList);

    if (nodeSize) {
      hnd = (HND_PTR)listMemReturnItem(findMatchingHnd, (char *)hndList, 
				       nodeHndList);
      if (hnd) {
	/* set ENode and CNode */
	if (!dispatch->treeNode->ECNode) {
	  dispatch->treeNode->ECNode = NEW(ECNODE_TYPE);
	}
	dispatch->treeNode->ECNode->ENode = ENode;
	dispatch->treeNode->ECNode->CNode = node;

	return hnd;
      }
    }
    node = node->parent;
  }

  /* Added by reids, 17-Feb-92: Kill task tree if no handler found */
  foundHnd = NULL;
  findName = "killTopLevelTaskTreeHandler";
  hashTableIterate(findHnd, handlerTable);
  return foundHnd;
}


/******************************************************************************
*
* FUNCTION: void addExceptionHnd(dispatch, addInfo)
*
* DESCRIPTION: 
* Add an exception handler to a node.
* If the node already has this handler then simply do not add it again,
* this is needde if retries call hnd that add exception handlers.
*
* INPUTS: 
* DISPATCH_PTR dispatch;
* ADD_EXC_PTR addInfo;
*
* OUTPUTS: void.
*
******************************************************************************/

int findHnd(key, hnd)
HND_KEY_PTR key;
HND_PTR hnd;
{
  if (STREQ(key->name, findName)) {
    foundHnd = hnd;
    return 0;
  }

  return 1;
}

/*********************************************************/

void addExceptionHnd(dispatch, addInfo)
DISPATCH_PTR dispatch;
ADD_EXC_PTR addInfo;
{
  DISPATCH_PTR addDispatch;
  TASK_TREE_NODE_PTR addNode;

  addNode = NULL;

  if (addInfo->ref->refId == NO_REF)
    addNode = tcaTaskTreeRootGlobal;
  else {
    addDispatch = DISPATCH_FROM_ID(addInfo->ref->refId);
    
    /* 12-May-91: fedor: is this correct? */
    if (addDispatch)
      addNode = addDispatch->treeNode;
  }

  if (addNode) {
    /* 12-May-91: fedor: major bad news!- new handler key prevents
     using simple hash look up */
    foundHnd = NULL;
    findName = addInfo->hndName;
    hashTableIterate(findHnd, handlerTable);

    if (!foundHnd)
      TCA_ERROR1("ERROR: addExcepHnd: Missing handler %s\n.", addInfo->hndName);

    if (!addNode->excHndList)
      addNode->excHndList = listCreate();

    if (!listMemberItem((char *)foundHnd, addNode->excHndList))
      listInsertItem((char *)foundHnd, addNode->excHndList);
  }

  /* A bit more efficient than using tcaFreeData */
  freeDataStructure(dispatch->msg->msgData->msgForm, (char *)addInfo);
}

/******************************************************************************
*
* FUNCTION: void addExceptionToMsgHnd(dispatch, addInfo)
*
* DESCRIPTION: 
* Add an exception handler to a class of messages.
* If the message already has this handler then simply do not add it again.
* Use "findOrRegisterMessage" in case the message is not yet registered.
*
* INPUTS: 
* DISPATCH_PTR dispatch;
* ADD_EXC_MSG_PTR addInfo;
*
* OUTPUTS: void.
*
******************************************************************************/
void addExceptionToMsgHnd(dispatch, addInfo)
DISPATCH_PTR dispatch;
ADD_EXC_MSG_PTR addInfo;
{
  MSG_PTR msg;

  msg = findOrRegisterMessage(addInfo->msgName);

  foundHnd = NULL;
  findName = addInfo->hndName;
  hashTableIterate(findHnd, handlerTable);
  
  if (!foundHnd)
    TCA_ERROR1("ERROR: addExcepHnd: Missing handler %s\n.", addInfo->hndName);

  if (!msg->excepList) {
    msg->excepList = listCreate();
  }
  if (!listMemberItem((char *)foundHnd, msg->excepList)) {
    listInsertItemFirst((char *)foundHnd, msg->excepList);
  }
  /* A bit more efficient than using tcaFreeData */
  freeDataStructure(dispatch->msg->msgData->msgForm, (char *)addInfo);
}

/******************************************************************************
*
* FUNCTION: void addMsgExceptions(msg, taskTreeNode)
*
* DESCRIPTION: 
* If the message has any message-wide exceptions, add them to the exception
* list of the individual task tree node.
*
* INPUTS: 
* MSG_PTR msg;
* TASK_TREE_NODE_PTR taskTreeNode;
*
* OUTPUTS: void.
*
******************************************************************************/
void addMsgExceptions(msg, taskTreeNode)
MSG_PTR msg;
TASK_TREE_NODE_PTR taskTreeNode;
{
  HND_PTR exceptionHnd;

  if (msg->excepList) {
    if (!taskTreeNode->excHndList) {
      taskTreeNode->excHndList = listCreate();
    }
    exceptionHnd = (HND_PTR)listFirst(msg->excepList);
    while (exceptionHnd) {
      /* Insert last so as to maintain the same order as original excepList */
      listInsertItemLast((char *)exceptionHnd, taskTreeNode->excHndList);
      exceptionHnd = (HND_PTR)listNext(msg->excepList);
    }
  }
}

/******************************************************************************
*
* FUNCTION: void reExecMsg(dispatch, hnd)
*
* DESCRIPTION:
* Does the work of resending a message.
*
* The "Now" point is moved back and the node is temporally severed from its
* children to prevent the confusion of moving time backwards.
*
* The sub tree of the node is killed. And the message is resent based
* on the method in Long-Ji Lin's original code.
*
* If hnd is not NULL then it is reset. This allows this code to be
* used to resend a retry and a by passed handler with the same dispatch.
*
* Selection of a handler is ignored in recvMsg for exception class 
* dispatchs if it is already set.

* INPUTS: 
* DISPATCH_PTR dispatch;
* HND_PTR hnd;
*
* OUTPUTS: void. 
* calls processInactiveTests which will call processActiveMessage.
*
* NOTES:
* 29-May-91: fedor: If the node to reExec is not an exception class
* - most retries. Then recvMsg will reselect the handler - this may 
* be bad if multiple handlers are registered for the same message.
* Then again this may be good so I left it and wrote this cute note.
*
******************************************************************************/

void reExecMsg(dispatch, hnd)
DISPATCH_PTR dispatch;
HND_PTR hnd;
{
  int retracted;

  resourceRemoveAttending(dispatch);

  if (hnd) {
   dispatch->hnd = hnd;
   dispatch->des = hnd->hndOrg;
   dispatch->desId = hnd->hndOrg->sd;
  }

  retracted = retractEndOfConstraints(dispatch->treeNode);

  /* RTG Oct-7-92 : do all the qlattice assertions inside the same 
   * DELAY_RULES call to avoid inconsistencies in the qlattice.
   */
  DELAY_RULES({ 
    QRetract(Now, ">", dispatch->treeNode->handlingInterval->end,
	     "Completed Message");
    QRetract(Now, ">=", dispatch->treeNode->handlingInterval->start, 
	     "Scheduler");
    (void)listIterate(removeParentChildRelations, (char *)dispatch->treeNode, 
		      dispatch->treeNode->children);

    AssertHoldingConstraint(dispatch->treeNode);});

  if (retracted) 
    restoreEndOfRelations(dispatch->treeNode);
  
  killSubTaskTree(dispatch->treeNode);

  DISPATCH_SET_STATUS(AllocatedDispatch, dispatch);

  dispatch->class = dispatch->msg->msgData->class;

  tapWhenSent(dispatch);
  setUpBeforeTaps(dispatch);

  /* 22-May-91: fedor: should note that the class information gets
     munged to the failure class and failure class data on multiple times
     through */ 

  processInactiveTests(dispatch);

  setUpDuringAndAfterTaps(dispatch);
}


/******************************************************************************
*
* FUNCTION: void retryExcepHnd(dispatch, ref)
*
* DESCRIPTION: 
* Retry the specified ref. 
*
* If the node to retry is not the exception node (most cases) then mark
* the exception as handled so that killing the sub tree at a parent node
* will complete.
*
* If the retry node is the exception node then do not mark as handled
* in case there are temporal constraints that will get activated
* inadvertently (unlikely). This also avoid extra resource work (minor).
*
* The passed dispatch will be cleaned up as a constraint dispatch.
*
* INPUTS: 
* DISPATCH_PTR dispatch;
* TCA_REF_PTR *ref;
*
* OUTPUTS: void. calls reExecMsg.
*
******************************************************************************/

void retryExcepHnd(dispatch, ref)
DISPATCH_PTR dispatch;
TCA_REF_PTR *ref;
{
  DISPATCH_PTR retry, exception;

  retry = DISPATCH_FROM_ID((*ref)->refId);

  exception = DISPATCH_FROM_ID(dispatch->pRef);

  retry->treeNode->retries++;

  reExecMsg(retry, (HND_PTR)NULL);
  
  Log_Message1("  Retry  %15s:\n", exception->msg->msgData->name);
  
  if (retry != exception)
    processHandledMessage(exception);

  tcaRefFree(*ref); /* Assumes the data is a simple TCA_REF_PTR */
}


/******************************************************************************
*
* FUNCTION: void byPassExcepHnd(dispatch, ref)
*
* DESCRIPTION: 
* By pass the current exception by selecting a different exception handler.
* byPassHnd marks where in the list of handlers the previous search stopped.
*
* INPUTS: 
* DISPATCH_PTR dispatch;
* TCA_REF_PTR *ref;
*
* OUTPUTS: void. calls reExecMsg to continue.
*
******************************************************************************/

void byPassExcepHnd(dispatch, ref)
DISPATCH_PTR dispatch;
TCA_REF_PTR *ref;
{
  HND_PTR hnd;
  DISPATCH_PTR exception;

  exception = DISPATCH_FROM_ID((*ref)->refId);

  byPassHnd = exception->hnd;
  hnd = chooseExceptionHandler(exception);

  Log_Message1("  ByPass %15s:\n", exception->msg->msgData->name);

  if (hnd) {
    reExecMsg(exception, hnd);
  }
  else {
    processHandledMessage(exception);

    Log_Message("WARNING: Unable to find a handler for the exception.\n");
    Log_Message("WARNING: Killing task tree under root node.\n");
    killSubTaskTree(tcaTaskTreeRootGlobal);
  }

  /* the exception for byPassExcepHnd will be freed because its a constraint */

  tcaRefFree(*ref); /* Assumes the data is a simple TCA_REF_PTR */
}


/******************************************************************************
*
* FUNCTION: void getInfoExcepHnd(dispatch, ref)
*
* DESCRIPTION:
* Return information about an exception.
*
* INPUTS: 
* DISPATCH_PTR dispatch;
* TCA_REF_PTR *ref;
*
* OUTPUTS: void. calls centralReply.
*
* NOTES:
* 29-May-91: fedor:
* Another place where a consistent null TCA_REF_PTR would be useful.
*
******************************************************************************/

void getInfoExcepHnd(dispatch, ref)
DISPATCH_PTR dispatch;
TCA_REF_PTR *ref;
{
  ECNODE_PTR ECNode;
  TCA_REF_PTR nullRef;
  EXC_INFO_TYPE excInfo;
  DISPATCH_PTR exception;

  exception = DISPATCH_FROM_ID((*ref)->refId);

  ECNode = exception->treeNode->ECNode;

  if (ECNode) {
    excInfo.numOfRetries = ECNode->CNode->retries;

    excInfo.ERef = tcaRefCreate(ECNode->ENode->dispatch->msg,
				ECNode->ENode->dispatch->msg->msgData->name,
				ECNode->ENode->dispatch->locId);

    if (IS_ROOT_NODE(ECNode->CNode)) {
      excInfo.CRef = CREATE_NULL_REF();
    } else {
      excInfo.CRef = tcaRefCreate(ECNode->CNode->dispatch->msg,
				  ECNode->CNode->dispatch->msg->msgData->name,
				  ECNode->CNode->dispatch->locId);
    }

    centralReply(dispatch, (char *)&excInfo);

    tcaRefFree(excInfo.ERef);
    tcaRefFree(excInfo.CRef);
  } else {
    excInfo.numOfRetries = 0;

    /* 17-May-91: fedor: should implement null tca_ref_ptr transfer */
    nullRef = CREATE_NULL_REF();
    excInfo.ERef = nullRef;
    excInfo.CRef = nullRef;

    centralReply(dispatch, (char *)&excInfo);

    tcaRefFree(nullRef);
  }

  tcaRefFree(*ref); /* Assumes the data is a simple TCA_REF_PTR */
}


/******************************************************************************
*
* FUNCTION: void exceptionIntialize()
*
* DESCRIPTION: 
* Initialize exception handling. Remeber to init byPassHnd.
*
* INPUTS: none.
*
* OUTPUTS: void.
*
******************************************************************************/

void exceptionIntialize()
{
  byPassHnd = NULL;

  centralRegisterMessage("addExcepMsg", InformClass,
			 "{TCA_REF_PTR, string}", (char *)NULL);
  centralRegisterHandler("addExcepMsg", "addExceptionHnd", addExceptionHnd);

  centralRegisterMessage("addExceptions", InformClass,
			 "{string, string}", (char *)NULL);
  centralRegisterHandler("addExceptions", "addExceptionToMsgHnd", 
			 addExceptionToMsgHnd);

  centralRegisterMessage("retryExcepMsg", InformClass, "TCA_REF_PTR", (char *)NULL);
  centralRegisterHandler("retryExcepMsg", "retryExcepHnd", retryExcepHnd);

  centralRegisterMessage("byPassExcepMsg", InformClass, 
			 "TCA_REF_PTR", (char *)NULL);
  centralRegisterHandler("byPassExcepMsg", "byPassExcepMsg", byPassExcepHnd);
 
  centralRegisterMessage("getInfoExcepMsg", QueryClass, "TCA_REF_PTR",
			 "{TCA_REF_PTR, TCA_REF_PTR, int}");
  centralRegisterHandler("getInfoExcepMsg", "getInfoExcepHnd", 
			 getInfoExcepHnd);

  /* Added by reids, 17-Feb-92: Handles all exceptions that fall through 
     to top of task tree */
  centralRegisterMessage("killTaskTreeException", ExceptionClass,
			 (char *)NULL, (char *)NULL);
  centralRegisterHandler("killTaskTreeException", "killTopLevelTaskTreeHandler",
			 killTopLevelTaskTreeHandler);

  Add_Message_To_Ignore("byPassExcepMsg");
  Add_Message_To_Ignore("retryExcepMsg");
}
