/******************************************************************************
*
* PROJECT: Carnegie Mellon Planetary Rover Project
*          Task Control Architecture 
* 
* (c) Copyright 1991 Christopher Fedor and Reid Simmons.  All rights reserved.
* 
* MODULE: key
*
* FILE: key.c
*
* ABSTRACT:
* 
* Collection of hash and key functions for table lookups.
*
* REVISION HISTORY
*
* 16-Aug-90 Christopher Fedor, School of Computer Science, CMU
* created.
*
******************************************************************************/

#ifdef VXWORKS
#include "stdioLib.h"
#else
#include "stdio.h"
#endif
#include "ctype.h"

#include "tcaInternal.h"
#include "key.h"


/******************************************************************************
*
* FUNCTION: int intHashFunc(i)
*
* DESCRIPTION: Return int value as hash value.
*
* INPUTS: int *i;
*
* OUTPUTS: int
*
******************************************************************************/

int intHashFunc(i)
int *i;
{
  return(*i);
}


/******************************************************************************
*
* FUNCTION: int intKeyEqFunc(a, b)
*
* DESCRIPTION: Return comparison of the value of int pointers a and b
*
* INPUTS: 
* int *a, *b;
*
* OUTPUTS: int
*
******************************************************************************/

int intKeyEqFunc(a, b)
int *a, *b;
{
  return(*a == *b);
}


/******************************************************************************
*
* FUNCTION: int strHashFunc(s)
*
* DESCRIPTION: Calculate a simple hash of a NULL terminated string.
*
* INPUTS: char *s
*
* OUTPUTS: int
*
******************************************************************************/

int strHashFunc(s)
char *s;
{
  int sum, i;

  sum = 0;
  for (i=0;s[i] != '\0';i++) 
    if (isupper(s[i]))
      sum += tolower(s[i]);
    else
      sum += s[i];

  return sum;
}


/******************************************************************************
*
* FUNCTION: int strKeyEqFunc(a, b)
*
* DESCRIPTION: 
* Return 1 (TRUE) if strings a and b are equal. Returns 0 (FASLE) otherwise.
* Case insensitive compare.
*
* INPUTS: char *a, *b
*
* OUTPUTS: int
*
******************************************************************************/

int strKeyEqFunc(a, b)
char *a, *b;
{
  int i;
  char a1, b1;

  i = 0;
  while (a[i] != '\0' && b[i] != '\0') {
    if (isupper(a[i]))
      a1 = tolower(a[i]);
    else 
      a1 = a[i];
    if (isupper(b[i]))
      b1 = tolower(b[i]);
    else 
      b1 = b[i];
    
    if (a1 != b1)
      return FALSE;

    i++;
  }

  if (a[i] == b[i])
    return TRUE;
  else
    return FALSE;
}


/******************************************************************************
*
* FUNCTION: int hndHashFunc(hndKey)
*
* DESCRIPTION: Calculate a simple hash of a NULL terminated string and sd.
*
* INPUTS: HND_KEY_PTR hndKey;
*
* OUTPUTS: int
*
******************************************************************************/

int hndHashFunc(hndKey)
HND_KEY_PTR hndKey;
{
  char *s;
  int sum, i;

  sum = hndKey->sd;
  s = hndKey->name;

  for (i=0;s[i] != '\0';i++) 
    if (isupper(s[i]))
      sum += tolower(s[i]);
    else
      sum += s[i];

  return sum;
}


/******************************************************************************
*
* FUNCTION: int hndKeyEqFunc(hndKeyA, hndKeyB)
*
* DESCRIPTION: 
* Return 1 (TRUE) if hndKeyA and hndKeyB are equal. 
* Returns 0 (FASLE) otherwise. Case insensitive compare.
*
* INPUTS: HND_KEY_PTR hndKeyA, hndKeyB;
*
* OUTPUTS: int
*
******************************************************************************/

int hndKeyEqFunc(hndKeyA, hndKeyB)
HND_KEY_PTR hndKeyA, hndKeyB;
{
  int i;
  char a1, b1, *a, *b;

  if (hndKeyA->sd != hndKeyB->sd)
    return FALSE;

  a = hndKeyA->name;
  b = hndKeyB->name;

  i = 0;
  while (a[i] != '\0' && b[i] != '\0') {
    if (isupper(a[i]))
      a1 = tolower(a[i]);
    else 
      a1 = a[i];
    if (isupper(b[i]))
      b1 = tolower(b[i]);
    else 
      b1 = b[i];
    
    if (a1 != b1)
      return FALSE;

    i++;
  }

  if (a[i] == b[i])
    return TRUE;
  else
    return FALSE;
}


/******************************************************************************
*
* FUNCTION: int classHashFunc(className)
*
* DESCRIPTION: Class hash function.
*
* INPUTS: TCA_MSG_CLASS_TYPE *className;
*
* OUTPUTS: int
*
******************************************************************************/

int classHashFunc(className)
TCA_MSG_CLASS_TYPE *className;
{
  return ((int)*className);
}


/******************************************************************************
*
* FUNCTION: int classEqFunc(classA, classB)
*
* DESCRIPTION: Class key equal function.
*
* INPUTS: TCA_MSG_CLASS_TYPE *classA, *classB;
*
* OUTPUTS: int
*
******************************************************************************/

int classEqFunc(classA, classB)
TCA_MSG_CLASS_TYPE *classA, *classB;
{
  return (*classA == *classB);
}
