/*****************************************************************************
* 
* PROJECT: Carnegie Mellon Planetary Rover Project
*
* (c) Copyright 1991 Christopher Fedor and Reid Simmons.  All rights reserved.
* 
* MODULE: logging
*
* FILE: logging.c
*
* ABSTRACT:
* Facilities for logging TCA data on the terminal or in a log file.
*
* REVISION HISTORY
*
*  3-Aug-89 Reid Simmons, School of Computer Science, CMU
* Added code for ignoring registration msgs.
*
*  2-Aug-89 Reid Simmons, School of Computer Science, CMU
* Created.
*
*
* Functions:
* In the following, The <n> stands for the number of arguments
* the function takes (unlike printf, the do not take arbitrary
* numbers of arguments, perhaps someday that will be rectified).
*
*   Log<n> (format_string, [args]) :
*     Logs unconditionally onto the terminal and into the log file
*     file (if it is open).  Flushes log file.
*
*   Log_Message<n> (format_string, [args]) :
*     Logs onto the terminal if Terminal_Log_Message is not 0;
*     Logs into the Log_File if File_Log_Message is not 0.
*
*   Log_Status<n> (format_string, [args]) :
*     Logs onto the terminal if Terminal_Log_Status is not 0;
*     Logs into the Log_File if File_Log_Status is not 0.
*     Flushes log file.
*
*   Log_Time (indent) :
*     Logs the time (hh:mm:sec.milli), indenting "indent" spaces.
*     Logs onto the terminal if Terminal_Log_Time is not 0;
*     Logs into the Log_File if File_Log_Time is not 0.
*
*   Log_Data (format, data, indent) :
*     Logs the "data" using the given "format", indenting each line
*     by "indent".  Assumes a line length of 80 and truncates after
*     5 lines of output (these can be changed by setting the
*     variables Line_Length and Print_Length, respectively).
*     Logs onto the terminal if Terminal_Log_Data is not 0;
*     Logs into the Log_File if File_Log_Data is not 0
*
*   Start_Ignore_Logging () :
*     Starting now, don't log messages or status reports
*     on the terminal if Terminal_Log_Ignore is TRUE, and
*     don't log in file if File_Log_Ignore is TRUE.
*
*   End_Ignore_Logging () :
*     Reinstate logging messages and status reports
*
*   Start_Terminal_Logging () :
*     Prints out the TCA header and version number.
*
*   Start_File_Logging () :
*     If either File_Log_Messages or File_Log_Status are non-zero,
*     opens the file given by Log_File_Name and prints out the TCA
*     header and version number.
*     Sets signal functions to trap errors and close log file
*     Gives user option to enter initial comments to file.
*
*   End_File_Logging () :
*     Gives user option to enter final comments to file.
*     Close the Log_File (if open).
*
*
*****************************************************************************/


#include "stdio.h"
/* #include "signal.h"*/
#include "sys/types.h"
#ifdef SGI
#include "timeb.h"
#else
#include "sys/timeb.h"
#endif
#include "sys/time.h"
extern time_t time(); /* Is not in sys/time.h!! */

#include "tcaInternal.h"
#include "formatters.h"

#include "logging.h"
#include "printData.h"

int Terminal_Log_Messages = TRUE;

int Terminal_Log_Status = TRUE;

int Terminal_Log_Time = FALSE;

int Terminal_Log_Data = FALSE;

int Terminal_Log_Ignore = TRUE;

int Terminal_Log_HandleSummary = FALSE;

int File_Log_Messages = FALSE;

int File_Log_Status = FALSE;

int File_Log_Time = FALSE;

int File_Log_Data = FALSE;

int File_Log_Ignore = FALSE;

int File_Log_HandleSummary = FALSE;

char Log_File_Name[MAX_LOG_FILE_NAME_LENGTH];

FILE *Log_File = NULL;



void Log (format_string)
     char *format_string;
{
  printf(format_string);

  if (File_Log_Messages || File_Log_Status) {
    fprintf(Log_File, format_string);
    fflush(Log_File);
  }
}

void Log1 (format_string, arg1)
     char *format_string, *arg1;
{
  printf(format_string, arg1);

  if (File_Log_Messages || File_Log_Status) {
    fprintf(Log_File, format_string, arg1);
    fflush(Log_File);
  }
}

void Log2 (format_string, arg1, arg2)
     char *format_string, *arg1, *arg2;
{
  printf(format_string, arg1, arg2);

  if (File_Log_Messages || File_Log_Status) {
    fprintf(Log_File, format_string, arg1, arg2);
    fflush(Log_File);
  }
}

void Log3 (format_string, arg1, arg2, arg3)
     char *format_string, *arg1, *arg2, *arg3;
{
  printf(format_string, arg1, arg2, arg3);

  if (File_Log_Messages || File_Log_Status) {
    fprintf(Log_File, format_string, arg1, arg2, arg3);
    fflush(Log_File);
  }
}

void LogHandleSummary(arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8)
long arg1, arg5, arg6, arg7, arg8;
float arg2, arg3, arg4;
{
  if (Terminal_Log_HandleSummary)
    printf("%5d: Msg: %2.2f%%, Mon: %2.2f%%, Wait %2.2f%%, Msg: (%d, %d) Ave: (%d, %d)\n", 
	    arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8);

  if (File_Log_HandleSummary) {
    fprintf(Log_File, "%5d: Msg: %2.2f%%, Mon: %2.2f%%, Wait %2.2f%%, Msg: (%d, %d) Ave: (%d, %d)\n", 
	    arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8);
    fflush(Log_File);
  }
}


void Log_Message (format_string)
     char *format_string;
{
  if (Terminal_Log_Messages)
    printf(format_string);

  if (File_Log_Messages) {
    fprintf(Log_File, format_string);
    fflush(Log_File);
  }
}

void Log_Message1 (format_string, arg1)
     char *format_string, *arg1;
{
  if (Terminal_Log_Messages)
    printf(format_string, arg1);

  if (File_Log_Messages) {
    fprintf(Log_File, format_string, arg1);
    fflush(Log_File);
  }
}

void Log_Message2 (format_string, arg1, arg2)
     char *format_string, *arg1, *arg2;
{
  if (Terminal_Log_Messages)
    printf(format_string, arg1, arg2);

  if (File_Log_Messages) {
    fprintf(Log_File, format_string, arg1, arg2);
    fflush(Log_File);
  }
}

void Log_Message3 (format_string, arg1, arg2, arg3)
     char *format_string, *arg1, *arg2, *arg3;
{
  if (Terminal_Log_Messages)
    printf(format_string, arg1, arg2, arg3);

  if (File_Log_Messages) {
    fprintf(Log_File, format_string, arg1, arg2, arg3);
    fflush(Log_File);
  }
}


void Log_Status (format_string)
     char *format_string;
{
  if (Terminal_Log_Status)
    printf(format_string);

  if (File_Log_Status) {
    fprintf(Log_File, format_string);
    fflush(Log_File);
  }
}

void Log_Status1 (format_string, arg1)
     char *format_string, *arg1;
{
  if (Terminal_Log_Status)
    printf(format_string, arg1);

  if (File_Log_Status) {
    fprintf(Log_File, format_string, arg1);
    fflush(Log_File);
  }
}

void Log_Status2 (format_string, arg1, arg2)
     char *format_string, *arg1, *arg2;
{
  if (Terminal_Log_Status)
    printf(format_string, arg1, arg2);

  if (File_Log_Status) {
    fprintf(Log_File, format_string, arg1, arg2);
    fflush(Log_File);
  }
}

void Log_Status3 (format_string, arg1, arg2, arg3)
     char *format_string, *arg1, *arg2, *arg3;
{
  if (Terminal_Log_Status)
    printf(format_string, arg1, arg2, arg3);

  if (File_Log_Status) {
    fprintf(Log_File, format_string, arg1, arg2, arg3);
    fflush(Log_File);
  }
}


void Log_Time (indent)
     int indent;
{ struct timeb time_block;
  struct tm *local_time;
  int hundredths;

  if (Terminal_Log_Time || File_Log_Time) {
    ftime(&time_block);
    local_time = localtime(&(time_block.time));
    hundredths = time_block.millitm/10;

    if (Terminal_Log_Time)
      printf("%*d:%02d:%02d.%02d", indent+2, local_time->tm_hour, 
	     local_time->tm_min, local_time->tm_sec, hundredths);

    if (File_Log_Time) {
      fprintf(Log_File, "%*d:%02d:%02d.%02d", indent+2, local_time->tm_hour,
	      local_time->tm_min, local_time->tm_sec, hundredths);
      fflush(Log_File);
    }
  }
}

void Log_Data (Format, Data, indent)
     FORMAT_PTR Format;
     char *Data;
     int indent;
{ int tmp;

  if (Data) {
    tmp = indentGlobal;
    indentGlobal = indent;
    if (Terminal_Log_Data)
      Print_Formatted_Data(stdout, Format, Data);
    if (File_Log_Data)
      Print_Formatted_Data(Log_File, Format, Data);
    indentGlobal = tmp;
  }
  else {
    if (Terminal_Log_Data)
      printf("%*s\n", indent+7, "NO DATA");
    if (File_Log_Data) {
      fprintf(Log_File, "%*s\n", indent+7, "NO DATA");
      fflush(Log_File);
    }
  }
}


int tmp_term_msgs, tmp_term_status, tmp_term_time, tmp_term_data;
int tmp_file_msgs, tmp_file_status, tmp_file_time, tmp_file_data;


void Start_Ignore_Logging ()
{
  if (Terminal_Log_Ignore) {
    tmp_term_msgs = Terminal_Log_Messages;
    Terminal_Log_Messages = FALSE;
    tmp_term_status = Terminal_Log_Status;
    Terminal_Log_Status = FALSE;
    tmp_term_time = Terminal_Log_Time;
    Terminal_Log_Time = FALSE;
    tmp_term_data = Terminal_Log_Data;
    Terminal_Log_Data = FALSE;
  }
  if (File_Log_Ignore) {
    tmp_file_msgs = File_Log_Messages;
    File_Log_Messages = FALSE;
    tmp_file_status = File_Log_Status;
    File_Log_Status = FALSE;
    tmp_file_time = File_Log_Time;
    File_Log_Time = FALSE;
    tmp_file_data = File_Log_Data;
    File_Log_Data = FALSE;
  }
}

void End_Ignore_Logging ()
{
  if (Terminal_Log_Ignore) {
    Terminal_Log_Messages = tmp_term_msgs;
    Terminal_Log_Status = tmp_term_status;
    Terminal_Log_Time = tmp_term_time;
    Terminal_Log_Data = tmp_term_data;
  }
  if (File_Log_Ignore) {
    File_Log_Messages = tmp_file_msgs;
    File_Log_Status = tmp_file_status;
    File_Log_Time = tmp_file_time;
    File_Log_Data = tmp_file_data;
  }
}

void Start_Terminal_Logging ()
{
  printf("Task Control Server %d.%d (%s)\n",
	  TCA_VERSION_MAJOR, TCA_VERSION_MINOR, VERSION_DATE);
}

/*
int LogFileHandler (sig, code, scp)
     int sig, code;
     struct sigcontext *scp;
{
  if (sig == SIGTSTP)
    Log("Central stopped from terminal -- Have a nice day\n");
  else
    Log("Central died a horrible death -- Be happy, don't worry\n");

  End_File_Logging();

  tcaError((char *)NULL);
}
*/
#define LINE_LENGTH 80
#define COMMENT_LENGTH LINE_LENGTH-3

void add_comments_to_log_file ()
{ 
  char comment_string[COMMENT_LENGTH+1];
  static char astericks[] = "********************************************************************************";
  static char comments[] =  "*                                   Comments                                   *";
  static char spacer[] =    "*                                                                              *";
  int last_line_p, has_comments_p = FALSE;

  printf("Enter any comments in the log file (end with blank line)\n");
    
  do {
    (void)fgets(comment_string, COMMENT_LENGTH, stdin);
    last_line_p = (comment_string[0] == '\n'); /* the "newline" character */
    if (last_line_p) {
      if (has_comments_p)
	fprintf(Log_File, "%s\n%s\n\n", spacer, astericks);
    }
    else {
      if (!has_comments_p) {
	fprintf(Log_File, "\n%s\n%s\n%s\n", astericks, comments, spacer);
	has_comments_p = TRUE;
      }
      /* Replace "newline" character */
      comment_string[strlen(comment_string)-1] = '\0';
      fprintf(Log_File, "* %-*s*\n", LINE_LENGTH-3, comment_string);
    }
  }
  while (!last_line_p);
}

void Start_File_Logging ()
{ 
  time_t current_time;

  if (File_Log_Messages || File_Log_Status || File_Log_Data || 
      File_Log_Time || File_Log_HandleSummary) {
    Log_File = fopen(Log_File_Name, "w");
    if (!Log_File) {
      TCA_ERROR1("ERROR: Can not open file %s.\n", Log_File_Name);
    }

   printf("Logging to %s;\nTo add comments at end of session, exit using ^C\n",
	   Log_File_Name);
    
    current_time = time((time_t *)NULL);
    fprintf(Log_File, "Logging Task Control Server %d.%d (%s) on %s",
	    TCA_VERSION_MAJOR, TCA_VERSION_MINOR, VERSION_DATE, 
	    ctime(&current_time));

    add_comments_to_log_file();

    /****
     * Catch errors so the log file can be closed in a reasonable manner
     ****/
/*
9-Aug-89: fedor: comment out and push a minimal set to central

    (void)signal(SIGHUP, LogFileHandler);
    (void)signal(SIGQUIT, LogFileHandler);
    (void)signal(SIGILL, LogFileHandler);
    (void)signal(SIGHUP, LogFileHandler);
    (void)signal(SIGTRAP, LogFileHandler);
    (void)signal(SIGIOT, LogFileHandler);
    (void)signal(SIGEMT, LogFileHandler);
    (void)signal(SIGFPE, LogFileHandler);
    * (void)signal(SIGKILL, LogFileHandler); Can't be trapped *
    (void)signal(SIGBUS, LogFileHandler);
    (void)signal(SIGSEGV, LogFileHandler);
    (void)signal(SIGSYS, LogFileHandler);
    (void)signal(SIGPIPE, LogFileHandler);
    (void)signal(SIGALRM, LogFileHandler);
    (void)signal(SIGTERM, LogFileHandler);
    (void)signal(SIGURG, LogFileHandler);
    * signal(SIGSTOP, LogFileHandler); Can't be trapped *
    (void)signal(SIGTSTP, LogFileHandler);
    (void)signal(SIGCONT, LogFileHandler);
    (void)signal(SIGCHLD, LogFileHandler);
    (void)signal(SIGCLD, LogFileHandler);
    (void)signal(SIGTTIN, LogFileHandler);
    (void)signal(SIGTTOU, LogFileHandler);
    (void)signal(SIGIO, LogFileHandler);
    (void)signal(SIGPOLL, LogFileHandler);
    (void)signal(SIGXCPU, LogFileHandler);
    (void)signal(SIGXFSZ, LogFileHandler);
    (void)signal(SIGVTALRM, LogFileHandler);
    (void)signal(SIGPROF, LogFileHandler);
    (void)signal(SIGWINCH, LogFileHandler);
    (void)signal(SIGLOST, LogFileHandler);
    (void)signal(SIGUSR1, LogFileHandler);
    (void)signal(SIGUSR2, LogFileHandler);
*/
  }
}

void End_File_Logging ()
{
  if (Log_File != NULL) {
    add_comments_to_log_file();
    (void)fclose(Log_File);
    File_Log_Messages = File_Log_Status = File_Log_Data = File_Log_Time = FALSE;
  }
}
