/*****************************************************************************
* 
* PROJECT: Carnegie Mellon Planetary Rover Project
*
* (c) Copyright 1991 Christopher Fedor and Reid Simmons.  All rights reserved.
* 
* MODULE: logging
*
* FILE: logging.h
*
* ABSTRACT:
* External declarations for message and status logging facilities.
*
* REVISION HISTORY
*
*  6-Oct-89 Christopher Fedor, School of Computer Science, CMU
* Added message handle time summary.
*
*  3-Aug-89 Reid Simmons, School of Computer Science, CMU
* Added code for ignoring registration msgs.
*
*  2-Aug-89 Reid Simmons, School of Computer Science, CMU
* Created.
*
*****************************************************************************/

/*********************************************************************
 *                                                                   *
 * logging.h                                                         *
 *                                                                   *
 *                                                                   *
 * Functions:                                                        *
 *   In the following, The <n> stands for the number of arguments    *
 *   the function takes (unlike printf, the do not take arbitrary    *
 *   numbers of arguments, perhaps someday that will be rectified).  *
 *                                                                   *
 *   Log<n> (format_string, [args]) :                                *
 *     logs unconditionally onto the terminal and into the log file  *
 *     file (if it is open).                                         *
 *                                                                   *
 *   Log_Message<n> (format_string, [args]) :                        *
 *     logs onto the terminal if Terminal_Log_Message is not 0;      *
 *     logs into the Log_File if File_Log_Message is not 0;          *
 *                                                                   *
 *   Log_Status<n> (format_string, [args]) :                         *
 *     logs onto the terminal if Terminal_Log_Status is not 0;       *
 *     logs into the Log_File if File_Log_Status is not 0;           *
 *                                                                   *
 *   Log_Time (indent) :                                             *
 *     logs the time (hh:mm:sec.milli), indenting "indent" spaces.   *
 *     Logs onto the terminal if Terminal_Log_Time is not 0;         *
 *     Logs into the Log_File if File_Log_Time is not 0.             *
 *                                                                   *
 *   Log_Data (format, data, indent) :                               *
 *     logs the "data" using the given "format", indenting each line *
 *     by "indent".  Assumes a line length of 80 and truncates after *
 *     5 lines of output (these can be changed by setting the        *
 *     variables Line_Length and Print_Length, respectively).        *
 *     Logs onto the terminal if Terminal_Log_Data is not 0;         *
 *     Logs into the Log_File if File_Log_Data is not 0;             *
 *                                                                   *
 *   Start_Ignore_Logging () :                                       *
 *     starting now, don't log messages or status reports            *
 *     on the terminal if Terminal_Log_Ignore is TRUE, and           *
 *     don't log in file if File_Log_Ignore is TRUE.                 *
 *                                                                   *
 *   End_Ignore_Logging () :                                         *
 *     reinstate logging messages and status reports                 *
 *                                                                   *
 *   Start_Terminal_Logging () :                                     *
 *     Prints out the TCA header and version number.                 *
 *                                                                   *
 *   Start_File_Logging () :                                         *
 *     If either File_Log_Messages or File_Log_Status are non-zero,  *
 *     opens the file given by Log_File_Name and prints out the TCA  *
 *     header and version number.                                    *
 *     Sets signal functions to trap errors and close log file       *
 *     Gives user option to enter initial comments to file.          *
 *                                                                   *
 *   End_File_Logging () :                                           *
 *     Gives user option to enter final comments to file.            *
 *     Close the Log_File (if open).                                 *
 *                                                                   *
 *                                                                   *
 *********************************************************************/

extern int Terminal_Log_Messages; /* initially TRUE (0) */

extern int Terminal_Log_Status; /* initially TRUE (0) */

extern int Terminal_Log_Time; /* initially FALSE (0) */

extern int Terminal_Log_Data; /* initially FALSE (0) */

extern int Terminal_Log_Ignore; /* initially TRUE (0) */

extern int Terminal_Log_HandleSummary; /* initially FALSE */


extern int File_Log_Messages; /* initially FALSE (0) */

extern int File_Log_Status; /* initially FALSE (0) */

extern int File_Log_Time; /* initially FALSE (0) */

extern int File_Log_Data; /* initially FALSE (0) */

extern int File_Log_Ignore; /* initially TRUE (0) */

extern int File_Log_HandleSummary;  /* initially FALSE */

#define MAX_LOG_FILE_NAME_LENGTH 60

extern char Log_File_Name[MAX_LOG_FILE_NAME_LENGTH];

extern FILE *Log_File; /* initially NULL */


extern void Log (); /* format_string */

extern void Log1 (); /* format_string, arg1 */

extern void Log2 (); /* format_string, arg1, arg2 */

extern void Log3 (); /* format_string, arg1, arg2, arg3 */

extern void LogHandleSummary();

extern void Log_Message (); /* format_string */

extern void Log_Message1 (); /* format_string, arg1 */

extern void Log_Message2 (); /* format_string, arg1, arg2 */

extern void Log_Message3 (); /* format_string, arg1, arg2, arg3 */

extern void Log_Status (); /* format_string */

extern void Log_Status1 (); /* format_string, arg1 */

extern void Log_Status2 (); /* format_string, arg1, arg2 */

extern void Log_Status3 (); /* format_string, arg1, arg2, arg3 */

extern void Log_Time ();

extern void Log_Data (); /* data, indent */

extern void Start_Ignore_Logging ();

extern void End_Ignore_Logging ();

extern void Start_Terminal_Logging ();

extern void Start_File_Logging ();

extern void End_File_Logging ();

