/*****************************************************************************
* PROJECT: Carnegie Mellon Planetary Rover Project
*          Task Control Architecture
*
* (c) Copyright 1991 Christopher Fedor and Reid Simmons.  All rights reserved.
* 
* MODULE: monitors
*
* FILE: mon.c
*
* ABSTRACT:
*
* Monitors
*
* REVISION HISTORY:
*
* 17-Aug-91 Christopher Fedor, School of Computer Science, CMU
* Updated actMsg for command class actions to issue the correct class data.
* Changed global tplConstr value to simply malloc storage for class data.
*
* 15-Aug-91 Christopher Fedor, School of Computer Science, CMU
* Moved tplConstr from in reply monitor procedures to be a global
* so when it is used as class data it will not vanish with the stack.
*
*  9-Jul-91 Christopher Fedor, School of Computer Science, CMU
* Changed format tests to test with format string because the message
* may not have been parsed yet when the test is done.
*
* 19-Mar-91 Christopher Fedor, School of Computer Science, CMU
* Goofed on implementation of fireEvery. Reimplemented.
* fireEvery: fire action message every n times condition is satisfied.
*
* 16-Nov-90 Christopher Fedor, School of Computer Science, CMU
* Added cancelIfAttendingIntervalMonitor call for taskTree.c
* WARNING: Need to check interaction between stopping an interval
* monitor that is also a durable item.
*
* 24-Oct-90 Christopher Fedor, School of Computer Science, CMU
* Removed unnecessary queue include files.
*
* 13-Oct-90 Christopher Fedor, School of Computer Science, CMU
* Reimplemented fireEvery - to fire the action message multiple times.
*
* 12-Oct-90 Christopher Fedor, School of Computer Science, CMU
* Reimplemented polling monitors.
*
*  9-Oct-90 Christopher Fedor, School of Computer Science, CMU
* Reimplemented durable items.
*
*  8-Oct-90 Christopher Fedor, School of Computer Science, CMU
* Reimplemented demon monitor.
*
*  6-Aug-90 Christopher Fedor, School of Computer Science, CMU
* Replaced module monitor code with monitor behavior routines now
* in behaviors.c Reunited this split module to a single central
* monitor module.
*
* 12-Apr-90 Long-Ji Lin at School of Computer Science, CMU
* Added code to check for potential deadlocks caused by inappropriate 
* expiration time specified for interval monitors.
*
* 30-Mar-90 Christopher Fedor, School of Computer Science, CMU
* Revised to software standards.
*
* 26-Mar-90 Long-Ji Lin, School of Computer Science, CMU
* Modified code to allow options of interval monitors to be specified at run
* time rather than at registration time.
* Changed the way interval monitors are handled, so we can use interval
* monitors just like goals to specify temporal constraints of other
* messages.
* Reimplemented tcaIntervalMonitor() and added
* tcaIntervalMonitorWithConstraints();
*
* 12-Nov-89 Reid Simmons, School of Computer Science, CMU
* Changed way monitors are registered, so the name of the monitor is used, 
* instead of the generic name (e.g. "PointMonitor").
*
*  9-Nov-89 Reid Simmons, School of Computer Science, CMU
* Fix to kill point monitors correctly; no longer necessary to register
* messages before registering the monitor that uses them.
*
* 20-Oct-89 Reid Simmons, School of Computer Science, CMU
* Added a "duration" loop to time the extent of dispatches.  
* Used to implement a "tcaDelay" command message, and adding a "duration"
* option to interval monitors.
*
* 26-Sep-89 Reid Simmons, School of Computer Science, CMU
* Streamlined calling of monitors.
* Added options to interval monitors.
*
*  2-Aug-89 Reid Simmons, School of Computer Science, CMU
* Changed "printf"s to use logging facility.
*
* 12-May-89 Reid Simmons, School of Computer Science, CMU
* Split off module monitor code from monitors.c
*
*  1-May-89 Reid Simmons, School of Computer Science, CMU
* created.
*
*****************************************************************************/

#include "stdio.h"
#include <sys/types.h>
#include <sys/time.h>
extern time_t time(); /* Is not in sys/time.h!! */

#include "tcaInternal.h"
#include "hash.h"
#include "idtable.h"
#include "list.h"
#include "logging.h"
#include "key.h"

#include "datamsg.h"
#include "dispatch.h"
#include "sendMsg.h"

#include "tms.h"
#include "cells.h"
#include "rules.h"
#include "qlattice.h"

#include "taskTree.h"
#include "tplConstr.h"
#include "mon.h"

extern void parseMsg();

extern QUANTITY_PTR Time_Of();
extern QUANTITY_PTR Now;

extern HASH_TABLE_PTR messageTable;

extern int tcaServerGlobal;
extern MODULE_PTR tcaServerModGlobal;

static int lastTime = 0;
static LIST_PTR  durableItemList, pollingMonitorList;
static HASH_TABLE_PTR monitorTable, demonMonitorTable;


/******************************************************************************
*
* FUNCTION: DURABLE_ITEM_PTR addDurableItem(duration, intMon, dispatch)
*
* DESCRIPTION: Creates a durableItem adds it to the list and returns it.
*
* INPUTS: 
* int duration;
* INTERVAL_MON intMon;
* DISPATCH_PTR dispatch;
*
* OUTPUTS: DURABLE_ITEM_PTR
*
******************************************************************************/

DURABLE_ITEM_PTR addDurableItem(duration, intMon, dispatch)
int duration;
INTERVAL_MONITOR_DATA_PTR intMon;
DISPATCH_PTR dispatch;
{
  DURABLE_ITEM_PTR durableItem;

  durableItem = NEW(DURABLE_ITEM_TYPE);
  
  durableItem->endTime = (int)time((time_t *)NULL)+duration;
  durableItem->dispatch = dispatch;
  durableItem->intMon = intMon;

  listInsertItem((char *)durableItem, durableItemList);

  return durableItem;
}


/******************************************************************************
*
* FUNCTION: void spliceOutDurableItem(durableItem)
*
* DESCRIPTION: Remove durableItem from durableItemList and reclaim memory.
*
* INPUTS: DURABLE_ITEM_PTR durableItem;
*
* OUTPUTS: void.
*
******************************************************************************/

void spliceOutDurableItem(durableItem)
DURABLE_ITEM_PTR durableItem;
{
  listDeleteItem((char *)durableItem, durableItemList);
  tcaFree((char *)durableItem);
}


/******************************************************************************
*
*  POLLING MONITOR
*
******************************************************************************/

/******************************************************************************
*
* FUNCTION: void spliceOutPollingMonitor(intMon)
*
* DESCRIPTION: Remove intMon from the list of polling monitors and free it.
*
* INPUTS: INTERVAL_MONITOR_DATA_PTR intMon;
*
* OUTPUTS: void.
*
******************************************************************************/

void spliceOutPollingMonitor(intMon)
INTERVAL_MONITOR_DATA_PTR intMon;
{
  listDeleteItem((char *)intMon, pollingMonitorList);
  tcaFree((char *)intMon);
}


/******************************************************************************
*
* FUNCTION: void stopPollingMonitor(intMon)
*
* DESCRIPTION: 
* Remove the instance of the polling monitor intMon and mark the monitor
* as handled.
*
* INPUTS: INTERVAL_MONITOR_DATA_PTR intMon;
*
* OUTPUTS: void.
*
******************************************************************************/

extern void terminatePollingMonitor();

void stopPollingMonitor(intMon)
INTERVAL_MONITOR_DATA_PTR intMon;
{ 
  if (intMon->expirationTime)
     removeOutRule(terminatePollingMonitor,
		   get_relation(Now, intMon->expirationTime));

  if (intMon->fireTime == ALREADY_FIRED) {
    /* keep poller around until the current condition message returns */
    intMon->monInfo = NULL;
  }
  else
    spliceOutPollingMonitor(intMon);

  dispatchUpdateAndDisplay(HandledDispatch, intMon->dispatch);
  CompletionConstraints(intMon->dispatch);
}


/******************************************************************************
*
* FUNCTION: void terminatePollingMonitor(ruleName, relNode, intMon)
*
* DESCRIPTION: Rule function to remove a polling monitor.
*
* INPUTS: 
* char *ruleName; 
* TMS_NODE_PTR *relNode; 
* INTERVAL_MONITOR_DATA_PTR intMon;
*
* OUTPUTS: void.
*
******************************************************************************/

void terminatePollingMonitor(ruleName, relNode, intMon)
char *ruleName; 
TMS_NODE_PTR relNode; 
INTERVAL_MONITOR_DATA_PTR intMon;
{ 
  if (Less_Than_Now_Relp(relNode) && Relp_Still_Out(relNode)) {
    if (intMon->intMonOps->options->duration >= 0)
      spliceOutDurableItem(intMon->durableItem);
    stopPollingMonitor(intMon);
  }
}


/******************************************************************************
*
* FUNCTION: void replyPollingMonitor(dispatch, intMon)
*
* DESCRIPTION: Reply function from a polling monitor condition message.
*
* INPUTS: 
* DISPATCH_PTR dispatch;
* INTERVAL_MONITOR_DATA_PTR intMon;
*
* OUTPUTS: void.
*
******************************************************************************/

void replyPollingMonitor(dispatch, intMon)
DISPATCH_PTR dispatch;
INTERVAL_MONITOR_DATA_PTR intMon;
{
  int *tplConstr;
  MSG_PTR actMsg;
  BLOCK_COM_PTR comData;
  DISPATCH_PTR parentDispatch;
  
  if (!intMon->monInfo) {
    /* monitor already ended */
    spliceOutPollingMonitor(intMon);
    /* 12-May-91: fedor: can not free here becuase it will interfere with
       resourceProcessPending in recvMsg.c */
    /* dispatchFree(dispatch);*/
    return;
  }
  
  if (dataMsgTestMsgData(DISPATCH_RES_DATA(dispatch))) {

    intMon->condSatisfiedCounter++;

    parentDispatch = DISPATCH_FROM_ID(DISPATCH_PARENT_REF(dispatch));

    actMsg = (MSG_PTR)hashTableFind(intMon->monInfo->actionName, messageTable);

    if (!actMsg) {
      TCA_ERROR1("ERROR: replyPointMonitor: missing action message: %s\n",
		 intMon->monInfo->actionName);
    }

    if (intMon->condSatisfiedCounter == intMon->intMonOps->options->fireEvery){
      intMon->condSatisfiedCounter = 0; /* reset counter */
      --intMon->maxFireCounter;

      if (actMsg->msgData->class = CommandClass) {
	comData = NEW(BLOCK_COM_TYPE);
	comData->waitFlag = FALSE;
	comData->tplConstr = NO_TPLCONSTR;
	sendDataMsg(actMsg, DISPATCH_RES_DATA(dispatch), (char *)comData,
		    parentDispatch);
      } else {
	tplConstr = NEW(int);
	*tplConstr = NO_TPLCONSTR;
	sendDataMsg(actMsg, DISPATCH_RES_DATA(dispatch), (char *)tplConstr,
		    parentDispatch);
      }
    }
  }


  /* 12-May-91: fedor: can not free here becuase it will interfere with
     resourceProcessPending in recvMsg.c */
  /* dispatchFree(dispatch);*/

  if (intMon->maxFireCounter == 0) 
    stopPollingMonitor(intMon);
  else
    intMon->fireTime = (int)time((time_t *)NULL)+intMon->intMonOps->options->period;
}


/******************************************************************************
*
* FUNCTION: void execPollingMonitor(intMon)
*
* DESCRIPTION: Start the condition message for the polling monitor intMon.
*
* INPUTS: INTERVAL_MONITOR_DATA_PTR intMon;
*
* OUTPUTS: void.
*
******************************************************************************/

void execPollingMonitor(intMon)
INTERVAL_MONITOR_DATA_PTR intMon;
{
  MSG_PTR condMsg;

 condMsg = 
   (MSG_PTR)hashTableFind(intMon->monInfo->conditionName, messageTable);

  if (!condMsg)
    TCA_ERROR1("ERROR: execPollingMonitor: missing condition message: %s\n",
	       intMon->monInfo->conditionName);

  intMon->fireTime = ALREADY_FIRED;

  sendDataMsgWithReply(condMsg, DISPATCH_MSG_DATA(intMon->dispatch), 
		       intMon->dispatch, replyPollingMonitor, (char *)intMon);
}


/******************************************************************************
*
* FUNCTION: void pollingMonStart(intMon)
*
* DESCRIPTION: 
* Create the termination rule and initialize the polling monitor intMon.
*
* INPUTS: 
* INTERVAL_MONITOR_DATA_PTR intMon;
*
* OUTPUTS: void.
*
******************************************************************************/

void pollingMonStart(intMon)
INTERVAL_MONITOR_DATA_PTR intMon;
{
  if (intMon->expirationTime)
    addOutRule(ruleCreateOut("Terminate Monitor", terminatePollingMonitor, 3, 
			     (char *)intMon), 
	       relationGetOrMake(Now, intMon->expirationTime));
  
  listInsertItem((char *)intMon, pollingMonitorList);

  if (intMon->intMonOps->options->initialWait) 
    intMon->fireTime = (int)time((time_t *)NULL)+intMon->intMonOps->options->period;
  else 
    execPollingMonitor(intMon);
}      


/******************************************************************************
*
*  DEMON MONITOR
*
******************************************************************************/

/******************************************************************************
*
* FUNCTION: void stopDemonMonitor(intMon)
*
* DESCRIPTION: 
*
* INPUTS:
*
* OUTPUTS:
*
******************************************************************************/

extern void terminateDemonMonitor();

void stopDemonMonitor(intMon)
INTERVAL_MONITOR_DATA_PTR intMon;
{ 
  MSG_PTR cancelMsg;
  int cancelKey = ALREADY_CANCELLED; 

  if (intMon->expirationTime)
    removeOutRule(terminateDemonMonitor, get_relation(Now, intMon->expirationTime));

  if (*(intMon->id) == NO_KEY)
    /* have yet to receive the reply to setUpMsg */
    intMon->id = &cancelKey;
  else {
    cancelMsg = (MSG_PTR)hashTableFind(intMon->monInfo->conditionName, 
				       messageTable);
    if (!cancelMsg)
      TCA_ERROR1("ERROR: stopDemonMonitor: missing cancel message: %s\n",
		 intMon->monInfo->conditionName);

    if (cancelMsg->msgData->class != InformClass)
      TCA_ERROR1("ERROR: Demon Cancel Message: %s: Should be InformClass.\n",
		 cancelMsg->msgData->name);

    if (!strKeyEqFunc("int", cancelMsg->msgFormatStr))
      TCA_ERROR1("ERROR: Demon Cancel Message: %s: Message format is not int\n",
		 cancelMsg->msgData->name);

    if (!cancelMsg->parsedFormats) {
      parseMsg(cancelMsg);
    }

    (void)centralSendMessage((TCA_REF_PTR)NULL, cancelMsg, 
			     (char *)intMon->id, (char *)NULL);

    dispatchUpdateAndDisplay(HandledDispatch, intMon->dispatch);
    CompletionConstraints(intMon->dispatch);

    intMon = (INTERVAL_MONITOR_DATA_PTR)hashTableRemove((char *)intMon->id,
							demonMonitorTable);
    
    tcaFree((char *)intMon);
  }
}


/******************************************************************************
*
* FUNCTION: void terminateDemonMonitor(ruleName, relNode, intMon)
*
* DESCRIPTION: 
*
* INPUTS:
*
* OUTPUTS:
*
******************************************************************************/

void terminateDemonMonitor(ruleName, relNode, intMon)
char *ruleName; 
TMS_NODE_PTR relNode; 
INTERVAL_MONITOR_DATA_PTR intMon;
{ 
  if (Less_Than_Now_Relp(relNode) && Relp_Still_Out(relNode)) {
    if (intMon->intMonOps->options->duration >= 0)
      spliceOutDurableItem(intMon->durableItem);
    stopDemonMonitor(intMon);
  }
}


/******************************************************************************
*
* FUNCTION: void fireDemonHnd(dispatch, id)
*
* DESCRIPTION: 
*
* INPUTS:
*
* OUTPUTS:
*
* NOTES:
* 6-Oct-90: fedor:
* test for data for action message first?
*
* 8-Oct-90: fedor
* problems if maxFireCounter wraps - blah!
*
******************************************************************************/

void fireDemonHnd(dispatch, id)
DISPATCH_PTR dispatch;
int *id;
{ 
  int *tplConstr;
  MSG_PTR actMsg;
  BLOCK_COM_PTR comData;
  INTERVAL_MONITOR_DATA_PTR intMon;

  intMon = (INTERVAL_MONITOR_DATA_PTR)hashTableFind((char *)id, demonMonitorTable);

  if (intMon) {

    intMon->condSatisfiedCounter++;

    actMsg = (MSG_PTR)hashTableFind(intMon->monInfo->actionName, 
				    messageTable);
    if (!actMsg)
      TCA_ERROR1("ERROR: fireDemonHnd: missing action message: %s\n",
		 intMon->monInfo->actionName);
      
    if (intMon->condSatisfiedCounter == intMon->intMonOps->options->fireEvery){
      intMon->condSatisfiedCounter = 0; /* reset counter */
      --intMon->maxFireCounter;

      if (actMsg->msgData->class = CommandClass) {
	comData = NEW(BLOCK_COM_TYPE);
	comData->waitFlag = FALSE;
	comData->tplConstr = NO_TPLCONSTR;
	sendDataMsg(actMsg, DISPATCH_RES_DATA(dispatch), (char *)comData,
		    intMon->dispatch);
      } else {
	tplConstr = NEW(int);
	*tplConstr = NO_TPLCONSTR;
	sendDataMsg(actMsg, DISPATCH_RES_DATA(dispatch), (char *)tplConstr,
		    intMon->dispatch);
      }
    }
    
      if (intMon->maxFireCounter == 0) 
	stopDemonMonitor(intMon);
    }

  /* 12-May-91: fedor: not sure yet if this one interfers 
     with resourceProcessPending */
  dispatchFree(dispatch);
}


/******************************************************************************
*
* FUNCTION: void replyDemonMonitor(dispatch, data)
*
* DESCRIPTION: 
* Store intMon in the demonMonitorTable under the returned key from setUpMsg.
*
* INPUTS:
* DISPATCH_PTR dispatch; 
* INTERVAL_MONITOR_DATA_PTR intMon;
*
* OUTPUTS: void.
*
******************************************************************************/

void replyDemonMonitor(dispatch, intMon)
DISPATCH_PTR dispatch;
INTERVAL_MONITOR_DATA_PTR intMon;
{
  if (*intMon->id != ALREADY_CANCELLED) {
    intMon->id = (int *)dispatchDecodeResponse(dispatch);
    (void)hashTableInsert((char *)intMon->id, sizeof(int), 
			  (char *)intMon, demonMonitorTable);
  }
  /* 12-Oct-90: fedor: more memory problems ... 
    free dispatch here causes it to crash when replying from the setUpMsg 
    in deliverResponse - the dispatch is NULL */
}


/******************************************************************************
*
* FUNCTION: void demonMonStart(dispatch, intMon)
*
* DESCRIPTION:
* Sets up an instance of a demon monitor and call set up message.
*
* INPUTS:
* DISPATCH_PTR dispatch; 
* INTERVAL_MONITOR_DATA_PTR intMon;
*
* OUTPUTS: void.
*
******************************************************************************/

void demonMonStart(dispatch, intMon)
DISPATCH_PTR dispatch;
INTERVAL_MONITOR_DATA_PTR intMon;
{
  MSG_PTR setUpMsg;

  if (intMon->expirationTime)
    addOutRule(ruleCreateOut("Terminate Monitor", terminateDemonMonitor, 
			     3, (char *)intMon),
	       relationGetOrMake(Now, intMon->expirationTime));
  
  setUpMsg = (MSG_PTR)hashTableFind(intMon->monInfo->setUpName,
				    messageTable);
  
  if (!setUpMsg)
    TCA_ERROR1("ERROR: demonMonStart: missing set up message: %s\n",
	       intMon->monInfo->setUpName);

  sendDataMsgWithReply(setUpMsg, DISPATCH_MSG_DATA(dispatch), dispatch,
		       replyDemonMonitor, (char *)intMon);
}


/******************************************************************************
*
* FUNCTION: void intervalMonitorHnd(dispatch, intMonOps)
*
* DESCRIPTION:
* Sets up an instance of an interval monitor.
*
* INPUTS:
* DISPATCH_PTR dispatch; 
* INT_MON_CLASS_PTR intMonOps;
*
* OUTPUTS: void.
*
* NOTES:
*
*  5-Oct-90: fedor
* AttendingConstraints were set in the original after the temporally inactive
* test. AttendingConstraints already got set before this handler was called.
* This may be a problem.
*
*  9-Oct-90: fedor
* Reid does not think it will be a problem if the monitor is never set up
* because expiration already occured even though AttendingConstraints have
* already been called. Could be another potential memory leak.
*
******************************************************************************/

void intervalMonitorHnd(dispatch, intMonOps)
DISPATCH_PTR dispatch; 
INTERVAL_MON_CLASS_PTR intMonOps;
{
  int key = NO_KEY;
  TCA_TIME_POINT_PTR CEnd; 
  INTERVAL_MONITOR_DATA_PTR intMon;

  intMon = NEW(INTERVAL_MONITOR_DATA_TYPE);
  intMon->id = &key;
  intMon->fireTime = 0;
  intMon->maxFireCounter = 0;
  intMon->condSatisfiedCounter = 0;
  intMon->dispatch = dispatch;

  /* 16-Nov-90: fedor: for taskTree kills - store the intMon on the 
     monitor dispatch. This has the potential for messing up 
     future calls to sendDataMsg. */
  dispatch->respData = (char *)intMon;

  intMon->durableItem = NULL;
  intMon->monInfo = 
    (MON_REG_DATA_PTR)hashTableFind(dispatch->msg->msgData->name, 
				    monitorTable);

  if (!intMon->monInfo)
    TCA_ERROR1("ERROR: intervalMonitorHnd: missing monitor info: %s\n",
	       dispatch->msg->msgData->name);

  CEnd = &(intMonOps->expire);
  intMon->expirationTime = (VALID_TIME(*CEnd)) ? Time_Of(CEnd) : NULL;
  
  intMon->intMonOps = intMonOps;

  if (intMon->expirationTime && QIs_True(Now, ">=", intMon->expirationTime))
 Log_Status1("Monitor %s is not scheduled because it is temporally inactive\n",
	     dispatch->msg->msgData->name);
    else {
      intMon->maxFireCounter = intMonOps->options->maxFire;

      if (intMonOps->options->duration > 0)
	intMon->durableItem = addDurableItem(intMonOps->options->duration,
					     intMon, dispatch);

      switch(dispatch->msg->msgData->class) {
      case DemonMonitorClass:
	demonMonStart(dispatch, intMon);
	break;
      case PollingMonitorClass:
	pollingMonStart(intMon);
	break;
      default:
	TCA_ERROR1("ERROR: intervalMonitorHnd: Unknown Class: %d\n",
		   dispatch->msg->msgData->class);
	break;
      }
    }
}

/* 07-Aug-91: Reid: Need to set the interval monitor start time *before*
   "processInactiveTests" is called in recvMessage */

void setIntervalMonitorStartTime (dispatch, intMonOps)
DISPATCH_PTR dispatch; 
INTERVAL_MON_CLASS_PTR intMonOps;
{
  TCA_TIME_POINT_PTR CStart; 
  TASK_TREE_NODE_PTR monitorNode;
  struct _QUANTITY_TYPE *MStart;

  monitorNode = dispatch->treeNode;

  MStart = monitorNode->handlingInterval->start;
  CStart = &(intMonOps->start);

  if (VALID_TIME(*CStart))
    QAssert(MStart, "=", Time_Of(CStart), "Monitor Constraints");
  else if (monitorNode->parent->handlingInterval)
    QAssert(MStart, "=", monitorNode->parent->handlingInterval->start, 
	    "Monitor Constraints");
}


/******************************************************************************
*
*  POINT MONITOR
*
******************************************************************************/

/******************************************************************************
*
* FUNCTION: void replyPointMonitor(dispatch, monInfo)
*
* DESCRIPTION:
*
* INPUTS:
* DISPATCH_PTR dispatch;
* MON_REG_DATA_PTR monInfo;
*
* OUTPUTS: void.
*
******************************************************************************/

void replyPointMonitor(dispatch, monInfo)
DISPATCH_PTR dispatch;
MON_REG_DATA_PTR monInfo;
{
  int *tplConstr;
  MSG_PTR actMsg;
  BLOCK_COM_PTR comData;
  DISPATCH_PTR parentDispatch;

  parentDispatch = DISPATCH_FROM_ID(DISPATCH_PARENT_REF(dispatch));

  if (!monInfo)
    TCA_ERROR1("ERROR: replyPointMonitor: missing monitor info: %s\n",
	       dispatch->msg->msgData->name);

  if (dataMsgTestMsgData(DISPATCH_RES_DATA(dispatch))) {
    actMsg = (MSG_PTR)hashTableFind(monInfo->actionName, messageTable);
    if (!actMsg)
      TCA_ERROR1("ERROR: replyPointMonitor: missing action message: %s\n",
		 monInfo->actionName);

    if (actMsg->msgData->class = CommandClass) {
      comData = NEW(BLOCK_COM_TYPE);
      comData->waitFlag = FALSE;
      comData->tplConstr = NO_TPLCONSTR;
      sendDataMsg(actMsg, DISPATCH_RES_DATA(dispatch), (char *)comData,
		  parentDispatch);
    } else {
      tplConstr = NEW(int);
      *tplConstr = NO_TPLCONSTR;
      sendDataMsg(actMsg, DISPATCH_RES_DATA(dispatch), (char *)tplConstr,
		  parentDispatch);
    }
  }

  processHandledMessage(parentDispatch);
}


/******************************************************************************
*
* FUNCTION: void pointMonitorHnd(dispatch, data)
*
* DESCRIPTION:
*
* INPUTS:
* DISPATCH_PTR dispatch;
* char *data;
*
* OUTPUTS: void.
*
******************************************************************************/

void pointMonitorHnd(dispatch, data)
DISPATCH_PTR dispatch;
char *data;
{
  MSG_PTR msg, condMsg;
  MON_REG_DATA_PTR monInfo;

  msg = dispatch->msg;

  monInfo = (MON_REG_DATA_PTR)hashTableFind(msg->msgData->name, monitorTable);

  if (!monInfo)
    TCA_ERROR1("ERROR: pointMonitorHnd: missing monitor info: %s\n",
	       msg->msgData->name);

  condMsg = (MSG_PTR)hashTableFind(monInfo->conditionName, messageTable);

  if (!condMsg)
    TCA_ERROR1("ERROR: pointMonitorHnd: missing condition message: %s\n",
	       monInfo->conditionName);

  sendDataMsgWithReply(condMsg, DISPATCH_MSG_DATA(dispatch), dispatch,
	      replyPointMonitor, (char *)monInfo);
}


/******************************************************************************
*
*  DELAY COMMAND
*  Use the duration loop to implement a delay command -- 
*  waits for a given duration, before returning success.
*
*  9-Oct-90: fedor: this whole delay needs more thought.
*
******************************************************************************/

void delayCommandHnd(dispatch, duration)
DISPATCH_PTR dispatch;
int *duration;
{ 
  (void)addDurableItem(*duration, (INTERVAL_MONITOR_DATA_PTR)NULL, dispatch);
}

int sameDurableItem(dispatch, durableItem)
DISPATCH_PTR dispatch;
DURABLE_ITEM_PTR durableItem;
{
  return (dispatch == durableItem->dispatch);
}

int abortDelayCommand(dispatch)
DISPATCH_PTR dispatch;
{ 
  DURABLE_ITEM_PTR delayItem;

  delayItem = (DURABLE_ITEM_PTR)listMemReturnItem(sameDurableItem, (char *)dispatch,
						  durableItemList);
  if (delayItem) {
    spliceOutDurableItem(delayItem);
    return TRUE;
  }

  return FALSE;
}


/******************************************************************************
*
*  MONITOR REGISTRATION AND INITIALIZATION
*
******************************************************************************/

/******************************************************************************
*
* FUNCTION: void parseMonitorMsg(msg)
*
* DESCRIPTION: 
* Check and parse message format information. This allows monitors to be
* defined before used.
*
* INPUTS: 
* MSG_PTR msg;
*
* OUTPUTS: void.
*
* NOTES: 
* 20-Jun-91: fedor: wow a year to finally do this!
*
******************************************************************************/

void parseMonitorMsg(msg)
MSG_PTR msg;
{
  MSG_PTR conditionMsg, setUpMsg, actMsg;

  if (msg->msgData->class == DemonMonitorClass) {
    setUpMsg = msgFind(msg->msgFormatStr);
    actMsg = msgFind(msg->resFormatStr);

    if (!setUpMsg)
      TCA_ERROR1("ERROR: parseMonitorMsg: Missing Message Definition: %s\n",
		 msg->msgFormatStr);

    if (setUpMsg->msgData->class != QueryClass)
      TCA_ERROR1("ERROR: Demon Set Up Message: %s: Should be QueryClass.\n",
		 msg->msgFormatStr);

    if (!strKeyEqFunc("int", setUpMsg->resFormatStr))
      TCA_ERROR1("ERROR: Demon Set Up Message: %s: Reply format is not int.\n",
		 msg->msgFormatStr);

    if (!actMsg)
      TCA_ERROR1("ERROR: parseMonitorMsg: Missing Message Definition: %s\n",
		 msg->resFormatStr);

    if (!setUpMsg->parsedFormats)
      parseMsg(setUpMsg);

    if (!actMsg->parsedFormats)
      parseMsg(actMsg);

    msg->msgData->msgForm = setUpMsg->msgData->msgForm;
    msg->msgData->resForm = actMsg->msgData->msgForm;
  }
  else {
    conditionMsg = msgFind(msg->msgFormatStr);
    
    if (!conditionMsg)
      TCA_ERROR1("ERROR: parseMonitorMsg: Missing Message Definition: %s\n",
		 msg->msgFormatStr);

    if (!conditionMsg->parsedFormats)
      parseMsg(conditionMsg);

    msg->msgData->msgForm = conditionMsg->msgData->msgForm;
    msg->msgData->resForm = conditionMsg->msgData->resForm;
  }

  msg->parsedFormats = TRUE;
}

/******************************************************************************
*
* FUNCTION: void regMonitorHnd(dispatch, regData)
*
* DESCRIPTION: 
* Handler for tcaRegisterPointMonitor, tcaRegisterPollingMonitor and
* tcaRegisterDemonMonitor.
*
* INPUTS: 
* DISPATCH_PTR dispatch; 
* MON_REG_DATA_PTR regData;
*
* OUTPUTS: none.
*
* NOTES:
*
* 19-Jun-90: fedor: 
* arrrrrgggg! have the system register the message and update the format 
* info later - since we no longer have the string format info to parse. 
*
* Need to allow message def to come later requires some 
* module information query.
*
* Need to test formats of key messages that they return what they are
* supposed to - setUpMsg for demons should return int.
*
* PointMonitor: 
* PollingMonitor:
* msgForm = condition message's msgForm
* resForm = condition message's resForm
*
* DemonMonitor:
* msgForm = setup message's msgForm - setUp QueryClass and reply format "int".
* resForm = action message's msgForm
* cancelMsg - InformClass whose format is "int".
*
******************************************************************************/

void regMonitorHnd(dispatch, regData)
DISPATCH_PTR dispatch; 
MON_REG_DATA_PTR regData;
{ 
  MSG_PTR monMsg;
  MON_REG_DATA_PTR oldRegData;

  centralRegisterMessage(regData->monitorName, regData->class, 
			 (char *)NULL, (char *)NULL);
  monMsg = (MSG_PTR)hashTableFind(regData->monitorName, messageTable);

  /* 22-Jun-91: fedor: reset monitor parse flag because
     centralRegisterMessage will flag it as parsed. */
  monMsg->parsedFormats = FALSE;

  switch (regData->class) {
  case PointMonitorClass:
    centralRegisterHandler(regData->monitorName, "pointMonitorHnd",
		       pointMonitorHnd);
    break;
  case DemonMonitorClass:
  case PollingMonitorClass:
    centralRegisterHandler(regData->monitorName, "intervalMonitorHnd",
		       intervalMonitorHnd);
    break;
  default:
    tcaError("ERROR: regMonitorHnd: unknown monitor class.\n");
    break;
  }

  /* 20-Jun-91: fedor: msgFormatStr and resFormatStr are not byte
     copied because regData is saved to check a defintion change. */

  if (regData->class == DemonMonitorClass) {
    monMsg->msgFormatStr = regData->setUpName;
    monMsg->resFormatStr = regData->actionName;
  }
  else {
    monMsg->msgFormatStr = regData->conditionName;
    monMsg->resFormatStr = regData->conditionName;
  }

  oldRegData = (MON_REG_DATA_PTR)hashTableInsert(regData->monitorName, 
						 strlen(regData->monitorName)+1,
						 (char *)regData, monitorTable);

  if (oldRegData) {
    Log1("WARNING: regMonitorHnd: Registration for %s updated.\n",
	 regData->monitorName);
    tcaFree((char *)oldRegData);
    /* 20-Jun-91: fedor: trigger a reparse on new definition. */
    monMsg->parsedFormats = FALSE;
  }
}


/******************************************************************************
*
* FUNCTION: int pollFn(currentTime, intMon)
*
* DESCRIPTION: 
*
* INPUTS:
*
* OUTPUTS:
*
******************************************************************************/

int pollFn(currentTime, intMon)
int *currentTime;
INTERVAL_MONITOR_DATA_PTR intMon;
{
  if (intMon->fireTime != ALREADY_FIRED && intMon->fireTime <= *currentTime) 
    execPollingMonitor(intMon);

  return TRUE;
}


/******************************************************************************
*
* FUNCTION: int durableFn(currentTime, durablleItem)
*
* DESCRIPTION: 
*
* INPUTS:
*
* OUTPUTS:
*
******************************************************************************/

int durableFn(currentTime, durableItem)
int *currentTime;
DURABLE_ITEM_PTR durableItem;
{
  if (durableItem->endTime <= *currentTime) {
    switch (durableItem->dispatch->class) {
    case CommandClass: 
      /* Currently, this will only be used by the "delay" command 
         If we make it more general, then look out for calling success twice */
      /* 9-Oct-90: fedor - this will do the success because failure 
	 is not done yet. */
      processHandledMessage(durableItem->dispatch);
      break;
    case DemonMonitorClass:
      stopDemonMonitor(durableItem->intMon);
      break;
    case PollingMonitorClass:
      stopPollingMonitor(durableItem->intMon);
      break;
    default: 
      tcaError("Only Commands and Interval Monitors can have durations");
    }
    spliceOutDurableItem(durableItem);
  }
  return TRUE;
}


/******************************************************************************
*
* FUNCTION: void monitorPoll()
*
* DESCRIPTION: 
*
* INPUTS: none.
*
* OUTPUTS: none.
*
* NOTES:
*  9-Oct-90: fedor:
* all time values in danger of wrapping - switch to unsigned?
*
******************************************************************************/

void monitorPoll()
{ 
  int currentTime;

  if (listLength(pollingMonitorList) || listLength(durableItemList)) {
    currentTime = (int)time((time_t *)NULL);
    if (currentTime > lastTime) {
      lastTime = currentTime;
      (void)listIterate(pollFn, (char *)&currentTime, pollingMonitorList);
      (void)listIterate(durableFn, (char *)&currentTime, durableItemList);
    }
  }
}


/******************************************************************************
*
* FUNCTION: void cancelIfAttendingIntervalMonitor(dispatch)
*
* DESCRIPTION: 
* Given a dispatch, test if it is attending and an interval monitor and
* cancel it. An interval monitor is a DemonMonitorClass or PollingMonitorClass.
*
* This provides an interface to taskTree for stopping monitors.
*
* INPUTS: DISPATCH_PTR dispatch;
*
* OUTPUTS: none.
*
* NOTES:
* 16-Nov-90: fedor: questions remain if you stop a monitor that
* is also a durable item.
*
******************************************************************************/

void cancelIfAttendingIntervalMonitor(dispatch)
DISPATCH_PTR dispatch;
{
  INTERVAL_MONITOR_DATA_PTR intMon;

  intMon = (INTERVAL_MONITOR_DATA_PTR)dispatch->respData;
  if (dispatch->status == AttendingDispatch && dispatch->respData) {
    if (dispatch->msg->msgData->class == DemonMonitorClass)
      stopDemonMonitor(intMon);
    else if (dispatch->msg->msgData->class == PollingMonitorClass)
      stopPollingMonitor(intMon);
  }
}


/******************************************************************************
*
* FUNCTION: void monitorInitialize()
*
* DESCRIPTION: Initialize central monitors.
*
* INPUTS: none.
*
* OUTPUTS: none.
*
******************************************************************************/

void monitorInitialize()
{
  durableItemList = listCreate();
  pollingMonitorList = listCreate();

  demonMonitorTable = hashTableCreate(10, intHashFunc, intKeyEqFunc);
  monitorTable = hashTableCreate(10, strHashFunc, strKeyEqFunc);

  centralRegisterMessage("regMonitorMsg", InformClass, 
			 "{int, string, string, string, string}", (char *)NULL);

  centralRegisterHandler("regMonitorMsg", "regMonitorHnd", regMonitorHnd);


  /* 13-Oct-90: fedor: this is a message for users to delay for a certain
     number of seconds - there should be a tca routine for this instead. */
  centralRegisterMessage("tcaDelayCommand", CommandClass, "int", (char *)NULL);
  centralRegisterHandler("tcaDelayCommand", 
			 "delayCommandHnd", delayCommandHnd);
  (void)addAbortHandler("tcaDelayCommand", abortDelayCommand);
}

