/******************************************************************************
*
* PROJECT: Carnegie Mellon Planetary Rover Project
*          Task Control Architecture 
* 
* (c) Copyright 1991 Christopher Fedor and Reid Simmons.  All rights reserved.
* 
* MODULE: monitors
*
* FILE: monMod.c
*
* ABSTRACT:
* 
* module monitor code.
*
* REVISION HISTORY
*
* 13-Oct-90 Christopher Fedor, School of Computer Science, CMU
* Added extra checks for user supplied monitor options.
*
*  3-Oct-90 Christopher Fedor, School of Computer Science, CMU
* created from monitor module routines that were in behviors.c
*
******************************************************************************/

#ifdef VXWORKS
#include "stdioLib.h"
#else
#include "stdio.h"
#endif

#include "tcaInternal.h"
#include "sendMsg.h"

#include "monMod.h" 

extern int testInconsistent();
extern void checkMessageClass();


/******************************************************************************
*
* FUNCTION: TCA_RETURN_VALUE_TYPE tcaPointMonitorWithConstraints(ref, name, 
*                                                                data, tpl)
*
* DESCRIPTION:
*
* INPUTS:
* TCA_REF_PTR ref;
* char *name;
* void *data; 
* int tpl;
*
* OUTPUTS: TCA_RETURN_VALUE_TYPE
*
******************************************************************************/

TCA_RETURN_VALUE_TYPE tcaPointMonitorWithConstraints(ref, name, data, tpl)
TCA_REF_PTR ref;
char *name;
void *data;
int tpl;
{ 
  MSG_PTR msg;

  if (testInconsistent(tpl))
    TCA_MOD_ERROR1("Inconsistent temporal constraints for monitor %s\n", name);

  msg = msgFind(name);
  checkMessageClass(msg, PointMonitorClass);
  
  (void)sendMessage(ref, msg, (char *)data, (char *)&tpl);

  return Success;
}


/******************************************************************************
*
* FUNCTION: TCA_RETURN_VALUE_TYPE tcaPointMonitor(name, data)
*
* DESCRIPTION:
*
* INPUTS: char *name;
*         void *data;
*
* OUTPUTS: TCA_RETURN_VALUE_TYPE
*
******************************************************************************/

TCA_RETURN_VALUE_TYPE tcaPointMonitor(name, data)
char *name;
void *data;
{ 
  MSG_PTR msg;

  msg = msgFind(name);
  checkMessageClass(msg, PointMonitorClass);

  return tcaPointMonitorWithConstraints((TCA_REF_PTR)NULL, name, data, SEQ_ACH);
}

/*************************************************************/

void tcaRegisterDemonMonitor(monitorName, setUpName, actName, cancelName)
char *monitorName, *setUpName, *actName, *cancelName;
{
  MON_REG_DATA_TYPE monRegData;

  monRegData.class = DemonMonitorClass;

  monRegData.monitorName = monitorName;
  monRegData.conditionName = cancelName;
  monRegData.actionName = actName;
  monRegData.setUpName = setUpName;

  (void)tcaInform("regMonitorMsg", (void *)&monRegData);
}


/*************************************************************/

void tcaRegisterPointMonitor(monitorName, condName, actName)
char *monitorName, *condName, *actName;
{
  MON_REG_DATA_TYPE monRegData;

  monRegData.class = PointMonitorClass;

  monRegData.monitorName = monitorName;
  monRegData.conditionName = condName;
  monRegData.actionName = actName;
  monRegData.setUpName = NULL;

  (void)tcaInform("regMonitorMsg", (void *)&monRegData);
}


/*****************************************************************************
*
* FUNCTION: void tcaRegisterPollingMonitor(monitorName, condName, actName)
*
* DESCRIPTION:
*
* INPUTS:
* char *monitorName, *condName, *actName;
*
* OUTPUTS: void.
*
*****************************************************************************/

void tcaRegisterPollingMonitor(monitorName, condName, actName)
char *monitorName, *condName, *actName;
{
  MON_REG_DATA_TYPE monRegData;

  monRegData.class = PollingMonitorClass;

  monRegData.monitorName = monitorName;
  monRegData.conditionName = condName;
  monRegData.actionName = actName;
  monRegData.setUpName = NULL; 

  (void)tcaInform("regMonitorMsg", (void *)&monRegData);
}


/*****************************************************************************
*
* FUNCTION: void initializeMonitorOptions(options)
*
* DESCRIPTION:
* Sets default values for interval monitors.
* Options and defaults are: 
* maxFire:     maximum times to fire action message; -1 indicates infinite
*              (default is -1)
* fireEvery:   fire action message every n times condition is satisfied.
*              (default is 1)
* duration:    maximum time, in seconds, that the monitor can run;
*	       -1 indicates infinite time (default is -1)
* period:      time between polls, in seconds (default is 10)
* initialWait: if TRUE, don't poll immediately (default is FALSE)
*
* INPUTS: INTERVAL_MONITOR_OPTIONS_PTR options;
*
* OUTPUTS: void
*
* NOTES:
* as a side effect the fields of options is initialized.
*
*****************************************************************************/

void initializeMonitorOptions(options)
INTERVAL_MONITOR_OPTIONS_PTR options;
{
  options->maxFire = -1;
  options->fireEvery = 1;
  options->duration = -1;
  options->period = 10;
  options->initialWait = FALSE;
}


/*****************************************************************************
*
* FUNCTION: INTERVAL_MONITOR_OPTIONS_PTR tcaCreateMonitorOptions()
*
* DESCRIPTION: Creates and initializes a monitor option.
*
* INPUTS: none.
*
* OUTPUTS: INTERVAL_MONITOR_OPTIONS_PTR
*
*****************************************************************************/

INTERVAL_MONITOR_OPTIONS_PTR tcaCreateMonitorOptions()
{
  INTERVAL_MONITOR_OPTIONS_PTR options;

  options = NEW(INTERVAL_MONITOR_OPTIONS_TYPE);

  initializeMonitorOptions(options);

  return options;
}


/*****************************************************************************
* 
* FUNCTION void checkForDeadLock(monitorName, expirationTime, options)
*
* DESCRIPTION: Display warning message about possible deadlock.
*
* INPUTS:
* char *monitorName;
* TCA_TIME_POINT_PTR expirationTime;
* INTERVAL_MONITOR_OPTIONS_PTR options;
*
* OUTPUTS: void
*
*****************************************************************************/

void checkForDeadlock(monitorName, expirationTime, options)
char *monitorName;
TCA_TIME_POINT_PTR expirationTime;
INTERVAL_MONITOR_OPTIONS_PTR options;
{
  if (expirationTime == &tcaDefaultTimeGlobal) {
    fprintf(stderr, "WARNING: interval monitor %s %s\n", monitorName,
	    (((options->maxFire <= 0) && (options->duration < 0)) ? 
	     "will last forever unless explicitly killed." :
	     "may last forever."));
#ifdef LISP
    fflush(stdout);
#endif /* LISP */
  }
}


/*****************************************************************************
*
* FUNCTION: TCA_RETURN_VALUE_TYPE tcaIntervalMonitorWithConstraints(ref, name,
*                                                data, start, expire, options)
*
* DESCRIPTION:
* IF ref == NULL, use the issuer as the monitor's parent.
* IF start == DEFAULT_TIME, the start time of the monitor is the 
*    start time of its parent.
* IF expire == DEFAULT_TIME, the monitor could last forever unless
*    appropriate options are specified.
* IF options == NULL, see initializeMonitorOptions() for default values.
*
* INPUTS:
* TCA_REF_PTR ref; 
* char *name;
* void *data;
* TCA_TIME_POINT_TYPE start, expire;
* INTERVAL_MONITOR_OPTIONS_PTR options;
*
* OUTPUTS: TCA_RETURN_VALUE
*
*****************************************************************************/

TCA_RETURN_VALUE_TYPE tcaIntervalMonitorWithConstraints(ref, name, data, start,
							expire, options)
TCA_REF_PTR ref; 
char *name;
void *data;
TCA_TIME_POINT_TYPE start, expire;
INTERVAL_MONITOR_OPTIONS_PTR options;
{
  MSG_PTR msg;
  INTERVAL_MON_CLASS_TYPE intMonOptions;
   
  msg = msgFind(name);

  if ((msg->msgData->class != PollingMonitorClass) &&
      (msg->msgData->class != DemonMonitorClass)) {
    TCA_MOD_ERROR1("Monitor %s must be a polling or demon-invoked monitor\n",
		   name);
  }

  if (options) {
    intMonOptions.options = options;
    if (options->maxFire < -1) {
      fprintf(stderr, "WARNING: Max Fire of monitor %s set to default of -1\n",
	      name);
#ifdef LISP
      fflush(stdout);
#endif /* LISP */
      intMonOptions.options->maxFire = -1;
    }
    if (options->fireEvery < 1) {
      fprintf(stderr, "WARNING: Fire Every of monitor %s set to default of 1\n",
	      name);
#ifdef LISP
      fflush(stdout);
#endif /* LISP */
      intMonOptions.options->fireEvery = 1;
    }
    if (options->duration < -1) {
      fprintf(stderr, "WARNING: Duration of monitor %s set to default of -1.\n",
	      name);
#ifdef LISP
      fflush(stdout);
#endif /* LISP */
      intMonOptions.options->duration = -1;
    }
    if (options->period < 1) {
      fprintf(stderr, 
	      "WARNING: Polling period of monitor %s rounded up to 1 second\n",
	      name);
#ifdef LISP
      fflush(stdout);
#endif /* LISP */
      intMonOptions.options->period = 1;
    }
    if (options->initialWait != FALSE && options->initialWait != TRUE) {
      fprintf(stderr, "WARNING: Initial wait of monitor %s set to FALSE\n",
	      name);
#ifdef LISP
      fflush(stdout);
#endif /* LISP */
      intMonOptions.options->initialWait = FALSE;
    }
  }
  else
    intMonOptions.options = tcaCreateMonitorOptions();
  
  checkForDeadlock(name, &expire, intMonOptions.options);

  intMonOptions.start = start;
  intMonOptions.expire = expire;
  
  return sendMessage(ref, msg, (char *)data, (char *)&intMonOptions);
}

#ifdef LISP

TCA_RETURN_VALUE_TYPE tcaIntMonWithConstraintsLISPC(ref, name, data,
    startClass, startINTclass, startINTmsgRef,
    expireClass, expireINTclass, expireINTmsgRef,
    maxFire, fireEvery, duration, period, initialWait)
TCA_REF_PTR ref; 
char *name, *data;
TCA_INTERVAL_CLASS_TYPE startClass, expireClass, startINTclass, expireINTclass;
int startINTmsgRef, expireINTmsgRef;
int maxFire, fireEvery, duration, period, initialWait;
{
  TCA_TIME_POINT_TYPE start, expire;
  TCA_INTERVAL_TYPE startInt, expireInt;

  INTERVAL_MONITOR_OPTIONS_TYPE options;

  options.maxFire = maxFire;
  options.fireEvery = fireEvery;
  options.duration = duration;
  options.period = period;
  options.initialWait = initialWait;

  startInt.class = startINTclass;
  startInt.msgRef = startINTmsgRef;

  expireInt.class = expireINTclass;
  expireInt.msgRef = expireINTmsgRef;

  start.class = startClass;
  start.interval = startInt;

  expire.class = expireClass;
  expire.interval = expireInt;

  return(tcaIntervalMonitorWithConstraints(ref, name, data, 
					   start, expire, &options));
}

#endif /* LISP */


/*****************************************************************************
*
* FUNCTION: TCA_RETURN_VALUE_TYPE tcaIntervalMonitor(name, data)
*
* DESCRIPTION:
*
* INPUTS: char *name;
*         void *data; 
*
* OUTPUTD: TCA_RETURN_VALUE
*
*****************************************************************************/
   
TCA_RETURN_VALUE_TYPE tcaIntervalMonitor(name, data)
char *name;
void *data; 
{
  return tcaIntervalMonitorWithConstraints((TCA_REF_PTR)NULL, name, data,
					   tcaDefaultTimeGlobal, 
					   tcaDefaultTimeGlobal, 
					   (INTERVAL_MONITOR_OPTIONS_PTR)NULL);
}


/*****************************************************************************
*
* FUNCTION: TCA_RETURN_VALUE_TYPE tcaFireDemon(name, demonUserId, data)
*
* DESCRIPTION:
*
* INPUTS:
* char *name;
* void *data;
* int demonUserId;
*
* OUTPUTS: TCA_RETURN_VALUE_TYPE
*
*****************************************************************************/

TCA_RETURN_VALUE_TYPE tcaFireDemon(name, demonUserId, data)
char *name;
void *data;
int demonUserId;
{ 
  MSG_PTR msg;

  msg = msgFind(name);
  
  return sendResponse((TCA_REF_PTR)NULL, msg, (char *)data, FireDemonClass, 
		      (char *)&demonUserId);
}


