/******************************************************************************
*
* PROJECT: Carnegie Mellon Planetary Rover Project
*          Task Control Architecture 
* 
* (c) Copyright 1991 Christopher Fedor and Reid Simmons.  All rights reserved.
* 
* MODULE: tca
*
* FILE: msgTap.c
*
* ABSTRACT: 
*
* Provide the capability to "tap" a message.
*
* One can set up a "tap" on a message so that whenever a specified condition
* holds, a "listening" message is sent the same data as the tapped message.
* The conditions under which the tap is sent relate mainly to the status of
* the tapped message (e.g., before it is handled, after a reply is received, 
* after the message, and its submessages, are achieved, etc.).
* The file "msgTapMon.c" gives a complete listing of the available conditions.
*
* REVISION HISTORY
*
* 13-Sep-91 Christopher Fedor, School of Computer Science, CMU
* Changed conditionName from an array to a function.
*
* 18-Aug-91 Christopher Fedor, School of Computer Science, CMU
* Changed tapClassData for commands to be of type BLOCK_COM_PTR.
*
* 17-Aug-91 Christopher Fedor, School of Computer Science, CMU
* Modified tapClassData to malloc the class data instead of passing an address.
*
*  8-Jul-91 Reid Simmons, School of Computer Science, CMU
* "cleanUpHandler" had to be enhanced considerably -- there were real bad
* interactions when it was called by a rule, and killed off a task tree before
* other rules had a chance to fire! 
*
* 24-Jun-91 Reid Simmons, School of Computer Science, CMU
* Streamlined use of rules, and added "cleanUpAfterAchieved" message.
*
* 23-Jun-91 Christopher Fedor, School of Computer Science, CMU
* Added findOrRegisterMessage calls to tapMessageHandler call to create 
* a temp message for missing messages. This call is used by handlers
* that are registered before their message. Because taps are activated
* only in response to a message being used, other routines will catch the
* missing message and update things before the tap call. The update occurs
* in selfRegisterMsg.
*
* 12-Mar-91 Reid Simmons, School of Computer Science, CMU
* Created.
*
******************************************************************************/

#include "stdio.h"

#include "tca.h"
#include "tcaInternal.h"
#include "logging.h"

#include "datamsg.h"
#include "hash.h"
#include "formatters.h"
#include "idtable.h"
#include "dispatch.h"
#include "list.h"

#include "tms.h"
#include "cells.h"
#include "rules.h"
#include "qlattice.h"
#include "tplConstr.h"
#include "taskTree.h"

#include "msgTap.h"

#include "res.h"

extern MODULE_PTR tcaServerModGlobal;

extern HASH_TABLE_PTR messageTable;

extern DISPATCH_PTR buildDispatchInternal (); /* Defined in recvMsg.c */

extern QUANTITY_PTR Now; /* Defined in tplConstr.c */
extern QUANTITY_PTR Time_Of(); /* Defined in tplConstr.c */

extern MSG_PTR findOrRegisterMessage();

extern void killTaskTree(); /* defined in taskTree.c */

static TCA_TIME_POINT_TYPE startHandling, endHandling;
static TCA_TIME_POINT_TYPE startAchievement, endAchievement;
static TCA_TIME_POINT_TYPE startPlanning, endPlanning;

static RULE_PTR beforeHandlingRule, beforeAchievingRule, whileAchievingRule;
static RULE_PTR afterAchievedRule, whenAchievedRule, beforePlanningRule;
static RULE_PTR whilePlanningRule, afterPlannedRule, whenPlannedRule;

/*****************************************************************
* This is used for handling the case where the "cleanUpHandler" is
* invoked from within a rule firing (e.g., a tap activation).
****************************************************************/
static TMS_NODE_PTR alwaysTrueNodeGlobal;
static RULE_PTR killLaterRuleGlobal = NULL;


/*****************************************************************************
* 
* FUNCTION: char *conditionName(cond)
*
* DESCRIPTION: Return a string for display use of the condition name.
*
* INPUTS: TAP_CONDITION_TYPE cond
*
* OUTPUTS: char *
*
*****************************************************************************/

char *conditionName(cond)
TAP_CONDITION_TYPE cond;
{
  switch(cond) {
  case WhenSent:
    return("WhenSent");
  case BeforeHandling:
    return("BeforeHandling");
  case WhileHandling:
    return("WhileHandling");
  case AfterHandled:
    return("AfterHandled");
  case AfterReplied:
    return("AfterReplied");
  case AfterSuccess:
    return("AfterSuccess");
  case AfterFailure:
    return("AfterFailure");
  case BeforeAchieving:
    return("BeforeAchieving");
  case WhileAchieving:
    return("WhileAchieving");
  case AfterAchieved:
    return("AfterAchieved");
  case WhenAchieved:
    return("WhenAchieved");
  case BeforePlanning:
    return("BeforePlanning");
  case WhilePlanning:
    return("WhilePlanning");
  case AfterPlanned:
    return("AfterPlanned");
  case WhenPlanned:
    return("WhenPlanned");
  case WhenKilled:
    return("WhenKilled");
  default:
    return("UNKNOWN CONDITION");
  }
}


/*****************************************************************************
* 
* FUNCTION: int sameTap(tap1, tap2)
*
* DESCRIPTION: Do the two taps have the same conditions and messages?
*
* INPUTS: TAP_PTR tap1, tap2
*
* OUTPUTS: TRUE/FALSE
*
*****************************************************************************/

int sameTap(tap1, tap2)
TAP_PTR tap1, tap2;
{
  return ((tap1->condition == tap2->condition) &&
	  (tap1->listeningMsg == tap2->listeningMsg));
}

/******************************************************************************
*
* FUNCTION: addTapInfoToDispatch(tappedDispatch, allocateExtraTaps)
* 
* DESCRIPTION: Add a TAP_INFO_TYPE structure to the tapInfo field of the 
*  tappedDispatch (if one was not there already).  If allocateExtraTaps is
*  TRUE, start a list of extra taps for this dispatch ("extra taps" apply only
*  to the dispatch itself, not to all messages of that type; they are added
*  using "tcaTapReference").
*
* INPUTS: DISPATCH_PTR tappedDispatch;
*         int allocateExtraTaps; (TRUE/FALSE)
*
* OUTPUTS: modifies the tapInfo field of the tappedDispatch
*
******************************************************************************/

void addTapInfoToDispatch(tappedDispatch, allocateExtraTaps)
DISPATCH_PTR tappedDispatch; 
int allocateExtraTaps;
{
  TAP_INFO_PTR tapInfo;

  if (!tappedDispatch->tapInfo) {
    tapInfo = NEW(TAP_INFO_TYPE);
    tapInfo->tappedDispatch = NULL;
    tapInfo->extraTaps = NULL;
    tappedDispatch->tapInfo = tapInfo;
  }
  if (allocateExtraTaps && !tappedDispatch->tapInfo->extraTaps) {
    tappedDispatch->tapInfo->extraTaps = listCreate();
  }
}


/******************************************************************************
*
* FUNCTION: freeTapInfo(tapInfo)
* 
* DESCRIPTION: Free the tap info structure and the "extraTaps" list
*
* INPUTS: TAP_INFO_PTR tapInfo
*
* OUTPUTS: void
*
******************************************************************************/

void freeTapInfo(tapInfo)
TAP_INFO_PTR tapInfo;
{
  if (tapInfo) {
    if (tapInfo->extraTaps) 
      listFree(tapInfo->extraTaps);
    tcaFree((char *)tapInfo);
  }
}


/******************************************************************************
*
* FUNCTION: 
* 
* DESCRIPTION: 
*
* INPUTS: 
*
* OUTPUTS: 
*
******************************************************************************/


/* Returns TRUE if the tapped and listening messages are the right
   types for the given tap condition */

int checkTapMessageClasses(tapCondition, tappedMsg, listeningMsg)
TAP_CONDITION_TYPE tapCondition; 
MSG_PTR tappedMsg, listeningMsg;
{
  TCA_MSG_CLASS_TYPE tappedClass, listeningClass;
  int queryType;

  tappedClass = tappedMsg->msgData->class;
  listeningClass = listeningMsg->msgData->class;

  if (listeningClass == QueryClass) {
    Log_Message1("TAP NOT ADDED: the listening message %s must be a query\n",
		 listeningMsg->msgData->name);
    return FALSE;
  }

  switch (tapCondition) {
  case WhenSent: 
  case BeforeHandling: 
  case WhileHandling:
  case AfterHandled: 
    return TRUE;

  case AfterReplied: 
    queryType = (tappedClass == QueryClass || tappedClass == MultiQueryClass);
    if (!queryType) {
     Log_Message2("TAP NOT ADDED: %s must be a query for the '%s' condition\n",
	   tappedMsg->msgData->name, conditionName(tapCondition));
    } 
    return queryType; 

  case AfterSuccess:
  case AfterFailure: 
    if ((tappedClass == GoalClass) || (tappedClass == CommandClass)) {
      return TRUE;
    } else {
Log_Message2("TAP NOT ADDED: %s must be a goal or command for '%s' condition\n"
	   ,tappedMsg->msgData->name, conditionName(tapCondition));
      return FALSE;
    }

  case BeforeAchieving:
  case WhileAchieving:
  case AfterAchieved:
  case WhenAchieved: 
  case WhenKilled: 
    if ((tappedClass == GoalClass) || (tappedClass == CommandClass) ||
	(tappedClass == PointMonitorClass) || 
	(tappedClass == PollingMonitorClass) || 
	(tappedClass == DemonMonitorClass)){
      return TRUE;
    } else {
   Log_Message2("TAP NOT ADDED: %s must be a goal, command or monitor for '%s' condition\n",
	   tappedMsg->msgData->name, conditionName(tapCondition));
      return FALSE;
    }

  case BeforePlanning:
  case WhilePlanning:
  case AfterPlanned:
  case WhenPlanned: 
    if (tappedClass != GoalClass) {
      Log_Message2("TAP NOT ADDED: %s must be a goal for '%s' condition\n",
	   tappedMsg->msgData->name, conditionName(tapCondition));
    }
    return (tappedClass == GoalClass);

  default: 
    Log_Message1("DID NOT FIND %d\n", (char *)(int)tapCondition);
    return FALSE;
  }
}


/******************************************************************************
*
* FUNCTION: 
* 
* DESCRIPTION: 
*
* INPUTS: 
*
* OUTPUTS: 
*
******************************************************************************/

/* Return the nominal class data associated with messages of the given class */

char *tapClassData(class)
TCA_MSG_CLASS_TYPE class;
{
  int *tplConstr;
  BLOCK_COM_PTR comData;
  MULTI_QUERY_CLASS_PTR mqData;

  switch (class) {
  case GoalClass:
  case PointMonitorClass: 
    tplConstr = NEW(int);
    *tplConstr = NO_TPLCONSTR;
    return (char *)tplConstr;

  case CommandClass:
    comData = NEW(BLOCK_COM_TYPE);
    comData->waitFlag = FALSE;
    comData->tplConstr = NO_TPLCONSTR;
    return (char *)comData;

  case PollingMonitorClass:
  case DemonMonitorClass: 
    tcaError("tapClassData: Monitor class data not implemented");

  case QueryClass:
  case InformClass:
  case BroadcastClass:
  case FailureClass: 
    return NULL;

  case MultiQueryClass:
    mqData = NEW(MULTI_QUERY_CLASS_TYPE);
    mqData->num = mqData->max = 0;
    return (char *)mqData;

  default: 
    tcaError("tapClassData: Class type not handled"); return NULL;
  }
}


/******************************************************************************
*
* FUNCTION: int checkTaps(msg, tap)
* 
* DESCRIPTION: 
* List iterator for messages with taps that were registered after the tap
* was registered.
*
* INPUTS: 
* MSG_PTR msg; - the message being tapped.
* TAP_PTR tap;
*         
* OUTPUTS: 1 - force entire list iteration.
*
******************************************************************************/

int checkTaps(msg, tap)
MSG_PTR msg;
TAP_PTR tap;
{
  if (!checkTapMessageClasses(tap->condition, msg, tap->listeningMsg))
    listDeleteItem((char *)tap, msg->tapList);

  return 1;
}


/******************************************************************************
*
* FUNCTION: tapMessageHandler(dispatch, tapData)
* 
* DESCRIPTION: Handle the tcaTapMessage call
*
* INPUTS: DISPATCH_PTR dispatch;
*         TAP_MSG_PTR tapData;
*
* OUTPUTS: none
*
******************************************************************************/

void tapMessageHandler(dispatch, tapData)
DISPATCH_PTR dispatch; TAP_MSG_PTR tapData;
{
  TAP_PTR tap;
  MSG_PTR tappedMsg, listeningMsg;

  tappedMsg = findOrRegisterMessage(tapData->tappedMsg);
  listeningMsg = findOrRegisterMessage(tapData->listeningMsg);

  if (tappedMsg->msgData->class == HandlerRegClass ||
      listeningMsg->msgData->class == HandlerRegClass ||
      checkTapMessageClasses(tapData->condition, tappedMsg, listeningMsg)){
    tap = NEW(TAP_TYPE);
    tap->condition = tapData->condition;
    tap->listeningMsg = listeningMsg;
    if (!tappedMsg->tapList) {
      tappedMsg->tapList = listCreate();
    }
    if (listMemReturnItem(sameTap, (char *)tap, tappedMsg->tapList)) {
      Log_Message3("Tap %s for %s with condition %s already inserted\n", 
		   tapData->listeningMsg, tapData->tappedMsg,
		   conditionName(tapData->condition));
      tcaFree((char *)tap);
    } else {
      listInsertItemLast((char *)tap, tappedMsg->tapList);
    }
  }
  /* Just free the top level structure, because the string names might
     be saved in "findOrRegisterMessage" */
  tcaFree((char *)tapData);
}


/******************************************************************************
*
* FUNCTION: tapReferenceHandler(dispatch, tapData)
* 
* DESCRIPTION: Handle the tcaTapReference call
*
* INPUTS: DISPATCH_PTR dispatch;
*         TAP_REF_PTR tapData;
*
* OUTPUTS: none
*
******************************************************************************/

void tapReferenceHandler(dispatch, tapData)
DISPATCH_PTR dispatch; TAP_REF_PTR tapData;
{
  TAP_PTR tap;
  DISPATCH_PTR tapDispatch;
  MSG_PTR tappedMsg, listeningMsg;

  tappedMsg = tapData->tappedRef->msg;
  listeningMsg = findOrRegisterMessage(tapData->listeningMsg);

  if (listeningMsg->msgData->class == HandlerRegClass ||
      checkTapMessageClasses(tapData->condition, tappedMsg, listeningMsg)) {
    tap = NEW(TAP_TYPE);
    tap->condition = tapData->condition;
    tap->listeningMsg = listeningMsg;
    tapDispatch = DISPATCH_FROM_ID(tapData->tappedRef->refId);
    addTapInfoToDispatch(tapDispatch, TRUE);
    listInsertItemLast((char *)tap, tapDispatch->tapInfo->extraTaps);
  }
  /* Don't free the listeningMsg string, since it might be used in
     "findOrRegisterMessage" */
  tcaRefFree(tapData->tappedRef);
  tcaFree((char *)tapData);
}


/******************************************************************************
*
* FUNCTION: findTappedReferenceHandler(dispatch, tapData)
* 
* DESCRIPTION: Handle the tcaFindTappedReference call
*
* INPUTS: DISPATCH_PTR dispatch;
*         TCA_REF_PTR listeningRefPtr;
*
* OUTPUTS: none
*
******************************************************************************/

void findTappedReferenceHandler(dispatch, listeningRefIdPtr)
DISPATCH_PTR dispatch; int *listeningRefIdPtr;
{
  TCA_REF_PTR tappedRef; 
  DISPATCH_PTR listeningDispatch, tappedDispatch;

  listeningDispatch = DISPATCH_FROM_ID(*listeningRefIdPtr);
  if (!listeningDispatch->tapInfo || 
      !listeningDispatch->tapInfo->tappedDispatch) {
    centralReply(dispatch, (char *)NULL);
  } else {
    tappedDispatch = listeningDispatch->tapInfo->tappedDispatch;
    tappedRef = tcaRefCreate((MSG_PTR)NULL, tappedDispatch->msg->msgData->name,
			     tappedDispatch->locId);
    reserveDispatch(tappedDispatch);
    centralReply(dispatch, (char *)&tappedRef);
    tcaRefFree(tappedRef);
  }
  tcaFree((char *)listeningRefIdPtr); /* Use simple free: only 1 int */
}


/******************************************************************************
*
* FUNCTION: void removeTapHandler(dispatch, tapData)
* 
* DESCRIPTION: Handle the tcaRemoveTap call
*
* INPUTS: 
* DISPATCH_PTR dispatch;
* TAP_MSG_PTR tapData;
*
* OUTPUTS: none
*
******************************************************************************/

void removeTapHandler(dispatch, tapData)
DISPATCH_PTR dispatch; 
TAP_MSG_PTR tapData;
{
  TAP_TYPE tap;
  MSG_PTR tappedMsg, listeningMsg;

  tappedMsg  = (MSG_PTR)hashTableFind(tapData->tappedMsg, messageTable);

  if (!tappedMsg) {
    Log_Message1("TAP NOT REMOVED: Message %s not registered\n",
		 tapData->tappedMsg);
  } else {
    listeningMsg = (MSG_PTR)hashTableFind(tapData->listeningMsg, messageTable);
    if (!listeningMsg) {
      Log_Message1("TAP NOT REMOVED: Message %s not registered\n", 
		   tapData->listeningMsg);
    } else if (!tappedMsg->tapList) {
      Log_Message1("%s does not currently have taps", tapData->tappedMsg);
    } else {
      tap.condition = tapData->condition;
      tap.listeningMsg = listeningMsg;
      listTestDeleteItem(sameTap, (char *)&tap, tappedMsg->tapList);
    }
  }
  /* A bit more efficient than using tcaFreeData. 
   * It's OK to free the whole structure here, because the code above
   * only uses "hashTableFind" -- the message names are not stored anywhere
   */
  freeDataStructure(dispatch->msg->msgData->msgForm, (char *)tapData);
}


/*****************************************************************************
*
* FUNCTION: cleanUpLater(ruleName, node, killList)
* 
* DESCRIPTION: Kill all the nodes in the killList after the 
*              other rules have run.  This is a bit hairy to prevent nodes
*              from being killed multiple times.
*
* INPUTS: 
*
* OUTPUTS: Returns TRUE (needed by listIterate)
*
******************************************************************************/

static int removeChildrenFromKillList (killList, treeNode)
     LIST_PTR killList; TASK_TREE_NODE_PTR treeNode; 
{
  listDeleteItem((char *)treeNode->dispatch, killList);
  (void)listIterate(removeChildrenFromKillList, (char *)killList, 
		    treeNode->children);
  return TRUE;
}

void cleanUpLater (ruleName, node, killList)
     char *ruleName; TMS_NODE_PTR node; LIST_PTR killList;
{
  DISPATCH_PTR dispatchToKill;

  dispatchToKill = (DISPATCH_PTR)listFirst(killList);
  while (dispatchToKill) {
    /* Avoid killing nodes multiple times */
    (void)listIterate(removeChildrenFromKillList, (char *)killList,
		      dispatchToKill->treeNode->children);

    killTaskTree(dispatchToKill->locId);
    dispatchToKill = (DISPATCH_PTR)listNext(killList);
  }
  listFree(killList);

  killLaterRuleGlobal = NULL;
}

/*****************************************************************************
*
* FUNCTION: cleanUpHandler(listeningDispatch, dummy)
* 
* DESCRIPTION: Find the associated tapped reference and kill its task tree.
*
* INPUTS: 
* DISPATCH_PTR listeningDispatch;
* char *dummy;
*
* OUTPUTS: none
*
******************************************************************************/

void cleanUpHandler(listeningDispatch, dummy)
DISPATCH_PTR listeningDispatch; 
char *dummy;
{
  DISPATCH_PTR tappedDispatch;

  tappedDispatch = listeningDispatch->tapInfo->tappedDispatch;

  if (!tappedDispatch->treeNode) {
    Log_Status1("ERROR: Trying to clean up a dispatch (%s) with no task tree node",
		tappedDispatch->msg->msgData->name);
  } else if (isDelayedGlobal) {
    /* Delay the killing of the dispatch until the current set of rules have run */
    if (!killLaterRuleGlobal) {
      killLaterRuleGlobal = ruleCreateNode("Delay Kill Rule", cleanUpLater, 5,
					   alwaysTrueNodeGlobal, 
					   (char *)listCreate());
      addRuleToNode(killLaterRuleGlobal, alwaysTrueNodeGlobal);
    }
    listInsertItemLast((char *)tappedDispatch,
		       (LIST_PTR)killLaterRuleGlobal->parameters);
  } else {
    killTaskTree(tappedDispatch->locId);
  }
}


/******************************************************************************
*
* FUNCTION: int tapOK(tap)
* 
* DESCRIPTION: 
* Returns true if tap message was registered and has a valid handler.
*
* INPUTS: 
* TAP_PTR tap;
*
* OUTPUTS: int
*
* NOTES:
* A handler will have a resource with capacity of 0 if the server created it.
*
******************************************************************************/

int tapOK(tap)
TAP_PTR tap;
{
  HND_PTR hnd;
  MSG_PTR msg;
  int flag;
  LIST_PTR hndList;

  msg = tap->listeningMsg;

  if (msg->msgData->class == HandlerRegClass) {
    Log_Message1("WARNING: Tap Message Ignored: %s Not Registered.\n",
		 msg->msgData->name);
    return 0;
  }

  hndList = msg->hndList;

  /* 18-Oct-91: fedor: blah! central handlers should
     have a central resource as well - need to merge this idea
     with hndSelect/chooseMsgHanlder in recvMsg */

  flag = 0;
  hnd = (HND_PTR)listFirst(hndList);
  while (hnd && !flag) {
    if (hnd->hndOrg == tcaServerModGlobal && hnd->hndProc) 
      flag = 1;
    else if (hnd->resource && hnd->resource->capacity)
      flag = 1;
    hnd = (HND_PTR)listNext(hndList);
  }

  if (!flag) {
    Log_Message1("WARNING: Tap Message Ignored: %s No Handler.\n",
		 msg->msgData->name);
  }
    
  return flag;
}


/******************************************************************************
*
* FUNCTION: DISPATCH_PTR createListeningDispatch (tappedDispatch, dataMsg, tap)
* 
* DESCRIPTION: 
*
* INPUTS: 
* DISPATCH_PTR tappedDispatch; 
* DATA_MSG_PTR dataMsg; 
* TAP_PTR tap;
*
* OUTPUTS: DISPATCH_PTR
*
******************************************************************************/

DISPATCH_PTR createListeningDispatch (tappedDispatch, dataMsg, tap)
DISPATCH_PTR tappedDispatch; 
DATA_MSG_PTR dataMsg; 
TAP_PTR tap;
{
  DISPATCH_PTR listeningDispatch;
  TCA_MSG_CLASS_TYPE listeningClass;

  if (!tapOK(tap))
    return NULL;

  listeningClass = tap->listeningMsg->msgData->class;
  listeningDispatch = buildDispatchInternal(tap->listeningMsg, dataMsg, 
					    tapClassData(listeningClass),
					    (DISPATCH_PTR)NULL, NULL, (char *)NULL);
  if (taskTreeClass(listeningDispatch->class)) {
    Initialize_TaskTreeNode(listeningDispatch, NO_REF);
  }
  addTapInfoToDispatch(listeningDispatch, FALSE);
  listeningDispatch->tapInfo->tappedDispatch = tappedDispatch;
  return listeningDispatch;
}


/******************************************************************************
*
* FUNCTION: void sendListeningDispatch (tappedDispatch, dataMsg, tap)
* 
* DESCRIPTION: 
*
* INPUTS: 
* DISPATCH_PTR tappedDispatch; 
* DATA_MSG_PTR dataMsg; 
* TAP_PTR tap;
*
* OUTPUTS: 
*
******************************************************************************/

void sendListeningDispatch (tappedDispatch, dataMsg, tap)
DISPATCH_PTR tappedDispatch; 
DATA_MSG_PTR dataMsg; 
TAP_PTR tap;
{
  DISPATCH_PTR listeningDispatch;
  MSG_PTR listeningMsg;

  listeningDispatch = createListeningDispatch(tappedDispatch, dataMsg, tap);

  if (!listeningDispatch)
    return;

  listeningMsg = DISPATCH_MSG(listeningDispatch);
  recvMessage(listeningDispatch, listeningMsg->msgData->class,
	      listeningDispatch->classData);
}


/******************************************************************************
*
* FUNCTION: void sendListeningDispatchAndConstrain(tappedDispatch, tap, 
*                    tappedTimePoint, relationship, listeningTimePoint)
* 
* DESCRIPTION: 
*
* INPUTS: 
* DISPATCH_PTR tappedDispatch; 
* TAP_PTR tap;
* TCA_TIME_POINT_PTR tappedTimePoint, listeningTimePoint; 
* char *relationship;
*
* OUTPUTS: void
*
******************************************************************************/

void sendListeningDispatchAndConstrain (tappedDispatch, tap, tappedTimePoint,
					relationship, listeningTimePoint)
DISPATCH_PTR tappedDispatch; 
TAP_PTR tap;
TCA_TIME_POINT_PTR tappedTimePoint, listeningTimePoint; 
char *relationship;
{
  DISPATCH_PTR listeningDispatch;
  MSG_PTR listeningMsg;
  DATA_MSG_PTR tappedMsgData;
  QUANTITY_PTR tappedQuantity, listeningQuantity;

  tappedMsgData = DISPATCH_MSG_DATA(tappedDispatch);

  listeningDispatch = createListeningDispatch(tappedDispatch, tappedMsgData, 
					      tap);

  if (!listeningDispatch)
    return;

  listeningMsg = DISPATCH_MSG(listeningDispatch);

  if (taskTreeClass(listeningDispatch->class)) {
    tappedTimePoint->interval.msgRef = tappedDispatch->locId;
    tappedQuantity = Time_Of(tappedTimePoint);
    listeningTimePoint->interval.msgRef = listeningDispatch->locId;
    listeningQuantity = Time_Of(listeningTimePoint);
    QAssert(tappedQuantity, relationship, listeningQuantity, "Tap Constraint");
  }
  recvMessage(listeningDispatch, listeningMsg->msgData->class, 
	      listeningDispatch->classData);
}


/******************************************************************************
*
* FUNCTION: int sendOnCondition(condition, dispatch, tap)
* 
* DESCRIPTION: 
*
* INPUTS: 
* TAP_CONDITION_TYPE condition; 
* DISPATCH_PTR dispatch; 
* TAP_PTR tap; 
*
* OUTPUTS: int
*
******************************************************************************/

int sendOnCondition(condition, dispatch, tap)
TAP_CONDITION_TYPE condition; 
DISPATCH_PTR dispatch; 
TAP_PTR tap; 
{
  if (tap->condition == condition) {
    sendListeningDispatch(dispatch, DISPATCH_MSG_DATA(dispatch), tap);
  }
  return TRUE;
}


/******************************************************************************
*
* FUNCTION: void tapOnCondition(sendOnConditionFn, dispatch)
* 
* DESCRIPTION: 
*
* INPUTS: 
* int (*sendOnConditionFn)(); 
* DISPATCH_PTR dispatch;
*
* OUTPUTS: void
*
******************************************************************************/

void tapOnCondition(sendOnConditionFn, dispatch)
int (*sendOnConditionFn)(); 
DISPATCH_PTR dispatch;
{
  LIST_PTR tapsList, extraTapsList;

  tapsList = DISPATCH_TAPS(dispatch);

  if (tapsList) {
    (void)listIterateFromFirst(sendOnConditionFn, (char *)dispatch, tapsList);
  }

  extraTapsList = DISPATCH_EXTRA_TAPS(dispatch);

  if (extraTapsList) {
    (void)listIterateFromFirst(sendOnConditionFn, (char *)dispatch, extraTapsList);
  }
}


/******************************************************************************
*
* FUNCTION: void tapActivation(ruleName, relationshipNode, activationFn)
* 
* DESCRIPTION: 
*
* INPUTS: 
* char *ruleName;
* TMS_NODE_PTR relationshipNode;
* int (*activationFn)();
*
* OUTPUTS: void
*
******************************************************************************/

extern void tapActivation(); /* defined below */

void removeTapRule(activationFn, relation)
int (*activationFn)(); 
CELL_PTR relation;
{
  RULE_PTR rule;

  rule = (RULE_PTR)listFirst(relation->outRules);
  while (rule) {
    if ((rule->fn == tapActivation) &&
        (rule->parameters == (char *)activationFn)) {
      listDeleteItem((char *)rule, relation->outRules);
      return;
    }
    rule = (RULE_PTR)listNext(relation->outRules);
  }
}

void tapActivation(ruleName, relationshipNode, activationFn)
char *ruleName;
TMS_NODE_PTR relationshipNode;
int (*activationFn)();
{
  DISPATCH_PTR dispatch;
  QUANTITY_PTR quantity;
  TCA_TIME_POINT_PTR timePoint;

  if (Less_Than_Now_Relp(relationshipNode) && 
      Relp_Still_Out(relationshipNode)) {
    /* Remove the tapActivation rule to avoid having it fire again */
    removeTapRule(activationFn, relationshipNode->cell);

    quantity = relation_other_arg(relationshipNode->cell, Now);
    timePoint = (TCA_TIME_POINT_PTR)(quantity->name);
    dispatch = DISPATCH_FROM_ID(timePoint->interval.msgRef);
    tapOnCondition(activationFn, dispatch);
  }
}

/****************************************************************************/

int sendWhenSentTap(dispatch, tap)
DISPATCH_PTR dispatch; 
TAP_PTR tap;
{ 
  return sendOnCondition(WhenSent, dispatch, tap); 
}

/****************************************************************************/

void tapWhenSent(dispatch)
DISPATCH_PTR dispatch;
{ 
  tapOnCondition(sendWhenSentTap, dispatch); 
}

/****************************************************************************/

int sendBeforeHandlingTap(dispatch, tap)
DISPATCH_PTR dispatch; 
TAP_PTR tap;
{ 
  return sendOnCondition(BeforeHandling, dispatch, tap); 
}

/****************************************************************************/

void tapBeforeHandling(dispatch)
DISPATCH_PTR dispatch;
{ 
  tapOnCondition(sendBeforeHandlingTap, dispatch); 
}

/****************************************************************************/

int sendWhileHandlingTap(dispatch, tap)
DISPATCH_PTR dispatch; 
TAP_PTR tap;
{ 
  return sendOnCondition(WhileHandling, dispatch, tap); 
}

/****************************************************************************/

void tapWhileHandling(dispatch)
DISPATCH_PTR dispatch;
{ 
  tapOnCondition(sendWhileHandlingTap, dispatch); 
}

/****************************************************************************/

int sendAfterHandledTap(dispatch, tap)
DISPATCH_PTR dispatch; 
TAP_PTR tap;
{ 
  return sendOnCondition(AfterHandled, dispatch, tap);
}

/****************************************************************************/

void tapAfterHandled(dispatch)
DISPATCH_PTR dispatch;
{ 
  tapOnCondition(sendAfterHandledTap, dispatch); 
}

/****************************************************************************/

int sendAfterSuccessTap(dispatch, tap)
DISPATCH_PTR dispatch; 
TAP_PTR tap;
{ 
  return sendOnCondition(AfterSuccess, dispatch, tap);
}

/****************************************************************************/

void tapAfterSuccess(dispatch)
DISPATCH_PTR dispatch;
{ 
  tapOnCondition(sendAfterSuccessTap, dispatch); 
}

/****************************************************************************/

int sendWhenSuccess(dispatch, tap)
DISPATCH_PTR dispatch; 
TAP_PTR tap;
{ 
  if (tap->condition == WhenSuccess) {
    sendListeningDispatchAndConstrain(dispatch, tap, 
				      &endAchievement, ">=", &endAchievement);
  }
  return TRUE;
}

/****************************************************************************/

void tapWhenSuccess(dispatch)
DISPATCH_PTR dispatch;
{ 
  tapOnCondition(sendWhenSuccess, dispatch); 
}
/****************************************************************************/

int sendAfterFailureTap(dispatch, tap)
DISPATCH_PTR dispatch; 
TAP_PTR tap;
{ 
;  return sendOnCondition(AfterFailure, dispatch, tap); 
}

/****************************************************************************/

void tapAfterFailure(dispatch)
DISPATCH_PTR dispatch;
{ 
  tapOnCondition(sendAfterFailureTap, dispatch); 
}
/****************************************************************************/

int sendWhenFailure(dispatch, tap)
DISPATCH_PTR dispatch; 
TAP_PTR tap;
{ 
  if (tap->condition == WhenFailure) {
    sendListeningDispatchAndConstrain(dispatch, tap, 
				      &endAchievement, ">=", &endAchievement);
  }
  return TRUE;
}

/****************************************************************************/

void tapWhenFailure(dispatch)
DISPATCH_PTR dispatch;
{ 
  tapOnCondition(sendWhenFailure, dispatch); 
}
/****************************************************************************/

int sendAfterRepliedTap(dispatch, tap)
DISPATCH_PTR dispatch; 
TAP_PTR tap;
{
  if (tap->condition == AfterReplied) {
    sendListeningDispatch(dispatch, DISPATCH_RES_DATA(dispatch), tap);
  }
  return TRUE;
}

/****************************************************************************/

void tapAfterReplied(dispatch)
DISPATCH_PTR dispatch;
{ 
  tapOnCondition(sendAfterRepliedTap, dispatch); 
}

/****************************************************************************/

int sendBeforeHandlingTapAndConstrain(dispatch, tap)
DISPATCH_PTR dispatch; 
TAP_PTR tap;
{
  if (tap->condition == BeforeHandling) {
    sendListeningDispatchAndConstrain(dispatch, tap,
				      &startHandling, ">=", &endHandling);
  }
  return TRUE;
}

/****************************************************************************/

/*****************************************************************
 * Return FALSE if the rule is added (to halt the list iteration)
 ****************************************************************/
int addBeforeHandlingRule(dispatch, tap)
DISPATCH_PTR dispatch; 
TAP_PTR tap;
{
  if (tap->condition == BeforeHandling) {
    addOutRule(beforeHandlingRule,
	       relationGetOrMake(Now, 
				 dispatch->treeNode->handlingInterval->start));
    return FALSE;
  } else {
    return TRUE;
  }
}

/****************************************************************************/

int sendBeforeAchievingTapAndConstrain(dispatch, tap)
DISPATCH_PTR dispatch; 
TAP_PTR tap;
{
  if (tap->condition == BeforeAchieving) {
    sendListeningDispatchAndConstrain(dispatch, tap, 
			      &startAchievement, ">=", &endAchievement);
  }
  return TRUE;
}

/****************************************************************************/

int sendWhileAchievingTap(dispatch, tap)
DISPATCH_PTR dispatch; 
TAP_PTR tap;
{ 
  return sendOnCondition(WhileAchieving, dispatch, tap); 
}

/****************************************************************************/

int sendWhenAchievedTapAndConstrain(dispatch, tap)
DISPATCH_PTR dispatch; 
TAP_PTR tap;
{
  if (tap->condition == WhenAchieved) {
    sendListeningDispatchAndConstrain(dispatch, tap, 
				      &endAchievement, ">=", &endAchievement);
  }
  return TRUE;
}

/****************************************************************************/

int sendAfterAchievedTap(dispatch, tap)
DISPATCH_PTR dispatch; 
TAP_PTR tap;
{ 
  return sendOnCondition(AfterAchieved, dispatch, tap); 
}

/****************************************************************************/

/*****************************************************************
 * Return FALSE if the rule is added (to halt the list iteration)
 ****************************************************************/

int addBeforeAchievingRule(dispatch, tap)
DISPATCH_PTR dispatch; 
TAP_PTR tap;
{
  QUANTITY_PTR startAchievementQty;

  if (tap->condition == BeforeAchieving) {
    startAchievementQty = dispatch->treeNode->achievementInterval->start;
    if (QIs_True(Now, "<", startAchievementQty)) {
      addOutRule(beforeAchievingRule, relationGetOrMake(Now, 
							startAchievementQty));
      return FALSE;
    } else {
      (void)sendBeforeAchievingTapAndConstrain(dispatch, tap);
    }
  }
  return TRUE;
}

/****************************************************************************/

int addWhileAchievingRule(dispatch, tap)
DISPATCH_PTR dispatch; 
TAP_PTR tap;
{
  QUANTITY_PTR startAchievementQty;

  if (tap->condition == WhileAchieving) {
    startAchievementQty = dispatch->treeNode->achievementInterval->start;
    if (QIs_True(Now, "<", startAchievementQty)) {
      addOutRule(whileAchievingRule, relationGetOrMake(Now, 
						       startAchievementQty));
      return FALSE;
    } else {
      (void)sendWhileAchievingTap(dispatch, tap);
    }
  }
  return TRUE;
}

/****************************************************************************/

int addAfterAchievedRule(dispatch, tap)
DISPATCH_PTR dispatch; 
TAP_PTR tap;
{
  QUANTITY_PTR endAchievementQty;

  if (tap->condition == AfterAchieved) {
    endAchievementQty = dispatch->treeNode->achievementInterval->end;
    if (QIs_True(Now, "<", endAchievementQty)) {
      addOutRule(afterAchievedRule, relationGetOrMake(Now, endAchievementQty));
      return FALSE;
    } else {
      (void)sendAfterAchievedTap(dispatch, tap);
    }
  }
  return TRUE;
}

/****************************************************************************/

int addWhenAchievedRule(dispatch, tap)
DISPATCH_PTR dispatch; 
TAP_PTR tap;
{
  QUANTITY_PTR endAchievementQty;

  if (tap->condition == WhenAchieved) {
    endAchievementQty = dispatch->treeNode->achievementInterval->end;
    if (QIs_True(Now, "<", endAchievementQty)) {
      addOutRule(whenAchievedRule, relationGetOrMake(Now, endAchievementQty));
      return FALSE;
    } else {
      (void)sendWhenAchievedTapAndConstrain(dispatch, tap);
    }
  }
  return TRUE;
}

/****************************************************************************/

int sendBeforePlanningTapAndConstrain(dispatch, tap)
DISPATCH_PTR dispatch; 
TAP_PTR tap;
{
  if (tap->condition == BeforePlanning) {
    sendListeningDispatchAndConstrain(dispatch, tap,
				      &startPlanning, ">=", &endPlanning);
  }
  return TRUE;
}

/****************************************************************************/

int sendWhilePlanningTap(dispatch, tap)
DISPATCH_PTR dispatch; 
TAP_PTR tap;
{ 
  return sendOnCondition(WhilePlanning, dispatch, tap); 
}

/****************************************************************************/

int sendWhenPlannedTapAndConstrain(dispatch, tap)
DISPATCH_PTR dispatch; 
TAP_PTR tap;
{
  if (tap->condition == WhenPlanned) {
    sendListeningDispatchAndConstrain(dispatch, tap, 
				      &endPlanning, ">=", &endPlanning);
  }
  return TRUE;
}

/****************************************************************************/

int sendAfterPlannedTap(dispatch, tap)
DISPATCH_PTR dispatch; 
TAP_PTR tap;
{ 
  return sendOnCondition(AfterPlanned, dispatch, tap); 
}

/****************************************************************************/

/*****************************************************************
 * Return FALSE if the rule is added (to halt the list iteration)
 ****************************************************************/

int addBeforePlanningRule(dispatch, tap)
DISPATCH_PTR dispatch; 
TAP_PTR tap;
{
  QUANTITY_PTR startPlanningQty;

  if (tap->condition == BeforePlanning) {
    startPlanningQty = dispatch->treeNode->planningInterval->start;
    if (QIs_True(Now, "<", startPlanningQty)) {
      addOutRule(beforePlanningRule, relationGetOrMake(Now, startPlanningQty));
      return FALSE;
    } else {
      (void)sendBeforePlanningTapAndConstrain(dispatch, tap);
    }
  }
  return TRUE;
}

/****************************************************************************/

int addWhilePlanningRule(dispatch, tap)
DISPATCH_PTR dispatch; 
TAP_PTR tap;
{
  QUANTITY_PTR startPlanningQty;

  if (tap->condition == WhilePlanning) {
    startPlanningQty = dispatch->treeNode->planningInterval->start;
    if (QIs_True(Now, "<", startPlanningQty)) {
      addOutRule(whilePlanningRule, relationGetOrMake(Now, startPlanningQty));
      return FALSE;
    } else {
      (void)sendWhilePlanningTap(dispatch, tap);
    }
  }
  return TRUE;
}

/****************************************************************************/

int addAfterPlannedRule(dispatch, tap)
DISPATCH_PTR dispatch; 
TAP_PTR tap;
{
  QUANTITY_PTR endPlanningQty;

  if (tap->condition == AfterPlanned) {
    endPlanningQty = dispatch->treeNode->planningInterval->end;
    if (QIs_True(Now, "<", endPlanningQty)) {
      addOutRule(afterPlannedRule, relationGetOrMake(Now, endPlanningQty));
      return FALSE;
    } else {
      (void)sendAfterPlannedTap(dispatch, tap);
    }
  }
  return TRUE;
}

/****************************************************************************/

int addWhenPlannedRule(dispatch, tap)
DISPATCH_PTR dispatch; 
TAP_PTR tap;
{
  QUANTITY_PTR endPlanningQty;

  if (tap->condition == WhenPlanned) {
    endPlanningQty = dispatch->treeNode->planningInterval->end;
    if (QIs_True(Now, "<", endPlanningQty)) {
      addOutRule(whenPlannedRule, relationGetOrMake(Now, endPlanningQty));
      return FALSE;
    } else {
      (void)sendWhenPlannedTapAndConstrain(dispatch, tap);
    }
  }
  return TRUE;
}

/****************************************************************************/

/*****
 * The dispatch should be a goal, command, or monitor 
 *****/

void setUpTapRule(addRuleFn, dispatch)
int (*addRuleFn)(); 
DISPATCH_PTR dispatch;
{
  /* listIterate returns TRUE if all the elements 
     of the list passed the test */
  if (listIterateFromFirst(addRuleFn, (char *)dispatch, DISPATCH_TAPS(dispatch))) {
    (void)listIterateFromFirst(addRuleFn, (char *)dispatch, 
			       DISPATCH_EXTRA_TAPS(dispatch));
  }
}

/****************************************************************************/

void setUpBeforeTaps(dispatch)
DISPATCH_PTR dispatch;
{
  setUpTapRule(addBeforeHandlingRule, dispatch);
  setUpTapRule(addBeforeAchievingRule, dispatch);

  if (dispatch->class == GoalClass) 
    setUpTapRule(addBeforePlanningRule, dispatch);
}

/****************************************************************************/

void setUpDuringAndAfterTaps(dispatch)
DISPATCH_PTR dispatch;
{
  setUpTapRule(addWhileAchievingRule, dispatch);
  setUpTapRule(addAfterAchievedRule, dispatch);
  setUpTapRule(addWhenAchievedRule, dispatch);

  if (dispatch->class == GoalClass) {
    setUpTapRule(addWhilePlanningRule, dispatch);
    setUpTapRule(addAfterPlannedRule, dispatch);
    setUpTapRule(addWhenPlannedRule, dispatch);
  }
}

/****************************************************************************/

void setUpWhenKilledTaps(dispatch)
DISPATCH_PTR dispatch;
{
  if (DISPATCH_TAPS(dispatch) || DISPATCH_EXTRA_TAPS(dispatch)) {
    Log_Message("setUpWhenKilledTaps not yet implemented\n");
  }
}

/****************************************************************************/
/* Currently not used (RGS: 11/11/92)
void removeWiretapRules (treeNode)
TASK_TREE_NODE_PTR treeNode;
{
  CELL_PTR relation;

  relation = get_relation(Now, treeNode->handlingInterval->start);
  if (relation) {
    removeTapRule(sendBeforeHandlingTapAndConstrain, relation);
  }

  relation = get_relation(Now, treeNode->achievementInterval->start);
  if (relation) {
    removeTapRule(sendBeforeAchievingTapAndConstrain, relation);
    removeTapRule(sendWhileAchievingTap, relation);
  }

  relation = get_relation(Now, treeNode->achievementInterval->end);
  if (relation) {
    removeTapRule(sendAfterAchievedTap, relation);
    removeTapRule(sendWhenAchievedTapAndConstrain, relation);
  }

  relation = get_relation(Now, treeNode->planningInterval->start);
  if (relation) {
    removeTapRule(sendBeforePlanningTapAndConstrain, relation);
    removeTapRule(sendWhilePlanningTap, relation);
  }

  relation = get_relation(Now, treeNode->planningInterval->end);
  if (relation) {
    removeTapRule(sendAfterPlannedTap, relation);
    removeTapRule(sendWhenPlannedTapAndConstrain, relation);
  }
}
*/

/******************************************************************************
*
* FUNCTION: void tapInitialize()
* 
* DESCRIPTION: Register all the user-interface functions for message tapping
*
* INPUTS: none
*
* OUTPUTS: none
*
******************************************************************************/

void tapInitialize()
{
  centralRegisterMessage("tapMessage", InformClass,
			 TAP_MSG_INFORM_FORM, (char *)NULL);
  centralRegisterHandler("tapMessage", "tapMessageHandler", tapMessageHandler);

  centralRegisterMessage("tapReference", InformClass,
			 TAP_REF_INFORM_FORM, (char *)NULL);
  centralRegisterHandler("tapReference", "tapReferenceHandler",
			 tapReferenceHandler);

  centralRegisterMessage("removeTap", InformClass, 
			 TAP_MSG_INFORM_FORM, (char *)NULL);
  centralRegisterHandler("removeTap", "removeTapHandler", removeTapHandler);

  centralRegisterMessage("findTappedReference", QueryClass, 
			 "int", "TCA_REF_PTR");
  centralRegisterHandler("findTappedReference", "findTappedReferenceHandler",
			 findTappedReferenceHandler);

  centralRegisterMessage("cleanUpAfterAchieved", InformClass,
			 (char *)NULL, (char *)NULL);
  centralRegisterHandler("cleanUpAfterAchieved", "cleanUpHandler", cleanUpHandler);

  Add_Message_To_Ignore("tapMessage");
  Add_Message_To_Ignore("tapReference");
  Add_Message_To_Ignore("removeTap");
  Add_Message_To_Ignore("findTappedReference");

  startHandling.class = StartPoint; 
  startHandling.interval.class = HandlingInterval;
  endHandling.class = EndPoint; 
  endHandling.interval.class = HandlingInterval;

  startAchievement.class = StartPoint; 
  startAchievement.interval.class = AchievementInterval;
  endAchievement.class = EndPoint; 
  endAchievement.interval.class = AchievementInterval;

  startPlanning.class = StartPoint; 
  startPlanning.interval.class = PlanningInterval;
  endPlanning.class = EndPoint; 
  endPlanning.interval.class = PlanningInterval;

  beforeHandlingRule = ruleCreateOut("Before Handling", tapActivation, 2,
			     (char *)sendBeforeHandlingTapAndConstrain);
  beforeHandlingRule->used = 1000; /* Never free, Reid: 10-Jul-91 */

  beforeAchievingRule = ruleCreateOut("Before Achieving", tapActivation, 2,
				 (char *)sendBeforeAchievingTapAndConstrain);
  beforeAchievingRule->used = 1000; /* Never free, Reid: 10-Jul-91 */


  whileAchievingRule = ruleCreateOut("While Achieving", tapActivation, 4,
				     (char *)sendWhileAchievingTap);
  whileAchievingRule->used = 1000; /* Never free, Reid: 10-Jul-91 */

  afterAchievedRule = ruleCreateOut("After Achieved", tapActivation, 4,
				    (char *)sendAfterAchievedTap);
  afterAchievedRule->used = 1000; /* Never free, Reid: 10-Jul-91 */
 
  whenAchievedRule = ruleCreateOut("When Achieved", tapActivation, 1,
				   (char *)sendWhenAchievedTapAndConstrain);
  whenAchievedRule->used = 1000; /* Never free, Reid: 10-Jul-91 */

  beforePlanningRule = ruleCreateOut("Before Planning", tapActivation, 2,
			     (char *)sendBeforePlanningTapAndConstrain);
  beforePlanningRule->used = 1000; /* Never free, Reid: 10-Jul-91 */

  whilePlanningRule = ruleCreateOut("While Planning", tapActivation, 4,
				    (char *)sendWhilePlanningTap);
  whilePlanningRule->used = 1000; /* Never free, Reid: 10-Jul-91 */

  afterPlannedRule = ruleCreateOut("After Planned", tapActivation, 4,
				   (char *)sendAfterPlannedTap);
  afterPlannedRule->used = 1000; /* Never free, Reid: 10-Jul-91 */

  whenPlannedRule = ruleCreateOut("When Planned", tapActivation, 1,
				  (char *)sendWhenPlannedTapAndConstrain);
  whenPlannedRule->used = 1000; /* Never free, Reid: 10-Jul-91 */

  /* Used within "cleanUpHandler" */ 
  alwaysTrueNodeGlobal = tmsCreateNode("Always True");
  tmsAssertNode(alwaysTrueNodeGlobal, "Always True");
}
