/******************************************************************************
*
* PROJECT: Carnegie Mellon Planetary Rover Project
*          Task Control Architecture 
*
* (c) Copyright 1991 Christopher Fedor and Reid Simmons.  All rights reserved.
*
* MODULE: behaviors
*
* FILE: behaviors.c
*
* ABSTRACT:
* 
* tca behavior level.
*
* REVISION HISTORY
*
*  9-Oct-92 Reid Simmons, School of Computer Science, CMU
* Replaced "tcaAddConstraint" by "tcaInform" 
*
*  2-Jul-91 Reid Simmons, School of Computer Science, CMU
* Hacked memory management -- closed some holes.
*
* 27-May-91 Christopher Fedor, School of Computer Science, CMU
* Changed msgFind to pass back hndName to cache msgInfo from the msgInfoHnd
* instead of from deliverDispatch to avoid timing errors.
*
* 30-Jan-91 Christopher Fedor, School of Computer Science, CMU
* Added fflush(stdout) to printf for module code calls from lisp
*
*  3-Oct-90 Christopher Fedor, School of Computer Science, CMU
* Removed logging and changed call from tcaError to tcaModError.
*
*  3-Oct-90 Christopher Fedor, School of Computer Science, CMU
* Moved module monitor code from behaviors.c to monModule.c
*
*  3-Sep-90 Christopher Fedor, School of Computer Science, CMU
* Replaced msgQuery and FindMessage with msgFind.
* Moved msg.c comments to behaviors and removed msg.c
* Moved msgQeury and FindMessage to behaviors.c. Updated message
* lookup to handle HandlerRegisteredClass and multiple messages for
* a single handler.
*
* 16-Aug-90 Christopher Fedor, School of Computer Science, CMU
* moved Message_Class_Name from com.c to behaviors.c
*
*  7-Apr-90 Christopher Fedor, School of Computer Science, CMU
* Revised to Software Standards.
*
*  1-Dec-89 Long-Ji Lin, School of Computer Science, CMU
* Made variable "responseIssuedGlobal" global.
*
* 27-Nov-89 Reid Simmons, School of Computer Science, CMU
* Added non-blocking query pairs -- tcaQuerySend and tcaQueryReceive.
*
* 17-aug-89 Reid Simmons, School of Computer Science, CMU
* Moved error message from FindMessage to MsgQuery, so that MsgQuery can
* be called to check whether a message has been registered.
* Added exception handlers.
*
* 14-Aug-89 Reid Simmons, School of Computer Science, CMU
* Moved "CreateReference" to tasktree.c
*
*  2-Aug-89 Reid Simmons, School of Computer Science, CMU
* Added checks to ensure that routines are called with the 
* correct message class.
*
* 28-Jun-89 Reid Simmons, School of Computer Science, CMU
* Added check to see if response (Reply, Success, Failure) is sent by handlers.
*
*  9-May-89 Christopher Fedor, School of Computer Science, CMU
* Added change to SelfRegisterMsg Calls. Probably broke modularization.
*
*  3-May-89 Christopher Fedor, School of Computer Science, CMU
* Fixed Failure to send the descriptive string.
* Changed Failure to be called at top level with NULL as a message ref.
* Changed route_module.c to behaviors.c
*
*  1-May-89 Reid Simmons, School of Computer Science, CMU
* Added code for sending temporal constraints with goal messages.  
* Re-modularized.
*
* 28-Apr-89 Reid Simmons, School of Computer Science, CMU
* Added code for sending the parent reference (dispatch) with messages.
*
*  7-Mar-89 Christopher Fedor, School of Computer Science, CMU
* Moved user level routines to route_module.c Created ClassQuery, MsgQuery.
*
*  6-Mar-89 Christopher Fedor, School of Computer Science, CMU
* created behaviors.
* 
*  1-Dec-88 Christopher Fedor, School of Computer Science, CMU
* created message cache.
*
******************************************************************************/

#include "stdio.h"

#include "tcaInternal.h"
#include "formatters.h"
#include "datamsg.h"
#include "hash.h"
#include "idtable.h"
#include "dispatch.h"

#include "sendMsg.h"

/* Route Tables */
extern ID_TABLE_PTR msgIdTable;
extern ID_TABLE_PTR hndIdTable;

extern HASH_TABLE_PTR messageTable, classFormatTable;

extern int tcaServerGlobal;

extern TCA_RETURN_VALUE_TYPE waitForReply();  /* defined in comModule.c */
extern TCA_RETURN_VALUE_TYPE waitForReply3(); /* defined in comModule.c */

extern MSG_PTR msgCreate();

extern void establishDirect();

extern void classDataFree();

#ifdef LISP
TCA_REF_PTR lispRefSaveGlobal = NULL;
#endif

/*****************************************************/

/* Id of the dispatch currently being handled */
int parentRefGlobal = -1; 

/* TRUE only if handler issues a Succes/Failure/Reply message */
int responseIssuedGlobal = FALSE;

int inconsistentConstraintsGlobal[2] = {PLAN_FIRST+DELAY_PLANNING, 0};

int (*lispBufferSizeGlobal)();
int (*lispEncodeMsgGlobal)();
int (*lispDecodeMsgGlobal)();
int (*lispReplyMsgGlobal)();
int (*lispExitGlobal)();

void tcaSetSizeEncodeDecode(bufferSize, encode, decode, lispReply, lispExit)
int (*bufferSize)(), (*encode)(), (*decode)(), (*lispReply)(), (*lispExit)();
{
  lispBufferSizeGlobal = bufferSize;
  lispEncodeMsgGlobal = encode;
  lispDecodeMsgGlobal = decode;
  lispReplyMsgGlobal = lispReply;
  lispExitGlobal = lispExit;
}


/******************************************************************************
*
* FUNCTION: MSG_PTR msgHndFind(name, hndName)
*           MSG_PTR msgFind(name)
*
* DESCRIPTION: 
* Performs a message lookup for a message of name. If the message is not
* found the central server is queried and the information is cached.
* If the message is still not found or the information returned from the
* central server is not complete an error is issued. 
*
* INPUTS: 
* char *name, *hndName;
*
* OUTPUTS: MSG_PTR
*
******************************************************************************/

MSG_PTR msgHndFind(name, hndName)
char *name, *hndName;
{
  MSG_PTR msg;
  MSG_DATA_PTR msgData;

  MSG_ASK_TYPE msgAsk;

  msg = (MSG_PTR)hashTableFind(name, messageTable);

  if (!msg || msg->msgData->class == HandlerRegClass) {
    if (tcaServerGlobal != CENTRAL_SERVER_ID) {
      msgAsk.msgName = name;
      msgAsk.hndName = hndName;

      msgData = NEW(MSG_DATA_TYPE);

      (void)tcaQuery("msgInfoMsg", (void *)&msgAsk, (void *)&msgData);
      if (!msgData)
	TCA_MOD_ERROR1("ERROR: msgFind: Query for message %s returned NULL.\n",
		       name);

      if (msgData->class != HandlerRegClass) {
	msg = msgCreate(msgData);
	if (!hndName && msgData->refId == -1)
	  establishDirect(msg);
	return msg;
      }
    }
    TCA_MOD_ERROR1("ERROR: msgFind: Message %s used before registered.\n", 
		   name);
  }

  return msg;
}

MSG_PTR msgFind(name)
char *name;
{
  return msgHndFind(name, (char *)NULL);
}

MSG_PTR msgFind2(name, hndName)
char *name, *hndName;
{
  MSG_PTR msg;
  MSG_DATA_PTR msgData;

  MSG_ASK_TYPE msgAsk;

  msg = (MSG_PTR)hashTableFind(name, messageTable);

  if (!msg || msg->msgData->class == HandlerRegClass) {
      msgAsk.msgName = name;
      msgAsk.hndName = hndName;

      (void)tcaQuery("msgInfoMsg", (void *)&msgAsk, (void *)&msgData);
      if (!msgData)
	return NULL;

      if (msgData->class != HandlerRegClass) {
	msg = msgCreate(msgData);
	if (!hndName && msgData->refId == -1)
	  establishDirect(msg);
	return msg;
      }
      return NULL;
    }
  return msg;
}
  


/******************************************************************************
*
* FUNCTION: char *messageClassName(class)
*
* DESCRIPTION:
* Return the name of the class (a string). Used for display purposes
*
* INPUTS: TCA_MSG_CLASS_TYPE class;
*
* OUTPUTS: char *
*
******************************************************************************/

char *messageClassName(class)
TCA_MSG_CLASS_TYPE class;
{ 
  switch(class) {
  case QueryClass: 
    return("Query");
  case GoalClass: 
    return("Goal");
  case CommandClass: 
    return("Command");
  case PointMonitorClass: 
    return("PMonitor");
  case DemonMonitorClass: 
  case PollingMonitorClass: 
    return("IMonitor");
  case InformClass: 
    return("Inform");
  case ExceptionClass: 
    return("Exception");
  case BroadcastClass:
    return("Broadcast");
  case MultiQueryClass:
    return("MQuery");
  default: 
    return ("UNKNOWN CLASS");
  }
}


/******************************************************************************
*
* FUNCTION: void checkMessageClass(msg, class)
*
* DESCRIPTION:
* Check to make sure that functions, such as Query, are called
* only with messages of that class.
*
* INPUTS:
* MSG_PTR msg;
* TCA_MSG_CLASS_TYPE class;
*
* OUTPUTS: void. Message is logged.
*
******************************************************************************/

void checkMessageClass(msg, class)
MSG_PTR msg;
TCA_MSG_CLASS_TYPE class;
{ 
  char *name, *use;

  if (msg->msgData->class != class) {
    name = messageClassName(class);

    switch (msg->msgData->class) {
    case QueryClass: 
      use = "tcaQuery"; 
      break;
    case GoalClass:
      use = "tcaExpandGoal"; 
      break;
    case CommandClass:
      use = "tcaExecuteCommand";
      break;
    case PointMonitorClass:
      use = "tcaPointMonitor";
      break;
    case InformClass: 
      use = "tcaInform"; 
      break;
    case BroadcastClass: 
      use = "tcaBroadcast"; 
      break;
    case MultiQueryClass: 
      use = "tcaMultiQuery"; 
      break;
    }

    TCA_MOD_ERROR3("%s is not a %s class message.  Use function '%s' instead\n",
		   msg->msgData->name, name, use);
  }
}


/******************************************************************************
*
* FUNCTION: void execHnd(directInfo, dataMsg)
*
* DESCRIPTION:
*
* INPUTS:
* DIRECT_INFO_PTR directInfo;
* DATA_MSG_PTR dataMsg;
*
* OUTPUTS: void.
*
******************************************************************************/

void execHnd(directInfo, dataMsg)
DIRECT_INFO_PTR directInfo;
DATA_MSG_PTR dataMsg;
{ 
  MSG_PTR msg;
  HND_PTR hnd;
  char *data, *name, **pointerToName;
  TCA_REF_PTR tcaRef;
  TCA_MSG_CLASS_TYPE class;
  CLASS_FORM_PTR classForm;
  int tmpParentRef, tmpResponse;

#ifdef LISP
  BUFFER_TYPE buffer;
#endif

  hnd = (HND_PTR)idTableItem(dataMsg->intent, hndIdTable);

  if (dataMsg->classTotal) {
    class = ExecHndClass;
    classForm = (CLASS_FORM_PTR)hashTableFind((char *)&class, classFormatTable);
    if (!classForm)
      tcaModError("ERROR: execHnd: missing ExecHndClass class format.");

    pointerToName = (char **)dataMsgDecodeClass(classForm->format, dataMsg);
    
    name = (char *)(*pointerToName);

    /* printf("execHnd: name: %s\n", name); */
#ifdef LISP
    /*     fflush(stdout); */
#endif

    msg = msgHndFind(name, hnd->hndData->hndName);
    classDataFree(class, (char *)pointerToName);
    hnd->msg = msg;
  }
  else
    msg = hnd->msg;

  if (!msg) {
    msg = msgHndFind(hnd->hndData->msgName, hnd->hndData->hndName);
    hnd->msg = msg;
    msg->direct = TRUE;
    msg->sd = directInfo->sd;
    if (msg->msgData->refId != -1)
      tcaModError("ERROR: execHnd: Sanity Check Failed!");
  }

  tcaRef = tcaRefCreate(msg, msg->msgData->name, dataMsg->msgRef);

#ifndef LISP

  data = dataMsgDecodeMsg(msg->msgData->msgForm, dataMsg);

#else
  buffer.bstart = 0;
  buffer.buffer = NULL;

  if (msg->msgData->msgForm && dataMsg->msgTotal) {
    buffer.buffer = dataMsg->msgData;
    data = (char *)(*lispDecodeMsgGlobal)(msg->msgData->msgForm, &buffer);
  }
  else
    data = (char *)(*lispDecodeMsgGlobal)(0, NULL);

#endif

  dataMsgFree(dataMsg);

  tmpParentRef = parentRefGlobal;
  tmpResponse = responseIssuedGlobal;
  parentRefGlobal = tcaRef->refId;
  responseIssuedGlobal = FALSE;

  (*hnd->hndProc)(tcaRef, data);

#ifndef VXWORKS

  if (!responseIssuedGlobal && RESPONSE_EXPECTED(msg)) {
    fprintf(stderr, "WARNING: No ");
#ifdef LISP
    fflush(stderr);
#endif
    if (TWO_WAY_MSG(msg)) {
      fprintf(stderr, "Reply");
#ifdef LISP
      fflush(stderr);
#endif
    }
    else {
      fprintf(stderr, "Success/Failure");
#ifdef LISP
      fflush(stderr);
#endif
    }
    fprintf(stderr, " sent for message %s;\nResource will probably remain locked\n",
	    msg->msgData->name);
#ifdef LISP
    fflush(stderr);
#endif
  }

#endif

  /* Send a "success" reponse to constraint messages -- new for version 7.
     Gets rid of race conditions that existed in old version (reids) */
  if (ONE_WAY_MSG(msg)) {
    (void)sendResponse(tcaRef, (MSG_PTR)NULL, (char *)NULL, 
		       SuccessClass, (char *)NULL);
  }

  tcaRefFree(tcaRef);

  parentRefGlobal = tmpParentRef;
  responseIssuedGlobal = tmpResponse;
}


/******************************************************************************
*
* FUNCTION: TCA_RETURN_VALUE_TYPE tcaReply(dispatch, data)
*
* DESCRIPTION:
*
* INPUTS:
* DISPATCH_PTR dispatch;
* char *data;
*
* OUTPUTS: TCA_RETURN_VALUE_TYPE
*
******************************************************************************/

TCA_RETURN_VALUE_TYPE tcaReply(ref, data)
TCA_REF_PTR ref;
char *data;
{
  TCA_RETURN_VALUE_TYPE status;

  responseIssuedGlobal = TRUE;
  status = sendResponse(ref, (MSG_PTR)NULL, data, ReplyClass, (char *)NULL);
  
  return status;
}


/******************************************************************************
*
* FUNCTION: TCA_RETURN_VALUE_TYPE tcaNullReply(ref)
*
* DESCRIPTION:
*
* INPUTS: TCA_REF_PTR ref;
*
* OUTPUTS: TCA_RETURN_VALUE_TYPE
*
******************************************************************************/

TCA_RETURN_VALUE_TYPE tcaNullReply(ref)
TCA_REF_PTR ref;
{
  return tcaReply(ref, (char *)NULL);
}


/******************************************************************************
*
* FUNCTION: TCA_RETURN_VALUE_TYPE tcaSuccess(ref)
*
* DESCRIPTION:
*
* INPUTS: TCA_REF_PTR ref;
*
* OUTPUTS: TCA_RETURN_VALUE_TYPE
*
******************************************************************************/

TCA_RETURN_VALUE_TYPE tcaSuccess(ref)
TCA_REF_PTR ref;
{
  TCA_RETURN_VALUE_TYPE status;

  responseIssuedGlobal = TRUE;

  status = sendResponse(ref, (MSG_PTR)NULL, (char *)NULL, 
			SuccessClass, (char *)NULL);
  return status;
}


/******************************************************************************
*
* FUNCTION: TCA_RETURN_VALUE_TYPE tcaFailure(ref, description, data)
*
* DESCRIPTION:
*
* INPUTS:
* TCA_REF_PTR ref;
* char *description;
* void *data;
*
* OUTPUTS: TCA_RETURN_VALUE_TYPE
*
******************************************************************************/

TCA_RETURN_VALUE_TYPE tcaFailure(ref, description, data)
TCA_REF_PTR ref;
char *description;
void *data;
{
  MSG_PTR msg;
  int tplConstr;
  TCA_RETURN_VALUE_TYPE status;

  responseIssuedGlobal = TRUE;

  /* 12-May-91: fedor: this double sending of messages is silly */
  msg = msgFind2(description, (char *)NULL);
  if (msg) {
    if (msg->msgData->class != ExceptionClass)
      TCA_MOD_ERROR1("%s is not an exception class message for tcaFailure.\n",
		     msg->msgData->name);

    tplConstr = NO_TPLCONSTR;
    (void)sendMessage((TCA_REF_PTR)NULL, msg, (char *)data, (char *)&tplConstr);
  }

  status = sendResponse(ref, (MSG_PTR)NULL, (char *)NULL,
			FailureClass, (char *)&description);
  return status;
}


/******************************************************************************
*
* FUNCTION: TCA_RETURN_VALUE_TYPE tcaQuerySend(name, query, ref)
*
* DESCRIPTION:
* Send a non-blocking query, which doesn't wait for a reply.
* Must be paired with a corresponding tcaQueryReceive, with the same ref.
*
* INPUTS:
* char *name;
* void *query;
* TCA_REF_PTR *ref;
*
* OUTPUTS: TCA_RETURN_VALUE_TYPE
*
******************************************************************************/

TCA_RETURN_VALUE_TYPE tcaQuerySend(name, query, ref)
char *name;
void *query;
TCA_REF_PTR *ref;
{
  int refId;
  MSG_PTR msg;
  TCA_RETURN_VALUE_TYPE returnValue;

  msg = msgFind(name);
  checkMessageClass(msg, QueryClass);

  refId = nextSendMessageRef();
  returnValue = sendMessage((TCA_REF_PTR)NULL, msg, (char *)query, (char *)NULL);

  *ref = tcaRefCreate(msg, name, refId);
#ifdef LISP
  lispRefSaveGlobal = *ref;
#endif

  return returnValue;
}

/*****************************************************************
 * Return to LISP the reference created for a tcaQuerySend or tcaMultiQuery.
 * There may be a cleaner way to do this, but this seems the easiest
 * and quickest to implement (reids 11/13/92)
 ****************************************************************/
#ifdef LISP
TCA_REF_PTR tcaQuerySendRefLISPC ()
{
  return lispRefSaveGlobal;
}
#endif

/******************************************************************************
*
* FUNCTION: TCA_RETURN_VALUE_TYPE tcaQueryReceive(ref, reply)
*
* DESCRIPTION:
* ref is a TCA_REF_PTR created by tcaQuerySend.
* reply is a pointer to (already allocated) data.
*
* Waits for a reply to a tcaQuerySend.  The ref passed must be the same as the
* one gotten from the tcaQuerySend function, but the send's and receive's do
* not have to be in the same order. That is, you can send query1, send query2, 
* receive query2, receive query1.  TCA makes sure the right replies get paired
* with the right receives.
*
* INPUTS:
* TCA_REF_PTR ref;
* void *reply;
*
* OUTPUTS: TCA_RETURN_VALUE_TYPE
*
******************************************************************************/

TCA_RETURN_VALUE_TYPE tcaQueryReceive(ref, reply)
TCA_REF_PTR ref;
void *reply;
{
  TCA_RETURN_VALUE_TYPE returnValue;

  returnValue = waitForReply(ref, (char *)reply);

  tcaRefFree(ref);

  return returnValue;
}


/******************************************************************************
*
* FUNCTION: TCA_RETURN_VALUE_TYPE tcaQuery(name, query, reply)
*
* DESCRIPTION:
*
* INPUTS: char *name;
*         void *query, *reply;
*
* OUTPUTS: TCA_RETURN_VALUE_TYPE
*
* NOTES
* 3-May-89: fedor:
* ReturnValue always returns Success from sendMessage - otherwise
* sendMessage will exit. A hint of continued error handling problems.
*
******************************************************************************/

TCA_RETURN_VALUE_TYPE tcaQuery(name, queryData, replyData)
char *name;
void *queryData, *replyData;
{
  int refId;
  MSG_PTR msg;
  TCA_REF_PTR ref;
  TCA_RETURN_VALUE_TYPE returnValue;

  msg = msgFind(name);
  checkMessageClass(msg, QueryClass);

  refId = nextSendMessageRef();
  returnValue = sendMessage((TCA_REF_PTR)NULL, msg, 
			    (char *)queryData, (char *)NULL);

  ref = tcaRefCreate(msg, name, refId);

  if (returnValue == Success) 
    returnValue = waitForReply(ref, (char *)replyData);

  tcaRefFree(ref);

  return returnValue;
}


/******************************************************************************
*
* FUNCTION: int testInconsistent(tplConstraints)
*
* DESCRIPTION: Check if all bits of the inconsistent constraint are set.
*
* INPUTS: int tplConstraints;
*
* OUTPUTS: int
*
* NOTES:
* Should this be moved to tplConstr?
*
******************************************************************************/

int testInconsistent(tplConstraints)
int tplConstraints;
{
  int index;

  for (index=0; inconsistentConstraintsGlobal[index] != 0; index++)
    if ((inconsistentConstraintsGlobal[index] & tplConstraints) ==
	 inconsistentConstraintsGlobal[index]) {
      return TRUE;
    }
  return FALSE;
}


/******************************************************************************
*
* FUNCTION: TCA_RETURN_VALUE_TYPE tcaExpandGoalWithConstraints(ref, name, 
*                                                              data, tplConstr)
*
* DESCRIPTION:
*
* INPUTS:
* TCA_REF_PTR ref;
* char *name;
* void *data; 
* int tplConstr;
*
* OUTPUTS: TCA_RETURN_VALUE_TYPE
*
******************************************************************************/

TCA_RETURN_VALUE_TYPE tcaExpandGoalWithConstraints(ref, name, data, tplConstr)
TCA_REF_PTR ref;
char *name;
void *data; 
int tplConstr;
{ 
  MSG_PTR msg;

  if (testInconsistent(tplConstr))
    TCA_MOD_ERROR1("Inconsistent temporal constraints for goal %s\n", name);

  msg = msgFind(name);
  checkMessageClass(msg, GoalClass);
  return sendMessage(ref, msg, (char *)data, (char *)&tplConstr);
}


/******************************************************************************
*
* FUNCTION: TCA_RETURN_VALUE_TYPE tcaExpandGoal(name, data)
*
* DESCRIPTION:
*
* INPUTS: char *name, *data;
*
* OUTPUTS: TCA_RETURN_VALUE_TYPE
*
******************************************************************************/

TCA_RETURN_VALUE_TYPE tcaExpandGoal(name, data)
char *name, *data;
{ 
  MSG_PTR msg;
  int tplConstr;

  msg = msgFind(name);
  checkMessageClass(msg, GoalClass);

  tplConstr = SEQ_ACH;

  return sendMessage((TCA_REF_PTR)NULL, msg, (char *)data, (char *)&tplConstr);
}


/******************************************************************************
*
* FUNCTION: TCA_RETURN_VALUE_TYPE tcaInform(name, data)
*
* DESCRIPTION:
*
* INPUTS: char *name;
*         void *data;
*
* OUTPUTS: TCA_RETURN_VALUE_TYPE
*
******************************************************************************/

TCA_RETURN_VALUE_TYPE tcaInform(name, data)
char *name;
void *data;
{ 
  MSG_PTR msg;

  msg = msgFind(name);
  checkMessageClass(msg, InformClass);

  return sendMessage((TCA_REF_PTR)NULL, msg, (char *)data, (char *)NULL);
}

/******************************************************************************
*
* FUNCTION: TCA_RETURN_VALUE_TYPE tcaAddConstraint(name, data)
*
* DESCRIPTION: THIS FUNCTION SHOULD EVENTUALLY BE REPLACED, AND "tcaInform"
*              USED EXCLUSIVELY INSTEAD.
*
* INPUTS: char *name;
*         void *data;
*
* OUTPUTS: TCA_RETURN_VALUE_TYPE
*
******************************************************************************/

TCA_RETURN_VALUE_TYPE tcaAddConstraint(name, data)
char *name;
void *data;
{ 
  return tcaInform(name, data);
}


/******************************************************************************
*
* FUNCTION: TCA_RETURN_VALUE_TYPE tcaExecuteCommand(name, data)
*
* DESCRIPTION:
*
* INPUTS: char *name;
*         void *data;
*
* OUTPUTS: TCA_RETURN_VALUE_TYPE
*
******************************************************************************/

TCA_RETURN_VALUE_TYPE tcaExecuteCommand(name, data)
char *name;
void *data;
{ 
  MSG_PTR msg;
  BLOCK_COM_TYPE blockCom;

  msg = msgFind(name);
  checkMessageClass(msg, CommandClass);

  blockCom.waitFlag = 0;
  blockCom.tplConstr = SEQ_ACH;

  return sendMessage((TCA_REF_PTR)NULL, msg, (char *)data, (char *)&blockCom);
}


/******************************************************************************
*
* FUNCTION: TCA_RETURN_VALUE_TYPE tcaExecuteCommandWithConstraints(ref, name, 
*                                                              data, tplConstr)
*
* DESCRIPTION:
*
* INPUTS:
* TCA_REF_PTR ref; 
* char *name;
* void *data; 
* int tplConstr;
*
* OUTPUTS: TCA_RETURN_VALUE_TYPE
*
******************************************************************************/

TCA_RETURN_VALUE_TYPE tcaExecuteCommandWithConstraints(ref, name, data, 
						       tplConstr)
TCA_REF_PTR ref; 
char *name;
void *data;
int tplConstr;
{ 
  MSG_PTR msg;
  BLOCK_COM_TYPE blockCom;

  if (testInconsistent(tplConstr))
    TCA_MOD_ERROR1("Inconsistent temporal constraints for command %s\n", name);

  msg = msgFind(name);
  checkMessageClass(msg, CommandClass);

  blockCom.waitFlag = 0;
  blockCom.tplConstr = tplConstr;
  return sendMessage(ref, msg, (char *)data, (char *)&blockCom);
}


/******************************************************************************
*
* FUNCTION: TCA_RETURN_VALUE_TYPE tcaWaitForCommand(name, data)
*
* DESCRIPTION:
*
* INPUTS: char *name;
*         void *data;
*
* OUTPUTS: TCA_RETURN_VALUE_TYPE
*
* NOTES:
* 14-Aug-91: fedor: should check return valuse from 
* sendMessage and waitForReply.
*
******************************************************************************/

TCA_RETURN_VALUE_TYPE tcaWaitForCommand(name, data)
char *name;
void *data;
{ 
  int refId;
  MSG_PTR msg;
  TCA_REF_PTR tcaRef;
  BLOCK_COM_TYPE blockCom;
  TCA_MSG_CLASS_TYPE replyClass;

  msg = msgFind(name);
  checkMessageClass(msg, CommandClass);

  blockCom.waitFlag = 1;
  blockCom.tplConstr = SEQ_ACH;

  refId = nextSendMessageRef();
  (void)sendMessage((TCA_REF_PTR)NULL, msg, (char *)data, (char *)&blockCom);

  tcaRef = tcaRefCreate(msg, name, refId);

  (void)waitForReply(tcaRef, (char *)&replyClass);

  tcaRefFree(tcaRef);

  if (replyClass == SuccessClass)
    return Success;
  else
    return Failure;
}


/******************************************************************************
*
* FUNCTION: TCA_RETURN_VALUE_TYPE tcaWaitForCommandWithConstraints(ref, name, 
*                                                              data, tplConstr)
*
* DESCRIPTION:
*
* INPUTS:
* TCA_REF_PTR ref; 
* char *name;
* void *data; 
* int tplConstr;
*
* OUTPUTS: TCA_RETURN_VALUE_TYPE
*
* NOTES:
* 14-Aug-91: fedor: should check return valuse from 
* sendMessage and waitForReply.
*
******************************************************************************/

TCA_RETURN_VALUE_TYPE tcaWaitForCommandWithConstraints(ref, name, data, 
						       tplConstr)
TCA_REF_PTR ref; 
char *name;
void *data;
int tplConstr;
{ 
  int refId;
  MSG_PTR msg;
  TCA_REF_PTR tcaRef;
  BLOCK_COM_TYPE blockCom;
  TCA_MSG_CLASS_TYPE replyClass;

  if (testInconsistent(tplConstr))
    TCA_MOD_ERROR1("Inconsistent temporal constraints for command %s\n", name);
 
  msg = msgFind(name);
  checkMessageClass(msg, CommandClass);
  
  blockCom.waitFlag = 1;
  blockCom.tplConstr = tplConstr;
  
  refId = nextSendMessageRef();
  (void)sendMessage(ref, msg, (char *)data, (char *)&blockCom);
  
  tcaRef = tcaRefCreate(msg, name, refId);

  (void)waitForReply(tcaRef, (char *)&replyClass);

  tcaRefFree(tcaRef);

  if (replyClass == SuccessClass)
    return Success;
  else
    return Failure;
}

/******************************************************************************
*
* FUNCTION: TCA_RETURN_VALUE_TYPE tcaBroadcast(name, data)
*
* DESCRIPTION: Broadcast a one-way ("inform"-type) message to *all* modules
*              that registered a handler for that particular message.
*
* INPUTS: char *name;
*         void *data;
*
* OUTPUTS: TCA_RETURN_VALUE_TYPE
*
******************************************************************************/

TCA_RETURN_VALUE_TYPE tcaBroadcast(name, data)
char *name;
void *data;
{ 
  MSG_PTR msg;

  msg = msgFind(name);
  checkMessageClass(msg, BroadcastClass);

  return sendMessage((TCA_REF_PTR)NULL, msg, (char *)data, (char *)NULL);
}

/******************************************************************************
*
* FUNCTION: TCA_RETURN_VALUE_TYPE tcaMultiQuery(name, query, max, refPtr)
*
* DESCRIPTION:
* Send a "MultiQuery" to all handlers that can handle the message.
* "max" is the maximum number of responses requested.
* "refPtr" is set to an identifier to be used with a corresponding tcaMultiReceive.
*
* INPUTS: char *name;
*         void *query;
*         int max;
*         TCA_REF_PTR *refPtr;
*
* OUTPUTS: TCA_RETURN_VALUE_TYPE
*
* NOTE: "refPtr" is a *pointer* to a TCA_REF_PTR.
*       Although one can set the maximum number of responses to be received,
*       it is not guaranteed that that number will be received by tcaMultiReceive:
*       the number of responses is MIN("max", number of registered handlers).
*
******************************************************************************/

TCA_RETURN_VALUE_TYPE tcaMultiQuery(name, query, max, refPtr)
char *name;
void *query;
int max;
TCA_REF_PTR *refPtr;
{
  int refId;
  MSG_PTR msg;
  MULTI_QUERY_CLASS_TYPE multiQueryClassData;
  TCA_RETURN_VALUE_TYPE returnValue;

  msg = msgFind(name);
  checkMessageClass(msg, MultiQueryClass);

  refId = nextSendMessageRef();
  multiQueryClassData.num = 0;
  multiQueryClassData.max = max;
  returnValue = sendMessage((TCA_REF_PTR)NULL, msg, (char *)query,
			    (char *)&multiQueryClassData);

  *refPtr = tcaRefCreate(msg, name, refId);
#ifdef LISP
  lispRefSaveGlobal = *refPtr;
#endif

  return returnValue;
}

/******************************************************************************
*
* FUNCTION: TCA_RETURN_VALUE_TYPE tcaMultiReceive(ref, reply, timeout)
*
* DESCRIPTION: ref is a TCA_REF_PTR created by tcaMultiQuery.
*              reply is a pointer to (already allocated) data.
*
* Waits for replies to a tcaMultiQuery.  The ref passed must be the same as
*    the one gotten from the tcaMultiQuery function.
* "timeout" is number of seconds to wait for a reply before returning.  If no
*    message is received in that time, the return value is "TimeOut". The defined
*    value WAITFOREVER indicates not to return until the reply is received .
* To receive all the responses to a multi-query, loop doing a tcaMultiReceive.
*    When no more responses will be forthcoming, the return value of 
*    tcaMultiReceive is "NullReply".
* For example:  while (tcaMultiReceive(mqRef, WAITFOREVER, &reply) != NullReply) {
*                 -- Process and/or save reply data here --
*                 }
*
* INPUTS: TCA_REF_PTR ref;
*         void *reply;
*         long timeout;
*
* OUTPUTS: TCA_RETURN_VALUE_TYPE
*          Can have the values: 
*          "Success" -- a reply was received
*          "NullReply" -- indicates that no more replies will be forthcoming.
*		          This can come about either because the "max" number
*                         of replies were received (see tcaMultiSend), or all 
*                         the handlers registered for that message replied.
*          "TimeOut" -- "timeout" seconds elapsed without a message being received.
*
* NOTE: The responses are received in the order that the modules handle the query.
*       If "max" < number registered handlers, then the "max" fastest replies 
*       will be reported.
*
******************************************************************************/

TCA_RETURN_VALUE_TYPE tcaMultiReceive(ref, reply, timeout)
TCA_REF_PTR ref;
void *reply;
long timeout;
{
  TCA_RETURN_VALUE_TYPE returnValue;

  returnValue = waitForReply3(ref, (char *)reply, 0, timeout);

  if (returnValue == NullReply) 
    tcaRefFree(ref);

  return returnValue;
}
