/******************************************************************************
*
* PROJECT: Carnegie Mellon Planetary Rover Project
*          Task Control Architecture 
* 
* (c) Copyright 1991 Christopher Fedor and Reid Simmons.  All rights reserved.
* 
* MODULE: communications
*
* FILE: comServer.c
*
* ABSTRACT:
* 
* Communication Routines - Central Server
*
* REVISION HISTORY
*
* 6-May-92 Christopher Fedor, School of Computer Science, CMU
* Added while loop to select if select dies on an EINTR then retry select.
*
* 31-Oct-91 Christopher Fedor, School of Computer Science, CMU
* Moved resourceProcessPending of the implied resource from 
* the place where temp resources were removed to waitUntilReady release
* so that a reconnecting module has a chance to initialize before handling
* incoming messages.
*
* 28-Oct-91 Christopher Fedor, School of Computer Science, CMU
* Added repliesPending and alive field to MODULE_PTR to 
* track replies to modules that have exited and clean up pending replies.
* Added he calls moduleFree, pendingReplyAdd, pendingReplyRemove.
*
* 24-Jun-91 Christopher Fedor, School of Computer Science, CMU
* Update refId field for HandlerRegClass in selfRegisterMsg since the
* msg was never added to the msgIdTable. This is very old code and the
* whole registration of messages should be rethought because msgData is
* not needed by a module until a msgFind succeeds. selfRegisterMsg got
* moved to server only code and still has alot of old dependencies.
*
* 20-Jun-91 Christopher Fedor, School of Computer Science, CMU
* Moved parsing to msgInfoHnd. Added a parsedFormats toggle to msg struct.
* User messages will get parsed when they are first asked for.
* This corrects an error which wouldn't allow monitor messages to get parsed
* unless a tcaWaitUntilReady was issued. the wait until ready idea also did
* not work for a single module - which would need not use that call.
*
* 11-Jun-91 Christopher Fedor, School of Computer Science, CMU
* Changed parsing of message formatters to occur before releasing the
* wait until ready holds. This should allow tcaRegisterNamedFormatters time
* to complete before using them.
*
* 13-Mar-91 Reid Simmons, School of Computer Science, CMU
*  Added initialization routine for message tapping.
*
* 11-Dec-90 Christopher Fedor, School of Computer Science, CMU
* Added timing information. Also added function to remove 
* messages waiting for handlers.
*
* 25-Aug-90 Christopher Fedor, School of Computer Science, CMU
* Comments from no longer needed general com routines moved to comServer.c
* Revised comServer.c to software standards.
*
* 18-Aug-90 Christopher Fedor, School of Computer Science, CMU
* Simplified common routines. Service lookup routines were removed
* they should be put back when the dust settles - for now the port
* number is fixed and defined in tcaInternal.h
*
* 16-Aug-90 Christopher Fedor, School of Computer Science, CMU
* Moved Message_Class_Name to behaviors.c
* Removed initReplyHnd which is no longer used.
* Revised to Software Standards.
*  
*  3-Dec-89 Long-Ji Lin, School of Computer Science, CMU
* Commented out dataMsgFree. Added call to Exception_Layer_Initialize.
*
* 14-Aug-89 Reid Simmons, School of Computer Science, CMU
* Move "CreateRefHnd" to tasktree.c
*
* 11-Aug-89 Reid Simmons, School of Computer Science, CMU
* Changed "goaltree" to "tasktree".
*
*  2-Aug-89 Reid Simmons, School of Computer Science, CMU
* Changed "printf"s to use logging facility.
*
* 10-May-89 Christopher Fedor, School of Computer Science, CMU
* InitReplyHnd - reserve a slot for reply.
*
*  9-May-89 Reid Simmons, School of Computer Science, CMU
* Added call to implement polling monitors.
*
* 28-Apr-89 Reid Simmons, School of Computer Science, CMU
* Added code to initialize temporal constraints.
* Initialize "RootNode" field in TC_ModInfo.
*
* 11-Mar-89 Christopher Fedor, School of Computer Science, CMU
* Rewrote InitInternalFormatters to use tcaRegisterLengthFormatter.
* Removed com_server.c routines found in Reid Simmons communcations.c
*
*  9-Mar-89 Reid Simmons, School of Computer Science, CMU
* Seperated Module/Server Common communciation routines.
*
*  6-Mar-89 Christopher Fedor, School of Computer Science, CMU
* tried to seperate module/server communications
*
* 23-Feb-89 Christopher Fedor, School of Computer Science, CMU
* why yet again of course!
*
*  1-Dec-88 Christopher Fedor, School of Computer Science, CMU
* created.
*
******************************************************************************/

#include "stdio.h"

#include "errno.h"
extern int errno;

#include "sys/types.h"
#include "sys/timeb.h"
#include "sys/socket.h"
#include "netinet/in.h"
#include "netdb.h"
#include "ctype.h"

#include "sys/time.h"

#include "basics.h"
#include "tcaInternal.h"
#include "formatters.h"
#include "logging.h"

#include "datamsg.h"
#include "hash.h"
#include "idtable.h"

#include "dispatch.h"

#include "tplConstr.h"

#include "taskTree.h"

#include "list.h"
#include "tms.h"
#include "cells.h"
#include "rules.h"

#include "key.h"
#include "res.h"
#include "msgTap.h"

#define PROTOCOL_NAME "tcp"

#define TCP 6
#define BACKLOG 5

int serverPortGlobal;

extern int tcaServerGlobal;
extern fd_set tcaConnectionListGlobal;

extern TCA_VERSION_TYPE versionGlobal;

extern ID_TABLE_PTR msgIdTable;
extern HASH_TABLE_PTR messageTable;
extern HASH_TABLE_PTR handlerTable;

extern ID_TABLE_PTR hndIdTable;
extern ID_TABLE_PTR dispatchTable;

extern int checkTaps();

extern MSG_PTR msgCreate();
extern HND_PTR selfRegisterHnd();

extern HASH_TABLE_PTR resourceTable, classFormatTable;

extern void tcaModuleInitialize(); /* defined in comModule.c */
extern void classInitialize(); /* defined in recvMsg.c */
extern void monitorInitialize(); /* defined in mon.c */

extern void CompletionConstraints();

extern void msgInfoMsgSend();

extern void registerClass();

extern void tapInitialize();
extern void exceptionIntialize();

extern void monitorPoll();
extern void parseMonitorMsg();

extern void addFormatToTable();

extern FORMAT_PTR byteFormat;
extern FORMAT_PTR charFormat;
extern FORMAT_PTR shortFormat;
extern FORMAT_PTR intFormat;
extern FORMAT_PTR longFormat;
extern FORMAT_PTR floatFormat;
extern FORMAT_PTR doubleFormat;

extern FORMAT_PTR ParseFormatString();

extern void (*dPrintSTR_FN)();
extern void (*dPrintFORMAT_FN)();
extern void (*dPrintMAP_FN)();
extern void (*dPrintTCA_FN)();
extern void (*dPrintCHAR_FN)();
extern void (*dPrintSHORT_FN)();
extern void (*dPrintLONG_FN)();
extern void (*dPrintINT_FN)();
extern void (*dPrintFLOAT_FN)();
extern void (*dPrintDOUBLE_FN)();
extern void (*dPrintBYTE_FN)();
extern void (*dPrintTWOBYTE_FN)();

#include "printData.h"

/* 11-Dec-90: fedor ************ messy timing stuff ***************/

float msgPer, monPer, waitPer;

struct timeb tp;
long startTime, endTime, diffTime, avgTime, mTotal;
long avgSize, byteSize, byteTotal;

long startLoop, endLoop, startMon, endMon;

long totalMsgRun;

float totalLoop, totalMon, totalWait, totalMsg;

/**************/

MODULE_PTR tcaServerModGlobal;
MOD_DATA_TYPE newModDataGlobal = {"**NEWMOD**", "**UNKNOWN**"};

int waitTotalGlobal = 0;
int waitExpectedGlobal = 0;

HASH_TABLE_PTR moduleTable;
LIST_PTR moduleList, waitList;

typedef struct {
  int sd;
  char *name;
} MODULE_KEY_TYPE, *MODULE_KEY_PTR;


/******************************************************************************
*
* FUNCTION: int modHashFunc(modKey)
*
* DESCRIPTION: Module data hash function.
*
* INPUTS: MODULE_KEY_PTR modKey;
*
* OUTPUTS: int
*
******************************************************************************/

int modHashFunc(modKey)
MODULE_KEY_PTR modKey;
{
  char *s;
  int sum, i;

  sum = modKey->sd;
  s = modKey->name;
  for (i=0;s[i] != '\0';i++) 
    if (isupper(s[i]))
      sum += tolower(s[i]);
    else
      sum += s[i];

  return sum;
}


/******************************************************************************
*
* FUNCTION: int modKeyEqFunc(modKeyA, modKeyB)
*
* DESCRIPTION: 
* Module key equal function. A module is equal if its name and socket
* origin number are equal. Case insensitive.
*
* INPUTS: MODULE_KEY_PTR modKeyA, modKeyB;
*
* OUTPUTS: int
*
******************************************************************************/

int modKeyEqFunc(modKeyA, modKeyB)
MODULE_KEY_PTR modKeyA, modKeyB;
{
  int i;
  char a1, b1;
  char *a, *b;

  if (modKeyA->sd != modKeyB->sd)
    return FALSE;

  a = modKeyA->name;
  b = modKeyB->name;

  i = 0;
  while (a[i] != '\0' && b[i] != '\0') {
    if (isupper(a[i]))
      a1 = tolower(a[i]);
    else 
      a1 = a[i];
    if (isupper(b[i]))
      b1 = tolower(b[i]);
    else 
      b1 = b[i];
    
    if (a1 != b1)
      return FALSE;

    i++;
  }

  if (a[i] == b[i])
    return TRUE;
  else
    return FALSE;
}


/******************************************************************************
*
* FUNCTION: MODULE_PTR moduleCreate(sd, modData)
*
* DESCRIPTION: 
* Create a new module. Initialize module parameters sd and modData.
*
* INPUTS:
* int sd;
* MOD_DATA_PTR modData;
*
* OUTPUTS: MODULE_PTR
*
******************************************************************************/

MODULE_PTR moduleCreate(sd, modData)
int sd;
MOD_DATA_PTR modData;
{
  MODULE_PTR module;

  module = NEW(MODULE_TYPE);
  module->sd = sd;
  module->wait = FALSE;
  module->port = -1;
  module->alive = TRUE;
  module->repliesPending = 0;
  module->modData = modData;
  module->hndList = listCreate();
  module->impliedResource = NULL;
  module->resourceList = listCreate();

  return module;
}


/******************************************************************************
*
* FUNCTION: void removeConnection(module)
*
* DESCRIPTION: 
* Removes a module in response to a close detected on a socket or
* an explicit call to close a module - tcaClose.
*
* INPUTS: 
* MODULE_PTR module;
*
* OUTPUTS: void. 
*
* REVISION HISTORY:
*
*  1-Oct-90 Christopher Fedor
* Reimplemented wait list for modules to make removing test for
* tcaWaitUntilReady condition easier.
*
* 27-Jul-89 Christopher Fedor
* Make sure the module left wasn't waiting in a tcaWaitUntilReady.
*
* 27-Apr-89 Christopher Fedor 
* Added Deregistration of Resources.
*
*  2-Jan-89 Reid Simmons
* Added Deregistration of Handlers for disconnecting modules.
*
* 11-Nov-89 Christopher Fedor
* Created.
*
******************************************************************************/

int removeHndFromMsg(hnd, msg)
HND_PTR hnd;
MSG_PTR msg;
{
  listDeleteItem((char *)hnd, msg->hndList);

  return TRUE;
}

void hndDelete(hnd)
HND_PTR hnd;
{
  HND_KEY_TYPE hndKey;

  hndKey.sd = hnd->sd;
  hndKey.name = hnd->hndData->hndName;

  listTestDeleteItemAll(removeHndFromMsg, (char *)hnd, hnd->msgList);
  listFree(hnd->msgList);

  (void)idTableRemove(hnd->localId, hndIdTable);
  (void)hashTableRemove((char *)&hndKey, handlerTable);

  if (hnd->hndData) {
    tcaFree((char *)hnd->hndData->msgName);
    tcaFree((char *)hnd->hndData->hndName);
    tcaFree((char *)hnd->hndData);
  }
  
  tcaFree((char *)hnd);
}

int removeDispatchFromWait(sd, dispatch)
int *sd;
DISPATCH_PTR dispatch;
{
  if (dispatch->orgId == *sd) {
    waitTotalGlobal--;
    dispatchFree(dispatch);
    return TRUE;
  }

  return FALSE;
}

void moduleFree(module)
MODULE_PTR module;
{
  if (module->modData) {
    if (module->modData->modName)
      tcaFree((char *)module->modData->modName);
    
    if (module->modData->hostName)
      tcaFree((char *)module->modData->hostName);
    
    tcaFree((char *)module->modData);
  }
    
  tcaFree((char *)module);

  module = NULL; /* possibly force some memory errors */
}

pendingReplyAdd(dispatch)
DISPATCH_PTR dispatch;
{
  if (dispatch->org)
    dispatch->org->repliesPending++;
}

pendingReplyRemove(dispatch)
DISPATCH_PTR dispatch;
{
  if (dispatch->org)
    dispatch->org->repliesPending--;
}

void removeConnection(module)
MODULE_PTR module;
{
  listDeleteItem((char *)module, moduleList);
  listTestDeleteItem(removeDispatchFromWait, (char *)&module->sd, waitList);

  listFreeAllItems(hndDelete, module->hndList);
  listFreeAllItems(resourceDelete, module->resourceList);

  listFree(module->hndList);
  listFree(module->resourceList);

  /*
   * Needed so that restarting the module doesn't interfere with
   * the old temporal constraints
   */
  removeModLastChild(tcaTaskTreeRootGlobal, module);

  module->alive = FALSE;

  if (!module->repliesPending)
    moduleFree(module);
}



/******************************************************************************
*
* FUNCTION: MODULE_PTR addConnection(sd, modData)
*
* DESCRIPTION: 
* Create a new module. 
* Add the sd to tcaConnectionListGlobal.
* Add the module to the list of active modules.
* Send the new module intial message set.
*
* INPUTS:
* int sd;
* MOD_DATA_PTR modData;
*
* OUTPUTS: MODULE_PTR
*
******************************************************************************/

MODULE_PTR addConnection(sd, modData)
int sd;
MOD_DATA_PTR modData;
{ 
  MODULE_PTR module;
  
  module = moduleCreate(sd, modData);

  FD_SET(module->sd, &tcaConnectionListGlobal);
  listInsertItem((char *)module, moduleList);

  msgInfoMsgSend(module->sd);

  return module;
}


/******************************************************************************
*
* FUNCTION: MODULE_PTR serverModCreate(sd)
*
* DESCRIPTION: 
* Create a new module for the central server.
* Add the sd to tcaConnectionListGlobal.
*
* INPUTS:
* int sd;
* MOD_DATA_PTR modData;
*
* OUTPUTS: MODULE_PTR
*
******************************************************************************/

MODULE_PTR serverModCreate(sd)
int sd;
{ 
  MODULE_PTR module;
  MOD_DATA_PTR modData;
  MODULE_KEY_TYPE modKey;

  modData = NEW(MOD_DATA_TYPE);
  modData->modName = (char *)tcaMalloc(sizeof(SERVER_MOD_NAME));
  bcopy(SERVER_MOD_NAME, modData->modName, sizeof(SERVER_MOD_NAME));
  modData->hostName = (char *)tcaMalloc(sizeof(char)*32);
  (void)gethostname(modData->hostName, 32);

  module = moduleCreate(sd, modData);
    
  FD_SET(sd, &tcaConnectionListGlobal);

  modKey.sd = sd;
  modKey.name = modData->modName;

  (void)hashTableInsert((char *)&modKey, sizeof(MODULE_KEY_TYPE), 
			(char *)module, moduleTable);

  listInsertItem((char *)module, moduleList);

  return module;
}


/**************************************************************************/

void NewModuleConnectHnd(dispatch, modData)
DISPATCH_PTR dispatch;
MOD_DATA_PTR modData;
{
  RESOURCE_PTR resource;
  TCA_VERSION_TYPE versionData;
  
  Log1(" Received a new connection: %d\n", (char *)dispatch->orgId);
  Log1("   modName : %s\n", modData->modName);
  Log1("   hostName: %s\n", modData->hostName);

  resource = resourceCreate(dispatch->orgId, modData->modName, 1);

  dispatch->org->modData = modData;
  dispatch->org->impliedResource = resource;

  listInsertItem((char *)resource, dispatch->org->resourceList);

  versionData.tcaMajor = TCA_VERSION_MAJOR;
  versionData.tcaMinor = TCA_VERSION_MINOR;

  centralReply(dispatch, (char *)&versionData);

  /* Don't free the modData: It is stored with the module */
}

/***********************************************************************/

void parseMsg(msg)
MSG_PTR msg;
{
  /* 20-Jun-91: fedor: parse formats added to allow messages
     to be registered before they are used and allow register named
     formatter to propagate across modules. */
  if (msg->msgFormatStr)
    msg->msgData->msgForm = ParseFormatString(msg->msgFormatStr);

  if (msg->resFormatStr)
    msg->msgData->resForm = ParseFormatString(msg->resFormatStr);      

  msg->parsedFormats = TRUE;
}


/******************************************************************************
*
* FUNCTION: MSG_PTR selfRegisterMsg(msgData)
*
* DESCRIPTION: 
*
* If a message with the same message name is found to exist in the hash table
* that message data is replaced with the new msgData and a warning message
* is issued. If this happens refId of the new msgData is set equal to the
* refId being replaced.
*
* If a message created by selfRegisterHnd is found it is simply replaced
* without warning.
*
* The newly created or updated message is returned.
*
* INPUTS:
* MSG_DATA_PTR msgData;
*
* OUTPUTS: MSG_PTR
*
* NOTES:
*
* 23-Jun-91: fedor: Always a good idea to trigger parse when info changes
* - this allows us to use HandlerRegClass for taps with missing msgs.
*
* 20-Jun-91: fedor: Toggle to reparse formatters - needed when modules
* come and go - really needed for selfquery msgs of modules that crash.
*
******************************************************************************/

MSG_PTR selfRegisterMsg(msgData)
MSG_DATA_PTR msgData;
{
  int localId;
  MSG_PTR msg;

  msg = (MSG_PTR)hashTableFind(msgData->name, messageTable);
  if (msg) {
    localId = msg->msgData->refId;
    if (msg->msgData->class != HandlerRegClass) {
      Start_Ignore_Logging();
      Log_Status1("Registration: Message %s Found. Updating.\n",
		  msgData->name);
      End_Ignore_Logging();
    } else {
      /* 24-Jun-91: fedor: update refId because a HandlerRegClass
	 did not get added to the msgIdTable - mostly to avoid the
	 msgIdTable in modules. This needs more work and a rewrite! */
      localId = idTableInsert((char *)msg, msgIdTable);
    }      

    tcaFree((char *)msg->msgData);
    msg->msgData = msgData;

    /* 24-Jun-91: fedor: not sure if this will force a parse everywhere
       - there may be some strong assumptions about message registration
       before wait until ready - 2nd registration elsewhere may not
       pick up correct parsed format info. */

    msg->parsedFormats = FALSE;

    if (msg->tapList && listLength(msg->tapList)) {
      (void)listIterate(checkTaps, (char *)msg, msg->tapList);
    }
  }
  else {
    msg = msgCreate(msgData);
    localId = idTableInsert((char *)msg, msgIdTable);
  }

  msg->msgData->refId = localId;

  return msg;
}

/***********************************************************************/

void registerMessageHnd(dispatch, msgRegData)
DISPATCH_PTR dispatch;
MSG_REG_DATA_PTR msgRegData;
{
  MSG_PTR msg;
  MSG_DATA_PTR msgData;

  msgData = NEW(MSG_DATA_TYPE);

  msgData->name = msgRegData->name;
  msgData->class = msgRegData->class;

  msgData->msgForm = NULL;
  msgData->resForm = NULL;

  msg = selfRegisterMsg(msgData);

  /* 3-Jul-91: Reid: Don't bother copying the format strings -- 
     just be careful not to delete msgRegData recursively! */
  if (msgRegData->msgForm) {
    msg->msgFormatStr = msgRegData->msgForm;
  }
  if (msgRegData->resForm) {
    msg->resFormatStr = msgRegData->resForm;
  }

  /* Just frees the top-level structure because the strings msgRegData
     points to (i.e., name, msgForm, resForm) are all stored. */
  tcaFree((char *)msgRegData);
}

/***********************************************************************/

void centralRegisterMessage(name, class, msgForm, resForm)
char *name;
TCA_MSG_CLASS_TYPE class;
char *msgForm, *resForm;
{
  MSG_PTR msg;
  MSG_DATA_PTR msgData;

  msgData = NEW(MSG_DATA_TYPE);

  msgData->name = name;
  msgData->class = class;

  if (msgForm) 
    msgData->msgForm = ParseFormatString(msgForm);
  else
    msgData->msgForm = NULL;

  if (resForm) 
    msgData->resForm = ParseFormatString(resForm);
  else
    msgData->resForm = NULL;

  msg = selfRegisterMsg(msgData);

  msg->parsedFormats = TRUE;
}

/***********************************************************************/

void centralRegisterNamedFormatter(formatterName, formatString)
char *formatterName, *formatString;
{ 
  addFormatToTable(formatterName, ParseFormatString(formatString));
}

/***********************************************************************/

void centralRegisterLengthFormatter(formatterName, length)
char *formatterName;
int length;
{ 
  int i;
  char s[11];

  for(i=0;i<11;i++)
    s[i] = '\0';

  (void)sprintf(s, "%d", length);

  centralRegisterNamedFormatter(formatterName, s);
}

/***********************************************************************/

void centralRegisterHandler(msgName, hndName, hndProc)
char *msgName, *hndName;
void (*hndProc)();
{
  HND_DATA_PTR hndData;

  if (!hndProc)
    TCA_ERROR1("ERROR: centralRegisterHandler: hndProc for %s is NULL.\n", 
	       hndName);

  hndData = NEW(HND_DATA_TYPE);

  hndData->refId = 0; /* will be updated by selfRegisterHandler */
  hndData->msgName = msgName;
  hndData->hndName = hndName;

  (void)selfRegisterHnd(0, tcaServerModGlobal, hndData, hndProc);
}

/***********************************************************************/

int addHndItem(hnd, dispatch)
HND_PTR hnd;
DISPATCH_PTR dispatch;
{
  dispatch->hnd = hnd;
  dispatch->des = hnd->hndOrg;
  dispatch->desId = hnd->hndOrg->sd;

  return 1;
}

int addPendItem(pendList, dispatch)
LIST_PTR pendList;
DISPATCH_PTR dispatch;
{
  listInsertItem((char *)dispatch, pendList);

  return 1;
}

void registerHandlerHnd(dispatch, hndData)
DISPATCH_PTR dispatch;
HND_DATA_PTR hndData;
{
  MSG_PTR msg;
  HND_PTR hnd;
  RESOURCE_PTR res;

  res = NULL;
  msg = (MSG_PTR)hashTableFind(hndData->msgName, messageTable);
  if (msg) {
    hnd = (HND_PTR)listFirst(msg->hndList);
    if (hnd) {
      res = hnd->resource;
      if (!res->capacity) {
	/* ok this handler is the created one */
	listDeleteItem((char *)hnd, msg->hndList);
	tcaFree((char *)hnd);
      }
    }
  }

  hnd = selfRegisterHnd(dispatch->orgId, dispatch->org, hndData, NULL);

  hnd->resource = dispatch->org->impliedResource;

  if (dispatch->org->port != -1) {
    msg->direct = TRUE;
  }

  if (!listMemberItem((char *)hnd, dispatch->org->hndList))
    listInsertItem((char *)hnd, dispatch->org->hndList);

  if (res && !res->capacity) {
    /* 12-May-91: fedor: make sure that this was the tmp resource and
       not simply the resource of the first hnd in the msg->hndList 
       - this can cause problems when registering multiple handlers
       for the same message. Should clean these ideas up. */
    listFree(res->attendingList);
    (void)listIterateFromFirst(addHndItem, (char *)hnd, res->pendingList);
    (void)listIterateFromFirst(addPendItem, 
			       (char *)dispatch->org->impliedResource->pendingList,
			       res->pendingList);
    listFree(res->pendingList);
    resourceDelete(res);

   /* 31-Oct-91: fedor: moved to waitFor release
      resourceProcessPendingRes(dispatch->org->impliedResource);*/
  }
  /* Don't free the hndData: It is stored with the handler */
}

/***********************************************************************/

void registerNamedFormHnd(dispatch, regData)
DISPATCH_PTR dispatch;
REG_DATA_PTR regData;
{
  addFormatToTable(regData->name, ParseFormatString(regData->format));
  
 /* Free the top level structure: Can the "format" string be freed as well? */
  tcaFree((char *)regData);
}

/***********************************************************************/

void msgInfoHnd(dispatch, msgAsk)
DISPATCH_PTR dispatch;
MSG_ASK_PTR msgAsk;
{
  HND_PTR hnd;
  HND_KEY_TYPE hndKey;

  MSG_PTR msg;
  MSG_DATA_PTR msgData;

  msgData = NULL;
  msg = (MSG_PTR)hashTableFind(msgAsk->msgName, messageTable);

  if (msg) {
    if (!msg->parsedFormats) {
      switch(msg->msgData->class) {
      case PollingMonitorClass:
      case PointMonitorClass:
      case DemonMonitorClass:
	parseMonitorMsg(msg);
	break;
      default:
	parseMsg(msg);
      }
    }

    msgData = msg->msgData;
    
    if (msg->direct && (msgData->class == InformClass)) {
      /* trigger establish direct connection */
      msgData->refId = -1;
    }

    if (msgAsk->hndName) {
      /* 27-May-91: fedor: flag hnd cached msg info */
      hndKey.sd = dispatch->orgId;
      hndKey.name = msgAsk->hndName;

      hnd = (HND_PTR)hashTableFind((char *)&hndKey, handlerTable);

      if (hnd)
	hnd->msg = msg;
      else
	Log_Message1("WARNING: msgInfoHnd: Missing handler: %s\n",
		     msgAsk->hndName);
    }
  }

  /* A bit more efficient than using tcaFreeData */
  freeDataStructure(dispatch->msg->msgData->msgForm, (char *)msgAsk);

  centralReply(dispatch, (char *)&msgData);
}

/***********************************************************************/

int waitReleaseFunc(go, dispatch)
int *go;
DISPATCH_PTR dispatch;
{
  centralReply(dispatch, (char *)go);
  resourceProcessPendingRes(dispatch->org->impliedResource);

  /* 9-Jul-91: Reid:
     The last received wait message dispatch is freed normally,
     since it is being handled.  The rest have to be freed here
     because they have been put on hold */
  dispatchFree(dispatch);
  return TRUE;
}  

void WaitHnd(dispatch, data)
DISPATCH_PTR dispatch;
char *data;
{
  int go = TRUE;

  waitTotalGlobal++;
  if (waitTotalGlobal < waitExpectedGlobal) 
    listInsertItem((char *)dispatch, waitList);
  else {
    centralReply(dispatch, (char *)&go);
    resourceProcessPendingRes(dispatch->org->impliedResource);
    listTestDeleteItemAll(waitReleaseFunc, (char *)&go, waitList);
  }
}

/**************************************************************************/

static LIST_PTR Message_Ignore_Set;

int Ignore_Logging_Message (Message)
MSG_PTR Message;
{
  return listMemberItem((char *)Message, Message_Ignore_Set);
}

void Add_Message_To_Ignore (name)
char *name;
{ 
  MSG_PTR Message;
  
  Message = (MSG_PTR)hashTableFind(name, messageTable);
  if (Message)
    listInsertItem((char *)Message, Message_Ignore_Set);
}


/**************************************************************************/

void tcaCloseHnd(dispatch, refId)
DISPATCH_PTR dispatch;
char *refId;
{
  Log1("tcaClose: Closed Connection Detected from: sd: %d:\n", 
       (char *)dispatch->org->sd);
  if (dispatch->org->modData) 
    Log2("tcaClose: Closing %s on %s\n", dispatch->org->modData->modName,
	 dispatch->org->modData->hostName);
  else
    Log("\n");

  FD_CLR(dispatch->org->sd, &tcaConnectionListGlobal);
  (void)shutdown(dispatch->org->sd, 2);
  (void)close(dispatch->org->sd);
  removeConnection(dispatch->org);
}

/**************************************************************************/

void classInfoHnd(dispatch, class)
DISPATCH_PTR dispatch;
TCA_MSG_CLASS_TYPE *class;
{
  CLASS_FORM_PTR classForm;

  classForm = (CLASS_FORM_PTR)hashTableFind((char *)class, classFormatTable);

  centralReply(dispatch, (char *)&classForm);

  tcaFree((char *)class); /* Use simple free: "TCA_MSG_CLASS_TYPE" is an integer */
}

/**************************************************************************/

void directResHnd(dispatch, direct)
DISPATCH_PTR dispatch;
DIRECT_PTR direct;
{
  if (strKeyEqFunc(dispatch->org->modData->modName, direct->name)) {
    dispatch->org->port = direct->port;
    Log_Message2("Direct Connection Established For: %s at %d\n",
		 direct->name, (char *)direct->port);
  }
  else {
    Log_Message1("Direct Connection: Not Made: %s did not originate call.\n",
		 direct->name);
  }
  /* Probably can free the whole thing here, but Chris would be the one to know */
  /*  freeDataStructure(dispatch->msg->msgData->msgForm, (char *)direct); */
}

/**************************************************************************/

void directInfoHnd(dispatch, name)
DISPATCH_PTR dispatch;
char **name;
{
  MSG_PTR msg;
  HND_PTR hnd;
  DIRECT_INFO_TYPE directInfo;

/*
  printf("name: %s\n", *name);
*/

  msg = (MSG_PTR)hashTableFind(*name, messageTable);

  hnd = (HND_PTR)listFirst(msg->hndList);

  directInfo.sd = hnd->hndOrg->sd;
  directInfo.port = hnd->hndOrg->port;
  directInfo.intent = hnd->hndData->refId;
  directInfo.host = hnd->hndOrg->modData->hostName;
  directInfo.module = hnd->hndOrg->modData->modName;
  
  centralReply(dispatch, (char *)&directInfo);

  tcaFree((char *)*name); /* "name" can be freed since it is just used for lookup */
}


/******************************************************************************
*
* FUNCTION: void serverMessagesInitialize()
*
* DESCRIPTION: 
* Initializes the message/handler system for the server and modules.
* Defines internal messages.
*
* INPUTS: none.
*
* OUTPUTS: void.
*
******************************************************************************/

void serverMessagesInitialize()
{
  centralRegisterLengthFormatter("TCA_MSG_CLASS_TYPE", 
				 sizeof(TCA_MSG_CLASS_TYPE));

  centralRegisterNamedFormatter("map", "fmat");

  centralRegisterMessage("msgInfoMsg", QueryClass, "{string, string}", 
		     "*{int, string, TCA_MSG_CLASS_TYPE, format, format}");
  centralRegisterHandler("msgInfoMsg", "msgInfoHnd", msgInfoHnd);

  centralRegisterMessage("newModuleConnectMsg", QueryClass, 
		     "{string, string}", "{int, int}");
  centralRegisterHandler("newModuleConnectMsg", "NewModuleConnectHnd",
		  NewModuleConnectHnd);

  centralRegisterMessage("classInfoMsg", QueryClass, 
		 "TCA_MSG_CLASS_TYPE", "*{TCA_MSG_CLASS_TYPE, format}");
  centralRegisterHandler("classInfoMsg", "classInfoHnd", classInfoHnd);

  centralRegisterMessage("registerMessageMsg", InformClass,
			 "{string, TCA_MSG_CLASS_TYPE, string, string}", 
			 (char *)NULL);
  centralRegisterHandler("registerMessageMsg", "registerMessageHnd",
			 registerMessageHnd);
  
  centralRegisterMessage("registerHandlerMsg", InformClass,
			 "{int, string, string}", (char *)NULL);
  centralRegisterHandler("registerHandlerMsg", "registerHandlerHnd",
			 registerHandlerHnd);

  centralRegisterMessage("registerNamedFormMsg", InformClass,
			 "{string, string}", (char *)NULL);
  centralRegisterHandler("registerNamedFormMsg", "registerNamedFormHnd",
			 registerNamedFormHnd);
		     
  centralRegisterMessage("waitMsg", QueryClass, (char *)NULL, "int");
  centralRegisterHandler("waitMsg", "WaitHnd", WaitHnd);

  centralRegisterMessage("tcaCloseMsg", InformClass, (char *)NULL, (char *)NULL);
  centralRegisterHandler("tcaCloseMsg", "tcaCloseHnd", tcaCloseHnd);

  centralRegisterMessage("directResMsg", InformClass,
			 "{int, string}", (char *)NULL);
  centralRegisterHandler("directResMsg", "directResHnd", directResHnd);

  centralRegisterMessage("directInfoMsg", QueryClass, 
			 "string", "{int, int, int, string, string}");
  centralRegisterHandler("directInfoMsg", "directInfoHnd", directInfoHnd);
}


/******************************************************************************
*
* FUNCTION: void printDataServerInitialize();
*
* DESCRIPTION: 
* Sets print data routines for the central server.
*
* INPUTS: none.
*
* OUTPUTS: void.
*
******************************************************************************/

void printDataServerInitialize()
{
  byteFormat = ParseFormatString("byte");
  charFormat = ParseFormatString("char");
  shortFormat = ParseFormatString("short");
  intFormat = ParseFormatString("int");
  longFormat = ParseFormatString("long");
  floatFormat = ParseFormatString("float");
  doubleFormat = ParseFormatString("double");

  dPrintSTR_FN = dPrintSTR;
  dPrintFORMAT_FN = dPrintFORMAT;
  dPrintMAP_FN = mapPrint;
  dPrintTCA_FN = dPrintTCA;
  dPrintCHAR_FN = printChar;
  dPrintSHORT_FN = printShort;
  dPrintLONG_FN = printLong;
  dPrintINT_FN = printInt;
  dPrintFLOAT_FN = printFloat;
  dPrintDOUBLE_FN = printDouble;
  dPrintBYTE_FN = printByte;
  dPrintTWOBYTE_FN = printTwoByte;
}


/******************************************************************************
*
* FUNCTION: int serverInitialize(expectedMods)
*
* DESCRIPTION: 
* Initialize the server. Creates a socket, looks up the server port
* number and binds the socket to that port number, and listens to the
* socket. The machine name for the central server is assumed to be the
* machine on which it is running and so it needs not be specified.
* tcaServerModGlobal is initialized to the central server module.
*
* INPUTS:
* int expectedMods;
*
* OUTPUTS: 
* Returns FALSE if there was an error detected, TRUE otherwise.
* If no error was detected serverMod is initialized to the server module.
*
******************************************************************************/

int serverInitialize(expectedMods)
int expectedMods;
{
  int sd;
  int port;
  struct sockaddr_in server;

  tcaModuleInitialize();

  bzero((char *)&server, sizeof(struct sockaddr_in));

#ifdef SERVER_PORT
  port = serverPortGlobal;
#else
  if ((sp = getservbyname(SERVER_NAME, PROTOCOL_NAME)) == (char *)NULL)
    return FALSE;

  port = sp->s_port;
#endif

  if (!listenAtPort(port, &sd))
    return FALSE;

  FD_ZERO(&tcaConnectionListGlobal);

  moduleList = listCreate();
  moduleTable = hashTableCreate(10, modHashFunc, modKeyEqFunc);

  tcaServerModGlobal = serverModCreate(sd);

  waitExpectedGlobal = expectedMods;

  /* 29-Aug-90: fedor: blah! */
  tcaServerGlobal = CENTRAL_SERVER_ID; 

  printDataServerInitialize();

  classInitialize();

  resourceInitialize();

  serverMessagesInitialize();

/*  serverInterfaceInitialize();*/

  dispatchTable = idTableCreate("Dispatch Table", 10);

  Message_Ignore_Set = listCreate();

  /* Ignore the following registration and informational messages */
  Add_Message_To_Ignore("msgInfoMsg");
  Add_Message_To_Ignore("newModuleConnectMsg");
  Add_Message_To_Ignore("classInfoMsg");
  
  Add_Message_To_Ignore("registerMessageMsg");
  Add_Message_To_Ignore("registerHandlerMsg");
  Add_Message_To_Ignore("registerNamedFormMsg");

  Add_Message_To_Ignore("registerResourceMsg");

  Add_Message_To_Ignore("regMonitorMsg");

  Add_Message_To_Ignore("waitMsg");
  
  taskTreeInitialize();

  monitorInitialize();

  tapInitialize();

  exceptionIntialize();

  waitList = listCreate();

  Start_File_Logging();
  Start_Terminal_Logging();

  Log1("Expecting %d\n", (char *)expectedMods);

  return TRUE;
}


/******************************************************************************
*
* FUNCTION: int iterateDataMsgRecv(readMask, module)
*
* DESCRIPTION: 
* This defines an iterate function to call dataRecvMsg 
* on each module that has input pending.
*
* INPUTS:
* fd_set readMask;
* MODULE_PTR module;
*
* OUTPUTS: TRUE - continue to next module.
*
******************************************************************************/

int iterateDataMsgRecv(readMask, module)
fd_set *readMask;
MODULE_PTR module;
{
  int sd;
  DATA_MSG_PTR dataMsg;
 
  if (FD_ISSET(module->sd, readMask)) {
    if (module == tcaServerModGlobal) {
      /* server open a new connection */
      sd = accept(module->sd, (struct sockaddr *)NULL, (int *)NULL);
      module = addConnection(sd, &newModDataGlobal);
    }

    ftime(&tp);
    startTime = 1000*tp.time + tp.millitm;
    
    switch(dataMsgRecv(module->sd, &dataMsg)) {
    case StatOK:

      byteSize = dataMsg->classTotal + dataMsg->msgTotal;
      byteTotal = byteTotal + byteSize;

      recvMessageBuild(module, dataMsg);
      break;
    case StatError:
      perror("ERROR: iterateDataMsgRecv ");
      /* fall through to close module */
    case StatEOF:
      /* 28-Aug-90: fedor: recycle dataMsg */
      Log_Status1("Closed Connection Detected from: sd: %d: \n",(char *)module->sd);
      /* 28-Aug-90: fedor: shouldn't need this test. */
      if (module->modData)
	Log2(" Closing %s on %s\n", module->modData->modName,
	     module->modData->hostName);
      else
	Log("\n");
      FD_CLR(module->sd, &tcaConnectionListGlobal);
      (void)shutdown(module->sd, 2);
      (void)close(module->sd);
      removeConnection(module);
      break;
    default:
      tcaError("ERROR: iterateDataMsgRecv: Unknown Status:");
    }

    ftime(&tp);
    endTime = 1000*tp.time + tp.millitm;

    diffTime = endTime - startTime;

    totalMsg = totalMsg + diffTime;

    mTotal++;

    avgTime = totalMsg/mTotal;

    avgSize = byteTotal/mTotal;
    LogHandleSummary( mTotal, msgPer, monPer, waitPer,
		     byteSize, diffTime, avgSize, avgTime);

  }
  return TRUE;
}


/******************************************************************************
*
* FUNCTION: void listenLoop()
*
* DESCRIPTION: 
* Central Clearing House for Incoming Messages.
* Grabs an available message and acts on it.
* Loops until quit. Need to define quit.
*
* INPUTS: none.
*
* OUTPUTS: void.
*
* NOTES:
* 20-Apr-89: fedor: dataMsgRecv should not return StatOK with a NULL dataMsg.
* need to insure dataMsg values 
*
* 27-Aug-90: fedor: need to put timing info back.
*
******************************************************************************/

void listenLoop()
{
  int stat;
  fd_set readMask; 
  struct timeval oneSecond;

  /******************/

  totalMsgRun = 0;

  startTime = endTime = totalMsg = diffTime =  mTotal = avgTime = 0;

  startLoop = endLoop = totalLoop = startMon = endMon = totalMon = 0;

  totalWait = 0;

  avgSize = byteTotal = 0;

  msgPer = monPer = waitPer = 0;

  /**************/

  /* dataMsgDisplayStats();*/

  oneSecond.tv_sec = 1;
  oneSecond.tv_usec = 0;

  for(;;) {

    readMask = tcaConnectionListGlobal;

    do {
      stat = select(FD_SETSIZE, &readMask, (fd_set *)NULL, (fd_set *)NULL,
		    &oneSecond);
    } while (stat < 0 && errno == EINTR);

    if (stat < 0)
      tcaError("ERROR: listenLoop: select: ");

    byteSize = 0;
    ftime(&tp);
    startLoop = 1000*tp.time + tp.millitm;

    startMon = startLoop;

    monitorPoll();

    ftime(&tp);
    endMon = 1000*tp.time + tp.millitm;

    totalMon = totalMon + (endMon - startMon);

    (void)listIterateFromFirst(iterateDataMsgRecv, (char *)&readMask, moduleList);

    ftime(&tp);
    endLoop = 1000*tp.time + tp.millitm;

    totalLoop = totalLoop + (endLoop - startLoop);

    totalWait = totalLoop - totalMsg - totalMon;

    msgPer = totalMsg / totalLoop * 100;
    monPer = totalMon / totalLoop * 100;
    waitPer = totalWait /totalLoop * 100;
  }
}
