/******************************************************************************
*
* PROJECT: Carnegie Mellon Planetary Rover Project
*          Task Control Architecture 
* 
* (c) Copyright 1991 Christopher Fedor and Reid Simmons.  All rights reserved.
* 
* MODULE: datamsg
*
* FILE: datamsg.c
*
* ABSTRACT:
* 
* A datamsg represents the encoded form of a message and is used to perform
* the actual sending and receiving of messages.
*
* REVISION HISTORY
*
* 6-May-92 Christopher Fedor, School of Computer Science, CMU
* Added EWOULDBLOCK detection - for now simply continue to try to write
* where it left off. THis may not be correct for all applications and can
* cause the central server to simply wedge.
*
*  9-Jul-1991 Reid Simmons, School of Computer Science, CMU
* Commented out stuff to store data message buffers.
*
*  8-Jul-1991 Reid Simmons, School of Computer Science, CMU
* Added routine to pre-allocate a datamsg buffer
*
* 25-Jun-91 Christopher Fedor, School of Computer Science, CMU
* Replaced fixed number of datamsg buffers with Reid's datamsg code.
* Just postponing the eventual memory lossage.
*
* 11-Jun-91 Christopher Fedor, School of Computer Science, CMU
* Set a fixed number of datamsg buffers.
*
* 30-Jan-91 Christopher Fedor, School of Computer Science, CMU
* Added fflush(stdout) to printf for module code calls from lisp
*
* 17-Sep-90 Christopher Fedor, School of Computer Science, CMU
* Added stats for tracking datamsg memory usage.
*
*  5-Jul-90 Christopher Fedor, School of Computer Science, CMU
* Added refCount for dealing with datamsg garbage collection.
*
*  4-Jun-90 Christopher Fedor, School of Computer Science, CMU
* Revised to Software Standards.
*
* 29-May-90 Christopher Fedor, School of Computer Science, CMU
* Revised double encoding of datamsg to support class data for tca 5.x
*
* 26-Apr-89 Christopher Fedor, School of Computer Science, CMU
* Changed to support double encoding.
*
* 23-Feb-89 Christopher Fedor, School of Computer Science, CMU
* Interface to Communications by dataMsgs
*
*    Dec-88 Christopher Fedor, School of Computer Science, CMU
* Created.
*
* 15-Nov-88 Christopher Fedor  created readNBytes
*
* NOTES:
*
* 26-Apr-89 Christopher Fedor, School of Computer Science, CMU
* Need to insure that central will not block on socket reads and writes. 
*
******************************************************************************/

#include <stdio.h>

#include "basics.h"
#include "tcaInternal.h"
#include "formatters.h"
#include "list.h"

#include "datamsg.h"

#ifndef VXWORKS
#include <errno.h>
extern int errno;
#endif

extern int bufferSize();
extern void encodeData();
extern char *decodeData();

extern int (*lispBufferSizeGlobal)();
extern int (*lispEncodeMsgGlobal)();

extern int tcaServerGlobal;

LIST_PTR dataMsgBufferList;
int DMFree, DMTotal, DMmin, DMmax;

char LispFlagGlobal;

#ifndef LISP

int lisp_value(a)
int a;
{
  perror("datamsg: ERROR: call to lisp_value");
  exit(1);
}

#endif


/******************************************************************************
*
* FUNCTION: TCA_RETURN_STATUS_TYPE readNBytes(sd, buf, nbytes)
*
* DESCRIPTION:
* Read nbytes of data from sd into buf. Continue to read until 
* nbytes have been read.
*
* INPUTS: 
* int sd;
* char *buf;
* int nbytes;
*
* OUTPUTS: TCA_RETURN_STATUS
*
* NOTES:
*
* buf is a preallocated pointer to storage equal to nbytes.
* Propagation of low level errors - failures on read/writes still an issue.
*
******************************************************************************/

TCA_RETURN_STATUS_TYPE readNBytes(sd, buf, nbytes)
int sd;
char *buf;
int nbytes;
{
  int amountRead, amountToRead;

  amountToRead = nbytes;
  for(;;){
    amountRead = read(sd, buf, amountToRead);
    if (amountRead < 0)
      return StatError;
    if (!amountRead)
      return StatEOF;
    amountToRead -= amountRead;
    if (!amountToRead)
      return StatOK;
    buf += amountRead;
  }
}    


/******************************************************************************
*
* FUNCTION: TCA_RETURN_STATUS_TYPE writeNBytes(sd, buf, nbytes)
*
* DESCRIPTION: This routine primarily calls the system function write.
*
* INPUTS: 
* int sd;
* char *buf;
* int nbytes;
*
* OUTPUTS: TCA_RETURN_STATUS_TYPE
*
******************************************************************************/

TCA_RETURN_STATUS_TYPE writeNBytes(sd, buf, nbytes)
int sd;
char *buf;
int nbytes;
{
  int amountWritten = 0;

  while (nbytes > 0) {
    amountWritten = write(sd, buf, nbytes);
    if (amountWritten < 0) {
#ifdef VXWORKS
      return StatError;
#else
      if (errno == EWOULDBLOCK) {
	fprintf(stderr, "WARNING: writeNBytes: EWOULDBLOCK: trying again!\n");
	usleep(1);
      } else {
	return StatError;
      }
#endif
    } else {
      nbytes -= amountWritten;
      buf += amountWritten;
    }
  }
  return StatOK;
}

/*****************************************************************************/

DATA_MSG_BUF_PTR dataMsgBufferCreate(size)
int size;
{
  DATA_MSG_BUF_PTR dataMsgBuf;

  dataMsgBuf = NEW(DATA_MSG_BUF_TYPE);

  dataMsgBuf->size = size;
  dataMsgBuf->allocated = FALSE;
  dataMsgBuf->dataMsg = (DATA_MSG_PTR)tcaMalloc((unsigned)size);

  return dataMsgBuf;
}

/*****************************************************************************/

void dataMsgInitialize()
{
  DMFree = 0;
  DMTotal = 0;
  DMmin = 2000000000;
  DMmax = -1;

  dataMsgBufferList = listCreate();
}

/*****************************************************************************/

/***
int findUnAllocatedSize(size, dataMsgBuf)
int *size;
DATA_MSG_BUF_PTR dataMsgBuf;
{
  if (!dataMsgBuf->allocated && *size <= dataMsgBuf->size)
    return TRUE;
  else
    return FALSE;
}
*****/

/*****************************************************************************/
/* June 21, 1991 Reid: Changed this somewhat to find the unallocated buffer
 * that is closest in size to what is needed.  Involves using a global var.
 */
/* Currently not used (RGS: 11/11/92) */
#if 0
static DATA_MSG_BUF_PTR bufferToAlloc;

int findUnAllocatedSize(size, dataMsgBuf)
int *size;
DATA_MSG_BUF_PTR dataMsgBuf;
{
  if (!dataMsgBuf->allocated && *size <= dataMsgBuf->size &&
   /* Within Central: Do not use buffer if more than twice the size needed */
   (tcaServerGlobal != CENTRAL_SERVER_ID || dataMsgBuf->size <= *size+(*size))
      && (!bufferToAlloc || dataMsgBuf->size < bufferToAlloc->size)) {
    bufferToAlloc = dataMsgBuf;
    return (dataMsgBuf->size != *size);
  }
  return TRUE;
}
#endif
/*****************************************************************************/

DATA_MSG_PTR dataMsgAlloc(size)
int size;
{
  DATA_MSG_PTR dataMsg;

  dataMsg = (DATA_MSG_PTR)tcaMalloc((unsigned)size);

  DMTotal++;
  
  if (size > DMmax) DMmax = size;
  if (size < DMmin) DMmin = size;

  return (dataMsg);
}

/* 9-Jul-91: Reid: Don't bother saving dataMsg's */
#if 0
{
  DATA_MSG_BUF_PTR dataMsgBuf;

  bufferToAlloc = NULL;
  (void)listIterateFromFirst(findUnAllocatedSize, (char *)&size, dataMsgBufferList);
  dataMsgBuf = bufferToAlloc;

  if (!dataMsgBuf) {
    dataMsgBuf = dataMsgBufferCreate(size);
    listInsertItem((char *)dataMsgBuf, dataMsgBufferList);
  }

  dataMsgBuf->allocated = TRUE;

  DMTotal++;
  
  if (size > DMmax)
    DMmax = size;

  if (size < DMmin)
    DMmin = size;

  return(dataMsgBuf->dataMsg);
}
#endif


/*****************************************************************************/
/* Currently not used (RGS: 11/11/92)
int findDataMsg(dataMsg, dataMsgBuf)
DATA_MSG_PTR dataMsg;
DATA_MSG_BUF_PTR dataMsgBuf;
{
  if (dataMsgBuf->dataMsg == dataMsg)
    return TRUE;
  else
    return FALSE;
}
*/
/*****************************************************************************/

void dataMsgDeAlloc(dataMsg)
DATA_MSG_PTR dataMsg;
{
  tcaFree((char *)dataMsg);

  DMFree++;
}

/* 9-Jul-91: Reid: Don't bother saving dataMsg's */
#if 0
{
  DATA_MSG_BUF_PTR dataMsgBuf;
  
  dataMsgBuf = (DATA_MSG_BUF_PTR)listMemReturnItem(findDataMsg, dataMsg,
						   dataMsgBufferList);

  if (!dataMsgBuf)
    tcaModError("dataMsgDeAlloc: missing dataMsgBuf");

  if (dataMsgBuf->allocated)
    dataMsgBuf->allocated = FALSE;
  else {
    fprintf(stderr,
	    "WARNING: dataMsgDeAlloc: called on unallocated dataMsg buffer.\n");
#ifdef LISP
    (void)fflush(stderr);
#endif
  }

  DMFree++;
}
#endif

/******************************************************************************
*
* FUNCTION: TCA_RETURN_STATUS_TYPE dataMsgRecv(sd, dataMsg)
*
* DESCRIPTION:
* Creates a dataMsg and initialize it with information from the socket
* specified by the value sd. If there is an error in reading the
* information dataMsg will be initialized to NULL. The communication
* status is returned.
*
* INPUTS:
* int sd;
* DATA_MSG_PTR *dataMsg;
*
* OUTPUTS: TCA_RETURN_STATUS_TYPE
*
******************************************************************************/

TCA_RETURN_STATUS_TYPE dataMsgRecv(sd, dataMsg)
int sd;
DATA_MSG_PTR *dataMsg;
{
  char *data;
  int amount;
  TCA_RETURN_STATUS_TYPE status;

  struct {
    int classTotal;
    int msgTotal;
  } totals;

  *dataMsg = NULL;

  status = readNBytes(sd, (char *)&totals, 2*sizeof(int));
  if (status != StatOK) {
    *dataMsg = NULL;
    return status;
  }

  *dataMsg = dataMsgAlloc((totals.classTotal+totals.msgTotal + sizeof(int) 
			   + 2*sizeof(char *) + 7*sizeof(int)));
  (*dataMsg)->refCount = 0; 

  (*dataMsg)->classTotal = totals.classTotal;
  (*dataMsg)->msgTotal = totals.msgTotal;

  (*dataMsg)->msgData = (char *)*dataMsg + sizeof(int) + 2*sizeof(char *) + 
                        7*sizeof(int);
  (*dataMsg)->classData = (char *)*dataMsg + sizeof(int) + 2*sizeof(char *) + 
                          7*sizeof(int) + totals.msgTotal;

  data = (char *)*dataMsg + sizeof(int) + 2*sizeof(char *) + 2*sizeof(int);

  amount = totals.classTotal + totals.msgTotal + 5*sizeof(int);

  status = readNBytes(sd, data, amount);

  return status;
}


/******************************************************************************
*
* FUNCTION: TCA_RETURN_STATUS_TYPE dataMsgSend(sd, dataMsg)
*
* DESCRIPTION:
* Sends the information in dataMsg to the socket specified by sd. The
* status of the communication will be returned.
*
* INPUTS:
* int sd;
* DATA_MSG_PTR dataMsg;
*
* OUTPUTS: TCA_RETURN_STATUS_TYPE
*
******************************************************************************/

TCA_RETURN_STATUS_TYPE dataMsgSend(sd, dataMsg)
int sd;
DATA_MSG_PTR dataMsg;
{
  int amount;
  char *sendInfo;

  amount = dataMsg->msgTotal + dataMsg->classTotal + 7*sizeof(int);
  
  sendInfo = (char *)dataMsg + sizeof(int) + 2*sizeof(char *);

  return(writeNBytes(sd, sendInfo, amount));
}


/******************************************************************************
*
* FUNCTION: 
*
* DESCRIPTION:
* Creates a new dataMsg and encodes data for sending based on Format. If
* data, or Format or both are NULL then a new dataMsg is still created
* but only Header Information is initialized.
*
* INPUTS: 
*
* OUTPUTS:
*
******************************************************************************/

DATA_MSG_PTR dataMsgCreate(parentRef, intent, classId, dispatchRef, msgRef,
			   msgFormat, msgData, classFormat, classData)
int parentRef;
int intent;
int classId;
int dispatchRef;
int msgRef;
FORMAT_PTR msgFormat, classFormat;
char *msgData, *classData;
{
  BUFFER_TYPE buffer;
  DATA_MSG_PTR dataMsg;
  int classTotal, msgTotal;

  msgTotal = 0;

  if (msgData && msgFormat) {
    if (msgData == &LispFlagGlobal) 
      (*lispBufferSizeGlobal)(&msgTotal, msgFormat, lisp_value(0));
    else
      msgTotal = bufferSize(msgFormat, msgData);      
  }

  if (classData && classFormat)
    classTotal = bufferSize(classFormat, classData);
  else 
    classTotal = 0;

  dataMsg = dataMsgAlloc(classTotal + msgTotal + sizeof(int) +
			 2*sizeof(char *) + 7*sizeof(int));

  dataMsg->refCount = 0;

  dataMsg->msgTotal = msgTotal;
  dataMsg->classTotal = classTotal;
  
  dataMsg->parentRef = parentRef;
  dataMsg->intent = intent;
  dataMsg->classId = classId;
  dataMsg->dispatchRef = dispatchRef;
  dataMsg->msgRef = msgRef;

  if (msgTotal) {
    dataMsg->msgData = (char *)dataMsg + sizeof(int) + 2*sizeof(char *) +
                       7*sizeof(int);

    if (msgData == &LispFlagGlobal) {
      buffer.buffer = dataMsg->msgData;
      buffer.bstart = 0;
      (*lispEncodeMsgGlobal)(msgFormat, lisp_value(0), &buffer);
    } else
      encodeData(msgFormat, msgData, dataMsg->msgData, 0);
  }
  else
    dataMsg->msgData = NULL;

  if (classTotal) {
    dataMsg->classData = (char *)dataMsg + sizeof(int) + 2*sizeof(char *) +
                         7*sizeof(int) + msgTotal;
    encodeData(classFormat, classData, dataMsg->classData, 0);
  }
  else
    dataMsg->classData = NULL;

  return dataMsg;
}


/******************************************************************************
*
* FUNCTION: char *dataMsgDecodeMsg(msgFormat, dataMsg)
*
* DESCRIPTION:
* Returns the message data defined by msgFormat from the information
* encoded in dataMsg. If there is no data to decode, dataMsgDecodeMsg
* returns NULL.
*
* INPUTS:
* FORMAT_PTR msgFormat; 
* DATA_MSG_PTR dataMsg;
*
* OUTPUTS: char *
*
******************************************************************************/

char *dataMsgDecodeMsg(msgFormat, dataMsg)
FORMAT_PTR msgFormat; 
DATA_MSG_PTR dataMsg;
{ 
  if (!msgFormat || !dataMsg->msgTotal) 
    return NULL;
  else 
    return decodeData(msgFormat, dataMsg->msgData, 0);
}


/******************************************************************************
*
* FUNCTION: char *dataMsgDecodeClass(classFormat, dataMsg)
*
* DESCRIPTION:
* Returns the message data defined by classFormat from the information
* encoded in dataMsg. If there is no data to decode, dataMsgDecodeMsg
* returns NULL.
*
* INPUTS:
* FORMAT_PTR classFormat;
* DATA_MSG_PTR dataMsg;
*
* OUTPUTS: char *
*
******************************************************************************/

char *dataMsgDecodeClass(classFormat, dataMsg)
FORMAT_PTR classFormat; 
DATA_MSG_PTR dataMsg;
{ 
  if (!classFormat || !dataMsg->classTotal) 
    return NULL;
  else 
    return decodeData(classFormat, dataMsg->classData, 0);
}


/******************************************************************************
*
* FUNCTION: void dataMsgFree(dataMsg)
*
* DESCRIPTION: Recycle a datamsg.
*
* INPUTS: DATA_MSG_PTR dataMsg
*
* OUTPUTS: none.
*
******************************************************************************/

void dataMsgFree(dataMsg)
DATA_MSG_PTR dataMsg;
{
  if (dataMsg) {
    dataMsg->refCount--;
    if (dataMsg->refCount < 1) {
      dataMsgDeAlloc(dataMsg);
    }
  }
}


/******************************************************************************
*
* FUNCTION: int dataMsgTestMsgData(dataMsg)
*
* DESCRIPTION: 
* Returns TRUE if the dataMsg number of message bytes is greater than zero.
* Returns FALSE otherwise or if dataMsg is NULL.
*
* INPUTS: DATA_MSG_PTR dataMsg
*
* OUTPUTS: int.
*
******************************************************************************/

int dataMsgTestMsgData(dataMsg)
DATA_MSG_PTR dataMsg;
{
  if (dataMsg && dataMsg->msgTotal > 0)
    return 1;
  else
    return 0;
}


/******************************************************************************
*
* FUNCTION: DATA_MSG_PTR dataMsgReplaceClassData(format, data, dataMsg)
*
* DESCRIPTION:
* Replaces class information on an existing dataMsg with data.
*
* INPUTS:
* FORMAT_PTR format;
* char *data;
* DATA_MSG_PTR dataMsg;
*
* OUTPUTS: DATA_MSG_PTR
*
******************************************************************************/

DATA_MSG_PTR dataMsgReplaceClassData(format, data, dataMsg)
FORMAT_PTR format;
char *data;
DATA_MSG_PTR dataMsg;
{
  DATA_MSG_PTR newDataMsg;
  int existingTotal, classTotal;

  existingTotal = dataMsg->msgTotal+sizeof(int)+2*sizeof(char *)+7*sizeof(int);

  if (data && format)
    classTotal = bufferSize(format, data);
  else
    classTotal = 0;

  newDataMsg = dataMsgAlloc(existingTotal+classTotal);

  /* 3-Sep-90: fedor: this should work because class info is in the
     dataMsg struture as the last item. */

  bcopy((char *)dataMsg, (char *)newDataMsg, existingTotal);

  if (classTotal) {
    newDataMsg->classData = (char *)newDataMsg+sizeof(int)+2*sizeof(char *) +
      7*sizeof(int) + dataMsg->msgTotal;
    encodeData(format, data, newDataMsg->classData, 0);
  }
  else
    newDataMsg->classData = NULL;

  /* reset msgData pointer */
  if (newDataMsg->msgTotal) 
    newDataMsg->msgData = (char *)dataMsg+sizeof(int)+2*sizeof(char *)+
      7*sizeof(int);

  /* set refCount */
  newDataMsg->refCount = 0;

  newDataMsg->classTotal = classTotal;

  dataMsgFree(dataMsg);

  return newDataMsg;
}


/******************************************************************************
*
* FUNCTION: void dataMsgDisplayStats()
*
* DESCRIPTION:
* Display datamsg memory usage stats.
*
* INPUTS: none.
*
* OUTPUTS: none.
*
******************************************************************************/

int sizeDM;
int numAllocatedDM;

int countDataMessages(allocatedP, dataMsgBuf)
     int allocatedP; DATA_MSG_BUF_PTR dataMsgBuf;
{
  if (dataMsgBuf->allocated == allocatedP) {
    numAllocatedDM++;
    sizeDM += dataMsgBuf->size;
  }
  return TRUE;
}

void dataMsgDisplayStats()
{
  fprintf(stderr, "Data Msg Buffer Stats\n");
  fprintf(stderr, "Total Number of Allocs  : %d\n", (char *)DMTotal);
  fprintf(stderr, "Total Number of Frees  : %d\n", (char *)DMFree);
  fprintf(stderr, "Min Request  : %d\n", (char *)DMmin);
  fprintf(stderr, "Max Request  : %d\n", (char *)DMmax);
  fprintf(stderr, "Buf List Size: %d\n", (char *)listLength(dataMsgBufferList));

  numAllocatedDM = sizeDM = 0;
  (void)listIterateFromFirst(countDataMessages, (char *)TRUE, 
			     dataMsgBufferList);
  fprintf(stderr, "  %d allocated, with %d total bytes\n", 
	  (char *)numAllocatedDM, (char *)sizeDM);
  numAllocatedDM = sizeDM = 0;
  (void)listIterateFromFirst(countDataMessages, (char *)FALSE, 
			     dataMsgBufferList);
  fprintf(stderr, "  %d deallocated, with %d total bytes\n", 
	  (char *)numAllocatedDM, (char *)sizeDM);

#ifdef LISP
  (void)fflush(stderr);
#endif
}

/* Give some more details */
void dataMsgStats()
{
  DATA_MSG_BUF_PTR dataMsgBuf;

  dataMsgDisplayStats();
  dataMsgBuf = (DATA_MSG_BUF_PTR)listFirst(dataMsgBufferList);
  while (dataMsgBuf) {
    fprintf(stderr, "Size: %d; Allocated: %d\n", 
	    (char *)dataMsgBuf->size, (char *)dataMsgBuf->allocated);
    dataMsgBuf = (DATA_MSG_BUF_PTR)listNext(dataMsgBufferList);
  }
}

/******************************************************************************
*
* FUNCTION: TCA_RETURN_VALUE_TYPE tcaAllocateDataBuffer()
*
* DESCRIPTION:
* Pre-allocate a data message buffer of the requested size
*
* INPUTS: int size -- number of bytes to pre-allocate.
*
* OUTPUTS: Returns "Success" if the buffer could be allocated, 
*          "Failure" otherwise
*
* NOTES: Currently, since "dataMsgBufferCreate" has no return value, 
*        "tcaAllocateDataBuffer" always returns Success (or bombs
*        out within "dataMsgBufferCreate").
******************************************************************************/
TCA_RETURN_VALUE_TYPE tcaAllocateDataBuffer(size)
     int size;
{
  DATA_MSG_BUF_PTR dataMsgBuf;

  /* Creates an unallocated data message buffer */
  dataMsgBuf = dataMsgBufferCreate(size);
  listInsertItem((char *)dataMsgBuf, dataMsgBufferList);

  if (size > DMmax) DMmax = size;
  if (size < DMmin) DMmin = size;

  return Success;
}
