/******************************************************************************
*
* PROJECT: Carnegie Mellon Planetary Rover Project
*          Task Control Architecture 
* 
* (c) Copyright 1991 Christopher Fedor and Reid Simmons.  All rights reserved.
* 
* MODULE: exception handling
*
* FILE: excepMod.c
*
* ABSTRACT:
* 
* module exception code.
*
* REVISION HISTORY
*
* 31-Dec-91 Reid Simmons, School of Computer Science, CMU
* Added facility to add exception handlers to class of messages, 
*  rather than just a single task tree node.
*
* 21-May-91 Christopher Fedor, School of Computer Science, CMU
* Revised to verion tca 5.x
*
* 25-Nov-89 Long-Ji Lin, School of Computer Science, CMU
* created.
*
******************************************************************************/

#include "stdio.h"

#include "tcaInternal.h"
#include "sendMsg.h"

#include "exception.h"

extern int responseIssuedGlobal;


/******************************************************************************
*
* FUNCTION: void tcaAddExceptionHandler(ref, hndName)
*
* DESCRIPTION: Adds the named handler to the task tree at node ref.
*
* INPUTS:
* TCA_REF_PTR ref;
* char *hndName;
*
* OUTPUTS: void.
*
******************************************************************************/

void tcaAddExceptionHandler(ref, hndName)
TCA_REF_PTR ref;
char *hndName;
{
  MSG_PTR msg;
  ADD_EXC_TYPE addInfo;

  addInfo.ref = ref;
  addInfo.hndName = hndName;

  msg = msgFind("addExcepMsg");
  (void)sendMessage((TCA_REF_PTR)NULL, msg, (char *)&addInfo, (char *)NULL);
}


/******************************************************************************
*
* FUNCTION: void tcaAddExceptionHandlerToMessage(msgName, hndName)
*
* DESCRIPTION: Adds the named exception handler to *all* named message instances.
*
* INPUTS:
* char *msgName;
* char *hndName;
*
* OUTPUTS: void.
*
******************************************************************************/

void tcaAddExceptionHandlerToMessage(msgName, hndName)
char *msgName, *hndName;
{
  MSG_PTR msg;
  ADD_EXC_MSG_TYPE addInfo;

  addInfo.msgName = msgName;
  addInfo.hndName = hndName;

  msg = msgFind("addExceptions");
  (void)sendMessage((TCA_REF_PTR)NULL, msg, (char *)&addInfo, (char *)NULL);
}


/******************************************************************************
*
* FUNCTION: void tcaGetExceptionInfo(ref, info, selection)
*
* DESCRIPTION: Return information about an exception.
*
* INPUTS:
*
* OUTPUTS: void.
*
* NOTES: 
* 12-May-91: fedor: This should be passed in automaticaly on an exception.
*
******************************************************************************/

void tcaGetExceptionInfo(ref, info)
TCA_REF_PTR ref;
EXC_INFO_PTR info;
{
  (void)tcaQuery("getInfoExcepMsg", (void *)&ref, (void *)info);
}

#ifdef LISP

EXC_INFO_PTR tcaGetExcepC(ref)
TCA_REF_PTR ref;
{
  EXC_INFO_PTR info;

  info = NEW(EXC_INFO_TYPE);
  (void)tcaQuery("getInfoExcepMsg", (char *)&ref, (char *)info);
  
  return info;
}

TCA_REF_PTR tcaGetERef(info)
EXC_INFO_PTR info;
{
  return info->ERef;
}

TCA_REF_PTR tcaGetCRef(info)
EXC_INFO_PTR info;
{
  return info->CRef;
}

int tcaGetRetries(info)
EXC_INFO_PTR info;
{
  return info->numOfRetries;
}

void tcaFreeExcInfo(info)
EXC_INFO_PTR info;
{
  cfree(info);
}

#endif


/******************************************************************************
*
* FUNCTION: void tcaRetry(ref)
*
* DESCRIPTION: Kill the task tree under node ref and resend message at of ref.
*
* INPUTS:
* TCA_REF_PTR ref;
*
* OUTPUTS: void.
*
******************************************************************************/

void tcaRetry(ref)
TCA_REF_PTR ref;
{
  MSG_PTR msg;

  responseIssuedGlobal = TRUE;

  msg = msgFind("retryExcepMsg");
  (void)sendMessage((TCA_REF_PTR)NULL, msg, (char *)&ref, (char *)NULL);
}


/******************************************************************************
*
* FUNCTION: void tcaByPassException(ref)
*
* DESCRIPTION: Notify the server that the exception can not be handled.
* The exception will then be passed to the next exception or propagate up
* the task tree.
*
* Should only be called from an exception handler.
*
* INPUTS:
* TCA_REF_PTR ref;
*
* OUTPUTS: void.
*
* NOTES: 12-May-91: fedor: How about a call to tcaFailure instead?
*
******************************************************************************/

void tcaByPassException(ref)
TCA_REF_PTR ref;
{
  MSG_PTR msg;

  responseIssuedGlobal = TRUE;

  msg = msgFind("byPassExcepMsg");
  (void)sendMessage((TCA_REF_PTR)NULL, msg, (char *)&ref, (char *)NULL);
}
