/*****************************************************************************
* PROJECT: Carnegie Mellon Planetary Rover Project
*          Task Control Architecture
*
* (c) Copyright 1991 Christopher Fedor and Reid Simmons.  All rights reserved.
* 
* MODULE: formatters
*
* FILE: formatters.c
*
* ABSTRACT:
*
* Data Format Routines.
*
* REVISION HISTORY:
*
*  9-Jul-91 Christopher Fedor, School of Computer Science, CMU
* Removed isIntFormat test routine because we now have string formats to test
* and this allows testing of formats without first parsing it.
*
*  2-Jul-91 Reid Simmons at School of Computer Science, CMU
* Added code for freeing class data.
*
* 30-Jan-91 Christopher Fedor, School of Computer Science, CMU
* Added fflush(stdout) to printf for module code calls from lisp
*
* 23-Oct-90 Christopher Fedor at School of Computer Science, CMU
* Moved isIntFormat from mon.c to here - it logically belongs here 
* even though it increases module code a little.
* format needs a complete rewrite! flush SIZES_TYPE, flush the union
* - saves little space and makes debugging very difficult ... additional 
* problems.
*
* 15-Oct-90 Christopher Fedor at School of Computer Science, CMU
* Moved tcaRegisterLengthFormatter and tcaRegisterNamedFormatter
* to module routines.
*
* 18-Jul-90 Reid Simmons at School of Computer Science, CMU
* Added code for freeing data structures.
*
* 10-Apr-90 Christopher Fedor at School of Computer Science, CMU
* Changed ifdef SUN3 to ifndef SUN4.
*
*  9-Apr-90 Reid Simmons at School of Computer Science, CMU
* Modified "alignField" to work for both Sun3 and Sun4 (Sparc) versions.
*
*  2-Apr-90 Christopher Fedor at School of Computer Science, CMU
* Revised to software standards.
*
* 11-Mar-89 Christopher Fedor at School of Computer Science, CMU
* Added tcaRegisterLengthFormatter
*
*  9-Feb-89 Reid Simmons at School of Computer Science, CMU
* Split into 2 files -- "PrimFmttrs" and "Formatters"; changed how 
* structured formatters are represented.  Added parser to take
* formatter string and create TC_FORMAT_PTR structures.
* INCOMPATIBLE CHANGE WITH MAJOR VERSION 1.x
*
*  6-Feb-89 Reid Simmons at School of Computer Science, CMU
* Added structured formatters.
*
*  3-Feb-89 Reid Simmons at School of Computer Science, CMU
* Standardized C and LISP interface to formatter functions.
*
*  2-Jan-89 Reid Simmons at School of Computer Science, CMU
* (1.1) Added INT and FLOAT formats for LISP interface.
*
*    Dec-88 Christopher Fedor at School of Computer Science, CMU
* created.
*
*****************************************************************************/


#include "stdio.h" 

#include "tcaInternal.h"
#include "hash.h"
#include "formatters.h"
#include "logging.h"

#ifdef LISP
#include "lisp.h"
#endif


typedef struct{
  int buffer, data;
} SIZES_TYPE;

extern int STR_Trans();
extern int FORMAT_Trans();

extern int UCMAT_Trans();
extern int CMAT_Trans();
extern int SMAT_Trans();
extern int IMAT_Trans();
extern int LMAT_Trans();
extern int FMAT_Trans();
extern int DMAT_Trans();

extern int SIUCMAT_Trans();
extern int SICMAT_Trans();
extern int SISMAT_Trans();
extern int SIIMAT_Trans();
extern int SILMAT_Trans();
extern int SIFMAT_Trans();
extern int SIDMAT_Trans();

extern int INT_Trans();
extern int CHAR_Trans();
extern int SHORT_Trans();
extern int LONG_Trans();

extern int FLOAT_Trans();
extern int BOOLEAN_Trans();
extern int DOUBLE_Trans();
extern int BYTE_Trans();
extern int TWOBYTE_Trans();

extern int TCA_REF_PTR_Trans();

HASH_TABLE_PTR formatNamesTable;

extern HASH_TABLE_PTR classFormatTable;
     
int (*TransTable[MAXFORMATTERS+1])();

/* If there is no formatter, it means that the ptr is recursive (self-ptr) */
#define CHOOSE_PTR_FORMAT(format, parentFormat) \
  (((format)->formatter.f) ? ((format)->formatter.f) : parentFormat)

#define ODDP(x) ((x) & 1)

#ifdef LISP

void blockCopyToArray(buffer, array, amount)
BUFFER_PTR buffer;
char *array;
int amount;
{
  bcopy(buffer->buffer+buffer->bstart, array, amount);
  buffer->bstart += amount;
}

void blockCopyFromArray(buffer, array, amount)
BUFFER_PTR buffer;
char *array;
int amount;
{
  bcopy(array, buffer->buffer+buffer->bstart, amount);
  buffer->bstart += amount;
}

int tcaRefNameLength(ref)
TCA_REF_PTR ref;
{
  if (ref && ref->name)
    return strlen(ref->name);
  else
    return 0;
}

char *tcaRefNameLisp(ref, length)
TCA_REF_PTR ref;
int length;
{
  bcopy(ref->name, (char *)Vecdata(SymbolValue(lisp_value(0))), length);

  return (char *)SymbolValue(lisp_value(0));
}

extern char LispFlagGlobal;

char *returnLispFlagGlobal()
{
  return &LispFlagGlobal;
}

void setIntValue(item, value)
int *item, value;
{
  *item = value;
}

int formatPrimitiveProc(format)
FORMAT_PTR format;
{
  return(format->formatter.i);
}

int formatType(format)
FORMAT_PTR format;
{
  return(format->type);
}

FORMAT_PTR formatChoosePtrFormat(format, parentFormat)
FORMAT_PTR format, parentFormat;
{
  return(CHOOSE_PTR_FORMAT(format, parentFormat));
}

/* 14-Feb-91: fedor: format array sillyness from StructFMT loops */

FORMAT_ARRAY_PTR formatFormatArray(format)
FORMAT_PTR format;
{
  return(format->formatter.a);
}

int formatFormatArrayMax(formatArray)
FORMAT_ARRAY_PTR formatArray;
{
  return(formatArray[0].i);
}

FORMAT_PTR formatFormatArrayItem(formatArray, i)
FORMAT_ARRAY_PTR formatArray;
int i;
{
  return(formatArray[i].f);
}

int formatGetInt(buffer)
BUFFER_PTR buffer;
{
  int i;

  bcopy(buffer->buffer+buffer->bstart, &i, sizeof(int));
  buffer->bstart += sizeof(int);

  return i;
}

char formatGetChar(buffer)
BUFFER_PTR buffer;
{
  char c;

  bcopy(buffer->buffer+buffer->bstart, &c, sizeof(char));
  buffer->bstart += sizeof(char);

  return c;
}

double formatGetDouble(buffer)
BUFFER_PTR buffer;
{
  double d;

  bcopy(buffer->buffer+buffer->bstart, &d, sizeof(double));
  buffer->bstart += sizeof(double);

  return d;
}

void formatPutInt(i, buffer)
int i;
BUFFER_PTR buffer;
{
  bcopy(&i, buffer->buffer+buffer->bstart, sizeof(int));
  buffer->bstart += sizeof(int);
}

void formatPutChar(c, buffer)
char c;
BUFFER_PTR buffer;
{
  bcopy(&c, buffer->buffer+buffer->bstart, sizeof(char));
  buffer->bstart += sizeof(char);
}

float formatGetFloat(buffer)
BUFFER_PTR buffer;
{
  float f;

  buffer->bstart += 
    FLOAT_Trans(Decode, &f, 0, buffer->buffer, buffer->bstart);

  return f;
}

void formatPutFloat(f, buffer)
float f;
BUFFER_PTR buffer;
{
  /* 14-Mar-91: fedor: this sillyness forces a pointer to a float
     without changing that float into a double. This is needed
     because FLOAT_Trans will transfer only 4 bytes but the conversion
     to a double will only transfer the first 4 bytes of the correct value. */

  float f2;

  f2 = f;
  buffer->bstart += 
    FLOAT_Trans(Encode, &f2, 0, buffer->buffer, buffer->bstart);
}


void formatPutDouble(d, buffer)
double d;
BUFFER_PTR buffer;
{
  bcopy(&d, buffer->buffer+buffer->bstart, sizeof(double));
  buffer->bstart += sizeof(double);
}

#endif

FORMAT_PTR createIntegerFormat(type, integer)
FORMAT_CLASS_TYPE type; 
int integer;
{ 
  FORMAT_PTR format;

  format = NEW_FORMATTER();
  format->type = type;
  format->formatter.i = integer;
  return format;
}

FORMAT_PTR createPrimitiveFormat(format_number) 
int format_number;
{ 
  return createIntegerFormat(PrimitiveFMT, format_number);
}

int formatterHashFN(key)
char *key;
{ 
  int i, sum;

  for(i=0, sum=0; key[i] != '\0'; i++) 
    sum += key[i];
  return sum;
}

int formatterEqFN(key1, key2) 
char *key1, *key2;
{ 
  return(STREQ(key1, key2));
}

/* KeyLength is calculated assuming "name" is a null-terminated string */
void addFormatToTable(name, format)
char *name; 
FORMAT_PTR format;
{ 
  (void)hashTableInsert((char *)name, 1+strlen(name), 
			(char *)format, formatNamesTable);
}


/*****************************************************************************
*
* FUNCTION: void registerPrimitiveFormat(formatter, formatNumber, 
*                                           translationFunction)
*
* DESCRIPTION:
*
* INPUTS:
* char *formatter;
* int formatNumber, (*translationFunction)();
*
* OUTPUTS: void.
*
*****************************************************************************/

void registerPrimitiveFormat(formatter, formatNumber, translationFunction)
char *formatter;
int formatNumber, (*translationFunction)();
{ 
  if ((formatNumber < 1) || (formatNumber > MAXFORMATTERS)) {
    TCA_MOD_ERROR2("RegisterPrimitiveFormatter: 0 < formatter number: %d < %d\n",
		   formatNumber, (MAXFORMATTERS+1));
  } else if (TransTable[formatNumber]) {
    TCA_MOD_ERROR1("RegisterFormatter: Formatter number %d already in use", 
		   formatNumber);
  } else {
    TransTable[formatNumber] = translationFunction;
    addFormatToTable(formatter, createPrimitiveFormat(formatNumber));
  }
}


/*****************************************************************************
*
* FUNCTION: void formatInitialize()
*
* DESCRIPTION:
*
* INPUTS: none.
*
* OUTPUTS: void.
*
*****************************************************************************/

void formatInitialize()
{ 
  int i;
  
  formatNamesTable = hashTableCreate(50, formatterHashFN, formatterEqFN);
  
  for(i=0; i< MAXFORMATTERS; i++) 
    TransTable[i] = NULL;
    
  registerPrimitiveFormat("string", STR_FMT, STR_Trans);
  registerPrimitiveFormat("char", CHAR_FMT, CHAR_Trans);
  registerPrimitiveFormat("short", SHORT_FMT, SHORT_Trans);
  registerPrimitiveFormat("long", LONG_FMT, LONG_Trans);
  registerPrimitiveFormat("int", INT_FMT, INT_Trans);
  registerPrimitiveFormat("float", FLOAT_FMT, FLOAT_Trans);
  registerPrimitiveFormat("boolean", BOOLEAN_FMT, BOOLEAN_Trans);
  registerPrimitiveFormat("double", DOUBLE_FMT, DOUBLE_Trans);
  registerPrimitiveFormat("byte", BYTE_FMT, BYTE_Trans);
  registerPrimitiveFormat("twobyte", TWOBYTE_FMT, TWOBYTE_Trans);

  registerPrimitiveFormat("format", FORMAT_FMT, FORMAT_Trans);

  registerPrimitiveFormat("ucmat", UCMAT_FMT, UCMAT_Trans);
  registerPrimitiveFormat("cmat", CMAT_FMT, CMAT_Trans);
  registerPrimitiveFormat("smat", SMAT_FMT, SMAT_Trans);
  registerPrimitiveFormat("imat", IMAT_FMT, IMAT_Trans);
  registerPrimitiveFormat("lmat", LMAT_FMT, LMAT_Trans);
  registerPrimitiveFormat("fmat", FMAT_FMT, FMAT_Trans);
  registerPrimitiveFormat("dmat", DMAT_FMT, DMAT_Trans);

  registerPrimitiveFormat("siucmat", SIUCMAT_FMT, SIUCMAT_Trans);
  registerPrimitiveFormat("sicmat", SICMAT_FMT, SICMAT_Trans);
  registerPrimitiveFormat("sismat", SISMAT_FMT, SISMAT_Trans);
  registerPrimitiveFormat("siimat", SIIMAT_FMT, SIIMAT_Trans);
  registerPrimitiveFormat("silmat", SILMAT_FMT, SILMAT_Trans);
  registerPrimitiveFormat("sifmat", SIFMAT_FMT, SIFMAT_Trans);
  registerPrimitiveFormat("sidmat", SIDMAT_FMT, SIDMAT_Trans);

  registerPrimitiveFormat("TCA_REF_PTR", TCA_REF_PTR_FMT, TCA_REF_PTR_Trans);

  /* 
    Added 18-Aug-89, used in HERO LISP system. "ubyte" (unsigned byte) is
    treated the same as a regular C byte.
    */
  registerPrimitiveFormat("ubyte", UBYTE_FMT, BYTE_Trans);
}


/*****************************************************************************
*
* FUNCTION: int fixedArraySize(formatArray) 
*
* DESCRIPTION:
* First element in a FORMAT_ARRAY_PTR is the size of the array.
* For a FixedArrayFMT (and VarArrayFMT) the second element is the 
* formatter, the rest of the elements are the dimensions.
*
* INPUTS: FORMAT_ARRAY_PTR formatArray;
*
* OUTPUTS: int
*
*****************************************************************************/

int fixedArraySize(formatArray) 
FORMAT_ARRAY_PTR formatArray;
{
  int i, arraySize;

  for(i=2, arraySize=1; i<formatArray[0].i; i++)
    arraySize *= formatArray[i].i;
  return arraySize;
}


/*****************************************************************************
*
* FUNCTION: int varArraySize(formatArray, parentFormat, dataStruct, dStart)
*
* DESCRIPTION:
*
* INPUTS:
* FORMAT_ARRAY_PTR formatArray; 
* FORMAT_PTR parentFormat;
* DATA_PTR dataStruct; 
* int dStart;
*
* OUTPUTS: int
*
*****************************************************************************/

int varArraySize(formatArray, parentFormat, dataStruct, dStart)
FORMAT_ARRAY_PTR formatArray; 
FORMAT_PTR parentFormat;
DATA_PTR dataStruct; 
int dStart;
{ 
  int i, j, arraySize, size;
  FORMAT_ARRAY_PTR parentStructArray;
  int currentPlace, sizePlace, foundPlace, offset;

  parentStructArray = parentFormat->formatter.a;
  foundPlace = 0;
  for(currentPlace=1;!foundPlace;currentPlace++)
    foundPlace = ((parentStructArray[currentPlace].f->type == VarArrayFMT) &&
		  (parentStructArray[currentPlace].f->formatter.a
		   == formatArray));
  currentPlace--;
  
  for(i=2, arraySize=1; i < formatArray[0].i; i++) {
    sizePlace = formatArray[i].i;
    offset = 0;
    if (currentPlace < sizePlace)
      for (j=currentPlace; j < sizePlace; j++)
	offset += dataStructureSize(parentStructArray[j].f);
    else 
      for (j=sizePlace; j < currentPlace; j++)
	offset -= dataStructureSize(parentStructArray[j].f);
    bcopy(dataStruct+dStart+offset, (char *)&size, sizeof(int));
    arraySize *= size;
  }

  return arraySize;
}


/*****************************************************************************
*
* FUNCTION: int elementSize(format) 
*
* DESCRIPTION:
* Returns the size (ALength) of the format's element.
* If the format is a structured type, returns 0 unless all the elements
* have the same length.
*
* INPUTS: FORMAT_PTR format;
*
* OUTPUTS: int
*
*****************************************************************************/

int elementSize(format)
FORMAT_PTR format;
{
  int firstSize, i;

  switch (format->type) {
  case LengthFMT: 
    return format->formatter.i;
  case PrimitiveFMT:
    return (TransTable[format->formatter.i])(ALength, NULL, 0, NULL, 0);
  case PointerFMT:
  case VarArrayFMT: 
    return sizeof(DATA_PTR);
  case FixedArrayFMT: 
    return elementSize(format->formatter.a[1].f);
  case StructFMT:
    firstSize = elementSize(format->formatter.a[1].f);
    if (firstSize != 0) {
      for (i=2; i<format->formatter.a[0].i; i++) {
	if (firstSize != elementSize(format->formatter.a[i].f))
	  return 0;
      }
    }
    return firstSize;
  default: return 0;
  }
}


/*****************************************************************************
*
* FUNCTION: int fixedLengthFormat(format) 
*
* DESCRIPTION:
* Returns TRUE (1) if the Format contains no pointers.
* (i.e., it is composed only of simple types).
*
* INPUTS: FORMAT_PTR format;
*
* OUTPUTS: int
*
*****************************************************************************/

int fixedLengthFormat(format) 
FORMAT_PTR format;
{ 
  int i, subelementSize;

  switch(format->type) {
  case LengthFMT: 
    return TRUE;
  case PrimitiveFMT:
    return (TransTable[format->formatter.i])(SimpleType, NULL, 0, NULL, 0);
  case PointerFMT:
  case VarArrayFMT: 
    return FALSE;
  case FixedArrayFMT: 
    return fixedLengthFormat(format->formatter.a[1].f);
  case StructFMT:
    /* For compatibility between Sun3's and Sun4's, in addition to all
     *  being simple types, the subelements of a structure must be the
     *  same size in order for the structure to be considered "fixed length" 
     */
    for (i=1; i<format->formatter.a[0].i; i++) {
      if (!fixedLengthFormat(format->formatter.a[i].f)) {
	return FALSE;
      }
      else if (i==1) { /* First element */
	subelementSize = elementSize(format->formatter.a[1].f);
	if (subelementSize == 0) return FALSE;
      }
      else if (subelementSize != elementSize(format->formatter.a[i].f))
	return FALSE;
    }
    return TRUE;
  default: return FALSE;
  }
}


/*****************************************************************************
*
* FUNCTION: int mostRestrictiveElement(format) 
*
* DESCRIPTION:
* Returns the longest element (ALength) of the format.
*
* INPUTS: FORMAT_PTR format;
*
* OUTPUTS: int
*
*****************************************************************************/

#ifdef SUN4
int mostRestrictiveElement(format)
FORMAT_PTR format;
{
  int maxSize, nextSize, i;

  switch (format->type) {
  case LengthFMT: return format->formatter.i;

  case PrimitiveFMT: 
    return (TransTable[format->formatter.i])(RLength, NULL, 0, NULL, 0);

  case PointerFMT:
  case VarArrayFMT: 
    return sizeof(DATA_PTR);

  case FixedArrayFMT: 
    return mostRestrictiveElement(format->formatter.a[1].f);

  case StructFMT:
    maxSize = mostRestrictiveElement(format->formatter.a[1].f);
    for (i=2; i<format->formatter.a[0].i; i++) {
      nextSize = mostRestrictiveElement(format->formatter.a[i].f);
      if (nextSize > maxSize) maxSize = nextSize;
    }
    return maxSize;
  }
}
#endif


/*****************************************************************************
*
* FUNCTION: int alignField(format, currentField, currentDataSize)
*
* DESCRIPTION:
* Returns either "currentDataSize" or 1 + "currentDataSize", to reflect
* how C would align fields in a structure.
*
* Sun3 version:
* This function works on the (empirical) model that the C compiler used on the
* Sun3s aligns fields in structures on word boundaries (unless the next 
* field in the structure is an odd number of bytes).
*
* Sun4 version:
* Fields must be aligned on the "appropriate" boundaries - e.g., ints on 4 byte
* boundaries, doubles on 8 byte boundaries.  Structures are padded at the end 
* to align with the most restrictive (longest) field in the structure.
*
* INPUTS:
* FORMAT_PTR format; 
* int currentField, currentDataSize;
*
* OUTPUTS: 
*
*****************************************************************************/

int alignField(format, currentField, currentDataSize)
FORMAT_PTR format; 
int currentField, currentDataSize;
{ 
#ifndef SUN4
  int nextField;
  FORMAT_PTR nextFormat;
  FORMAT_ARRAY_PTR formatArray; 

  if (!ODDP(currentDataSize)) 
    return currentDataSize;
  else {
    formatArray = format->formatter.a;
    nextField = 1 + currentField;
    if (nextField == formatArray[0].i)
      /* end of structure */
      return 1+currentDataSize;
    else {
      nextFormat = formatArray[nextField].f;
      if (((nextFormat->type == LengthFMT) && 
	   ODDP(nextFormat->formatter.i)) ||
	  ((nextFormat->type == PrimitiveFMT) &&
   ODDP((TransTable[nextFormat->formatter.i])(ALength, NULL, 0, NULL, 0))))
	return currentDataSize;
      else return 1+currentDataSize;
    }
  }
#endif

#ifdef SUN4
  int nextField, appropriateSize, rem;
  FORMAT_ARRAY_PTR formatArray;   

  formatArray = format->formatter.a;
  nextField = 1+currentField;
  if (nextField == formatArray[0].i) {
    /* end of structure; pad to appropriate boundary of longest subelement */
    appropriateSize = mostRestrictiveElement(format);
  }
  else {
    /* on Sparc, element must start on boundary compatible with size of largest
       element within the sub-structure */
    appropriateSize = mostRestrictiveElement(formatArray[nextField].f);
  }
  /* Round up to the next even multiple of "appropriateSize" */
  rem = currentDataSize % appropriateSize;
  if (rem != 0) 
    currentDataSize += appropriateSize - rem;

  return currentDataSize;
#endif
}


/*****************************************************************************
*
* FUNCTION: FORMAT_PTR optimizeFormatter(format) 
*
* DESCRIPTION:
* If the formatter (or any sub-formatter of it) is a fixed length format, 
* replace the definition with the appropriate length format.
* For example, "{{int, float}, string}" gets optimized to "{4, 4, string}"
* and "{int, float, [int:3]}" gets optimized to "20".
*
* Returns the new formatter, or the original one if no change occurs.
*
* INPUTS: FORMAT_PTR format;
*
* OUTPUTS: FORMAT_PTR
*
*****************************************************************************/

FORMAT_PTR optimizeFormatter(format) 
FORMAT_PTR format;
{ 
  int i;
  FORMAT_ARRAY_PTR formatArray;

  if (fixedLengthFormat(format))
    return createIntegerFormat(LengthFMT, dataStructureSize(format));
  else {
    switch(format->type) {
    case StructFMT: 
      formatArray = format->formatter.a;
      for(i=1;i < formatArray[0].i;i++) 
	formatArray[i].f = optimizeFormatter(formatArray[i].f);
      break;
    case PointerFMT:
      if (format->formatter.f)
	format->formatter.f = optimizeFormatter(format->formatter.f);
      break;
    case VarArrayFMT:
    case FixedArrayFMT: 
      format->formatter.a[1].f = optimizeFormatter(format->formatter.a[1].f);
    }
    return format;
  }
}


/*****************************************************************************
*
* FUNCTION: SIZES_TYPE bufferSize1(format, dataStruct, dStart, parentFormat)
*
* DESCRIPTION:
* "ParentFormat" is needed by SelfPtr ("*!") and VarArray ("<..>") formatters.
*  Both these formatters can only be embedded in a Struct format ("{...}".
*
* INPUTS:
* FORMAT_PTR format, parentFormat;
* DATA_PTR dataStruct; 
* int dStart; 
*
* OUTPUTS: SIZES_TYPE
*
*****************************************************************************/

SIZES_TYPE bufferSize1(format, dataStruct, dStart, parentFormat)
FORMAT_PTR format, parentFormat;
DATA_PTR dataStruct; 
int dStart; 
{ 
  SIZES_TYPE sizes;
  int (*formatProc)();
  DATA_PTR structPtr;
  FORMAT_PTR nextFormat;
  FORMAT_ARRAY_PTR formatArray;
  int i, bufferSize, currentData, arraySize, structStart, elements;

  bufferSize = 0;
  currentData = dStart;

  switch (format->type) {
  case LengthFMT:
    bufferSize += format->formatter.i;
    currentData += format->formatter.i;
    break;
  case PrimitiveFMT: 
    formatProc = TransTable[format->formatter.i];
    bufferSize += formatProc(ELength, dataStruct, currentData, NULL, 0);
    currentData += formatProc(ALength, NULL, 0, NULL, 0);
    break;
  case PointerFMT:
    structPtr = REF(DATA_PTR, dataStruct, currentData);
    bufferSize += sizeof(char);
    if (structPtr) { 
      nextFormat = CHOOSE_PTR_FORMAT(format, parentFormat);
      sizes = bufferSize1(nextFormat, structPtr, 0, (FORMAT_PTR)NULL);
      bufferSize += sizes.buffer;
    }
    currentData += sizeof(DATA_PTR);
    break;
  case StructFMT:
    formatArray = format->formatter.a;
    for(i=1;i < formatArray[0].i;i++) {
      sizes = bufferSize1(formatArray[i].f, dataStruct, currentData, format);
      bufferSize += sizes.buffer;
      currentData = alignField(format, i, currentData+sizes.data);
    }
    break;
  case FixedArrayFMT:
    formatArray = format->formatter.a;
    arraySize = fixedArraySize(formatArray);
    nextFormat = formatArray[1].f;
    if (fixedLengthFormat(nextFormat)) {
      elements = arraySize * dataStructureSize(nextFormat);
      bufferSize += elements;
      currentData += elements;
    }
    else {
      for(i=0;i < arraySize;i++) {
	sizes = bufferSize1(nextFormat, dataStruct, currentData, (FORMAT_PTR)NULL);
	bufferSize += sizes.buffer;
	currentData += sizes.data;
      }
    }
    break;
  case VarArrayFMT:
    formatArray = format->formatter.a;
    nextFormat = formatArray[1].f;
    arraySize = varArraySize(formatArray, parentFormat, 
				dataStruct, currentData);
    bufferSize += sizeof(int); /* for the size of the array */
    if (fixedLengthFormat(nextFormat))
      bufferSize += arraySize * dataStructureSize(nextFormat);
    else {
      structPtr = REF(DATA_PTR, dataStruct, currentData);
      structStart = 0;
      for(i=0;i < arraySize;i++) {
	sizes = bufferSize1(nextFormat, structPtr, structStart, (FORMAT_PTR)NULL);
	bufferSize += sizes.buffer;
	structStart += sizes.data;
      }
    }
    currentData += sizeof(int); /* skip over the pointer to the array */
  }

  sizes.buffer = bufferSize;
  sizes.data = currentData - dStart;
  return sizes;
}


/*****************************************************************************
*
* FUNCTION: SIZES_TYPE transferToBuffer(format, dataStruct, dStart, buffer, 
*                                       bStart, parentFormat)
*
* DESCRIPTION:
*
* INPUTS:
* FORMAT_PTR format, parentFormat;
* DATA_PTR dataStruct; 
* int dStart, bStart; 
* DATA_PTR buffer;
*
* OUTPUTS: SIZES_TYPE
*
*****************************************************************************/

SIZES_TYPE transferToBuffer(format, dataStruct, dStart, buffer, 
			    bStart, parentFormat)
FORMAT_PTR format, parentFormat;
DATA_PTR dataStruct; 
int dStart, bStart; 
DATA_PTR buffer;
{ 
  SIZES_TYPE sizes;
  int (*formatProc)();
  DATA_PTR structPtr;
  FORMAT_PTR nextFormat;
  FORMAT_ARRAY_PTR formatArray;
  int i, currentByte, currentData, structStart, arraySize, elements;

  currentByte = bStart;
  currentData = dStart;

  switch(format->type) {
  case LengthFMT:
    TO_BUFFER_AND_ADVANCE(dataStruct+currentData, buffer, currentByte, 
			   format->formatter.i);
    currentData += format->formatter.i;
    break;
  case PrimitiveFMT:
    formatProc = TransTable[format->formatter.i];
    currentByte += formatProc(Encode, dataStruct, currentData,
				buffer, currentByte);
    currentData += formatProc(ALength, NULL, 0, NULL, 0);
    break;
  case PointerFMT:
    structPtr = REF(DATA_PTR, dataStruct, currentData);
    /* Z means data, 0 means NULL*/
    buffer[currentByte] = (structPtr) ? 'Z' : '\0'; 
    
    currentByte += sizeof(char);
    currentData += sizeof(DATA_PTR);
    if (structPtr) { 
      nextFormat = CHOOSE_PTR_FORMAT(format, parentFormat);
      sizes = transferToBuffer(nextFormat, structPtr, 0, 
			       buffer, currentByte, (FORMAT_PTR)NULL);
      currentByte += sizes.buffer;
    }
    break;
  case StructFMT:
    formatArray = format->formatter.a;
    for(i=1;i < formatArray[0].i;i++) {
      sizes = transferToBuffer(formatArray[i].f, dataStruct,
			       currentData, buffer, currentByte, format);
      currentByte += sizes.buffer;
      currentData = alignField(format, i, currentData+sizes.data);
    }
    break;
  case FixedArrayFMT:
    formatArray = format->formatter.a;
    arraySize = fixedArraySize(formatArray);
    nextFormat = formatArray[1].f;
    if (fixedLengthFormat(nextFormat)) {
      elements = arraySize * dataStructureSize(nextFormat);
      bcopy(dataStruct+currentData, buffer+currentByte, elements);
      currentByte += elements;
      currentData += elements;
    }
    else {
      for(i=0;i < arraySize;i++) {
	sizes = transferToBuffer(nextFormat, dataStruct, currentData,
				 buffer, currentByte, (FORMAT_PTR)NULL);
	currentByte += sizes.buffer;
	currentData += sizes.data;
      }
    }
    break;
  case VarArrayFMT:
    formatArray = format->formatter.a;
    arraySize = varArraySize(formatArray, parentFormat,
				dataStruct, currentData);
    nextFormat = formatArray[1].f;
    TO_BUFFER_AND_ADVANCE(&arraySize, buffer, currentByte, sizeof(int));
    structPtr = REF(DATA_PTR, dataStruct, currentData);
    if (fixedLengthFormat(nextFormat)) {
      elements = arraySize * dataStructureSize(nextFormat);
      bcopy(structPtr, buffer+currentByte, elements);
      currentByte += elements;
    }
    else {
      structStart = 0;
      for(i=0;i < arraySize;i++) {
	sizes = transferToBuffer(nextFormat, structPtr, structStart,
				 buffer, currentByte, (FORMAT_PTR)NULL);
	currentByte += sizes.buffer;
	structStart += sizes.data;
      }
    }
    currentData += sizeof(DATA_PTR);
  }

  sizes.buffer = currentByte - bStart;
  sizes.data = currentData - dStart;
  return sizes;  
}


/*****************************************************************************
*
* FUNCTION: int dataStructureSize(format) 
*
* DESCRIPTION:
*
* INPUTS: FORMAT_PTR format;
*
* OUTPUTS: int
*
*****************************************************************************/

int dataStructureSize(format) 
FORMAT_PTR format;
{ 
  int i, size;
  FORMAT_ARRAY_PTR formatArray;

  size = 0;

  switch(format->type) {
  case LengthFMT: 
    size += format->formatter.i; 
    break;
  case PrimitiveFMT:
    size += TransTable[format->formatter.i](ALength, NULL, 0, NULL, 0);
    break;
  case PointerFMT: 
  case VarArrayFMT: 
    size += sizeof(DATA_PTR); 
    break;
  case StructFMT: 
    formatArray = format->formatter.a;
    for(i=1;i < formatArray[0].i;i++) 
      size = alignField(format, i, size+dataStructureSize(formatArray[i].f));
    break;
  case FixedArrayFMT:
    formatArray = format->formatter.a;
    size += fixedArraySize(formatArray) * 
      dataStructureSize(formatArray[1].f);
    break;
  }

  return size;
}


/*****************************************************************************
*
* FUNCTION: 
*
* DESCRIPTION:
*
* INPUTS:
*
* OUTPUTS: 
*
*****************************************************************************/

SIZES_TYPE transferToDataStructure(format, dataStruct, dStart, buffer, bStart, 
				  parentFormat)
FORMAT_PTR format, parentFormat;
DATA_PTR dataStruct; 
int dStart, bStart;
DATA_PTR buffer;
{ 
  SIZES_TYPE sizes;
  int (*formatProc)();
  char ptrVal;
  DATA_PTR newStruct;
  FORMAT_PTR nextFormat;
  FORMAT_ARRAY_PTR formatArray;
  int i, currentByte, currentData, structStart, arraySize, elements;

  currentByte = bStart;
  currentData = dStart;

  switch(format->type) {
  case LengthFMT:
    FROM_BUFFER_AND_ADVANCE(dataStruct+currentData, buffer, currentByte, 
			     format->formatter.i);
    currentData += format->formatter.i;
    break;
  case PrimitiveFMT:
    formatProc = TransTable[format->formatter.i];
    currentByte += formatProc(Decode, dataStruct, currentData,
				buffer, currentByte);
    currentData += formatProc(ALength, NULL, 0, NULL, 0);
    break;
  case PointerFMT:
    FROM_BUFFER_AND_ADVANCE(&ptrVal, buffer, currentByte, sizeof(char));
    if (ptrVal == '\0') 
      newStruct = NULL;
    else {
      nextFormat = CHOOSE_PTR_FORMAT(format, parentFormat);    
      newStruct = (DATA_PTR)tcaMalloc((unsigned)dataStructureSize(nextFormat));
      sizes = transferToDataStructure(nextFormat, newStruct, 0,
				      buffer, currentByte, (FORMAT_PTR)NULL);
      currentByte += sizes.buffer;
    }
  TO_BUFFER_AND_ADVANCE(&newStruct, dataStruct, currentData, sizeof(DATA_PTR));
    break;
  case StructFMT:
    formatArray = format->formatter.a;
    for(i=1;i < formatArray[0].i;i++) {
      sizes = transferToDataStructure(formatArray[i].f, dataStruct, 
				      currentData, buffer, currentByte, format);
      currentByte += sizes.buffer;
      currentData = alignField(format, i, currentData+sizes.data);
    }
    break;
  case FixedArrayFMT:
    formatArray = format->formatter.a;
    arraySize = fixedArraySize(formatArray);
    nextFormat = formatArray[1].f;
    if (fixedLengthFormat(nextFormat)) {
      elements = arraySize * dataStructureSize(nextFormat);
      bcopy(buffer+currentByte, dataStruct+currentData, elements);
      currentByte += elements;
      currentData += elements;
    }
    else {
      for(i=0;i < arraySize;i++) {
	sizes = transferToDataStructure(nextFormat, dataStruct, currentData,
					buffer, currentByte, (FORMAT_PTR)NULL);
	currentByte += sizes.buffer;
	currentData += sizes.data;
      }
    }
    break;
  case VarArrayFMT:
    formatArray = format->formatter.a;
    FROM_BUFFER_AND_ADVANCE(&arraySize, buffer, currentByte, sizeof(int));
    nextFormat = formatArray[1].f;
    newStruct = ((arraySize == 0) ? NULL :
		 (DATA_PTR)tcaMalloc((unsigned)(arraySize*
						dataStructureSize(nextFormat))));
  TO_BUFFER_AND_ADVANCE(&newStruct, dataStruct, currentData, sizeof(DATA_PTR));
    if (newStruct) {
      if (fixedLengthFormat(nextFormat)) {
	elements = arraySize * dataStructureSize(nextFormat);
	bcopy(buffer+currentByte, newStruct, elements);
	currentByte += elements;
      }
      else {
	structStart = 0;
	for(i=0;i < arraySize;i++) {
	  sizes = transferToDataStructure(nextFormat, newStruct, structStart, 
					  buffer, currentByte, (FORMAT_PTR)NULL);
	  currentByte += sizes.buffer;
	  structStart += sizes.data;
	}
      }
    }
  }

  sizes.buffer = currentByte - bStart;
  sizes.data = currentData - dStart;
  return sizes;  
}


/*************************************************************

  THESE FUNCTIONS FORM THE INTERFACE TO THE REST OF THE SYSTEM

*************************************************************/

int bufferSize(Format, DataStruct)
FORMAT_PTR Format; DATA_PTR DataStruct; 
{ 
  SIZES_TYPE sizes;

  sizes = bufferSize1(Format, DataStruct, 0, (FORMAT_PTR)NULL);
  return sizes.buffer;
}

/*************************************************************
		 
int dataStructureSize(Format) FORMAT_PTR Format;
{
  .
  . Function is defined (and used) above
  .
}

*************************************************************/

void encodeData(Format, DataStruct, Buffer, BStart)
FORMAT_PTR Format;
DATA_PTR DataStruct; 
DATA_PTR Buffer; 
int BStart;
{
  transferToBuffer(Format, DataStruct, 0, Buffer, BStart, (FORMAT_PTR)NULL);
}

/*************************************************************/

char *decodeData(Format, Buffer, BStart)
FORMAT_PTR Format; 
DATA_PTR Buffer; 
int BStart;
{
  char *DataStruct;

  DataStruct = tcaMalloc((unsigned)dataStructureSize(Format));

  transferToDataStructure(Format, DataStruct, 0, Buffer, BStart, (FORMAT_PTR)NULL);

  return DataStruct;
}


/*****************************************************************************
*
* FUNCTION: freeDataElements
*
* DESCRIPTION: Frees the data elements "malloc"ed by the equivalent call
*              to "TransferToDataStructure"
*
* INPUTS: format of data structure
*         pointer data structure itself
*         start of relevant part of data structure
*         format of parent data structure (or NULL)
*
* OUTPUTS: number of bytes processed in the top-level dataStruct.
*
*****************************************************************************/

int freeDataElements(format, dataStruct, dStart, parentFormat)
FORMAT_PTR format, parentFormat;
DATA_PTR dataStruct; 
int dStart;
{ 
  int (*formatProc)();
  DATA_PTR structPtr;
  FORMAT_PTR nextFormat;
  FORMAT_ARRAY_PTR formatArray;
  int size, i, currentData, structStart, arraySize;

  currentData = dStart;

  switch(format->type) {
  case LengthFMT:
    currentData += format->formatter.i;
    break;

  case PrimitiveFMT:
    formatProc = TransTable[format->formatter.i];
    (void)formatProc(DFree, dataStruct, currentData, NULL, 0);
    currentData += formatProc(ALength, NULL, 0, NULL, 0);
    break;

  case PointerFMT:
    structPtr = REF(DATA_PTR, dataStruct, dStart);
    currentData += sizeof(DATA_PTR);
    if (structPtr) {
      nextFormat = CHOOSE_PTR_FORMAT(format, parentFormat);    
      size = freeDataElements(nextFormat, structPtr, 0, (FORMAT_PTR)NULL);
      tcaFree((char *)structPtr);
    }
    break;

  case StructFMT:
    formatArray = format->formatter.a;
    for(i=1;i < formatArray[0].i;i++) {
    size = freeDataElements(formatArray[i].f, dataStruct, currentData, format);
      currentData = alignField(format, i, currentData+size);
    }
    break;

  case FixedArrayFMT:
    formatArray = format->formatter.a;
    arraySize = fixedArraySize(formatArray);
    nextFormat = formatArray[1].f;
    if (fixedLengthFormat(nextFormat)) {
      currentData += arraySize * dataStructureSize(nextFormat);
    }
    else {
      for(i=0;i < arraySize;i++) {
	size = freeDataElements(nextFormat, dataStruct, currentData, 
				(FORMAT_PTR)NULL);
	currentData += size;
      }
    }
    break;

  case VarArrayFMT:
    structPtr = REF(DATA_PTR, dataStruct, currentData);
    if (structPtr) {
      formatArray = format->formatter.a;
      arraySize = varArraySize(formatArray, parentFormat,
				 dataStruct, currentData);
      nextFormat = formatArray[1].f;
      if (!fixedLengthFormat(nextFormat)) {
	structStart = 0;
	for(i=0;i < arraySize;i++) {
	  size = freeDataElements(nextFormat, structPtr, structStart, 
				  (FORMAT_PTR)NULL);
	  structStart += size;
	}
      }
      tcaFree((char *)structPtr);
    }
    currentData += sizeof(DATA_PTR);
  }

  return currentData - dStart;
}

/*************************************************************/

void freeDataStructure(format, dataStruct)
FORMAT_PTR format; 
char *dataStruct;
{
  (void)freeDataElements(format, dataStruct, 0, (FORMAT_PTR)NULL);

  tcaFree(dataStruct);
}


/*****************************************************************************
*
* FUNCTION: void tcaFreeData(msgName, dataStruct)
*
* DESCRIPTION: 
* Using the queryForm format of the message, free the data structure
* and all of its allocated subelements
*
* INPUTS: 
* char *msgName; 
* void *dataStruct;
*
* OUTPUTS: void.
*
*****************************************************************************/

void tcaFreeData(msgName, dataStruct)
char *msgName; 
void *dataStruct;
{
  MSG_PTR msg;

  msg = msgFind(msgName);

  freeDataStructure(msg->msgData->msgForm, (char *)dataStruct);
}


/*****************************************************************************
*
* FUNCTION: void tcaFreeReply(msgName, replyData)
*
* DESCRIPTION: 
* Using the replyForm format of the message, free the all of the allocated
* subelements of the data structure, but do not free the data structure itself.
*
* INPUTS:
* char *msgName; 
* void *replyData;
*
* OUTPUTS: none
*
*****************************************************************************/

void tcaFreeReply(msgName, replyData)
char *msgName; 
void *replyData;
{
  MSG_PTR msg;

  msg = msgFind(msgName);

  (void)freeDataElements(msg->msgData->resForm, (char *)replyData,
			 0, (FORMAT_PTR)NULL);
}


/*****************************************************************************
*
* FUNCTION: void classDataFree(class, classData)
*
* DESCRIPTION: 
* Finds the class format associated with the class, and frees the data.
*
* INPUTS:
* TCA_MSG_CLASS_TYPE class;
* char *classData;
*
* OUTPUTS: none
*
*****************************************************************************/

void classDataFree(class, classData)
TCA_MSG_CLASS_TYPE class;
char *classData;
{
  CLASS_FORM_PTR classForm;

  if (classData) {
    classForm = (CLASS_FORM_PTR)hashTableFind((char *)&class, classFormatTable);
    if (!classForm) {
      tcaModError("ERROR: freeClassData: missing class format.");
    } else {
      freeDataStructure(classForm->format, classData);
    }
  }
}
