/******************************************************************************
*
* PROJECT: Carnegie Mellon Planetary Rover Project
*          Task Control Architecture 
* 
* (c) Copyright 1991 Christopher Fedor and Reid Simmons.  All rights reserved.
* 
* MODULE: registration
*
* FILE: reg.c
*
* ABSTRACT:
* 
* Register messages and handlers
*
* REVISION HISTORY
*
* 30-Jan-91 Christopher Fedor, School of Computer Science, CMU
* Added fflush(stdout) to printf for module code calls from lisp
*
* 28-Aug-90 Christopher Fedor, School of Computer Science, CMU
* Rewritten. Revised to Software Standards.
*
*  9-Nov-89 Reid Simmons, School of Computer Science, CMU
* Broke out the code to pre-register messages as a separate function
* ("findOrRegisterMessage").
*
*  2-Aug-89 Reid Simmons, School of Computer Science, CMU
* Changed "printf"s to use logging facilities.
*
*  1-Dec-88 Christopher Fedor, School of Computer Science, CMU
* created.
*
******************************************************************************/

#include "stdio.h"
#include "ctype.h"
#include "strings.h"

#include "basics.h"
#include "tcaInternal.h"
#include "formatters.h"
#include "datamsg.h"
#include "idtable.h"
#include "hash.h"
#include "list.h"

#include "res.h"

/* Route Tables */
extern ID_TABLE_PTR msgIdTable;
extern ID_TABLE_PTR hndIdTable;

extern HASH_TABLE_PTR messageTable, handlerTable;

extern int tcaServerGlobal;

/******************************************************************************
*
* FUNCTION: MSG_PTR msgCreate(msgData)
*
* DESCRIPTION: 
* Ceates a new message using msgData. The newly created message is
* stored in messageTable hash table stored by message name.  The message
* is also inserted into the idtable msgIdTable.
*
* INPUTS: MSG_DATA_PTR msgData;
*
* OUTPUTS: MSG_PTR
*
******************************************************************************/

MSG_PTR msgCreate(msgData)
MSG_DATA_PTR msgData;
{
  MSG_PTR msg;

  msg = NEW(MSG_TYPE);
  msg->direct = FALSE;
  msg->intent = 0;
  msg->sd = 0;
  msg->parsedFormats = FALSE;
  msg->msgData = msgData;
  msg->hndList = listCreate();
  msg->tapList = NULL;  
  msg->excepList = NULL;  

  /* 11-Jun-91: fedor: Blah! storing the parse string should 
     not be done here - but didnt want to mess with msgData for now ! */
  msg->msgFormatStr = NULL;
  msg->resFormatStr = NULL;

  (void)hashTableInsert(msgData->name, strlen(msgData->name)+1, 
			(char *)msg, messageTable);
  return msg;
}


/******************************************************************************
*
* FUNCTION: MSG_PTR findOrRegisterMessage(name)
*
* DESCRIPTION: 
* Find the message, if registered.  If not, create a message 
* record, setting the "class" to a special value to indicate that
* it is really not yet registered.
*
* INPUTS: char *name;
*
* OUTPUTS: MSG_PTR
*
******************************************************************************/

MSG_PTR findOrRegisterMessage(name)
char *name;
{
  MSG_PTR msg;
  MSG_DATA_PTR msgData;

  msg = (MSG_PTR)hashTableFind(name, messageTable);
  if (!msg) {
    msgData = NEW(MSG_DATA_TYPE);
    msgData->refId = 0;
    msgData->name = name;
    msgData->msgForm = NULL;
    msgData->resForm = NULL;
    msgData->class = HandlerRegClass;

    msg = msgCreate(msgData);
  }
  return msg;
}


/******************************************************************************
*
* FUNCTION: HND_PTR selfRegisterHnd(hndData, hndProc)
*
* DESCRIPTION: 
* selfRegisterHnd creates a new handler from hndData and updates the message
* associated with this handler to include this handler. If no message
* has been found a new message of class HandlerRegClass is created.
*
* The newly created Handler is inserted into the handler idtable hndIdTable
* and the handler hash table handlerTable. If a handler by the same name
* is found a warning is issued and the handler information is updated.
*
* If hndProc is NOT NULL then refId of hndData is updated to the localId
* returned by the idtable hndIdTable.
*
* The newly created or found handler is returned.
*
* INPUTS:
* HND_DATA_PTR hndData;
* int (*hndProc)();
*
* OUTPUTS: HND_PTR
*
******************************************************************************/

HND_PTR selfRegisterHnd(sd, hndOrg, hndData, hndProc)
int sd;
MODULE_PTR hndOrg;
HND_DATA_PTR hndData;
void (*hndProc)();
{
  int localId;
  MSG_PTR msg;
  HND_PTR hnd;

  HND_KEY_TYPE hndKey;

  msg = findOrRegisterMessage(hndData->msgName);
  
  hndKey.sd = sd;
  hndKey.name = hndData->hndName;

  hnd = (HND_PTR)hashTableFind((char *)&hndKey, handlerTable);
  if (!hnd) {
    hnd = NEW(HND_TYPE);
    hnd->sd = sd;
    hnd->hndOrg = hndOrg;
    hnd->hndProc = hndProc;
    hnd->hndData = hndData;
    hnd->resource = NULL;
    hnd->msgList = listCreate();

    (void)hashTableInsert((char *)&hndKey, sizeof(HND_KEY_TYPE), 
			  (char *)hnd, handlerTable);

    localId = idTableInsert((char *)hnd, hndIdTable);

    hnd->localId = localId;

    if (hndProc)
      hnd->hndData->refId = localId;
  }
#ifndef LISP
  else if (hndProc != hnd->hndProc) {
    fprintf(stderr, "WARNING: Procedure change ignored for existing handler %s.\n",
	    hnd->hndData->hndName);
  }
#endif 
/* 24-Jun-91: fedor: the warning is not meaningful for lisp because each
   re-register will cause a pointer change - lisp functions are not at
   static locations like c */

  /* 3-Sep-90: fedor: NULL forces initial module cache of a message. */
  hnd->msg = NULL;

  if (!listMemberItem((char *)hnd, msg->hndList))
    listInsertItem((char *)hnd, msg->hndList); 

  if (!listMemberItem((char *)msg, hnd->msgList))
    listInsertItem((char *)msg, hnd->msgList);

  return hnd;
}

/***************************************************************************/

void tcaRegisterMessage(name, class, msgForm, resForm)
char *name;
TCA_MSG_CLASS_TYPE class;
char *msgForm, *resForm;
{
  MSG_REG_DATA_TYPE msgRegData;

  msgRegData.name = name;
  msgRegData.class = class;
  msgRegData.msgForm= msgForm;
  msgRegData.resForm = resForm;  

  if (class == CommandClass)
    msgRegData.resForm = "int";
  
  (void)tcaInform("registerMessageMsg", (void *)&msgRegData);
}

/*************************************************************************/

void tcaRegisterHandler(msgName, hndName, hndProc)
char *msgName, *hndName;
void (*hndProc)();
{
  HND_DATA_PTR hndData;

  if (!hndProc)
    TCA_MOD_ERROR1("ERROR: tcaRegisterHandler: hndProc for %s is NULL.\n",
		   hndName);

  hndData = NEW(HND_DATA_TYPE);

  hndData->refId = 0; /* will be updated by selfRegisterHandler */
  hndData->msgName = msgName;
  hndData->hndName = hndName;

  (void)selfRegisterHnd(0, (MODULE_PTR)NULL, hndData, hndProc);

  (void)tcaInform("RegisterHandlerMsg", (void *)hndData);
}


/*****************************************************************************
*
* FUNCTION: void tcaRegisterNamedFormatter(formatterName, formatString)
*
* DESCRIPTION:
*
* INPUTS: char *formatterName, *formatString;
*
* OUTPUTS: void.
*
*****************************************************************************/

void tcaRegisterNamedFormatter(formatterName, formatString)
char *formatterName, *formatString;
{ 
  REG_DATA_TYPE regData;

  regData.name = formatterName;
  regData.format = formatString;

  (void)tcaInform("registerNamedFormMsg", (void *)&regData);
}


/*****************************************************************************
*
* FUNCTION: void tcaRegisterLengthFormatter(formatter, length)
*
* DESCRIPTION:
*
* INPUTS:
* char *formatter;
* int length;
*
* OUTPUTS: void.
*
*****************************************************************************/

void tcaRegisterLengthFormatter(formatter, length)
char *formatter;
int length;
{
  int i;
  char s[11];

  for(i=0;i<11;i++)
    s[i] = '\0';

  (void)sprintf(s, "%d", length);

  tcaRegisterNamedFormatter(formatter, s);
}
