/******************************************************************************
*
* PROJECT: Carnegie Mellon Planetary Rover Project
*          Task Control Architecture 
* 
* (c) Copyright 1991 Christopher Fedor and Reid Simmons.  All rights reserved.
* 
* MODULE: send message
*
* FILE: sendMsg.c
*
* ABSTRACT:
* 
* Module routines for initializing and using sendMsg/recvMsg pair.
*
* REVISION HISTORY
*
*  7-May-90 Christopher Fedor, School of Computer Science, CMU
* Created.
*
******************************************************************************/

#include "stdio.h"

#include "tcaInternal.h"

#include "formatters.h"

#include "datamsg.h"
#include "idtable.h"
#include "hash.h"

#include "dispatch.h"

extern int tcaServerGlobal;

extern int parentRefGlobal;

extern HASH_TABLE_PTR classFormatTable,  messageTable;

extern HASH_TABLE_PTR formatNamesTable;

extern MSG_PTR msgCreate();

int sendMessageRefGlobal = 1;

/*******************************************************/

int nextSendMessageRef()
{
  return sendMessageRefGlobal;
}

/*******************************************************/

TCA_RETURN_VALUE_TYPE sendMessage(ref, msg, msgData, classData)
TCA_REF_PTR ref;
MSG_PTR msg;
char *msgData, *classData;
{
  int refId, savedRef;
  DATA_MSG_PTR msgDataMsg;
  TCA_MSG_CLASS_TYPE class;

  CLASS_FORM_PTR classForm;

  FORMAT_PTR classFormat;

  classFormat = NULL;

  if (!msg) {
    tcaModError("ERROR: sendMessage: NULL msg.\n");
  }

  /* save the message ref to use in case msgFind triggers 
     a recursive call to sendMessage */

  savedRef = sendMessageRefGlobal;

  class = msg->msgData->class;

  classForm = (CLASS_FORM_PTR)hashTableFind((char *)&class, classFormatTable);

  if (classForm)
    classFormat = classForm->format;

  if (ref) 
    refId = ref->refId;
  else
    refId = NO_REF;

  if (msg->direct) {
    msgDataMsg = dataMsgCreate(parentRefGlobal, msg->intent, (int)class, 
			       refId, savedRef, msg->msgData->msgForm,
			       msgData, classFormat, classData);
    (void)dataMsgSend(msg->sd, msgDataMsg);
  } else {
    msgDataMsg = dataMsgCreate(parentRefGlobal, msg->msgData->refId, (int)class, 
			       refId, savedRef, msg->msgData->msgForm,
			       msgData, classFormat, classData);
    (void)dataMsgSend(tcaServerGlobal, msgDataMsg);
  }

  sendMessageRefGlobal++;

  dataMsgFree(msgDataMsg);

  return Success;
}

/*******************************************************/

TCA_RETURN_VALUE_TYPE sendResponse(ref, msg, resData, resClass, resClassData)
TCA_REF_PTR ref;
MSG_PTR msg;
char *resData;
TCA_MSG_CLASS_TYPE resClass;
char *resClassData;
{
  int intent, refId;
  FORMAT_PTR format;
  DATA_MSG_PTR dataMsg;
  CLASS_FORM_PTR classForm;
  FORMAT_PTR classFormat;

  classFormat = NULL;

  if (msg) {
    intent = msg->msgData->refId;
    format = msg->msgData->resForm;
  } else {
    intent = -1;
    if (ref) {
      if (!ref->msg) 
	ref->msg = msgFind(ref->name);
      format = ref->msg->msgData->resForm;
      msg = ref->msg;
    }
    else
      format = NULL;
  }

  classForm = (CLASS_FORM_PTR)hashTableFind((char *)&resClass, classFormatTable);

  if (classForm)
    classFormat = classForm->format;

  if (ref)
    refId = ref->refId;
  else
    refId = NO_REF;

  dataMsg = dataMsgCreate(parentRefGlobal, intent, (int)resClass, refId, -1,
			  format, resData, classFormat, resClassData);

  if (msg && msg->direct) {
    dataMsg->intent = QUERY_REPLY_INTENT;
    dataMsg->msgRef = refId;
    (void)dataMsgSend(msg->sd, dataMsg);
  }
  else
    (void)dataMsgSend(tcaServerGlobal, dataMsg);

  dataMsgFree(dataMsg);

  return Success;
}

/***************/

FORMAT_PTR formatterRecv(sd)
int sd;
{
  DATA_MSG_PTR dataMsg;
  FORMAT_PTR formatFormat, *pointerToFormat;

  formatFormat = (FORMAT_PTR)hashTableFind("format", formatNamesTable);
  if (!formatFormat)
    tcaModError("ERROR: formatterRecv: no formatFormat");

  switch(dataMsgRecv(sd, &dataMsg)){
  case StatOK:
    pointerToFormat = (FORMAT_PTR *)dataMsgDecodeMsg(formatFormat, dataMsg);
    dataMsgFree(dataMsg);
    return(*pointerToFormat);
  case StatError:
    tcaModError("ERROR: foramatterRecv: dataMsgRecv: StatError:"); return NULL;
  case StatEOF:
    tcaModError("ERROR: foramatterRecv: dataMsgRecv: StatEOF:"); return NULL;
  default:
    tcaModError("ERROR: foramatterRecv: dataMsgRecv: UNKNOWN:"); return NULL;
  }
}

/*******************************************************/

void msgInfoMsgInitialize()
{
  int refId;
  MSG_DATA_PTR msgData;

  (void)readNBytes(tcaServerGlobal, (char *)&refId, sizeof(int));

  msgData = NEW(MSG_DATA_TYPE);
  msgData->refId = refId;
  msgData->name = "msgInfoMsg";
  msgData->class = QueryClass;

  msgData->msgForm = formatterRecv(tcaServerGlobal);
  msgData->resForm = formatterRecv(tcaServerGlobal);

  (void)msgCreate(msgData);
}

/*******************************************************/

void classModInitialize()
{
  TCA_MSG_CLASS_TYPE class;
  CLASS_FORM_PTR classForm;

  /* 31-Aug-90: fedor: more and more difficult to do 
     query msg ref with a QueryClass format of "int" */
     
  class = GoalClass;
  (void)tcaQuery("classInfoMsg", (void *)&class, (void *)&classForm);
  (void)hashTableInsert((char *)&class, sizeof(TCA_MSG_CLASS_TYPE), 
			(char *)classForm, classFormatTable);

  class = CommandClass;
  (void)tcaQuery("classInfoMsg", (void *)&class, (void *)&classForm);
  (void)hashTableInsert((char *)&class, sizeof(TCA_MSG_CLASS_TYPE), 
			(char *)classForm, classFormatTable);

  class = ExceptionClass;
  (void)tcaQuery("classInfoMsg", (void *)&class, (void *)&classForm);
  (void)hashTableInsert((char *)&class, sizeof(TCA_MSG_CLASS_TYPE), 
			(char *)classForm, classFormatTable);

  class = PointMonitorClass;
  (void)tcaQuery("classInfoMsg", (void *)&class, (void *)&classForm);
  (void)hashTableInsert((char *)&class, sizeof(TCA_MSG_CLASS_TYPE),
			(char *)classForm, classFormatTable);

  class = PollingMonitorClass;
  (void)tcaQuery("classInfoMsg", (void *)&class, (void *)&classForm);
  (void)hashTableInsert((char *)&class, sizeof(TCA_MSG_CLASS_TYPE),
			(char *)classForm, classFormatTable);

  class = DemonMonitorClass;
  (void)tcaQuery("classInfoMsg", (void *)&class, (void *)&classForm);
  (void)hashTableInsert((char *)&class, sizeof(TCA_MSG_CLASS_TYPE),
			(char *)classForm, classFormatTable);

  class = ExecHndClass;
  (void)tcaQuery("classInfoMsg", (void *)&class, (void *)&classForm);
  (void)hashTableInsert((char *)&class, sizeof(TCA_MSG_CLASS_TYPE),
			(char *)classForm, classFormatTable);

  class = FailureClass;
  (void)tcaQuery("classInfoMsg", (void *)&class, (void *)&classForm);
  (void)hashTableInsert((char *)&class, sizeof(TCA_MSG_CLASS_TYPE),
			(char *)classForm, classFormatTable);

  class = FireDemonClass;
  (void)tcaQuery("classInfoMsg", (void *)&class, (void *)&classForm);
  (void)hashTableInsert((char *)&class, sizeof(TCA_MSG_CLASS_TYPE),
			(char *)classForm, classFormatTable);

  class = MultiQueryClass;
  (void)tcaQuery("classInfoMsg", (void *)&class, (void *)&classForm);
  (void)hashTableInsert((char *)&class, sizeof(TCA_MSG_CLASS_TYPE),
			(char *)classForm, classFormatTable);
}


