/******************************************************************************
*
* PROJECT: Carnegie Mellon Planetary Rover Project
*          Task Control Architecture 
* 
* (c) Copyright 1991 Christopher Fedor and Reid Simmons.  All rights reserved.
* 
* MODULE: tca
*
* FILE: tca.h
*
* ABSTRACT:
* 
*  Task Control Architecture
*
*  Include File
*
* REVISION HISTORY
*
* 13-Oct-92  Reid Simmons, School of Computer Science, CMU
* Changed ConstraintClass to InformClass (more informative name).
* Added BroadcastClass and MultiQueryClass messages.
* Added WhenSuccess and WhenFailure message tap conditions
*
****************************************************************************/

/* This flag ensures that multiple occurrences of the include file 
   don't cause problems*/
#ifndef INCtca
#define INCtca

/* Support for HandleMessage */
#define WAITFOREVER ((long)-1) /* must be less than 0 */

#ifdef Success
#undef Success
#endif

/* Communciation Return Values */
typedef enum {Success, TimeOut, Failure, WrongVersion, MsgUndefined,
	      WrongMsgClass, NullReply} TCA_RETURN_VALUE_TYPE;

/* Low Level Communication Status Values */
typedef enum {StatEOF, StatError, StatOK, StatSendEOF, StatSendError,
		StatRecvEOF, StatRecvError} TCA_RETURN_STATUS_TYPE;

/* Class Type */
typedef enum {UNKNOWN, HandlerRegClass, ExecHndClass, QueryClass, 
		GoalClass, CommandClass, InformClass, ExceptionClass,
		PollingMonitorClass, PointMonitorClass, DemonMonitorClass,
		ReplyClass, SuccessClass, FailureClass, FireDemonClass, 
		BroadcastClass, MultiQueryClass} TCA_MSG_CLASS_TYPE;
/* For upward compatibility */
#define ConstraintClass InformClass

/* Reference Status Return Values */
typedef enum { UnknownRefStatus, InactiveRef, PendingRef, ActiveRef,
	       HandledRef, PlannedRef, AchievedRef, KilledRef 
	     } TCA_REF_STATUS_TYPE;

typedef struct _TCA_REF *TCA_REF_PTR;

typedef struct { 
  int maxFire;     /* maximum times to fire action message;
		      -1 indicates infinite (default is -1) */
  int fireEvery;  /* fire action message every n times condition is satisfied.
		     (default is 1) */
  int duration;    /* maximum time, in seconds, that the monitor can run;
		      -1 indicates infinite time (default is -1) */
  int period;      /* time between polls, in seconds */
  int initialWait; /* if TRUE, don't poll immediately (default is FALSE) */
} INTERVAL_MONITOR_OPTIONS_TYPE, *INTERVAL_MONITOR_OPTIONS_PTR;

/*
  For representing temporal constraints within modules (central uses
  a different representation).
*/

typedef enum {NoTime, StartPoint, EndPoint} TCA_POINT_CLASS_TYPE;
typedef enum {NoInterval, HandlingInterval, AchievementInterval, 
		PlanningInterval} TCA_INTERVAL_CLASS_TYPE;

typedef enum {UnknownCondition, WhenSent, BeforeHandling, WhileHandling, 
		AfterHandled, AfterReplied, AfterSuccess, AfterFailure,
		BeforeAchieving, WhileAchieving, AfterAchieved, WhenAchieved,
		BeforePlanning, WhilePlanning, AfterPlanned, WhenPlanned,
		WhenSuccess, WhenFailure, WhenKilled} TAP_CONDITION_TYPE;

typedef struct {
  TCA_INTERVAL_CLASS_TYPE class;
  int msgRef;
} TCA_INTERVAL_TYPE, *TCA_INTERVAL_PTR;

typedef struct { 
  TCA_POINT_CLASS_TYPE class;
  TCA_INTERVAL_TYPE interval;
} TCA_TIME_POINT_TYPE, *TCA_TIME_POINT_PTR;

typedef struct _TIMELIST { 
  TCA_TIME_POINT_TYPE timePoint;
  struct _TIMELIST *Next;
} TCA_TIME_LIST_TYPE, *TCA_TIME_LIST_PTR;

typedef struct { 
  TCA_TIME_POINT_TYPE timePoint1;
  char *relationship;
  TCA_TIME_POINT_TYPE timePoint2;
} TCA_TPL_CONSTRAINT_TYPE, *TCA_TPL_CONSTRAINT_PTR;


typedef struct {
  TCA_REF_PTR ERef, CRef;
  int numOfRetries;		/* number of times of CNode being retried */
} EXC_INFO_TYPE, * EXC_INFO_PTR;

/***************************************************************************/

#define NO_TPLCONSTR     0

/* End(Achievement(Last_Node)) <= Start(Achievement(Node)) */
#define SEQ_ACH         01 

/* End(Planning(Last_Node)) <= Start(Planning(Node)) */
#define SEQ_PLANNING    02

/* End(Planning(Node)) <= Start(Achievement(Node)) */
#define PLAN_FIRST      04 

/* Start(Planning(Node)) = Start(Achievement(Node)) */
#define DELAY_PLANNING 010 

#define tcaRegisterGoalMessage(name, format) \
  tcaRegisterMessage(name, GoalClass, format, (char *)NULL)

#define tcaRegisterCommandMessage(name, format) \
  tcaRegisterMessage(name, CommandClass, format, (char *)NULL)

#define tcaRegisterInformMessage(name, format) \
  tcaRegisterMessage(name, InformClass, format, (char *)NULL)

/* Kept for upward compatibility -- use "tcaRegisterInformMessage" */
#define tcaRegisterConstraintMessage(name, format) \
  tcaRegisterInformMessage(name, format)

#define tcaRegisterQueryMessage(name, msgFormat, resFormat) \
  tcaRegisterMessage(name, QueryClass, msgFormat, resFormat)

#define tcaRegisterExceptionMessage(name, format) \
  tcaRegisterMessage(name, ExceptionClass, format, (char *)NULL)

#define tcaRegisterBroadcastMessage(name, format) \
  tcaRegisterMessage(name, BroadcastClass, format, (char *)NULL)

#define tcaRegisterMultiQueryMessage(name, msgFormat, resFormat) \
  tcaRegisterMessage(name, MultiQueryClass, msgFormat, resFormat)

#define VALID_TIME(timePoint) ((timePoint).class != NoTime)

#define TPL_ORDER(interval1, interval2) \
  tcaTplConstrain(tcaEndOf(interval1), "<=", tcaStartOf(interval2))

#define tcaRaiseException(a,b,c) tcaFailure(a,b,c)

/***************************************************************************/

extern TCA_REF_PTR tcaRootNodeGlobal;

/* Library Functions */

#ifdef __STDC__

/* 
 * "tcaInform" and "tcaAddConstraint" are equivalent -- the former name
 * is just more informative of what the function does
 */
extern TCA_RETURN_VALUE_TYPE tcaInform(char *name, void *data);
extern TCA_RETURN_VALUE_TYPE tcaAddConstraint(char *name, void *data);

extern TCA_RETURN_VALUE_TYPE tcaQuery(char *name, void *queryData, void *replyData);

extern TCA_RETURN_VALUE_TYPE tcaReply(TCA_REF_PTR ref, char *data);
extern TCA_RETURN_VALUE_TYPE tcaNullReply(TCA_REF_PTR ref);

extern TCA_RETURN_VALUE_TYPE tcaQuerySend(char *name, void *query,
					  TCA_REF_PTR *ref);
extern TCA_RETURN_VALUE_TYPE tcaQueryReceive(TCA_REF_PTR ref, void *reply);

extern TCA_RETURN_VALUE_TYPE tcaExpandGoal(char *name, char *data);

extern TCA_RETURN_VALUE_TYPE tcaExpandGoalWithConstraints(TCA_REF_PTR ref, 
							  char *name, 
							  void *data, 
							  int tplConstraints);

extern TCA_RETURN_VALUE_TYPE tcaExecuteCommand(char *name, void *data);

extern TCA_RETURN_VALUE_TYPE tcaExecuteCommandWithConstraints(TCA_REF_PTR ref, 
							  char *name, 
							  void *data, 
							  int tplConstraints);

extern TCA_RETURN_VALUE_TYPE tcaWaitForCommand(char *name, void *data);
extern TCA_RETURN_VALUE_TYPE tcaWaitForCommandWithConstraints(TCA_REF_PTR ref, 
							  char *name, 
							  void *data, 
							  int tplConstraints);

extern TCA_RETURN_VALUE_TYPE tcaSuccess(TCA_REF_PTR ref);
extern TCA_RETURN_VALUE_TYPE tcaFailure(TCA_REF_PTR ref, char *description,
					void *data);

extern TCA_RETURN_VALUE_TYPE tcaBroadcast(char *name, void *data);

extern TCA_RETURN_VALUE_TYPE tcaMultiQuery(char *name, void *query, int max, 
					   TCA_REF_PTR *refPtr);
extern TCA_RETURN_VALUE_TYPE tcaMultiReceive(TCA_REF_PTR ref, void *reply,
					     long timeout);

extern char *tcaServerMachine(void);
extern void tcaWaitUntilReady(void);

extern void tcaModuleListen(void);
extern TCA_RETURN_VALUE_TYPE tcaHandleMessage(long seconds);

extern void tcaConnectModule(char *moduleName, char *serverHost);
extern void tcaClose(void);

extern void tcaRegisterNamedFormatter(char *formName, char *formString);
extern void tcaRegisterLengthFormatter(char *formName, int length);

extern void tcaRegisterMessage(char *name, TCA_MSG_CLASS_TYPE class, 
			       char *msgForm, char *resForm);

extern void tcaRegisterHandler(char *msgName, char *hndName, 
			       void (*hndProc)());

extern void tcaRegisterResource(char *resName, int capacity);

extern TCA_REF_PTR tcaReserveResource(char *resName);
extern TCA_REF_PTR tcaReserveModResource(char *modName, char *resName);
extern void tcaCancelReservation(TCA_REF_PTR ref);

extern TCA_REF_PTR tcaLockResource(char *resName);
extern TCA_REF_PTR tcaLockModResource(char *modName, char *resName);
extern void tcaUnlockResource(TCA_REF_PTR ref);

extern void tcaAddHndToResource(char *hndName, char *resName);

extern void tcaRegisterPointMonitor(char *monitor, char *condition, char *act);
extern void tcaRegisterDemonMonitor(char *monitor, char *setUp, 
				    char *act, char *cancel);
extern void tcaRegisterPollingMonitor(char *monitor, char *condition, 
				      char *act);

extern TCA_RETURN_VALUE_TYPE tcaPointMonitor(char *name, void *data);
extern TCA_RETURN_VALUE_TYPE tcaPointMonitorWithConstraints(TCA_REF_PTR ref,
							    char *name,
							    void *data,
							    int tplConstraint);

extern INTERVAL_MONITOR_OPTIONS_PTR tcaCreateMonitorOptions(void);

extern TCA_RETURN_VALUE_TYPE tcaIntervalMonitor(char *name, void *data);
extern TCA_RETURN_VALUE_TYPE tcaIntervalMonitorWithConstraints(TCA_REF_PTR ref,
					    char *name, void *data,
					    TCA_TIME_POINT_TYPE start,
					    TCA_TIME_POINT_TYPE expire,
					    INTERVAL_MONITOR_OPTIONS_PTR ops);

extern TCA_RETURN_VALUE_TYPE tcaFireDemon(char *name, 
					  int demonUserId, void *data);

extern int tcaIsSameReference(TCA_REF_PTR ref1, TCA_REF_PTR ref2);

extern int tcaReferenceId(TCA_REF_PTR ref);
extern char *tcaReferenceName(TCA_REF_PTR ref);
extern char *tcaReferenceData(TCA_REF_PTR ref);
extern void tcaReferenceRelease(TCA_REF_PTR ref);
extern TCA_REF_STATUS_TYPE tcaReferenceStatus(TCA_REF_PTR ref);
extern void tcaRefFree(TCA_REF_PTR ref);

extern TCA_REF_PTR tcaAddChildReference(TCA_REF_PTR parent, char *name);
extern TCA_REF_PTR tcaCreateReference(char *name);
extern TCA_REF_PTR tcaFindParentReference(TCA_REF_PTR child);
extern TCA_REF_PTR tcaFindFirstChild(TCA_REF_PTR parent);
extern TCA_REF_PTR tcaFindLastChild(TCA_REF_PTR parent);
extern TCA_REF_PTR tcaFindNextChild(TCA_REF_PTR child);
extern TCA_REF_PTR tcaFindPreviousChild(TCA_REF_PTR child);
extern TCA_REF_PTR tcaFindFailedReference(TCA_REF_PTR exception);
extern TCA_REF_PTR tcaFindTopLevelReference(TCA_REF_PTR child);
extern TCA_REF_PTR tcaFindAnteReferenceByName(TCA_REF_PTR child, 
					      char *antecedentName);
extern TCA_REF_PTR tcaFindChildByName(TCA_REF_PTR parent, char *childName);

extern void tcaKillTaskTree(TCA_REF_PTR ref);
extern void tcaKillSubTaskTree(TCA_REF_PTR ref);

extern void tcaDisplayTaskTree(TCA_REF_PTR ref);

extern void tcaFreeData(char *msgName, void *data);
extern void tcaFreeReply(char *msgName, void *replyData);

extern TCA_TIME_POINT_TYPE tcaStartOf(TCA_INTERVAL_TYPE interval);

extern TCA_TIME_POINT_TYPE tcaEndOf(TCA_INTERVAL_TYPE interval);

extern TCA_INTERVAL_TYPE tcaHandlingOf(TCA_REF_PTR msgRef);

extern TCA_INTERVAL_TYPE tcaAchievementOf(TCA_REF_PTR msgRef);

extern TCA_INTERVAL_TYPE tcaPlanningOf(TCA_REF_PTR msgRef);

extern void tcaTplConstraint(TCA_TIME_POINT_TYPE timePoint1, 
			     char *relationship,
			     TCA_TIME_POINT_TYPE timePoint2);

extern TCA_TIME_POINT_TYPE tcaDefaultTimeGlobal;

extern void tcaRetry(TCA_REF_PTR ref);
extern void tcaByPassException(TCA_REF_PTR ref);
extern void tcaGetExceptionInfo(TCA_REF_PTR ref, EXC_INFO_PTR info);
extern void tcaAddExceptionHandler(TCA_REF_PTR ref, char *hndName);
extern void tcaAddExceptionHandlerToMessage(char *msgName, char *hndName);

extern TCA_RETURN_VALUE_TYPE tcaTapMessage(TAP_CONDITION_TYPE condition,
					   char *tappedMsg, char *listenMsg);

extern TCA_RETURN_VALUE_TYPE tcaTapReference(TAP_CONDITION_TYPE condition,
					     TCA_REF_PTR tappedRef,
					     char *listenMsg);

extern TCA_REF_PTR tcaFindTappedReference(TCA_REF_PTR listeningRef);

extern TCA_RETURN_VALUE_TYPE tcaRemoveTap(TAP_CONDITION_TYPE condition,
					   char *tappedMsg, char *listenMsg);

extern TCA_RETURN_VALUE_TYPE tcaCleanUpAfterAchieved(char *msgName);

extern void tcaRegisterFreeMemHnd(void (*func)(), int retry);
extern void tcaRegisterMallocHnd(char *(*func)(), int retry);

extern void tcaRegisterInitProc(void (*proc)());

#else

/* 
 * "tcaInform" and "tcaAddConstraint" are equivalent -- the former name
 * is just more informative of what the function does
 */
extern TCA_RETURN_VALUE_TYPE tcaInform();
extern TCA_RETURN_VALUE_TYPE tcaAddConstraint();

extern TCA_RETURN_VALUE_TYPE tcaQuery();

extern TCA_RETURN_VALUE_TYPE tcaReply();
extern TCA_RETURN_VALUE_TYPE tcaNullReply();

extern TCA_RETURN_VALUE_TYPE tcaQuerySend();
extern TCA_RETURN_VALUE_TYPE tcaQueryReceive();

extern TCA_RETURN_VALUE_TYPE tcaExpandGoal();
extern TCA_RETURN_VALUE_TYPE tcaExpandGoalWithConstraints();

extern TCA_RETURN_VALUE_TYPE tcaExecuteCommand();
extern TCA_RETURN_VALUE_TYPE tcaExecuteCommandWithConstraints();

extern TCA_RETURN_VALUE_TYPE tcaWaitForCommand();
extern TCA_RETURN_VALUE_TYPE tcaWaitForCommandWithConstraints();

extern TCA_RETURN_VALUE_TYPE tcaSuccess();
extern TCA_RETURN_VALUE_TYPE tcaFailure();

extern TCA_RETURN_VALUE_TYPE tcaBroadcast();

extern TCA_RETURN_VALUE_TYPE tcaMultiQuery();
extern TCA_RETURN_VALUE_TYPE tcaMultiReceive();

extern char *tcaServerMachine();
extern void tcaWaitUntilReady();

extern void tcaModuleListen();
extern TCA_RETURN_VALUE_TYPE tcaHandleMessage();

extern void tcaConnectModule();
extern void tcaClose();

extern void tcaRegisterNamedFormatter();
extern void tcaRegisterLengthFormatter();

extern void tcaRegisterMessage();
extern void tcaRegisterHandler();

extern void tcaRegisterResource();

extern TCA_REF_PTR tcaReserveResource();
extern TCA_REF_PTR tcaReserveModResource();
extern void tcaCancelReservation();

extern TCA_REF_PTR tcaLockResource();
extern TCA_REF_PTR tcaLockModResource();
extern void tcaUnlockResource();

extern void tcaAddHndToResource();

extern void tcaRegisterPointMonitor();
extern void tcaRegisterDemonMonitor();
extern void tcaRegisterPollingMonitor();

extern TCA_RETURN_VALUE_TYPE tcaPointMonitor();
extern TCA_RETURN_VALUE_TYPE tcaPointMonitorWithConstraints();

extern INTERVAL_MONITOR_OPTIONS_PTR tcaCreateMonitorOptions();

extern TCA_RETURN_VALUE_TYPE tcaIntervalMonitor();
extern TCA_RETURN_VALUE_TYPE tcaIntervalMonitorWithConstraints();

extern TCA_RETURN_VALUE_TYPE tcaFireDemon();

extern int tcaIsSameReference();

extern int tcaReferenceId();
extern char *tcaReferenceName();
extern char *tcaReferenceData();
extern void tcaReferenceRelease();
extern TCA_REF_STATUS_TYPE tcaReferenceStatus();

extern TCA_REF_PTR tcaAddChildReference();
extern TCA_REF_PTR tcaCreateReference();
extern TCA_REF_PTR tcaFindParentReference();
extern TCA_REF_PTR tcaFindFirstChild();
extern TCA_REF_PTR tcaFindLastChild();
extern TCA_REF_PTR tcaFindNextChild();
extern TCA_REF_PTR tcaFindPreviousChild();
extern TCA_REF_PTR tcaFindFailedReference();
extern TCA_REF_PTR tcaFindTopLevelReference ();
extern TCA_REF_PTR tcaFindAnteReferenceByName ();
extern TCA_REF_PTR tcaFindChildByName ();

extern void tcaKillTaskTree();
extern void tcaKillSubTaskTree();

extern void tcaDisplayTaskTree();

extern void tcaFreeData();
extern void tcaFreeReply();

extern TCA_TIME_POINT_TYPE tcaStartOf();

extern TCA_TIME_POINT_TYPE tcaEndOf();

extern TCA_INTERVAL_TYPE tcaHandlingOf();

extern TCA_INTERVAL_TYPE tcaAchievementOf();

extern TCA_INTERVAL_TYPE tcaPlanningOf();

extern TCA_TIME_POINT_TYPE tcaDefaultTimeGlobal;

extern void tcaRetry();
extern void tcaByPassException();
extern void tcaGetExceptionInfo();
extern void tcaAddExceptionHandler();

extern TCA_RETURN_VALUE_TYPE tcaTapMessage();

extern TCA_RETURN_VALUE_TYPE tcaTapReference();

extern TCA_REF_PTR tcaFindTappedReference();

extern TCA_RETURN_VALUE_TYPE tcaRemoveTap();

extern TCA_RETURN_VALUE_TYPE tcaCleanUpAfterAchieved();

extern void tcaRegisterFreeMemHnd();
extern void tcaRegisterMallocHnd();

extern void tcaRegisterInitProc();

#endif __STDC__

#endif INCtca
