/******************************************************************************
*
* PROJECT: Carnegie Mellon Planetary Rover Project
*          Task Control Architecture 
* 
* MODULE: tcaRef
*
* FILE: tcaRef.c
*
* ABSTRACT:
* 
* Implements TCA_REF_PTR
*
* REVISION HISTORY
*
* 14-Dec-91 Christopher Fedor, School of Computer Science, CMU
* Addded Rich Goodwin's tcaReferenceStatus.
*
* 14-Mar-91 Christopher Fedor, School of Computer Science, CMU
* Added tcaReferenceId for lisp interface to tpl constraints.
*
* 13-Aug-90 Christopher Fedor, School of Computer Science, CMU
* created.
*
******************************************************************************/

#include "stdio.h"

#include "tcaInternal.h"
#include "formatters.h"
#include "list.h"
#include "datamsg.h"

extern char LispFlagGlobal;

extern int dataStructureSize();

extern TCA_RETURN_VALUE_TYPE sendMessage();
extern TCA_RETURN_VALUE_TYPE waitForReply2();

LIST_PTR tcaRefFreeList;

tcaRefInitialize()
{
  tcaRefFreeList = listCreate();
}

/* 26-Jul-91: Reid: There was a memory leak here -- sometimes the name is
   newly malloc'd (see primFmtters) and sometimes an old copy is passed in.
   In either case, the name is not freed.  Solution is to always copy the
   string name in tcaRefCreate and free it in tcaRefFree
   */

TCA_REF_PTR tcaRefCreate(msg, name, refId)
MSG_PTR msg;
char *name;
int refId;
{
  TCA_REF_PTR tcaRef;
  int length;

  if (!listLength(tcaRefFreeList))
    tcaRef = NEW(TCA_REF_TYPE);
  else 
    tcaRef = (TCA_REF_PTR)listPopItem(tcaRefFreeList);

  tcaRef->refId = refId;    
  tcaRef->msg = msg;

  if (!name) {
    tcaRef->name = NULL;
  } else {
    length = strlen(name);
    tcaRef->name = tcaMalloc((unsigned)(length+1));
    bcopy(name, tcaRef->name, length);
    tcaRef->name[length] = '\0';
  }

  return tcaRef;
}

void tcaRefFree(tcaRef)
TCA_REF_PTR tcaRef;
{
  tcaRef->refId = 0;
  tcaRef->msg = NULL;

  tcaFree(tcaRef->name);
  tcaRef->name = NULL;

  listInsertItemFirst((char *)tcaRef, tcaRefFreeList);
}

/******************************************************************************
*
* FUNCTION: void tcaReferenceRelease(ref)
*
* DESCRIPTION: 
* Frees the reference and 
*
* INPUTS: TCA_REF_PTR ref;
*
* OUTPUTS: none
*
******************************************************************************/
void tcaReferenceRelease(ref)
     TCA_REF_PTR ref;
{
  if (ref) {
    (void)tcaInform("referenceRelease", (void *)&(ref->refId));

    tcaRefFree(ref);
  }
}

/******************************************************************************
*
* FUNCTION: char *tcaReferenceName(ref)
*
* DESCRIPTION: 
* Returns the name of the message associated with the reference.
*
* INPUTS: TCA_REF_PTR ref;
*
* OUTPUTS: char *
*
******************************************************************************/

char *tcaReferenceName(ref)
TCA_REF_PTR ref;
{
  return ref->name;
}


/******************************************************************************
*
* FUNCTION: int tcaReferenceId(ref)
*
* DESCRIPTION: 
* Done primarily for lisp.
* Returns the msg id of the message associated with the reference.
*
* INPUTS: TCA_REF_PTR ref;
*
* OUTPUTS: int
*
******************************************************************************/

int tcaReferenceId(ref)
TCA_REF_PTR ref;
{
  if (ref)
    return ref->refId;
  else
    return NO_REF;
}


/******************************************************************************
*
* FUNCTION: char *tcaReferenceData(ref)
*
* DESCRIPTION: 
* Returns a pointer to the data that was sent when the message
* that is associated with the reference was sent.
*
* INPUTS: TCA_REF_PTR ref;
*
* OUTPUTS: Pointer to data.
*
******************************************************************************/

char *tcaReferenceData(ref)
TCA_REF_PTR ref;
{
  int refId;
  MSG_PTR msg;
  char *msgData;
  TCA_REF_PTR waitRef;
  TCA_RETURN_VALUE_TYPE returnValue;

  msg = msgFind("referenceData");

  if (!ref->msg) {
    if (!ref->name) {
      /* 17-Jun-91: fedor: start enforcing correct refs */
      TCA_MOD_ERROR1("ERROR: tcaReferenceData: Badly Formed Reference: %d\n",
		     ref->refId);
    }
    ref->msg = msgFind(ref->name);
  }

  /* 17-Jun-91: fedor: check if any message form */
  if (!ref->msg->msgData->msgForm)
    return NULL;

  refId = nextSendMessageRef();
  returnValue = sendMessage((TCA_REF_PTR)NULL, msg, 
			    (char *)&ref->refId, (char *)NULL);

  if (returnValue != Success) 
    tcaModError("ERROR: tcaReferenceData: sendMessage Failed.\n");

  waitRef = tcaRefCreate(ref->msg, ref->name, refId);
  
  msgData = tcaMalloc((unsigned)dataStructureSize(ref->msg->msgData->msgForm));

  returnValue = waitForReply2(waitRef, msgData, 1);

  tcaRefFree(waitRef);

  if (returnValue == NullReply) {
    /* 17-Jun-91: fedor: if NullReply then nothing else was malloced. */
    cfree(msgData);
    return NULL;
  }
  else
    return msgData;
}

#ifdef LISP
/* 25-Jun-91: fedor: this is really a cheap shot at making ref data work
   for lisp - lets hope it works. */
int tcaRefDataLisp(ref)
TCA_REF_PTR ref;
{
  int refId;
  MSG_PTR msg;
  char *msgData;
  TCA_REF_PTR waitRef;
  TCA_RETURN_VALUE_TYPE returnValue;

  msg = msgFind("referenceData");

  if (!ref->msg) {
    if (!ref->name) {
      /* 17-Jun-91: fedor: start enforcing correct refs */
      TCA_MOD_ERROR1("ERROR: tcaReferenceData: Badly Formed Reference: %d\n",
		     ref->refId);
    }
    ref->msg = msgFind(ref->name);
  }

  /* 17-Jun-91: fedor: check if any message form */
  if (!ref->msg->msgData->msgForm)
    return 0;

  refId = nextSendMessageRef();
  returnValue = sendMessage((TCA_REF_PTR)NULL, msg, 
			    (char *)&ref->refId, (char *)NULL);

  if (returnValue != Success) 
    tcaModError("ERROR: tcaReferenceData: sendMessage Failed.\n");

  waitRef = tcaRefCreate(ref->msg, ref->name, refId);
  
  returnValue = waitForReply2(waitRef, &LispFlagGlobal, 1);

  tcaRefFree(waitRef);

  if (returnValue == NullReply) {
    /* 17-Jun-91: fedor: if NullReply then nothing else was malloced. */
    return 0;
  }
  else
    return 1;
}

#endif


/******************************************************************************
*
* FUNCTION:  TCA_REF_STATUS_TYPE tcaReferenceStatus(ref)
*
* DESCRIPTION: 
* Returns the msg status of the message associated with the reference.
*
* INPUTS: TCA_REF_PTR ref;
*
* OUTPUTS: TCA_REF_STATUS_TYPE
*
******************************************************************************/

TCA_REF_STATUS_TYPE tcaReferenceStatus(ref)
TCA_REF_PTR ref;
{
  TCA_REF_STATUS_TYPE status;

  (void)tcaQuery("referenceStatusMsg", (void *)&(ref->refId), (void *)&status);

  return status;
}
