/******************************************************************************
*
* PROJECT: Carnegie Mellon Planetary Rover Project
*          Task Control Architecture 
* 
* (c) Copyright 1991 Christopher Fedor and Reid Simmons.  All rights reserved.
* 
* MODULE: Formatters
*
* FILE: parseFmttrs.c
*
* ABSTRACT:
* Parser and Printer for Formatter Data Structures
*
* REVISION HISTORY
*
* 10-Feb-89 Reid Simmons, School of Computer Science, CMU
* Created.
*
******************************************************************************/

#include "stdio.h"
#include "basics.h"
#include "formatters.h"
#include "hash.h"
#include "lex.h"
#include "list.h"

#ifndef TRUE
#define TRUE  1
#endif

#ifndef FALSE
#define FALSE 0
#endif

extern TokenPtr TokenList;
extern int LexString();
extern int PrintToken();
extern int PrintTokenList();
extern TokenPtr NextToken();

extern FORMAT_PTR createIntegerFormat();

/* forward ref */
extern FORMAT_PTR Parse(); 

extern HASH_TABLE_PTR formatNamesTable;

static  char *ParseString = "";
static TokenPtr TokenListHead = NULL;


/*****************************************************************
 *  Mainly used by "Print_Formatter".  Defined here because      *
 *  "formatNamesTable" is defined in this file.              *
 *****************************************************************/

static FORMAT_PTR Search_Format;
static char *Found_Key;

int find_same_format (key, format)
char *key; FORMAT_PTR format;
{ 
  if ((format->type == Search_Format->type) &&
      (format->formatter.i == Search_Format->formatter.i)) {
    Found_Key = key;
    return FALSE;
  }
  else return TRUE;
}

char *search_hash_table_for_format(format)
FORMAT_PTR format;
{ 
  Found_Key = NULL;
  Search_Format = format;
  hashTableIterate(find_same_format, formatNamesTable);
  return Found_Key;
}

void FreeTokenList (TokenList) TokenPtr TokenList;
{ 
  TokenPtr LastToken;

  while (TokenList) {
    LastToken = TokenList;
    TokenList = TokenList->next;
    if (LastToken->Type == STR_TOK)
      tcaFree(LastToken->value.str);
    tcaFree((char *)LastToken);
  }
}

void ParserError(bad_token, expecting)
TokenPtr bad_token;
char *expecting;
{ 
  int i;

  (void)fprintf(stderr, "Format Parsing Error: Expected %s.\n%s\n", 
		expecting, ParseString);
  for (i=0; i<bad_token->Loc; i++) (void)fprintf(stderr, " ");
  tcaModError("^\n");
}

FORMAT_PTR new_f_formatter(type, formatter)
FORMAT_CLASS_TYPE type;
FORMAT_PTR formatter;
{ 
  FORMAT_PTR new_formatter;

  new_formatter = NEW_FORMATTER();
  new_formatter->type = type;
  new_formatter->formatter.f = formatter;
  return new_formatter;
}

/* First element in a FORMAT_ARRAY_PTR is the size of the array. */
FORMAT_PTR new_a_formatter(type, array_size)
FORMAT_CLASS_TYPE type; 
int array_size;
{ 
  FORMAT_PTR new_formatter;

  new_formatter = NEW_FORMATTER();
  new_formatter->type = type;
  new_formatter->formatter.a = NEW_FORMAT_ARRAY(array_size);
  new_formatter->formatter.a[0].i = array_size;
  return new_formatter;
}

FORMAT_PTR Struct_Format()
{ 
  FORMAT_PTR Form, subform;
  LIST_PTR format_list;
  int num_formats, i;

  format_list = listCreate();
  listInsertItem((char *)Parse(TRUE), format_list);
  num_formats = 1;

  while (TokenList->Type == COMMA_TOK) {
    (void)NextToken();
    listInsertItem((char *)Parse(TRUE), format_list);
    num_formats++;
  }

  Form = new_a_formatter(StructFMT, num_formats+1);
  /* Index from high to low since "format_list" 
     has formatters in reverse order */
  subform = (FORMAT_PTR)listFirst(format_list);
  for(i=num_formats;i>0;i--) {
    Form->formatter.a[i].f = subform;
    subform = (FORMAT_PTR)listNext(format_list);
  }

  return Form;
}

FORMAT_PTR Fixed_Array_Format()
{ 
  FORMAT_PTR Form, next_format;
  TokenPtr Token, tmp;
  int NumberOfIndexes, Continue, array_index;

  next_format = Parse(FALSE); /* Formatter */

  Token = NextToken();
  if (Token->Type != COLON_TOK)
    ParserError(Token, "':'");

  tmp = TokenList;
  NumberOfIndexes = 0;
  Continue = TRUE;
  do {
    if (tmp->Type != INT_TOK) 
      ParserError(tmp, "an integer value");
    else {
      NumberOfIndexes++;
      tmp = tmp->next;
      if (tmp->Type == COMMA_TOK) {
	tmp = tmp->next;
	Continue = TRUE;
      } else if (tmp->Type == RBRACK_TOK) 
	Continue = FALSE;
      else 
	ParserError(tmp, "a ',' or ']'");
    }
  } while (Continue);

  Form = new_a_formatter(FixedArrayFMT, NumberOfIndexes+2);
  Form->formatter.a[1].f = next_format;

  /* this time munch tokens */

  NumberOfIndexes += 2;
  for (array_index=2; array_index < NumberOfIndexes; array_index++) {
    Token = NextToken(); /* This is an INT_TOK */
    Form->formatter.a[array_index].i = Token->value.num;
    
    Token = NextToken(); /* This is a COMMA_TOK or a RBRACK_TOK */
  }

  return Form;
}

FORMAT_PTR Var_Array_Format()
{ 
  FORMAT_PTR Form, next_format;
  TokenPtr Token, tmp;
  int NumberOfIndexes, Continue, array_index;

  next_format = Parse(FALSE); /* Formatter */

  Token = NextToken();
  if (Token->Type != COLON_TOK)
    ParserError(Token, "':'");

  tmp = TokenList;
  NumberOfIndexes = 0;
  Continue = TRUE;
  do {
    if (tmp->Type != INT_TOK) 
      ParserError(tmp, "an integer value");
    else {
      NumberOfIndexes++;
      tmp = tmp->next;
      if (tmp->Type == COMMA_TOK) {
	tmp = tmp->next;
	Continue = TRUE;
      } else if (tmp->Type == GT_TOK) 
	Continue = FALSE;
      else 
	ParserError(tmp, "a ',' or '>'");
    }
  } while (Continue);

  Form = new_a_formatter(VarArrayFMT, NumberOfIndexes+2);
  Form->formatter.a[1].f = next_format;

  /* this time munch tokens */

  NumberOfIndexes += 2;
  for (array_index=2; array_index < NumberOfIndexes; array_index++) {
    Token = NextToken(); /* This is an INT_TOK */
    Form->formatter.a[array_index].i = Token->value.num;
    
    Token = NextToken(); /* This is a COMMA_TOK or a GT_TOK */
  }

  return Form;
}

FORMAT_PTR SelfPtr_Format()
{ 
  return new_f_formatter(PointerFMT, (FORMAT_PTR)NULL);
}

FORMAT_PTR Ptr_Format()
{ 
  return new_f_formatter(PointerFMT, Parse(FALSE));
}

FORMAT_PTR Length_Format(Token)
TokenPtr Token;
{ 
  return createIntegerFormat(LengthFMT, Token->value.num);
}

FORMAT_PTR Named_Format(Token)
TokenPtr Token;
{ 
  FORMAT_PTR Form;

  Form = (FORMAT_PTR)hashTableFind(Token->value.str, formatNamesTable);
  if (!Form) {
    TCA_MOD_ERROR2("Format %s is not registered\n%s\n",
		   Token->value.str, ParseString);
  }

  return Form;
}

FORMAT_PTR Primitive_Format(Token)
TokenPtr Token;
{ 
  return Named_Format(Token);
}

/*********************************************/

FORMAT_PTR ParseFormatString(FormString) 
char *FormString;
{ 
  FORMAT_PTR Formatter;

  ParseString = FormString;
  LexString(FormString);
  TokenListHead = TokenList;
  Formatter = Parse(FALSE);
  FreeTokenList(TokenListHead);
  return Formatter;
}


/* "within_struct_flag" is TRUE if the enclosing format is "{...}" */
FORMAT_PTR Parse(within_struct_flag) 
int within_struct_flag;
{ 
  TokenPtr Token;
  FORMAT_PTR ReturnForm;

  Token = NextToken();

  switch(Token->Type) {
  case LBRACE_TOK:
    ReturnForm = Struct_Format();
    Token = NextToken();
    if (Token->Type != RBRACE_TOK) ParserError(Token, "'}'");
    break;
  case LBRACK_TOK:
    ReturnForm = Fixed_Array_Format();
    break;
  case LT_TOK:
    if (!within_struct_flag)
      ParserError(Token,
	  "var array format '<..>' embedded within a structure '{..}'");
    else
      ReturnForm = Var_Array_Format();
    break;
  case STAR_TOK:
    if (TokenList->Type == BANG_TOK) {
      Token = NextToken();	/* eat BANG_TOK */ 
      if (!within_struct_flag) 
	ParserError(Token, 
		    "self pointer '*!' embedded within a structure '{..}'");
      else
	ReturnForm = SelfPtr_Format();
    }
    else
      ReturnForm = Ptr_Format();
    break;
  case INT_TOK:
    ReturnForm = Length_Format(Token);
    break;
  case STR_TOK:
    ReturnForm = Named_Format(Token);
    break;
  case PRIMITIVE_TOK:
    ReturnForm = Primitive_Format(Token);
    break;
  case EOS_TOK:
    ParserError(Token, 
		"additional tokens; premature end of string encountered");
    break;
  default:
    ParserError(Token, "a different token type");
  }

  return ReturnForm;
}
