/******************************************************************************
*
* PROJECT: Carnegie Mellon Planetary Rover Project
*          Task Control Architecture 
* 
* (c) Copyright 1991 Christopher Fedor and Reid Simmons.  All rights reserved.
* 
* MODULE: Printing Data.
*
* FILE: printData.c
*
* ABSTRACT:
* Print Out Data According to Formatter Specification.
*
* REVISION HISTORY
*
* 13-Sep-91 Christopher Fedor, School of Computer Science, CMU
* Made buffer a single global scratch area. Increased it to hold 400
* characters which should hold big ugly numbers without sprintf death.
* buffer as a global should only be used if the sprintf is followed by
* the printString routine. There is no recursion at this point so this should
* work. The termination character for the buffer is provided by sprintf -
* or one could explicitly set buffer[30] = '\0'.
* Made sure all fprintf routines had the correct number of arguments.
* Additional changes to software standards.
*
*  2-May-90 Christopher Fedor, School of Computer Science, CMU
* Added call to alignFied in Print_Structured_Data so that SUN4 version
* of central could correctly display data. Note: should look into having
* DPrint option of primFmttrs return correct number of bytes instead of
* a dummy of 1 to be consistent with other _TRANS operations.
* Perhaps I am just unsure why this was done.
*
* 28-Nov-89 Reid Simmons, School of Computer Science, CMU
* Added code to abort print out if PRINT_LENGTH exceeded.
*
* 24-Jul-89 Reid Simmons, School of Computer Science, CMU
* Created.
*
******************************************************************************/

#include "stdio.h"

#include "tcaInternal.h"
#include "logging.h"
#include "hash.h"

#include "formatters.h"
#include "tcaMatrix.h"

/*
* Number of characters to keep on a line following the beginning of a
* structured formatter. 10 was choosen arbitrarily. Might not always
* work
*/
#define DEFAULT_KEEP 10   

#define LINE_LENGTH  80  /* Number of characters per line. */
#define PRINT_LENGTH  5  /* Maximum number of lines printed. */
#define DEC_PLACES    2  /* Number of decimal places for floats and doubles. */

/* 14-Sep-91: fedor: used in logging.c */
int indentGlobal = 0;    /* Number of spaces to indent first line. */

int cursorPosGlobal = 0; /* Position of last printed character */
int lineNumGlobal   = 0; /* Number of lines printed out for current data */
int truncatedGlobal = 0; /* Non-zero only if ellipsis (...) has been printed */

/* 14-Sep-91: fedor: this is actually more space savings than
   many static buffers. sprintf should insure the ending '\0' termination.
   400 characters should be large enough to hold a max_normal vaule.
   Should only be used if the sprintf is done and then it is printed. */
char buffer[400]; /* main scratch buffer for printing */

extern int (*TransTable[])(); /* Defined in formatters.c */


/******************************************************************************
*
* FUNCTION: void newLine(stream)
*
* DESCRIPTION:
*
* INPUTS: FILE *stream;
*
* OUTPUTS: void.
*
******************************************************************************/

void newLine(stream)
FILE *stream;
{
  fprintf(stream, "\n");
  lineNumGlobal++;
  cursorPosGlobal = 0;
}


/******************************************************************************
*
* FUNCTION: void printTab(stream, tabPosition)
*
* DESCRIPTION: 
* Print spaces from current cursorPosGlobal to Tab_Position.
* Does nothing if position already passed.
*
* INPUTS: 
* FILE *stream;
* int tabPosition;
*
* OUTPUTS: void.
*
* NOTES: makes use of printf min field width.
*
******************************************************************************/

void printTab(stream, tabPosition)
FILE *stream;
int tabPosition;
{ 
  int spaces;

  if (!truncatedGlobal) {
    if (cursorPosGlobal < tabPosition) {
      spaces = tabPosition - cursorPosGlobal;
      fprintf(stream, "%*s", spaces, " ");
      cursorPosGlobal += spaces;
    }
  }
}


/******************************************************************************
*
* FUNCTION: void printSpace(stream)
*
* DESCRIPTION:
* Print out a space to the stream, unless at the end of a line,
* in which case, go to the next line.
*
* INPUTS: 
*
* OUTPUTS: 
*
******************************************************************************/

void printSpace(stream)
FILE *stream;
{
  if (!truncatedGlobal) {
    if (cursorPosGlobal == LINE_LENGTH) {
      newLine(stream);
      printTab(stream, indentGlobal);
    }

    fprintf(stream, "%c", ' ');
    cursorPosGlobal++;
  }
}


/******************************************************************************
*
* FUNCTION: void printString(stream, string, keepWithNext)
*
* DESCRIPTION:
* Print out the string to the stream.
* Move to next line if less than "keep_with_next" characters would
* remain on the current line after printing the string.
*
* INPUTS: 
* FILE *stream;
* char *string;
* int keepWithNext;
*
* OUTPUTS: void.
*
******************************************************************************/

void printString(stream, string, keepWithNext)
FILE *stream;
char *string;
int keepWithNext;
{ 
  int length;

  if (!truncatedGlobal) {
    if (string)
      length = strlen(string);
    else
      length = 0;

    if ((cursorPosGlobal + length + keepWithNext) > LINE_LENGTH) {
      newLine(stream);
      printTab(stream, indentGlobal+1);
    }
    if (lineNumGlobal < PRINT_LENGTH) {
      if (string) {
	fprintf(stream, "%s", string);
	cursorPosGlobal += length;
      }
      else {
	fprintf(stream, "NULL");
	cursorPosGlobal += 4;
      }
    }
    else {
      truncatedGlobal = 1;
      fprintf(stream, "...");
    }
  }
}


/******************************************************************************
*
* FUNCTION: void printInt(stream, intPtr, keepWithNext)
*
* DESCRIPTION:
* Print out the integer to the stream.
* Move to next line if less than "keep_with_next" characters would
* remain on the current line after printing the integer.
*
* INPUTS: 
* FILE *stream;
* int *intPtr;
* int keepWithNext;
*
* OUTPUTS: void.
*
******************************************************************************/

void printInt(stream, intPtr, keepWithNext)
FILE *stream;
int *intPtr;
int keepWithNext;
{ 
  (void)sprintf(buffer, "%d", *intPtr);
  printString(stream, buffer, keepWithNext);
}


/******************************************************************************
*
* FUNCTION: void printChar(stream, charPtr, keepWithNext)
*
* DESCRIPTION:
*
* INPUTS: 
* FILE *stream;
* char *charPtr;
* int keepWithNext;
*
* OUTPUTS: void.
*
******************************************************************************/

void printChar(stream, charPtr, keepWithNext)
FILE *stream;
char *charPtr;
int keepWithNext;
{ 
  (void)sprintf(buffer, "%c", *charPtr);
  printString(stream, buffer, keepWithNext);
}


/******************************************************************************
*
* FUNCTION: void printShort(stream, shortPtr, keepWithNext)
*
* DESCRIPTION:
*
* INPUTS: 
* FILE *stream;
* short *shortPtr;
* int keepWithNext;
*
* OUTPUTS: void.
*
******************************************************************************/

void printShort(stream, shortPtr, keepWithNext)
FILE *stream;
short *shortPtr;
int keepWithNext;
{ 
  (void)sprintf(buffer, "%d", *shortPtr);
  printString(stream, buffer, keepWithNext);
}


/******************************************************************************
*
* FUNCTION: void printLong(stream, longPtr, keepWithNext)
*
* DESCRIPTION:
*
* INPUTS: 
* FILE *stream;
* long *longPtr;
* int keepWithNext;
*
* OUTPUTS: void.
*
******************************************************************************/

void printLong(stream, longPtr, keepWithNext)
FILE *stream;
long *longPtr;
int keepWithNext;
{ 
  (void)sprintf(buffer, "%ld", *longPtr);
  printString(stream, buffer, keepWithNext);
}


/******************************************************************************
*
* FUNCTION: void printFloat(stream, floatPtr, keepWithNext)
*
* DESCRIPTION:
* Print out the floating_point number to the stream, with 
* accuracy of "DEC_PLACES".
* Move to next line if less than "keep_with_next" characters would
* remain on the current line after printing the float.
*
* INPUTS: 
* FILE *stream;
* float *floatPtr;
* int keepWithNext;
*
* OUTPUTS: void.
*
******************************************************************************/

void printFloat(stream, floatPtr, keepWithNext)
FILE *stream;
float *floatPtr;
int keepWithNext;
{ 
  (void)sprintf(buffer, "%.*f", DEC_PLACES, *floatPtr);
  printString(stream, buffer, keepWithNext);
}


/******************************************************************************
*
* FUNCTION: void printDouble(stream, doublePtr, keepWithNext)
*
* DESCRIPTION:
* Print out the double floating_point number to the stream, with
* accuracy of "DEC_PLACES".
* Move to next line if less than "keep_with_next" characters would
* remain on the current line after printing the double.
*
* INPUTS: 
* FILE *stream;
* double *doublePtr;
* int keepWithNext;
*
* OUTPUTS: void.
*
******************************************************************************/

void printDouble(stream, doublePtr, keepWithNext)
FILE *stream;
double *doublePtr;
int keepWithNext;
{ 
  (void)sprintf(buffer, "%.*lf", DEC_PLACES, *doublePtr);
  printString(stream, buffer, keepWithNext);
}


/******************************************************************************
*
* FUNCTION: void printHex(stream, intPtr, keepWithNext)
*
* DESCRIPTION:
*
* INPUTS: 
* FILE *stream;
* unsigned *intPtr;
* int KeepWithNext;
*
* OUTPUTS: void.
*
******************************************************************************/

void printHex(stream, intPtr, keepWithNext)
FILE *stream;
unsigned *intPtr;
int keepWithNext;
{
  (void)sprintf(buffer, "%#x\0", *intPtr);
  printString(stream, buffer, keepWithNext);
}


/******************************************************************************
*
* FUNCTION: void printByte(stream, bytePtr, keepWithNext)
*
* DESCRIPTION:
* Print out the byte (in hex) to the stream.
* Move to next line if less than "keep_with_next" characters would
* remain on the current line after printing the byte.
*
* INPUTS: 
* FILE *stream;
* char *bytePtr;
* int keepWithNext;
*
* OUTPUTS: void.
*
******************************************************************************/

void printByte(stream, bytePtr, keepWithNext)
FILE *stream;
char *bytePtr;
int keepWithNext;
{ 
  unsigned byte;

  byte = 0;
  bcopy(bytePtr, ((char *)&byte)+sizeof(int)-sizeof(char), sizeof(char));
  printHex(stream, &byte, keepWithNext);
}


/******************************************************************************
*
* FUNCTION: void printTwoByte(stream, twoBytePtr, keepWithNext)
*
* DESCRIPTION:
* Print out the double byte (in hex) to the stream.
* Move to next line if less than "keep_with_next" characters would
* remain on the current line after printing the bytes.
*
* INPUTS: 
* FILE *stream;
* char *twoBytePtr;
* int keepWithNext;
*
* OUTPUTS: void.
*
******************************************************************************/

void printTwoByte(stream, twoBytePtr, keepWithNext)
FILE *stream;
char *twoBytePtr;
int keepWithNext;
{ 
  unsigned twoByte;

  twoByte = 0;
  bcopy(twoBytePtr, ((char *)&twoByte)+sizeof(int)-2*sizeof(char), 
	2*sizeof(char));
  printHex(stream, &twoByte, keepWithNext);
}


/******************************************************************************
*
* FUNCTION: void printCommaSpace(stream)
*
* DESCRIPTION:
*
* INPUTS: FILE *stream;
*
* OUTPUTS: void.
*
******************************************************************************/

void printCommaSpace(stream)
FILE *stream;
{
  printString(stream, ",", 0);
  printSpace(stream);
}


/******************************************************************************
*
* FUNCTION: void startPrint(stream)
*
* DESCRIPTION:
*
* INPUTS: FILE *stream;
*
* OUTPUTS: void.
*
******************************************************************************/

void startPrint(stream)
FILE *stream;
{
  if (cursorPosGlobal != 0)
    fprintf(stream, "\n");
  cursorPosGlobal = 0;
  lineNumGlobal = 0;
  truncatedGlobal = 0;
  printTab(stream, indentGlobal);
}


/******************************************************************************
*
* FUNCTION: void endPrint(stream)
*
* DESCRIPTION:
*
* INPUTS: FILE *stream;
*
* OUTPUTS: void.
*
******************************************************************************/

void endPrint(stream)
FILE *stream;
{
  newLine(stream);
  truncatedGlobal = 0;
}


/******************************************************************************
*
* FUNCTION: int printArrayData(stream, format, dataArray, AStart, 
*                  formatArray, dimension, keepWithNext, startToken, endToken)
*
* DESCRIPTION:
*
* INPUTS: 
* FILE *stream;
* FORMAT_PTR format;
* char *dataArray;
* int AStart, dimension, keepWithNext;
* FORMAT_ARRAY_PTR formatArray;
* char *startToken, *endToken;
*
* OUTPUTS: int
*
******************************************************************************/

extern void Print_Structured_Data (); /* Defined below, used recursively */

int printArrayData(stream, format, dataArray, AStart, formatArray,
		     dimension, keepWithNext, startToken, endToken)
FILE *stream;
FORMAT_PTR format;
char *dataArray;
int AStart, dimension, keepWithNext;
FORMAT_ARRAY_PTR formatArray;
char *startToken, *endToken;
{ 
  int i, element_size, last_dimension_p, last_element_p;
  int vector_length, next_keep;

  printString(stream, startToken, DEFAULT_KEEP);

  last_dimension_p = (dimension+1 == formatArray[0].i);
  if (last_dimension_p)
    element_size = dataStructureSize(format);
  
  vector_length = formatArray[dimension].i;

  for (i=0; i<vector_length; i++) {
    if (truncatedGlobal) 
      return 0;
    else {
      last_element_p = (i+1 == vector_length);
      next_keep = ((last_element_p) ? keepWithNext+1 : 1);
      if (last_dimension_p) {
	Print_Structured_Data(stream, format, dataArray, AStart, 
			      (FORMAT_PTR)NULL, next_keep);
	AStart += element_size;
	if (!last_element_p)
	  printCommaSpace(stream);
      }
      else {
	AStart = printArrayData(stream, format, dataArray, AStart, 
				  formatArray, dimension+1, next_keep, 
				  startToken, endToken);
	if (!last_element_p)
	  printCommaSpace(stream);
      }
    }
  }

  printString(stream, endToken, keepWithNext);

  return AStart;
}


/******************************************************************************
*
* FUNCTION: FORMAT_ARRAY_PTR fix_format_array(varFormatArray, 
*                           parentStructArray, dataPtr, DStart)
*
* DESCRIPTION:
* Create a fixed format array that contains all the dimensions of
* the variable length array.
*
* INPUTS: 
* FORMAT_ARRAY_PTR varFormatArray, parentStructArray;
* DATA_PTR dataPtr;
* int DStart;
*
* OUTPUTS: 
*
******************************************************************************/

FORMAT_ARRAY_PTR fix_format_array(varFormatArray, parentStructArray,
				  dataPtr, DStart)
FORMAT_ARRAY_PTR varFormatArray, parentStructArray;
DATA_PTR dataPtr;
int DStart;
{ 
  FORMAT_ARRAY_PTR fixed_format_array;
  FORMAT_PTR current_formatter;
  int array_size, found_place, current_place, i, j, size_place, offset, size;

  found_place = 0;
  for (current_place=1; !found_place; current_place++) {
    current_formatter = parentStructArray[current_place].f;
    found_place = ((current_formatter->type == VarArrayFMT) &&
		   (current_formatter->formatter.a == varFormatArray));
  }
  current_place--;

  array_size = varFormatArray[0].i;
  fixed_format_array = (FORMAT_ARRAY_PTR)tcaMalloc((unsigned)array_size * 
						   sizeof(int));
  fixed_format_array[0].i = array_size;
  fixed_format_array[1].f = varFormatArray[1].f;
  for (i=2; i<array_size; i++) {
    size_place = varFormatArray[i].i;
    offset = 0;
    if (current_place < size_place)
      for (j=current_place; j<size_place; j++)
	offset += dataStructureSize(parentStructArray[j].f);
    else 
      for (j=size_place; j<current_place; j++)
	offset -= dataStructureSize(parentStructArray[j].f);
    bcopy(dataPtr+DStart+offset, (char *)&size, sizeof(int));
    fixed_format_array[i].i = size;
  }

  return fixed_format_array;
}


/******************************************************************************
*
* FUNCTION:
*
* DESCRIPTION:
* If there is no formatter, it means that the ptr is recursive (self-ptr)
*
* INPUTS: 
*
* OUTPUTS: 
*
******************************************************************************/

#define CHOOSE_PTR_FORMAT(Format, parentFormat) \
  (((Format)->formatter.f) ? ((Format)->formatter.f) : parentFormat)

void Print_Structured_Data (Stream, Format, Data_Ptr, DStart, parentFormat,
			    Keep_With_Next)
FILE *Stream;
FORMAT_PTR Format, parentFormat;
char *Data_Ptr;
int DStart, Keep_With_Next;
{ 
  int (*format_proc)();
  int dummy, i, last_element_p, next_keep, diff;
  DATA_PTR StructPtr, ArrayPtr;
  FORMAT_ARRAY_PTR format_array, fixed_format_array;

  switch (Format->type) {
  case LengthFMT:
    printString(Stream, "(", DEFAULT_KEEP);
    /* Print out all but the last few bytes of data */
    for (i=0; i<Format->formatter.i-sizeof(int); i+=sizeof(int)) {
      if (truncatedGlobal)
	return;
      else {
	printHex(Stream, (unsigned *)(Data_Ptr+DStart+i), 1);
	printCommaSpace(Stream);
      }
    }
    /* print what's remaining (from 1 to "sizeof(int)" bytes) */
    diff = Format->formatter.i - i;
    dummy = 0;
    bcopy(Data_Ptr+DStart+i, ((char *)&dummy)+sizeof(int)-diff, diff);
    printHex(Stream, (unsigned *)&dummy, Keep_With_Next+1);
    printString(Stream, ")", Keep_With_Next);
    /* 
      printString(Stream, "**length Formatter**", Keep_With_Next);
      */
    break;

  case PrimitiveFMT: 
    format_proc = TransTable[Format->formatter.i];
    dummy = format_proc(DPrint, Data_Ptr, DStart, Stream, Keep_With_Next);
    break;

  case PointerFMT:
    StructPtr = REF(DATA_PTR, Data_Ptr, DStart);
    if (StructPtr) { 
      printString(Stream, "*", DEFAULT_KEEP);
      Print_Structured_Data(Stream, CHOOSE_PTR_FORMAT(Format, parentFormat),
			    StructPtr, 0, (FORMAT_PTR)NULL, Keep_With_Next);
    }
    else
      printString(Stream, "NULL", Keep_With_Next);
    break;

  case StructFMT:
    printString(Stream, "{", DEFAULT_KEEP);
    format_array = Format->formatter.a;
    for (i=1;i<format_array[0].i;i++) {
      if (truncatedGlobal)
	return;
      else {
	last_element_p = (i+1 == format_array[0].i);
	next_keep = ((last_element_p) ? Keep_With_Next+1 : 1);
	Print_Structured_Data(Stream, format_array[i].f, Data_Ptr, DStart, 
			      Format, next_keep);
	DStart += dataStructureSize(format_array[i].f);
	DStart = alignField(Format, i, DStart);
	if (last_element_p)
	  printString(Stream, "}", Keep_With_Next);
	else
	  printCommaSpace(Stream);
      }
    }
    break;

  case FixedArrayFMT:
    (void)printArrayData(Stream, Format->formatter.a[1].f, Data_Ptr, DStart, 
			 Format->formatter.a, 2, Keep_With_Next, "[", "]");
    break;

  case VarArrayFMT:
    ArrayPtr = REF(DATA_PTR, Data_Ptr, DStart);
    format_array = Format->formatter.a;
    if (ArrayPtr) {
      fixed_format_array = fix_format_array(format_array, parentFormat->formatter.a,
					    Data_Ptr, DStart);
      (void)printArrayData(Stream, format_array[1].f, ArrayPtr, 0, 
			   fixed_format_array, 2, Keep_With_Next, "<", ">");
      tcaFree((char *)fixed_format_array);
    }
    else
      printString(Stream, "NULL", Keep_With_Next);
  }
}


/******************************************************************************
*
* FUNCTION: void Print_Formatted_Data(stream, format, dataPtr)
*
* DESCRIPTION:
* Print out some data (linearally) according to the given format.
* "Data_Ptr" is a pointer to some arbitrary data structure.
* "Format" is the format of the data
* "Stream" is where to print the data (e.g. stdout)
*
* INPUTS: 
* FILE *stream;
* FORMAT_PTR format;
* char **dataPtr;
*
* OUTPUTS: void.
*
******************************************************************************/

void Print_Formatted_Data(stream, format, dataPtr)
FILE *stream;
FORMAT_PTR format;
char *dataPtr;
{
  startPrint(stream);
  Print_Structured_Data(stream, format, dataPtr, 0, (FORMAT_PTR)NULL, 0);
  endPrint(stream);
}


/******************************************************************************
*
* FUNCTION: void Print_Data(formatString, dataPtr)
*
* DESCRIPTION:
* Print the data (pointed to be "Data_Ptr") according to the
* "Format_String" to the standard output (stdout).
*
* INPUTS: 
* char *formatString;
* char **dataPtr;
*
* OUTPUTS: void.
*
******************************************************************************/

void Print_Data(formatString, dataPtr)
char *formatString;
char *dataPtr;
{
  Print_Formatted_Data(stdout, ParseFormatString(formatString), dataPtr);
}


/******************************************************************************
*
* FUNCTION: void PrintArrayFormat(stream, startToken, formatArray, 
*                                 endToken, keepWithNext)
*
* DESCRIPTION:
*
* INPUTS: 
* FILE *stream;
* char *startToken, *endToken;
* FORMAT_ARRAY_PTR formatArray;
* int keepWithNext;
*
* OUTPUTS: void.
*
******************************************************************************/

void Print_Formatter1(); /* Defined below, used recursively */

void PrintArrayFormat(stream, startToken, formatArray, endToken, keepWithNext)
FILE *stream;
char *startToken, *endToken;
FORMAT_ARRAY_PTR formatArray;
int keepWithNext;
{ 
  int i, lastElementP;

  printString(stream, startToken, DEFAULT_KEEP);
  Print_Formatter1(stream, formatArray[1].f, keepWithNext);
  printString(stream, " : ", DEFAULT_KEEP);
  for (i=2;i<formatArray[0].i;i++) {
    if (truncatedGlobal)
      return;
    else {
      lastElementP = (i+1 == formatArray[0].i);
      printInt(stream, &(formatArray[i].i),
		((lastElementP) ? keepWithNext+1 : 1));
      if (lastElementP)
	printString(stream, endToken, keepWithNext);
      else
	printCommaSpace(stream);
    }
  }
}


/******************************************************************************
*
* FUNCTION: void Print_Formatter1(stream, format, keepWithNext)
*
* DESCRIPTION:
*
* INPUTS: 
* FILE *stream;
* FORMAT_PTR format;
* int keepWithNext;
*
* OUTPUTS: void.
*
******************************************************************************/

extern char *search_hash_table_for_format (); /* Defined in parse_fmttrs.c */

void Print_Formatter1(stream, format, keepWithNext)
FILE *stream;
FORMAT_PTR format;
int keepWithNext;
{ 
  int i, last_element_p;
  FORMAT_ARRAY_PTR format_array;
  char *named_format;

  named_format = search_hash_table_for_format(format);
  if (named_format && (named_format[0] != '\0')) 
    printString(stream, named_format, keepWithNext);
  else 
    switch(format->type) {
    case LengthFMT: 
      printInt(stream, &(format->formatter.i), keepWithNext);
      break;
    case PrimitiveFMT: 
      /* This case should be handled by searching the hash table */
      TCA_ERROR1("Unknown Primitive Format %d\n", format->formatter.i);

    case PointerFMT:
      printString(stream, "*", DEFAULT_KEEP);
      if (format->formatter.f)
	Print_Formatter1(stream, format->formatter.f, keepWithNext);
      else
	printString(stream, "!", keepWithNext);
      break;

    case StructFMT:
      format_array = format->formatter.a;
      printString(stream, "{", DEFAULT_KEEP);
      for (i=1; i<format_array[0].i; i++) {
	if (truncatedGlobal)
	  return;
	else {
	  last_element_p = (i+1 == format_array[0].i);
	  Print_Formatter1(stream, format_array[i].f,
			   ((last_element_p) ? keepWithNext+1 : 1));
	  if (last_element_p)
	    printString(stream, "}", keepWithNext);
	  else
	    printCommaSpace(stream);
	}
      }
      break;

    case FixedArrayFMT: 
      PrintArrayFormat(stream, "[", format->formatter.a, "]", keepWithNext);
      break;

    case VarArrayFMT: 
      PrintArrayFormat(stream, "<", format->formatter.a, ">", keepWithNext);
      break;
    }
}


/******************************************************************************
*
* FUNCTION: void Print_Formatter(stream, format)
*
* DESCRIPTION: Print a representation of the "Format" on the "Stream".
*
* INPUTS: 
* FILE *stream;
* FORMAT_PTR format;
*
* OUTPUTS: void.
*
******************************************************************************/

void Print_Formatter(stream, format)
FILE *stream;
FORMAT_PTR format;
{
  startPrint(stream);
  if (format) 
    Print_Formatter1(stream, format, 0);
  else
    printString(stream, "NULL Format", 0);
  endPrint(stream);
}

/************************************/

void dPrintSTR(stream, string, next)
FILE *stream;
char *string;
int next;
{
  printString(stream, string, next);
}

void dPrintFORMAT(stream, format, next)
FILE *stream;
FORMAT_PTR format;
int next;
{
  if (format)
    Print_Formatter1(stream, format, next);
  else
    printString(stream, "NULLFormat", next);
}


/******************************************************************************
*
* FUNCTION: void mapPrint(stream, map, format, keepWithNext, mapElements)
*
* DESCRIPTION:
*
* INPUTS: 
* FILE *stream;
* fmat *map;
* FORMAT_PTR format;
* int keepWithNext;
* char *mapElements;
*
* OUTPUTS: void
*
******************************************************************************/

void mapPrint(stream, map, format, keepWithNext, mapElements)
FILE *stream;
fmat *map;
FORMAT_PTR format;
int keepWithNext;
char *mapElements;
{ 
  FORMAT_ARRAY_TYPE mapFormatArray[4];

  mapFormatArray[0].i = 4;
  mapFormatArray[1].f = format;
  mapFormatArray[2].i = map->ub1 - map->lb1 + 1;
  mapFormatArray[3].i = map->ub2 - map->lb2 + 1;

  /* Print header: "[Map (lb1, ub1)(lb2, ub2): " */
  printString(stream, "[Map (", 5);
  printInt(stream, &(map->lb1), 1);
  printCommaSpace(stream);
  printInt(stream, &(map->ub1), 1);
  printString(stream, ")", 0);
  printString(stream, "(", 5);
  printInt(stream, &(map->lb2), 1);
  printCommaSpace(stream);
  printInt(stream, &(map->ub2), 2);
  printString(stream, "):", 0);
  printSpace(stream);

  (void)printArrayData(stream, mapFormatArray[1].f, mapElements,
		       /* &(map->el[map->lb1][map->lb2]),*/
		       0, mapFormatArray, 2, keepWithNext+1, "[", "]");
}

void dPrintTCA(stream, ref, next)
FILE *stream;
TCA_REF_PTR ref;
int next;
{
  printString(stream, "[TCA_REF (", 5);
  printInt(stream, &(ref->refId), 1);
  printString(stream, ", ", 3);
  printString(stream, ref->name, 10);
  printString(stream, ")]", 5);
}






