/**************************************************************************
* 
* PROJECT: Carnegie Mellon Planetary Rover Project
*          Task Control Architecture
*
* (c) Copyright 1991 Christopher Fedor and Reid Simmons.  All rights reserved.
* 
* MODULE: qlattice
*
* FILE: qlattice.h
*
* ABSTRACT:
*
* QUANTITY LATTICE STRUCTURES
* The upper and lower bounds are contextual-value "cells" which define
* the interval range of the quantity.  The upper and lower bounds default 
* to the values "[minus_infinity...plus_infinity]."
*
* The first four slots are mainly for bookkeeping and 
* doing searches in the lattice.
*
* The (real) interval value of a quantity is given by the next two slots, 
* each of which is a contextual_value "cell".  
*
* The "relations" slot is a list of relations with other quantities 
* (a relation is also a contextual_value "cell"
*
* The "arithmetic_fields" is used to hang on additional slots that are needed
* by the arithmetic package (qarith).
*
*
**************************************************************************/

typedef enum {No_Relationship, Less_Or_Equal, Not_Equal, 
	      Greater_Or_Equal, Greater, Equal, Less, Inconsistent, 
	      Real_Valued} QRELATION_TYPE;

typedef enum {Is_False, Is_True, Is_Unknown} TRI_VALUED_TYPE;

typedef struct { 
  float value;
  int limit;
} QBOUND_TYPE, *QBOUND_PTR;

/* 21-Nov-89: fedor: this needs more investigation to find how it is
   used to avoid much mallocing of memory */

/* 27-Nov-89: fedor: replaced SJ_TYPE lists with two justification slots */
typedef struct { 
  QRELATION_TYPE qRel1, qRel2;
  CELL_PTR justRel1, justRel2;
} SJ_TYPE;

typedef struct _QUANTITY_TYPE { 
  char *name;
  int qnumber;
  SJ_TYPE searchJust;
  QRELATION_TYPE mark;
  CELL_PTR lower_bound;
  CELL_PTR upper_bound;
  LIST_PTR relations;
  struct _arith_fields *arithmetic_fields;
} QUANTITY_TYPE, *QUANTITY_PTR;

typedef struct { 
  QUANTITY_PTR arg1;
  QUANTITY_PTR arg2;
} REL_DATA_TYPE, *REL_DATA_PTR;

extern void qlatticeInit();

extern QUANTITY_PTR createQuantity();
extern QUANTITY_PTR createRealQuantity();

extern QUANTITY_PTR Zero_Quant;
extern QUANTITY_PTR One_Quant;
extern QUANTITY_PTR Minus_One_Quant;

extern void Print_Quantity(); 
extern void Print_Relationship();

extern void QAssert();

extern char *QRelationship();

extern TRI_VALUED_TYPE QIs();

extern int QIs_True();

extern void QRetract();

#define QUNASSUME(q1, relationship, q2) \
  QRetract(q1, relationship, q2, "Assumption", NULL)

#define QASSUME(q1, relationship, q2) \
  QAssert(q1, relationship, q2, "Assumption", NULL)

#define RELATION_DATA(relation) ((REL_DATA_PTR)((relation)->datum))

#define REL_ARG1(relation) (RELATION_DATA(relation)->arg1)
#define REL_ARG2(relation) (RELATION_DATA(relation)->arg2)

#define CV_RELP(relp_supporter) ((QRELATION_TYPE)((relp_supporter)->datum))

#define CV_QBOUND(qbound_supporter) ((QBOUND_PTR)((qbound_supporter)->datum))

/*****************************************************************************
*  Returns the TMS node that encodes the information that "q1" stands in the
*  "relationship" to "q2".  Note that the node is returned even if it is "out" 
*     (ie, the relationship does not currently hold).
******************************************************************************/
extern TMS_NODE_PTR QNode();

/****************************************************************************
* Free the quantity.
* Eventually should be implemented to free all its associated cells and 
*  tms nodes, but does not do that at present.
*****************************************************************************/
extern void FreeQuantity ();


/* Auxilliary functions */

extern CELL_PTR get_relation();
extern CELL_PTR relationGetOrMake();
extern void removeQuantityRules();
extern QUANTITY_PTR relation_other_arg();
