/*************************************************************************

* (c) Copyright 1991 Christopher Fedor and Reid Simmons.  All rights reserved.
* 

    This file contains functions to generate and manipulate queues, stacks,
        and priority queues.  It also contains code for `remembering' lists
        of nodes.
   
    The abstractions defined are QUEUE, STACK, PRIORITY_QUEUE and MEMORY.
    Each abstraction `X' supports the functions CREATE_X, FREE_X, 
        CLEAR_X, and IS_X_EMPTY.
   
    QUEUE additionally supports ENQUEUE and DEQUEUE, and it supports
        HEAD_ENQUEUE for getting items in a LIFO manner.
   
    STACK additionally supports PUSH_STACK and POP_STACK.
   
    PRIORITY_QUEUE supports PRIORITY_ENQUEUE and PRIORITY_DEQUEUE.
   
    MEMORY supports REMEMBER and DO_MEMORY (which applys a function to each
       `remembered' item in a last in, first out manner).

    29-Jul-91 Reid Simmons   Rewrote entirely to use Chris' list package.
    26-Jul-91 Reid Simmons   Fixed "head_enqueue"
     9-Aug-89 Reid Simmons   Fixed "extend_queue".
     2-Aug-89 Reid Simmons   Changed "printf"s to use logging facility.
    26-Jun-89 Reid Simmons   Added code to make QUEUES, STACKS, and MEMORIES 
                             dynamically expandable.  
  *************************************************************************/

#include "stdio.h"

#include "basics.h"
#include "list.h"
#include "queue.h"

extern char *listPopItem(); /* defined in list.c */

/*
                .______________.
                .              .
                .    QUEUES    .
                .______________.

*/

queue_ptr Create_Queue() 
{ 
  queue_ptr new_queue;

  new_queue = NEW(queue_type);
  new_queue->list = listCreate();
  return new_queue;
}

/* Currently not used (RGS: 11/11/92)
void Free_Queue (queue)  queue_ptr queue;
{
  listFree(queue->list);
  tcaFree((char *) queue);
}
*/

/* This is actually quite efficient, but a simple listClear operation
   would be cleaner */
void Clear_Queue (queue)  queue_ptr queue;
{
  listFree(queue->list);
  queue->list = listCreate();
}

/* Currently not used (RGS: 11/11/92)
int Is_Queue_Empty (queue) queue_ptr queue;
{ 
  return (listLength(queue->list) == 0);
}
*/

void Enqueue (data, queue)
char *data; 
queue_ptr queue;
{ 
  listInsertItemLast(data, queue->list);
}

char *Dequeue (queue) queue_ptr queue;
{ 
  return listPopItem(queue->list);
}

void Head_Enqueue (data, queue)
char *data; queue_ptr queue;
{ 
  listInsertItemFirst(data, queue->list);
}


/*
                .______________.
                .              .
                .    STACKS    .
                .______________.

*/

stack_ptr Create_Stack()
{ stack_ptr new_stack;

  new_stack = NEW(stack_type);
  new_stack->list = listCreate();
  return new_stack;
}

/* Currently not used (RGS: 11/11/92)
void Free_Stack (stack)
stack_ptr stack;
{
  listFree(stack->list);
  tcaFree((char *) stack);
}
*/

/* This is actually quite efficient, but a simple listClear operation
   would be cleaner */
/* Currently not used (RGS: 11/11/92)
void Clear_Stack (stack)
stack_ptr stack;
{
  listFree(stack->list);
  stack->list = listCreate();
}
*/

/* Currently not used (RGS: 11/11/92)
int Is_Stack_Empty (stack)
stack_ptr stack;
{ 
  return (listLength(stack->list) == 0);
}
*/

void Push_Stack (data, stack)
char *data; 
stack_ptr stack;
{ 
  listInsertItemFirst(data, stack->list);
}

char *Pop_Stack(stack) 
stack_ptr stack;
{ 
  return listPopItem(stack->list);
}

#if 0
/* Currently not used (RGS: 11/11/92) */

/*
                ._______________________.
                .                       .
                .    PRIORITY_QUEUES    .
                ._______________________.


    INEFFICIENT IMPLEMENTATION OF PRIORITY-QUEUES
    Has the property that if two values are EQUAL, then the newest entry
      gets inserted ahead of the older entries (it's really like a PRIORITY-STACK 
      in that way!!!)
*/

priority_queue_ptr Create_Priority_Queue(valueLesspFn)
int (*valueLesspFn)();
{ priority_queue_ptr new_priority_queue;

  new_priority_queue = NEW(priority_queue_type);
  new_priority_queue->list = listCreate();
  new_priority_queue->valueLesspFn = valueLesspFn;
  return new_priority_queue;
}

void Free_Priority_Queue (priority_queue)
priority_queue_ptr priority_queue;
{
  listFree(priority_queue->list);
  tcaFree((char *) priority_queue);
}

/* This is actually quite efficient, but a simple listClear operation
   would be cleaner */
void Clear_Priority_Queue (priority_queue)
priority_queue_ptr priority_queue;
{
  listFree(priority_queue->list);
  priority_queue->list = listCreate();
}

int Is_Priority_Queue_Empty (priority_queue)
priority_queue_ptr priority_queue;
{ 
  return (listLength(priority_queue->list) == 0);
}

void Priority_Enqueue (data, priority_queue)
char *data; 
priority_queue_ptr priority_queue;
{
  listInsertItemFirst(data, priority_queue->list);
}

/* Inefficient, but makes use of already existing list package primitives */
char *Priority_Dequeue(priority_queue) 
priority_queue_ptr priority_queue;
{ 
  char *bestItem, *item;

  bestItem = listFirst(priority_queue->list);
  item = listNext(priority_queue->list);
  while (item) {
    if ((priority_queue->valueLesspFn)(bestItem, item)) {
      bestItem = item;
    }
    item = listNext(priority_queue->list);
  }
  return bestItem;
}
#endif

/*
                .________________.
                .                .
                .    MEMORIES    .
                .________________.

 `Memory' is a way of keeping track of a list of things.
*/

memory_ptr Create_Memory()
{ memory_ptr new_memory;

  new_memory = NEW(memory_type);
  new_memory->list = listCreate();
  return new_memory;
}

/* Currently not used (RGS: 11/11/92)
void Free_Memory (memory)
memory_ptr memory;
{
  listFree(memory->list);
  tcaFree((char *) memory);
}
*/

/* This is actually quite efficient, but a simple listClear operation
   would be cleaner */
void Clear_Memory (memory)
memory_ptr memory;
{
  listFree(memory->list);
  memory->list = listCreate();
}

/* Currently not used (RGS: 11/11/92)
int Is_Memory_Empty (memory)
memory_ptr memory;
{ 
  return (listLength(memory->list) == 0);
}
*/

void Remember (data, memory)
char *data; 
memory_ptr memory;
{ 
  listInsertItemFirst(data, memory->list);
}

/*
  Execute the function, iterating over the items in memory in
      a last in, first out manner.
*/

int applyMemoryFunc (memoryFn, item)
void (* memoryFn)();
char *item;
{
  memoryFn(item);
  return 1;
}

void Do_Memory (memory, memoryFn)
memory_ptr memory; 
void (* memoryFn)();
{ 
  (void)listIterateFromLast(applyMemoryFunc, (char *)memoryFn, memory->list);
}
