/******************************************************************************
*
* PROJECT: Carnegie Mellon Planetary Rover Project
*          Task Control Architecture 
* 
* (c) Copyright 1991 Christopher Fedor and Reid Simmons.  All rights reserved.
* 
* MODULE: Receive Message
*
* FILE: recvMsg.c
*
* ABSTRACT:
* 
* All messages enter through receive message. 
* Receive message provides the main thread of flow and control in tca.
*
* REVISION HISTORY
*
* 13-Oct-92  Reid Simmons, School of Computer Science, CMU
* Updated InformClass messages (used to be ConstraintClass) to issue a
* response.
* Added BroadcastClass and MultiQueryClass messages.
* Broke "recvMessage" into more manageable chunks.
*
* 14-Aug-91 Christopher Fedor, School of Computer Science, CMU
* Added Richard Goodwin's change to set tplConstr = NO_TPLCONSTR
* for PollingMonitorClass and DemonMonitorClass in recvMessage.
*
* 26-Jul-91  Reid Simmons, School of Computer Science, CMU
* Messages added to "toBeKilled" nodes were not being handled correctly.
*
* 02-Jul-91  Reid Simmons, School of Computer Science, CMU
* Fixed up some memory management problems, including freeing class data.
*
* 14-May-91 Christopher Fedor, School of Computer Science, CMU
* Moved dispatchUpdateAndDisplay to dispatch routine collection.
*
* 12-Mar-91  Reid Simmons, School of Computer Science, CMU
* Added calls to for tapping messages.
*
*  1-Nov-90 Christopher Fedor, School of Computer Science, CMU
* Rewritten for version tca 5.x
*
* 29-May-90 Christopher Fedor, School of Computer Science, CMU
* Replaced DReply with centralReply which performs a tcaReply for
* centrally handled queries.
*
* 23-Apr-90 Reid Simmons, School of Computer Science, CMU
* Added initial hacks for choosing between multiple handlers.
*
* 28-Nov-89 Long-Ji Lin, School of Computer Science, CMU
* Added code to dispatchServer() to choose exception handler 
* and to add task tree node for new message.
*
* 11-Aug-89 Reid Simmons, School of Computer Science, CMU
* Changed "goaltree" to "tasktree".
*
*  3-Aug-89 Reid Simmons, School of Computer Science, CMU
* Added code to (selectively) ignore logging 
* certain messages (registration msgs)
*
*  2-Aug-89 Reid Simmons, School of Computer Science, CMU
* Changed "printf"s to use logging facility.
*
*  2-Aug-89 Reid Simmons, School of Computer Science, CMU
* Added call to "Update_Last_Child".
*
* 10-July-89 Reid Simmons, School of Computer Science, CMU
* Split off from dispatch.c all the server (central) dependent routines.
*
*  9-May-89 Christopher Fedor, School of Computer Science, CMU
* Added "FakeReplyHnd" to DReply. DReply should go!
*
*  9-May-89 Reid Simmons, School of Computer Science, CMU
* Moved call to AddTplConstraints.
*
*  6-Mar-89 Christopher Fedor, School of Computer Science, CMU
* Divided to partition module/server calls.
*
* 24-Feb-89 Christopher Fedor, School of Computer Science, CMU
* Implemented ReplyFreeList as an idtable.
*
* 23-Feb-89 Christopher Fedor, School of Computer Science, CMU
* Replaced ReadDataMsg and WriteDataMsg.
*
*    Dec-88 Christopher Fedor, School of Computer Science, CMU
* created.
*
******************************************************************************/

#include "stdio.h"

#include "tcaInternal.h"
#include "formatters.h"
#include "logging.h"
#include "datamsg.h"
#include "idtable.h"

#include "dispatch.h"
#include "hash.h"
#include "list.h"
#include "queue.h"

#include "tms.h"
#include "cells.h"
#include "rules.h"   /* for DELAY_RULES #define */
#include "qlattice.h"

#include "tplConstr.h"
#include "taskTree.h"
#include "mon.h"
#include "msgTap.h"

#include "res.h"

#define ATTENDABLE(dispatch) \
  !QIs_True(Now, "<", dispatch->treeNode->handlingInterval->start)

extern int tcaDebugGlobal;

extern QUANTITY_PTR Now;

extern int tcaServerGlobal;
extern MODULE_PTR tcaServerModGlobal;

extern HASH_TABLE_PTR handlerTable;              
extern HASH_TABLE_PTR classFormatTable, messageTable;
extern HASH_TABLE_PTR formatNamesTable;

/* Route Tables */
extern ID_TABLE_PTR msgIdTable;

extern int sendMessageRefGlobal;

extern void fireDemonHnd();
extern void replyPollingMonitor();

extern int parentRefGlobal;

extern void centralRegisterNamedFormatter();
extern void centralRegisterLengthFormatter();

extern HND_PTR chooseExceptionHandler();

/*************************************************/

HND_PTR chooseMsgHandler(msg)
MSG_PTR msg;
{
  int size;
  LIST_PTR hndList;
  HND_PTR hnd, nextHnd, retHnd;

  hndList = msg->hndList;
  size = listLength(hndList);

  hnd = (HND_PTR)listFirst(hndList);

  if (!size) {
    size = 1;
    hnd = NEW(HND_TYPE);
    hnd->sd = 0;
    hnd->hndOrg = tcaServerModGlobal;
    hnd->hndProc = NULL;
    hnd->hndData = NULL;
    hnd->resource = resourceCreate(0, msg->msgData->name, 1);
    hnd->resource->capacity = 0;
    hnd->msgList = NULL;

    listInsertItem((char *)hnd, msg->hndList);
  }

  if (size > 1) {
    retHnd = hnd;
    nextHnd = (HND_PTR)listNext(hndList);
    while (nextHnd && 
	   STREQ(hnd->hndData->hndName, nextHnd->hndData->hndName)) {
      if (resourceAvailableRes(nextHnd->resource))
	retHnd = nextHnd;
      nextHnd = (HND_PTR)listNext(hndList);
    }
    hnd = !nextHnd ? retHnd : NULL;
  }

  return hnd;
}

/*************************************************/

void hndSelect(dispatch)
DISPATCH_PTR dispatch;
{
  HND_PTR hnd;

  if (dispatch->msg->msgData->class == ExceptionClass) {
    if (dispatch->hnd)
      return; /* assumes set by byPassExcepHnd or retryHnd */
    hnd = chooseExceptionHandler(dispatch);
  }
  else
    hnd = chooseMsgHandler(dispatch->msg);

  dispatch->hnd = hnd;

  if (!hnd)
    return;

  dispatch->des = hnd->hndOrg;
  dispatch->desId = hnd->hndOrg->sd;
}

/*************************************************/

processActiveMessage(dispatch)
DISPATCH_PTR dispatch;
{
  hndSelect(dispatch);
  /* Moved here from "recvMessage" because need to wait until handler
     selected, (reids 10-92) */ 
  if (dispatch->class == QueryClass && !resourceSelfQuery(dispatch))
    pendingReplyAdd(dispatch);

  if (DISPATCH_HANDLER_CHOOSEN(dispatch))
    processResourceAndDeliver(dispatch);
  else {
    TCA_ERROR1("ERROR: processActiveMessage: No Handler Choosen for %s.\n",
	       dispatch->msg->msgData->name);
  }
}

/*************************************************/

processResourceAndDeliver(dispatch)
DISPATCH_PTR dispatch;
{
  if (!dispatch->treeNode) {
    tapBeforeHandling(dispatch);
  }
  if (resourceAvailable(dispatch)) 
    processResAttendDeliver(dispatch);
  else {
    resourcePending(dispatch);
    PendingConstraints(dispatch);
    dispatchUpdateAndDisplay(PendingDispatch, dispatch);
  }
}

/*************************************************/

processResAttendDeliver(dispatch)
DISPATCH_PTR dispatch;
{
  resourceAttending(dispatch);
  AttendingConstraints(dispatch);
  deliverDispatch(dispatch);
  tapWhileHandling(dispatch);
  /* 02-Jul-91: Reid: 
     Just handled a central query or inform: It's OK to free the dispatch here */
  if (dispatch->class == ReplyClass ||
      (dispatch->class == InformClass && dispatch->des == tcaServerModGlobal))
    dispatchFree(dispatch);
}

/*************************************************/

processHandledMessage(dispatch)
DISPATCH_PTR dispatch;
{
  dispatchUpdateAndDisplay(HandledDispatch, dispatch);
  resourceRemoveAttending(dispatch);

  tapAfterHandled(dispatch);
  if (!ONE_WAY_MSG(dispatch->msg)) {
    switch (dispatch->class) {
    case SuccessClass: 
      tapWhenSuccess(dispatch); 
      tapAfterSuccess(dispatch); 
      break;
    case FailureClass: 
      tapWhenFailure(dispatch);
      tapAfterFailure(dispatch);
    }
  }

  resourceProcessPending(dispatch);
  CompletionConstraints(dispatch);
  HandleKillAfterAttendingNodes(dispatch);
  if (ONE_WAY_MSG(dispatch->msg))
    dispatchFree(dispatch);

  /* 14-May-91: fedor: exception class messages that are byPassed or retry
     should not be dispatch freed arbitrarly - may need to declare the first
     message handled and then resend the same dispatch through recvMessage */
}

/*************************************************/

/* 22-May-91:fedor: another level of flow and control for exception handling */
processInactiveTests(dispatch)
DISPATCH_PTR dispatch;
{
  if (NeedToKillMessage(dispatch)) {
    dispatch->status = InactiveDispatch;
    dispatch->treeNode->status = ((dispatch->refCount <= 0)
				  ? KilledNode : ToBeKilledNode);
  } else if (ATTENDABLE(dispatch)) 
    processActiveMessage(dispatch);
  else {
    InactiveConstraints(dispatch);
    dispatchUpdateAndDisplay(InactiveDispatch, dispatch);
  }
}

/*************************************************/

extern void recvMessage();

recvMessageBuild(module, dataMsg)
MODULE_PTR module;
DATA_MSG_PTR dataMsg;
{
  char *classData;
  MSG_PTR msg = NULL;

  DISPATCH_PTR dispatch;
  TCA_MSG_CLASS_TYPE class;
  CLASS_FORM_PTR classForm;

  FORMAT_PTR classFormat = NULL;

  if (dataMsg->intent != NO_REF)
    msg = (MSG_PTR)idTableItem(dataMsg->intent, msgIdTable);

  class = (TCA_MSG_CLASS_TYPE)dataMsg->classId;
  classForm = (CLASS_FORM_PTR)hashTableFind((char *)&class, classFormatTable);

  if (classForm)
    classFormat = classForm->format;

  classData = dataMsgDecodeClass(classFormat, dataMsg);

  if (dataMsg->dispatchRef != NO_REF) {
    dispatch = DISPATCH_FROM_ID(dataMsg->dispatchRef);
  } else {
    dispatch = dispatchAllocate();
    dispatch->msg = msg;

    dispatch->org = module;
    dispatch->orgId = module->sd;

    dispatch->refId = dataMsg->msgRef;
    dispatch->pRef = dataMsg->parentRef; 

    DISPATCH_SET_STATUS(AllocatedDispatch, dispatch);
  }

  if (dataMsg->msgRef == NO_REF)
    dispatchSetResData(dataMsg, dispatch);
  else 
    dispatchSetMsgData(dataMsg, dispatch);


  /* 22-May-91: fedor: NOTE: class and classData will change if this
     is a previously created dispatch. This may be undesired for
     behavior for a sendResponse call that refers to a previously
     created dispatch. */
  /*  3-Jul-91: reid: Clean up previous class data, 
      except for command class data, which is stored in blockCom */
  if (dispatch->classData && dispatch->class != CommandClass) 
    classDataFree(dispatch->class, dispatch->classData);
  dispatch->class = class;
  dispatch->classData = classData;

  recvMessage(dispatch, class, classData);
}

/*************************************************/

void recvTaskTreeMessage (dispatch, class, classData)
DISPATCH_PTR dispatch;
TCA_MSG_CLASS_TYPE class;
char *classData;
{
  int tplConstr;

  if (class == CommandClass) {
    dispatch->blockCom = (BLOCK_COM_PTR)classData;
    tplConstr = dispatch->blockCom->tplConstr;

    /* 28-Oct-91: fedor: Blah! Need some way to see if this blocking
       command is a self query style - or if that should be allowed? */
    if (dispatch->blockCom->waitFlag)
      pendingReplyAdd(dispatch);
  } else if (class == PollingMonitorClass || class == DemonMonitorClass) {
    tplConstr = NO_TPLCONSTR;
  } else
    tplConstr = *(int *)classData;

  if (!dispatch->treeNode) {
    Initialize_TaskTreeNode(dispatch, dispatch->pRef);
  } else if (DEAD_NODE(dispatch->treeNode)) {
    /* 29-Jul-91: Reid: Stop if this message has already been killed. */
    return;
  }

  addTplConstraints(dispatch, tplConstr);

  if ((class == PollingMonitorClass) || (class == DemonMonitorClass)) {
    setIntervalMonitorStartTime(dispatch, (INTERVAL_MON_CLASS_PTR)classData);
  }

  (void)Update_Last_Child(dispatch->treeNode);

  tapWhenSent(dispatch);
  setUpBeforeTaps(dispatch);

  processInactiveTests(dispatch);

  setUpDuringAndAfterTaps(dispatch);
}

/* 9-Oct-90: fedor - check durablFn in mon.c when implementing failure */
void recvSuccessFailureMessage (dispatch, class)
DISPATCH_PTR dispatch;
TCA_MSG_CLASS_TYPE class;
{
  DISPATCH_PTR replyDispatch;

  if (dispatch->blockCom && 
      dispatch->msg && dispatch->msg->msgData->class == CommandClass) {
    if (dispatch->blockCom->waitFlag) {
      replyDispatch = dispatchAllocate();
      replyDispatch->msg = dispatch->msg;

      replyDispatch->org = dispatch->org;
      replyDispatch->orgId = dispatch->orgId;

      replyDispatch->refId = dispatch->refId;
	
      DISPATCH_SET_STATUS(AttendingDispatch, replyDispatch);

      centralReply(replyDispatch, (char *)&class);
    }
    tcaFree((char *)dispatch->blockCom);
    dispatch->blockCom = NULL;
  }

  processHandledMessage(dispatch);
}

int processBroadcast(dispatch, hnd)
DISPATCH_PTR dispatch;
HND_PTR hnd;
{
  if (hnd != (HND_PTR)listFirst(dispatch->msg->hndList)) {
    dispatch = dispatchCopy(dispatch);
  }
  dispatch->hnd = hnd;
  dispatch->des = hnd->hndOrg;
  dispatch->desId = hnd->hndOrg->sd;
  processResourceAndDeliver(dispatch);
  return TRUE;
}

void recvBroadcastMessage (dispatch)
DISPATCH_PTR dispatch;
{
  tapWhenSent(dispatch);
  (void)listIterateFromLast(processBroadcast, (char *)dispatch, 
			    dispatch->msg->hndList);
}

/*
  The straightforward way to do this would be using "centralReply", 
  but that calls "recvMessage" recursively, which breaks things.
*/
void sendMultiQueryTermination (dispatch)
DISPATCH_PTR dispatch;
{
  static DATA_MSG_PTR nullReplyData = NULL;

  if (!nullReplyData) {
    nullReplyData = dataMsgCreate(0, QUERY_REPLY_INTENT, 0, 0, 0, (FORMAT_PTR)NULL,
				  (char *)NULL, (FORMAT_PTR)NULL, (char *)NULL);
  }

  nullReplyData->msgRef = dispatch->refId;

  if (!dispatch->org)
    tcaError("Error: sendMultiQueryTermination: No query origin.");
  else {
    if (dispatch->org->alive) {
      (void)dataMsgSend(dispatch->orgId, nullReplyData);
    } else if (!dispatch->org->repliesPending) {
      moduleFree(dispatch->org);
    }
  }
}

void recvMultiQueryMessage (dispatch)
DISPATCH_PTR dispatch;
{
  tapWhenSent(dispatch);
  /* 
   * Need to "park" the class data somewhere, so that it will not be freed
   * when a reply comes in.  Instead of adding another slot to "dispatch",
   * just reuse the "blockCom" slot
   */
  dispatch->blockCom = (BLOCK_COM_PTR)dispatch->classData;
  dispatch->classData = NULL;

  MULTI_QUERY_NUM(dispatch) = listLength(dispatch->msg->hndList);
  if (MULTI_QUERY_MAX(dispatch) > MULTI_QUERY_NUM(dispatch)) {
    MULTI_QUERY_MAX(dispatch) = MULTI_QUERY_NUM(dispatch);
  }

  if (MULTI_QUERY_MAX(dispatch) == 0) {
    sendMultiQueryTermination(dispatch);
    /* Free the class data, saved in "blockCom" slot */
    tcaFree((char *)dispatch->blockCom);
    dispatchFree(dispatch);
  } else {
    (void)listIterateFromLast(processBroadcast, (char *)dispatch, 
			      dispatch->msg->hndList);
  }
}

void multiQueryUpdate(dispatch, logStatus)
DISPATCH_PTR dispatch;
{
  --MULTI_QUERY_NUM(dispatch);
  --MULTI_QUERY_MAX(dispatch);
  if (MULTI_QUERY_MAX(dispatch) == 0) {
    sendMultiQueryTermination(dispatch);
  } else if (logStatus && MULTI_QUERY_MAX(dispatch) < 0) {
    Log_Status("    MultiQuery: Max replies already received; this reply not actually sent\n");
  }
  if (MULTI_QUERY_NUM(dispatch) == 0) {
    /* Free the class data, saved in "blockCom" slot */
    tcaFree((char *)dispatch->blockCom);
  }
}

void recvReplyMessage (dispatch)
DISPATCH_PTR dispatch;
{
  dispatchUpdateAndDisplay(HandledDispatch, dispatch);
  resourceRemoveAttending(dispatch);
  if (dispatch->msg->msgData->class != MultiQueryClass ||
      MULTI_QUERY_MAX(dispatch) > 0) {
    deliverResponse(dispatch);
    tapAfterReplied(dispatch);
    tapAfterHandled(dispatch);
  }

  if (dispatch->msg->msgData->class == MultiQueryClass) {
    multiQueryUpdate(dispatch, TRUE);
  }

  /* 31-Oct-90: fedor: problem with replies that are not simply
     replies to standard query messages ...etc. sooo we need to
     keep the dispatch around but perhaps not the datamsg.
     This just postpones the envetual memory lossage */

  dataMsgFree(dispatch->msgData);
  dataMsgFree(dispatch->resData);
  dispatch->msgData = NULL;
  dispatch->resData = NULL;
 
  /* 31-Oct-90: fedor: also need to do free before processing further
     otherwise we may loop and never free anything */

  resourceProcessPending(dispatch);

  /* 02-Jul-91: reid: Centrally handled dispatch queries are
     freed in "processResAttendDeliver" */
  if (dispatch->des != tcaServerModGlobal)
    dispatchFree(dispatch);
}

void recvMessage(dispatch, class, classData)
DISPATCH_PTR dispatch;
TCA_MSG_CLASS_TYPE class;
char *classData;
{
  if (tcaDebugGlobal) {
    idTablePrintInfo(dispatchTable);
    fprintf(stderr, "\n");
    tcaStats();
  }

  switch(class) {
  case QueryClass:
  case InformClass:
    tapWhenSent(dispatch);
    processActiveMessage(dispatch);
    break;
  case ReplyClass:
    recvReplyMessage(dispatch);
    break;

  case GoalClass:
  case PointMonitorClass:
  case DemonMonitorClass:
  case PollingMonitorClass:
  case ExceptionClass:
  case CommandClass:
    recvTaskTreeMessage(dispatch, class, classData);
    break;

  case BroadcastClass:
    recvBroadcastMessage(dispatch);
    break;

  case MultiQueryClass:
    recvMultiQueryMessage(dispatch);
    break;

  case SuccessClass:
  case FailureClass:
    recvSuccessFailureMessage(dispatch, class);
    break;

  case FireDemonClass:
    /* 8-Oct-90: fedor */
    /* printf("recvMessage: FireDemonClass: Attending\n");*/
    /* dispatchUpdateAndDisplay(AttendingDispatch, dispatch);*/
    Log_Message1("%-9s", "FireDemon");
    Log_Message1("%15d: ", (char *)*(int *)classData);
    Log_Message1("%10s", dispatch->org->modData->modName);
    Log_Message1(" --> %-15s (Sent)", tcaServerModGlobal->modData->modName);
    Log_Message("\n");
    fireDemonHnd(dispatch, (int *)classData);
    break;
  default:
    TCA_ERROR1("ERROR: recvMessage: class not implemented: %d\n", class);
  }
}

/*************************************************/

sendDataMsgWithReply(msg, dataMsg, parentDispatch, respProc, respData)
MSG_PTR msg;
DATA_MSG_PTR dataMsg;
DISPATCH_PTR parentDispatch;
void (*respProc)();
char *respData;
{
  sendDataMsgInternal(msg, dataMsg, (char *)NULL, parentDispatch, 
		      respProc, respData);
}

/*******************************************************/

sendDataMsg(msg, dataMsg, classData, parentDispatch)
MSG_PTR msg;
DATA_MSG_PTR dataMsg;
char *classData;
DISPATCH_PTR parentDispatch;
{
  sendDataMsgInternal(msg, dataMsg, classData, parentDispatch, NULL, (char *)NULL);
}

/*******************************************************/

DISPATCH_PTR buildDispatchInternal(msg, dataMsg, classData, parentDispatch, 
				   resProc, resData)
MSG_PTR msg;
DATA_MSG_PTR dataMsg;
char *classData;
DISPATCH_PTR parentDispatch;
void (*resProc)();
char *resData;
{
  int savedRef;
  DISPATCH_PTR dispatch;

  if (!msg) {
    tcaError("ERROR: sendDataMsg: NULL msg.\n");
  }

  /* save the message ref to use in case msgFind triggers 
     a recursive call to sendMessage */
  savedRef = sendMessageRefGlobal;
  sendMessageRefGlobal++;

  dispatch = dispatchAllocate();
  dispatch->msg = msg;

  dispatch->org = tcaServerModGlobal;
  dispatch->orgId = tcaServerGlobal;

  dispatch->refId = savedRef;

  dispatchSetMsgData(dataMsg, dispatch);

  DISPATCH_SET_STATUS(AllocatedDispatch, dispatch);
  
  if (resProc) {
    dispatch->respProc = resProc;
    dispatch->respData = resData;
  }

  if (parentDispatch)
    dispatch->pRef = parentDispatch->locId;

  dispatch->class = msg->msgData->class;
  dispatch->classData = classData;

  return dispatch;
}


/*******************************************************/

sendDataMsgInternal(msg, dataMsg, classData, parentDispatch, resProc, resData)
MSG_PTR msg;
DATA_MSG_PTR dataMsg;
char *classData;
DISPATCH_PTR parentDispatch;
void (*resProc)();
char *resData;
{
  DISPATCH_PTR dispatch;

  dispatch = buildDispatchInternal(msg, dataMsg, classData, parentDispatch, 
				   resProc, resData);
  recvMessage(dispatch, dispatch->class, classData);
}


/*******************************************************/

void formatterSend(sd, form)
int sd;
FORMAT_PTR form;
{
  DATA_MSG_PTR dataMsg;
  FORMAT_PTR formatFormat;

  formatFormat = (FORMAT_PTR)hashTableFind("format", formatNamesTable);
  if (!formatFormat)
    tcaError("ERROR: formatterSend: no formatFormat");

  dataMsg = dataMsgCreate(0, 0, 0, 0, 0, formatFormat, (char *)&form, 
			  (FORMAT_PTR)NULL, (char *)NULL);

  (void)dataMsgSend(sd, dataMsg);

  dataMsgFree(dataMsg);
}

/*******************************************************/

void msgInfoMsgSend(sd)
int sd;
{
  int refId;
  MSG_PTR msg;

  msg = (MSG_PTR)hashTableFind("msgInfoMsg", messageTable);

  refId = msg->msgData->refId;

  (void)writeNBytes(sd, (char *)&refId, sizeof(int));

  formatterSend(sd, msg->msgData->msgForm);
  formatterSend(sd, msg->msgData->resForm);
}

/*******************************************************/

void registerClass(className, format)
TCA_MSG_CLASS_TYPE className;
char *format;
{
  FORMAT_PTR form;
  CLASS_FORM_PTR classForm;

  form = ParseFormatString(format);
  classForm = (CLASS_FORM_PTR)hashTableFind((char *)&className, classFormatTable);

  if (classForm)
    classForm->format = form;
  else {
    classForm = NEW(CLASS_FORM_TYPE);
    classForm->className = className;
    classForm->format = form;

    (void)hashTableInsert((char *)&className, sizeof(TCA_MSG_CLASS_TYPE), 
			  (char *)classForm, classFormatTable);
  }
}

/*******************************************************/

TCA_RETURN_VALUE_TYPE centralSendMessage(ref, msg, msgData, classData)
TCA_REF_PTR ref;
MSG_PTR msg;
char *msgData, *classData;
{
  int refId, savedRef;
  DATA_MSG_PTR msgDataMsg;
  TCA_MSG_CLASS_TYPE class;

  CLASS_FORM_PTR classForm;

  FORMAT_PTR classFormat;

  classFormat = NULL;

  if (!msg) {
    tcaError("ERROR: sendMessage: NULL msg.\n");
  }

  /* save the message ref to use in case msgFind triggers 
     a recursive call to sendMessage */

  savedRef = sendMessageRefGlobal;

  class = msg->msgData->class;

  classForm = (CLASS_FORM_PTR)hashTableFind((char *)&class, classFormatTable);

  if (classForm)
    classFormat = classForm->format;

  if (ref) 
    refId = ref->refId;
  else
    refId = NO_REF;

  /* 8-Oct-90: fedor:
     parentRefGlobal problem!!! ** it is initialized to -1 in behaviors.c 
     so maybe this will not be a problem */

  msgDataMsg = dataMsgCreate(parentRefGlobal, msg->msgData->refId, 
			     (int)class, refId, savedRef, msg->msgData->msgForm,
			     msgData, classFormat, classData);

  sendMessageRefGlobal++;

  recvMessageBuild(tcaServerModGlobal, msgDataMsg);

  return Success;
}

/*******************************************************/

void classInitialize()
{
  registerClass(GoalClass, "int");
  registerClass(CommandClass, "{int, int}");
  registerClass(ExceptionClass, "int");
  registerClass(PointMonitorClass, "int");
  registerClass(ExecHndClass, "string");
  registerClass(FailureClass, "string");

  centralRegisterLengthFormatter("TCA_POINT_CLASS_TYPE", 
			     sizeof(TCA_POINT_CLASS_TYPE));

  centralRegisterLengthFormatter("TCA_INTERVAL_CLASS_TYPE", 
			     sizeof(TCA_INTERVAL_CLASS_TYPE));

  centralRegisterNamedFormatter("TCA_INTERVAL_TYPE",
			    "{TCA_INTERVAL_CLASS_TYPE, int}");

  centralRegisterNamedFormatter("TCA_TIME_POINT_TYPE",
			    "{TCA_POINT_CLASS_TYPE, TCA_INTERVAL_TYPE}");

  centralRegisterNamedFormatter("INTERVAL_MONITOR_OPTIONS_PTR",
			    "*{int, int, int, int, int}");

  centralRegisterNamedFormatter("INT_MON_CLASS_TYPE", 
  "{TCA_TIME_POINT_TYPE, TCA_TIME_POINT_TYPE, INTERVAL_MONITOR_OPTIONS_PTR}");

  registerClass(PollingMonitorClass, "INT_MON_CLASS_TYPE");
  registerClass(DemonMonitorClass, "INT_MON_CLASS_TYPE");
  registerClass(FireDemonClass, "int");

  registerClass(MultiQueryClass, "{int, int}");
}
