/******************************************************************************
*
* PROJECT: Carnegie Mellon Planetary Rover Project
*          Task Control Architecture 
* 
* (c) Copyright 1991 Christopher Fedor and Reid Simmons.  All rights reserved.
* 
* MODULE: resources
*
* FILE: res.c
*
* ABSTRACT:
* 
* Resources
*
* REVISION HISTORY
*
* 13-Sep-92, Reid Simmons, School of Computer Science, CMU
* When deleting a resource (when a module goes down), clear the handlers
* of the dispatches that are already active or queued.
*
* 28-Oct-91 Christopher Fedor, School of Computer Science, CMU
* Changed logic of tryResendingMessage to key on 'y' instead of != 'n'
* Fixed tryResendingMessage to try all active/pending resends. 
* Added readChar routine to improve interaction with getchar and user input.
*
* 13-Sep-91 Christopher Fedor, School of Computer Science, CMU
* Changed resourcePending to add new items to the last in the list for FIFO.
*
* 25-Jul-91, Reid Simmons, School of Computer Science, CMU
* Built on Chris' earlier work for resending messages after a module crash.
*
*  3-Jul-91, Reid Simmons, School of Computer Science, CMU
* Freed data used the centrally registered handlers. 
*
*  3-Dec-90 Christopher Fedor, School of Computer Science, CMU
* Moved self query test from recvMsg.c to resourceAvailable so that a 
* locked resource would not handle a self query style message.
*
*  3-Oct-90 Christopher Fedor, School of Computer Science, CMU
* Added resourceRemovePending routine to provide a resource interface
* to taskTree.c for kiling of taskTree nodes.
*
*  3-Oct-90 Christopher Fedor, School of Computer Science, CMU
* Removed repetitve code by creating resourceReserveLock and
* resourceModReserveLock. Changed resourceCreate to update capacity
* on existing resources and to force capacity to a minimum of 1.
*
*  1-Oct-90 Christopher Fedor, School of Computer Science, CMU
* Finished reservation and lock - lots of repetitve code to stamp out still.
*
* 25-Sep-90 Christopher Fedor, School of Computer Science, CMU
* Added tca 5.0 versions of reservation and lock.
*
* 24-Sep-90 Christopher Fedor, School of Computer Science, CMU
* Rewritten to support a list of equal name resources to implement
* locking and reservation.
*
* 26-Aug-90 Christopher Fedor, School of Computer Science, CMU
* New implmentation of resources.
*
* NOTES:
*  3-Oct-90: fedor
* Most (if not all) of the resource routines check for a NULL resource
* and simply return. This is desired because the central server does
* not have a resource (currently) and this avoids many special case
* tests elsewhere.
*
******************************************************************************/

#include "stdio.h"
#include "ctype.h"
#include "string.h"

#include "basics.h"
#include "tcaInternal.h"
#include "logging.h"
#include "list.h"
#include "hash.h"
#include "key.h"
#include "formatters.h"

#include "datamsg.h"
#include "idtable.h"
#include "dispatch.h"

#include "res.h"
#include "resMod.h"
#include "exception.h"

extern void multiQueryUpdate(); /* Defined in recvMsg.c */

extern LIST_PTR moduleList;
extern MODULE_PTR tcaServerModGlobal;

HASH_TABLE_PTR resourceTable;

int resendAfterCrashGlobal = FALSE;


/******************************************************************************
*
* FUNCTION: RESOURCE_PTR resourceCreate(sd, name, capacity)
*
* DESCRIPTION:
* Resources are unique by module. 
* If the same resource is requested to be created the capacity is simply
* updated. NULL is returned if the resource is found.
*
* A capacity of less than 1 is simply set to 1 and a message displayed.
*
* INPUTS: 
* int sd;
* char *name;
* int capacity;
*
* OUTPUTS: RESOURCE_PTR (NULL value if resource is updated.)
*
******************************************************************************/

int resEqFunc(sd, resource)
int *sd;
RESOURCE_PTR resource;
{
  return(resource->sd == *sd);
}

RESOURCE_PTR resourceCreate(sd, name, capacity)
int sd;
char *name;
int capacity;
{
  LIST_PTR resList;
  RESOURCE_PTR resource;

  if (capacity < 1) {
    Log2("WARNING: Capacity of %d:%s set to 1\n", (char *)sd, name);
    capacity = 1;
  }

  resList = (LIST_PTR)hashTableFind(name, resourceTable);

  if (resList) {
    resource = (RESOURCE_PTR)listMemReturnItem(resEqFunc, (char *)&sd, resList);
    if (resource) {
      Log3("WARNING: Capacity of %d:%s changed to %d\n", 
	   (char *)sd, name, (char *)capacity);
      return NULL;
    }
  }
  else {
    resList = listCreate();
    (void)hashTableInsert(name, strlen(name)+1, (char *)resList, resourceTable);
  }

  resource = NEW(RESOURCE_TYPE);
  
  resource->sd = sd;

  resource->name = (char *)tcaMalloc(sizeof(char)*(unsigned)(1+strlen(name)));
  (void)strcpy(resource->name, name);

  resource->capacity = capacity;
  resource->module = NULL;
  resource->pendingList = listCreate();
  resource->attendingList = listCreate();

  RESOURCE_SET_STATUS(ActiveResource, resource);

  listInsertItem((char *)resource, resList);
    
  return resource;
}


int readChar()
{
  int i, c;

  while ((i = getchar()) != '\n' && i != EOF)
    c = i;

  return c;
}



/******************************************************************************
*
* FUNCTION: int tryResendingMessage
*
* DESCRIPTION: 
* Ask to resend the message, and do so if the response is positive.
*
* INPUTS: char *param; (Ignored)
*         DISPATCH_PTR dispatch;
*
* OUTPUTS: Returns 1 to continue iterating through the list
******************************************************************************/

int tryResendingMessage(param, dispatch)
char *param;
DISPATCH_PTR dispatch;
{
  Log1("Resend %s message?", dispatch->msg->msgData->name);
  if (readChar() == 'y') {
    switch (dispatch->msg->msgData->class) {
    case QueryClass:
    case MultiQueryClass:
    case InformClass:
    case BroadcastClass:
      DISPATCH_SET_STATUS(AllocatedDispatch, dispatch);
      processActiveMessage(dispatch);
      break;
    case GoalClass:
    case CommandClass:
      dispatch->hnd = NULL;
      reExecMsg(dispatch, (HND_PTR)NULL);
      break;
    default: Log1("Cannot resend messages of %s class\n", 
		  messageClassName(dispatch->class));
    }
  } else {
    Log1("Not resending message: %s\n", dispatch->msg->msgData->name);
  }

  return 1; /* 28-Oct-91: fedor: test all */
}


/******************************************************************************
*
* FUNCTION: int clearHandler
*
* DESCRIPTION: Clear the handler slot of the dispatch (when the associated
*              resource is deleted)
*
* INPUTS: char *param; (Ignored)
*         DISPATCH_PTR dispatch;
*
* OUTPUTS: Returns 1 to continue iterating through the list
******************************************************************************/

int clearHandler(param, dispatch)
char *param;
DISPATCH_PTR dispatch;
{
  if (dispatch->msg && dispatch->msg->msgData) {
    Log1("Clearing handler for %s message\n", dispatch->msg->msgData->name);
  }
  dispatch->hnd = NULL;

  return TRUE; /* test all items */
}

/******************************************************************************
*
* FUNCTION: int multiQueryUpdateIterate
*
* DESCRIPTION: If the message being handled is a "MultiQuery", need to
*              update the data structure that indicates how many responses are
*              still outstanding.
*
* INPUTS: char *param; (Ignored)
*         DISPATCH_PTR dispatch;
*
* OUTPUTS: Returns 1 to continue iterating through the list
******************************************************************************/

int multiQueryUpdateIterate(param, dispatch)
char *param;
DISPATCH_PTR dispatch;
{
  if (dispatch->msg && dispatch->msg->msgData->class == MultiQueryClass) {
    multiQueryUpdate(dispatch, FALSE);
  }

  return TRUE; /* test all items */
}

/******************************************************************************
*
* FUNCTION: void resourceDelete(resource)
*
* DESCRIPTION:
* Remove the named resource from the resource table.
* Display a warning message if the pending or attending lists are not empty.
*
* INPUTS: RESOURCE_PTR resource;
*
* OUTPUTS: void.
*
* NOTES: If "resendAfterCrashGlobal" is true (the "-r" option to central),
*        then will ask to resend each message in the active and pending sets
*        of the resource. (25-Jul-91: Reid)
******************************************************************************/

void resourceDelete(resource)
RESOURCE_PTR resource;
{
  LIST_PTR resList;

  if (!resource)
    return;

  resList = (LIST_PTR)hashTableFind(resource->name, resourceTable);
  listDeleteItem((char *)resource, resList);

  if (resList && !listLength(resList)) {
    (void)hashTableRemove(resource->name, resourceTable);
    listFree(resList);
  }

  if (listLength(resource->attendingList)) {
    if (resendAfterCrashGlobal) {
      Log1("Resend active and/or pending messages for the %s resource?",
	   resource->name);
      if (readChar() == 'y') {
	(void)listIterate(tryResendingMessage, (char *)NULL, 
			  resource->attendingList);
	(void)listIterate(tryResendingMessage, (char *)NULL, resource->pendingList);
      }
      else {
	Log("Not resending messages.\n");
      }
    } else {
      (void)listIterate(clearHandler, (char *)NULL, resource->attendingList);
      (void)listIterate(clearHandler, (char *)NULL, resource->pendingList);

      (void)listIterate(multiQueryUpdateIterate, (char *)NULL, 
			resource->attendingList);
      (void)listIterate(multiQueryUpdateIterate, (char *)NULL, 
			resource->pendingList);
    }
    listFree(resource->attendingList);
    listFree(resource->pendingList);
  }

  tcaFree(resource->name);
  tcaFree((char *)resource);
}


/******************************************************************************
*
* FUNCTION: int resourceSelfQuery(dispatch)
*
* DESCRIPTION: Test if the message is a self query.
*
* INPUTS: DISPATCH_PTR dispatch
*
* OUTPUTS: int (TRUE or FALSE)
*
******************************************************************************/

int resourceSelfQuery(dispatch)
DISPATCH_PTR dispatch;
{
  if (dispatch->msg->msgData->class == QueryClass &&
      dispatch->desId == dispatch->orgId)
    return TRUE;
  else
    return FALSE;
}


/******************************************************************************
*
* FUNCTION: int resourceAvailableRes(resource)
*
* DESCRIPTION: Test if the resource has capacity for a handler.
*
* INPUTS: RESOURCE_PTR resource;
*
* OUTPUTS: int (TRUE or FALSE)
*
* NOTES:
* 1-Oct-90: fedor: resourceAvailableRes
* Used to select handlers from duplicate identical name handlers from
* duplicate modules.
* Sooo this may return true when the correct answer is false because the
* resource may be reserved or locked. This is ok because 
* processResourceAndDeliver will call resourceAvailable with the complete
* dispatch before sending - and that test will trap resource reserved and
* locked conditions.
*
* 3-Dec-90: fedor: 
* Now this is also used in the lock/reserve code to make sure the
* resource is not busy before setting the lock/reserve.
*
******************************************************************************/

int resourceAvailableRes(resource)
RESOURCE_PTR resource;
{
  if (resource)
    return(listLength(resource->attendingList) < resource->capacity);
  else
    return TRUE;
}


/******************************************************************************
*
* FUNCTION: int resourceAvailable(dispatch)
*
* DESCRIPTION: Test if the dispatch's handlers's resource is available.
*
* INPUTS: DISPATCH_PTR dispatch;
*
* OUTPUTS: int (TRUE or FALSE)
*
******************************************************************************/

int resourceAvailable(dispatch)
DISPATCH_PTR dispatch;
{
  RESOURCE_PTR resource;
  
  resource = dispatch->hnd->resource;

  if (!resource)
    return TRUE;

  if (resource->status == LockedResource)
    return FALSE;

  if (resource->status != ReservedResource)
    return(resourceAvailableRes(resource) || resourceSelfQuery(dispatch));
  else if (resource->module == dispatch->org)
    return(resourceAvailableRes(resource) || resourceSelfQuery(dispatch));
  else
    return FALSE;
}


/******************************************************************************
*
* FUNCTION: void resourceAttending(dispatch)
*
* DESCRIPTION: Add the dispatch to its handler's resource attending list.
*
* INPUTS: DISPATCH_PTR dispatch;
*
* OUTPUTS: void.
*
******************************************************************************/

void resourceAttending(dispatch)
DISPATCH_PTR dispatch;
{
  if (dispatch->hnd->resource)
    listInsertItem((char *)dispatch, dispatch->hnd->resource->attendingList);
}


/******************************************************************************
*
* FUNCTION: void resourceRemoveAttending(dispatch)
*
* DESCRIPTION: Remove the dispatch fom its handler's resource attending list.
*
* INPUTS: DISPATCH_PTR dispatch;
*
* OUTPUTS: void.
*
******************************************************************************/

void resourceRemoveAttending(dispatch)
DISPATCH_PTR dispatch;
{
  if (dispatch->hnd && dispatch->hnd->resource)
    listDeleteItem((char *)dispatch, dispatch->hnd->resource->attendingList);
}


/******************************************************************************
*
* FUNCTION: void resourceRemovePending(dispatch)
*
* DESCRIPTION: 
* Removes the dispatch from the dispatch's hnd resource pending set.
* Interface routine for taskTree killing of tree nodes.
*
* INPUTS: DISPATCH_PTR dispatch;
*
* OUTPUTS: void.
*
* NOTES:
* 24-Oct-90: fedor: little concerned over use of pending set to control
* availability of resource and other pending items.
*
******************************************************************************/

void resourceRemovePending(dispatch)
DISPATCH_PTR dispatch;
{
  if (dispatch->hnd && dispatch->hnd->resource)
    listDeleteItem((char *)dispatch, dispatch->hnd->resource->pendingList);
}


/******************************************************************************
*
* FUNCTION: void resourcePending(dispatch)
*
* DESCRIPTION: Add the dispatch to its handler's resource pending list.
*
* INPUTS: DISPATCH_PTR dispatch;
*
* OUTPUTS: void
*
******************************************************************************/

void resourcePending(dispatch)
DISPATCH_PTR dispatch;
{
  listInsertItemLast((char *)dispatch, dispatch->hnd->resource->pendingList);
}


/******************************************************************************
*
* FUNCTION: void resourceProcessPendingRes(resource)
*
* DESCRIPTION:
* Process a pending dispatch for this resource.
* If the resource is locked no dispatch is processed.
* If the resource is reserved then only a dispatch of the owner of the
* reservation is processed.
*
* Called by resourceProcessPending from flow and control and called by
* cancelation of a reservation or a lock to allow the resource pending
* dispatches to continue.
*
* INPUTS: RESOURCE_PTR resource;
*
* OUTPUTS: void.
*
******************************************************************************/

int sameModTest(module, dispatch)
MODULE_PTR module;
DISPATCH_PTR dispatch;
{
  return(dispatch->org == module);
}

void resourceProcessPendingRes(resource)
RESOURCE_PTR resource;
{
  DISPATCH_PTR pendingDispatch;

  if (resource) {
    if (resource->status == LockedResource)
      return;
    
    if (resource->status == ReservedResource) {
      pendingDispatch = (DISPATCH_PTR)listMemReturnItem(sameModTest, 
				 (char *)resource->module, resource->pendingList);
      listDeleteItem((char *)pendingDispatch, resource->pendingList);
    }
    else 
      pendingDispatch = (DISPATCH_PTR)listPopItem(resource->pendingList);

    if (pendingDispatch) 
      processResAttendDeliver(pendingDispatch);
  }
}


/******************************************************************************
*
* FUNCTION: void resourceProcessPending(dispatch)
*
* DESCRIPTION: 
* Calls resourceProcessPendingRes. Provides an interface to resources for
* flow and control that deals in terms of dispatches.
*
* INPUTS: DISPATCH_PTR dispatch;
*
* OUTPUTS: void.
*
******************************************************************************/

void resourceProcessPending(dispatch)
DISPATCH_PTR dispatch;
{
  if (dispatch->hnd)
    resourceProcessPendingRes(dispatch->hnd->resource);
}


/******************************************************************************
*
* FUNCTION: void registerResourceHnd(dispatch, addResForm)
*
* DESCRIPTION: Handler for tcaRegisterResource.
*
* INPUTS: 
* DISPATCH_PTR dispatch;
* ADD_RES_FORM_PTR addResForm;
*
* OUTPUTS: void.
*
******************************************************************************/

void registerResourceHnd(dispatch, addResForm)
DISPATCH_PTR dispatch;
ADD_RES_FORM_PTR addResForm;
{
  RESOURCE_PTR resource;

  resource = resourceCreate(dispatch->orgId, addResForm->resName, 
			    addResForm->capacity);
  
  if (resource)
    listInsertItem((char *)resource, dispatch->org->resourceList);

  /* A bit more efficient than using tcaFreeData */
  freeDataStructure(dispatch->msg->msgData->msgForm, (char *)addResForm);
}


/******************************************************************************
*
* FUNCTION: void addHndToResourceHnd(dispatch, addHndForm)
*
* DESCRIPTION: The handler for tcaAddHndToResource.
*
* INPUTS: 
* DISPATCH_PTR dispatch;
* ADD_HND_FORM_PTR addHndForm;
*
* OUTPUTS: void.
*
******************************************************************************/

int equalHndName(hndName, hnd)
char *hndName;
HND_PTR hnd;
{
  return(strKeyEqFunc(hndName, hnd->hndData->hndName));
}

int equalResName(resName, res)
char *resName;
RESOURCE_PTR res;
{
  return(strKeyEqFunc(resName, res->name));
}

void addHndToResourceHnd(dispatch, addHndForm)
DISPATCH_PTR dispatch;
ADD_HND_FORM_PTR addHndForm;
{
  HND_PTR hnd;
  RESOURCE_PTR resource;

  hnd = (HND_PTR)listMemReturnItem(equalHndName, addHndForm->hndName,
				   dispatch->org->hndList);

  if (!hnd)
    TCA_ERROR1("ERROR: tcaAddHndToResource: no handler %s\n", 
	       addHndForm->hndName);

  resource = (RESOURCE_PTR)listMemReturnItem(equalResName, addHndForm->resName,
				   dispatch->org->resourceList);  

  if (!resource)
    TCA_ERROR1("ERROR: tcaAddHndToResource: no resource %s\n", 
	       addHndForm->resName);

  if (hnd->resource != dispatch->org->impliedResource)
    Log_Status3("WARNING: Resource for %s is being changed from %s to %s.\n",
		hnd->hndData->hndName, hnd->resource->name, 
		resource->name);

  hnd->resource = resource;

  /* A bit more efficient than using tcaFreeData */
  freeDataStructure(dispatch->msg->msgData->msgForm, (char *)addHndForm);
}


/******************************************************************************
*
* FUNCTION: void resourceSetStatus(dispatch, resource, status)
*
* DESCRIPTION:
* Sets the reserved or locked status of a resource.
* Checks for multiple reservations of the same module
*
* INPUTS:
* DISPATCH_PTR dispatch;
* RESOURCE_PTR resource;
* RESOURCE_STATUS_TYPE status;
*
* OUTPUTS: void.
*
* NOTES: This handles the central reply for reservations and locks.
*
******************************************************************************/

void resourceSetStatus(dispatch, resource, status)
DISPATCH_PTR dispatch;
RESOURCE_PTR resource;
RESOURCE_STATUS_TYPE status;
{
  int reject;

  if (resource->status == ActiveResource && resourceAvailableRes(resource)) {
    reserveDispatch(dispatch);
    resource->module = dispatch->org;
    dispatch->resource = resource;
    RESOURCE_SET_STATUS(status, resource);
    centralReply(dispatch, (char *)&(dispatch->locId));
  } else if (status == ReservedResource && resource->module == dispatch->org) {
    /* multiple request for reservation from same module */
    reject = -1;
    centralReply(dispatch, (char *)&reject);
  }
  else {
    listInsertItem((char *)dispatch, resource->pendingList);
    dispatchUpdateAndDisplay(CentralPendingDispatch, dispatch);
  }
}


/******************************************************************************
*
* FUNCTION: void resourceReserveLock(dispatch, resName, status)
*
* DESCRIPTION:
* Called by the handlers for tcaReserveResource or tcaLockResource.
* The local id of the dispatch for this message becomes the reference for
* the reservation or lock. This dispatch can not be freed until the 
* reservation or lock is removed.
*
* If a reservation or lock is already in progress then this message is
* queued in the pending set of the resource to be sent again when
* the resource becomes available.
*
* INPUTS: 
* DISPATCH_PTR dispatch;
* char *resName;
* RESOURCE_STATUS_TYPE status;
*
* OUTPUTS: void.
*
******************************************************************************/

void resourceReserveLock(dispatch, resName, status)
DISPATCH_PTR dispatch;
char *resName;
RESOURCE_STATUS_TYPE status;
{
  LIST_PTR resList;
  RESOURCE_PTR resource;

  resList = (LIST_PTR)hashTableFind(resName, resourceTable);

  if (!resList)
    TCA_ERROR2("ERROR: %s: No Resource: %s\n", 
	       dispatch->msg->msgData->name, resName);
    
  if (listLength(resList) > 1)
    TCA_ERROR2("ERROR: %s: Resource %s is not unique. Specify module.\n", 
	       dispatch->msg->msgData->name, resName);

  resource = (RESOURCE_PTR)listFirst(resList);

  resourceSetStatus(dispatch, resource, status);
}


/******************************************************************************
*
* FUNCTION: void resourceModReserveLock(dispatch, addForm, status)
*
* DESCRIPTION: 
* Called by the handlers for tcaReserveModResource and tcaLockModResource.
* The local id of the dispatch for this message becomes the reference for
* the reservation or lock. This dispatch can not be freed until the 
* reservation or lock is removed.
*
* If a reservation or lock is already in progress then this message is
* queued in the pending set of the resource to be sent again when
* the resource becomes available.
*
* INPUTS: 
* DISPATCH_PTR dispatch;
* ADD_HND_FORM_PTR addForm;
* RESOURCE_STATUS_TYPE status;
*
* OUTPUTS: void.
*
******************************************************************************/

int modNameEq(modName, module)
char *modName;
MODULE_PTR module;
{
  return(strKeyEqFunc(modName, module->modData->modName));
}

void resourceModReserveLock(dispatch, addForm, status)
DISPATCH_PTR dispatch;
ADD_HND_FORM_PTR addForm;
RESOURCE_STATUS_TYPE status;
{
  LIST_PTR resList;
  MODULE_PTR module;
  RESOURCE_PTR resource;

  resList = (LIST_PTR)hashTableFind(addForm->resName, resourceTable);

  if (!resList)
    TCA_ERROR2("ERROR: %s: No Resource: %s\n", 
	       dispatch->msg->msgData->name, addForm->resName);

  module = (MODULE_PTR)listMemReturnItem(modNameEq, addForm->hndName, 
					 moduleList);
					 
  resource = (RESOURCE_PTR)listMemReturnItem(resEqFunc, (char *)&(module->sd),
					     resList);

  if (!resource)
    TCA_ERROR3("ERROR: %s: No Resource: %s: Registered by: %s\n", 
	       dispatch->msg->msgData->name, addForm->resName, 
	       addForm->hndName);
  
  resourceSetStatus(dispatch, resource, status);
}


/******************************************************************************
*
* FUNCTION: void reserveResourceHnd(dispatch, resName)
*
* DESCRIPTION: Handler for tcaReserveResource.
*
* INPUTS: 
* DISPATCH_PTR dispatch;
* char **resName;
*
* OUTPUTS: void.
*
******************************************************************************/

void reserveResourceHnd(dispatch, resName)
DISPATCH_PTR dispatch;
char **resName;
{
  resourceReserveLock(dispatch, *resName, ReservedResource);

  /* A bit more efficient than using tcaFreeData */
  freeDataStructure(dispatch->msg->msgData->msgForm, (char *)resName);
}


/******************************************************************************
*
* FUNCTION: void reserveModResourceHnd(dispatch, addForm)
*
* DESCRIPTION: Handler for tcaReserveModResource.
*
* INPUTS: 
* DISPATCH_PTR dispatch;
* ADD_HND_FORM_PTR addForm;
*
* OUTPUTS: void.
*
******************************************************************************/

void reserveModResourceHnd(dispatch, addForm)
DISPATCH_PTR dispatch;
ADD_HND_FORM_PTR addForm;
{
  resourceModReserveLock(dispatch, addForm, ReservedResource);

  /* A bit more efficient than using tcaFreeData */
  freeDataStructure(dispatch->msg->msgData->msgForm, (char *)addForm);
}


/******************************************************************************
*
* FUNCTION: void cancelReservationHnd(dispatch, ref)
*
* DESCRIPTION: 
* Handler for tcaCancelReservation.
* Also called by the handler for tcaUnlockResource.
*
* INPUTS:
* DISPATCH_PTR dispatch;
* int *ref; 
*
* OUTPUTS: void.
*
* NOTES: ref should be changed to a TCA_REF_PTR
*
******************************************************************************/

void cancelReservationHnd(dispatch, ref)
DISPATCH_PTR dispatch;
int *ref;
{
  DISPATCH_PTR reservation;

  reservation = DISPATCH_FROM_ID(*ref);

  reservation->resource->module = NULL;
  RESOURCE_SET_STATUS(ActiveResource, reservation->resource);

  resourceProcessPendingRes(reservation->resource);

  releaseDispatch(reservation);

  tcaFree((char *)ref); /* Use simple free: only 1 int */
}


/******************************************************************************
*
* FUNCTION: void lockResourceHnd(dispatch, resName)
*
* DESCRIPTION: Handler for tcaLockResource.
*
* INPUTS: 
* DISPATCH_PTR dispatch;
* char **resName;
*
* OUTPUTS: void.
*
******************************************************************************/

void lockResourceHnd(dispatch, resName)
DISPATCH_PTR dispatch;
char **resName;
{
  resourceReserveLock(dispatch, *resName, LockedResource);

  /* A bit more efficient than using tcaFreeData */
  freeDataStructure(dispatch->msg->msgData->msgForm, (char *)resName);
}


/******************************************************************************
*
* FUNCTION: void lockModResourceHnd(dispatch, addForm)
*
* DESCRIPTION: Handler for tcaLockModResource.
*
* INPUTS: 
* DISPATCH_PTR dispatch;
* ADD_HND_FORM_PTR addForm;
*
* OUTPUTS: void.
*
******************************************************************************/

void lockModResourceHnd(dispatch, addForm)
DISPATCH_PTR dispatch;
ADD_HND_FORM_PTR addForm;
{
  resourceModReserveLock(dispatch, addForm, LockedResource);

  /* A bit more efficient than using tcaFreeData */
  freeDataStructure(dispatch->msg->msgData->msgForm, (char *)addForm);
}


/******************************************************************************
*
* FUNCTION: void unlockResourceHnd(dispatch, ref)
*
* DESCRIPTION: Handler for tcaUnlockResource.
*
* INPUTS: 
* DISPATCH_PTR dispatch;
* int *ref;
*
* OUTPUTS: void.
*
* NOTES: ref should be changed to a TCA_REF_PTR
*
******************************************************************************/

void unlockResourceHnd(dispatch, ref)
DISPATCH_PTR dispatch;
int *ref;
{
  cancelReservationHnd(dispatch, ref);

  tcaFree((char *)ref); /* Use simple free: only 1 int */
}


/******************************************************************************
*
* FUNCTION: void resourceInitialize()
*
* DESCRIPTION: Initialize resources.
*
* INPUTS: none.
*
* OUTPUTS: none.
*
******************************************************************************/

void resourceInitialize()
{
  centralRegisterMessage("registerResourceMsg", InformClass, 
			 "{int, string}", (char *)NULL);
  centralRegisterHandler("registerResourceMsg", "registerResourceHnd", 
		  registerResourceHnd);

  centralRegisterMessage("addHndToResourceMsg", InformClass,
		  "{string, string}", (char *)NULL);
  centralRegisterHandler("addHndToResourceMsg", "addHndToResourceHnd",
		  addHndToResourceHnd);

  centralRegisterMessage("reserveResourceMsg", QueryClass, "string", "int");
  centralRegisterHandler("reserveResourceMsg", "reserveResourceHnd",
		  reserveResourceHnd);

  centralRegisterMessage("reserveModResourceMsg", QueryClass,
		  "{string, string}", "int");
  centralRegisterHandler("reserveModResourceMsg", "reserveModResourceHnd",
		  reserveModResourceHnd);

  centralRegisterMessage("cancelReservationMsg", InformClass,
			 "int", (char *)NULL);
  centralRegisterHandler("cancelReservationMsg", "cancelReservationHnd", 
		  cancelReservationHnd);

  centralRegisterMessage("lockResourceMsg", QueryClass, "string", "int");
  centralRegisterHandler("lockResourceMsg", "lockResourceHnd", 
		  lockResourceHnd);

  centralRegisterMessage("lockModResourceMsg", QueryClass,
		  "{string, string}", "int");
  centralRegisterHandler("lockModResourceMsg", "lockModResourceHnd",
		  lockModResourceHnd);

  centralRegisterMessage("unlockResourceMsg", InformClass, "int", (char *)NULL);
  centralRegisterHandler("unlockResourceMsg", "unlockResourceHnd", 
		  unlockResourceHnd);
  
  resourceTable = hashTableCreate(10, strHashFunc, strKeyEqFunc);
}
