/******************************************************************************
*
* PROJECT: Carnegie Mellon Planetary Rover Project
*          Task Control Architecture 
* 
* (c) Copyright 1991 Christopher Fedor and Reid Simmons.  All rights reserved.
* 
* MODULE: resources
*
* FILE: resMod.c
*
* ABSTRACT:
* 
* Resources - Module Routines.
*
* REVISION HISTORY
*
* 30-Jan-91 Christopher Fedor, School of Computer Science, CMU
* Added fflush(stdout) to printf for module code calls from lisp
*
* 24-Sep-90 Christopher Fedor, School of Computer Science, CMU
* Module resource routines - tca version 5.0
*
* 10-Jul-90 Christopher Fedor, School of Computer Science, CMU
* created.
*
******************************************************************************/

#ifdef VXWORKS
#include "stdioLib.h"
#else
#include "stdio.h"
#endif
#include "tcaInternal.h"

#include "resMod.h"

/******************************************************************************
*
* FUNCTION: void tcaRegisterResource(resName, capacity)
*
* DESCRIPTION: Define a new resource for a module.
*
* INPUTS: 
* char *resName;
* int capacity;
*
* OUTPUTS: void
*
******************************************************************************/

void tcaRegisterResource(resName, capacity)
char *resName;
int capacity;
{
  ADD_RES_FORM_TYPE addResForm;

  addResForm.capacity = capacity;
  addResForm.resName = resName;

  (void)tcaInform("registerResourceMsg", (void *)&addResForm);
}


/******************************************************************************
*
* FUNCTION: void tcaAddHndToResource(hndName, resName)
*
* DESCRIPTION: Assign a resource to a handler.
*
* INPUTS: char *hndName, *resName;
*
* OUTPUTS: void
*
******************************************************************************/

void tcaAddHndToResource(hndName, resName)
char *hndName, *resName;
{
  ADD_HND_FORM_TYPE addHndForm;

  addHndForm.hndName = hndName;
  addHndForm.resName = resName;

  (void)tcaInform("addHndToResourceMsg", (void *)&addHndForm);
}


/******************************************************************************
*
* FUNCTION: TCA_REF_PTR tcaReserveResource(resName)
*
* DESCRIPTION: Reserve a resource for use only by the reserving module.
*
* INPUTS: char *resName;
*
* OUTPUTS: TCA_REF_PTR
*
******************************************************************************/

TCA_REF_PTR tcaReserveResource(resName)
char *resName;
{
  int refId;
  MSG_PTR msg;
  TCA_REF_PTR ref;

  refId = 0;

  (void)tcaQuery("reserveResourceMsg", (void *)&resName, (void *)&refId);

  if (refId < 0) {
    fprintf(stderr, "WARNING: Ignoring reservation request.\n");
    fprintf(stderr, "%s is already reserved by this module.\n", resName);
    ref = CREATE_NULL_REF();
#ifdef LISP
    fflush(stdout);
#endif
  }
  else {
    msg = msgFind("reserveResourceMsg");
    ref = tcaRefCreate(msg, "reserveResourceMsg", refId);
  }

  return ref;
}


/******************************************************************************
*
* FUNCTION: TCA_REF_PTR tcaReserveModResource(modName, resName)
*
* DESCRIPTION: Reserve a resource by further specifying the module name.
*
* INPUTS:
* char *modName;
* char *resName;
*
* OUTPUTS: TCA_REF_PTR
*
******************************************************************************/

TCA_REF_PTR tcaReserveModResource(modName, resName)
char *modName;
char *resName;
{
  int refId;
  MSG_PTR msg;
  TCA_REF_PTR ref;
  ADD_HND_FORM_TYPE addForm;

  addForm.hndName = modName;
  addForm.resName = resName;

  refId = 0;

  (void)tcaQuery("reserveModResourceMsg", (void *)&addForm, (void *)&refId);

  if (refId < 0) {
    fprintf(stderr, "WARNING: Ignoring reservation request.\n");
    fprintf(stderr, "%s of module %s is already reserved by this module.\n", 
	    resName, modName);
    ref = CREATE_NULL_REF();
#ifdef LISP
    fflush(stdout);
#endif
  }
  else {
    msg = msgFind("reserveModResourceMsg");
    ref = tcaRefCreate(msg, "reserveModResourceMsg", refId);
  }

  return ref;
}


/******************************************************************************
*
* FUNCTION: void tcaCancelReservation(ref)
*
* DESCRIPTION: Cancel a reservation.
*
* INPUTS: TCA_REF_PTR ref
*
* OUTPUTS: void.
*
******************************************************************************/

void tcaCancelReservation(ref)
TCA_REF_PTR ref;
{
  if (STREQ(ref->name, "reserveResourceMsg") ||
      STREQ(ref->name, "reserveModResourceMsg")) {
    (void)tcaInform("cancelReservationMsg", (void *)&(ref->refId));
    tcaRefFree(ref);
  }
  else {
    tcaModError("ERROR: tcaCancelReservation: TCA_REF_PTR not a reservation.\n");
  }
}


/******************************************************************************
*
* FUNCTION: TCA_REF_PTR tcaLockResource(resName)
*
* DESCRIPTION: Lock a resource to prevent any module from using it.
*
* INPUTS: char *resName;
*
* OUTPUTS: TCA_REF_PTR
*
******************************************************************************/

TCA_REF_PTR tcaLockResource(resName)
char *resName;
{
  int refId;
  MSG_PTR msg;
  TCA_REF_PTR ref;

  (void)tcaQuery("lockResourceMsg", (void *)&resName, (void *)&refId);

  msg = msgFind("lockResourceMsg");

  ref = tcaRefCreate(msg, "lockResourceMsg", refId);

  return ref;
}


/******************************************************************************
*
* FUNCTION: TCA_REF_PTR tcaLockModResource(modName, resName)
*
* DESCRIPTION: Lock a resource - specified with module name.
*
* INPUTS: 
* char *modName;
* char *resName;
*
* OUTPUTS: TCA_REF_PTR
*
******************************************************************************/

TCA_REF_PTR tcaLockModResource(modName, resName)
char *modName;
char *resName;
{
  int refId;
  MSG_PTR msg;
  TCA_REF_PTR ref;
  ADD_HND_FORM_TYPE addForm;

  addForm.hndName = modName;
  addForm.resName = resName;

  (void)tcaQuery("lockModResourceMsg", (void *)&addForm, (void *)&refId);

  msg = msgFind("lockModResourceMsg");

  ref = tcaRefCreate(msg, "lockModResourceMsg", refId);

  return ref;
}


/******************************************************************************
*
* FUNCTION: void tcaUnlockResource(ref)
*
* DESCRIPTION: Cancel a resource lock.
*
* INPUTS: TCA_REF_PTR ref
*
* OUTPUTS: void.
*
******************************************************************************/

void tcaUnlockResource(ref)
TCA_REF_PTR ref;
{
  if (STREQ(ref->name, "lockResourceMsg") ||
      STREQ(ref->name, "lockModResourceMsg")) {
    (void)tcaInform("unlockResourceMsg", (void *)&(ref->refId));
    tcaRefFree(ref);
  }
  else {
    tcaModError("ERROR: tcaUnlockResource: TCA_REF_PTR not a lock.\n");
  }
}
