/*****************************************************************************
*
* PROJECT: Carnegie Mellon Planetary Rover Project
*          Task Control Architecture
*
* (c) Copyright 1991 Christopher Fedor and Reid Simmons.  All rights reserved.
* 
* MODULE: tms
*
* FILE: rules.h
*
* ABSTRACT:
* There are 6 types of rules -- 2 for tms_nodes, 2 for cells and 2 for
* maintaining consistency with the outside world.
*
* 1) Node_rule ---
*    <name fn priority tms_node>
*    Attached to a node.  
*    Fires the first time the node comes IN (first only once).
*    The "fn" is called with the tms_node.
*
* 2) Node_constraint_rule ---
*    <name fn priority consequent_cell antecedent_nodes>
*    Attached to one of the antecedent tms_nodes.
*    Fires when that node comes IN and if the rest of the
*    antecedents are IN. Otherwise, attaches itself to one of the OUT nodes,
*    waiting until all antecedent nodes are IN at the same time.
*    DOES NOT FIRE if the node it is currently attached to is itself derived
*    by "name" (i.e. the informant of the derivation is the same as the
*    "name"). NOTE: thus, "name" should be unique for each
*    invocation of this rule (see function ...).
*    The "fn" is called with the consequent cell and list of antecedent_nodes.
*
* 3) Cell_rule ---
*    <name fn priority>
*    Attached to a cell.
*    Fires when a new tms_node is added to the cell (see file "cells").
*    NOTE: the node does not have to be IN for the cell_rule to fire.
*    Adds a node_rule to the tms_node added.
*
* 4) Constraint_rule ---
*    <name fn priority consequent_cell antecedent_cells>
*    Attached to each of the antecendent cells.
*    Fires on the cross product of all antecendents (i.e. whenever a 
*    tms_node is added to an antecedent cell, create the cross product of 
*    that node and all other nodes of the other antecedents).
*    Adds a node_constraint_rule to the antecedent tms_nodes. 
* 
* 5) In_rule (for maintaining consistency with the outside world) ---
*    <name fn priority>
*    Attached to a cell.  
*    Fires whenever a node in that cell's contextual_value list comes IN, 
*    unless the cell is a ConstrainedCell, in which case the node must 
*    also be the most constrained. The "fn" is called with the node IN-ned.
*
* 6) Out_rule (for maintaining consistency with the outside world) ---
*    <name fn priority>
*    Attached to a cell.  
*    Fires whenever a node in that cell's contextual_value list goes OUT
*    (it is not fired if the node is initially OUT when the rule is added)
*    Argument is the node OUT-ted.
*
* In addition, all rules can be parameterized, so that the same "fn" can be
* used for different rule instances.  For example, a rule to apply a function
* to the value of the arguments can be parameterized with different functions,
* such as "max", "min", etc.
*
* For parameterized rules, the rule "fn" is called with the arguments as 
* indicated above, followed by a pointer to a structure containing the 
* parameters.
*
* REVISION HISTORY
*
* 03-Jul-1991 Reid Simmons at School of Computer Science, CMU
* Added a "used" slot to rule structures to facilitate memory management
*
*****************************************************************************/


typedef enum { NodeRule, NodeConstraintRule, CellRule, ConstraintRule, InRule, 
	       OutRule } RULE_CLASS_TYPE;


typedef struct { 
  CELL_PTR consequentCell;
  LIST_PTR antecedentList;
} ADDL_CONSTRAINT_TYPE, *ADDL_CONSTRAINT_PTR;

typedef struct {
  RULE_CLASS_TYPE class;
  char *name;
  void (*fn)();
  int priority;
  char *parameters;
  /* 2-Mar-90: fedor may want to try a union here */
  TMS_NODE_PTR node;
  LIST_PTR innedNodes; 
  LIST_PTR outtedNodes;
  ADDL_CONSTRAINT_PTR addlConstraints;
  int used; /* Number of applications of this rule */
} RULE_TYPE, *RULE_PTR;

/**********************************************************************
* Execute the 'body' without firing rules.
**********************************************************************/
#define DELAY_RULES(body)	\
  { int tmpDelay;		\
    tmpDelay = isDelayedGlobal; \
    isDelayedGlobal = TRUE;     \
    body;			\
    isDelayedGlobal = tmpDelay;	\
    (void)runRules(-1);		\
  }

extern int isDelayedGlobal;
extern int ruleTraceGlobal;

extern void rulesInit();
extern int runRules();
extern void fireCellRules();
extern void runInRules();
extern void runOutRules();
extern RULE_PTR ruleCreateNode();
extern RULE_PTR ruleCreateNodeConstraint();
extern RULE_PTR ruleCreateCell();
/* extern RULE_PTR ruleCreateConstraint(); */
extern RULE_PTR ruleCreateIn();
extern RULE_PTR ruleCreateOut();
extern void rulesFree();
extern void addRuleToNode();
/* extern void addRuleToCell(); */
extern void putRuleOnCell();
extern void addInRule();
extern void addOutRule();
/* extern void addConstraintRule(); */

extern void removeOutRule();
/* extern void removeInRule(); */
/* extern void removeCellRule(); */
/* extern void removeNodeRule(); */
