/******************************************************************************
*
* PROJECT: Carnegie Mellon Planetary Rover Project
*          Task Control Architecture 
*
* (c) Copyright 1991 Christopher Fedor and Reid Simmons.  All rights reserved.
* 
* MODULE: taskTree
*
* FILE: taskTree.c
*
* ABSTRACT:
* 
* Creation and manipulation of hierarchical task trees
* (containing goal, command, and monitor message nodes).
*
* REVISION HISTORY
*
* 15-Oct-92 Reid Simmons, School of Computer Science, CMU
* Converted "childrenList" to use the new "list" data structures.
* Changed the return values for "tcaReferenceStatus".
*
* 29-Sept-92 RTG : Modified FreeTaskTreeNode to Remove reference to the
* node from the parent node.  This was causing problems when doing
* findFirstChild when some of the children of a node had been killed.
*
* 31-Dec-91 Reid Simmons, School of Computer Science, CMU
* Added call to add all exception handlers to node that are associated
*  with the node's message in general.
*
* 14-Dec-91 Christopher Fedor, School of Computer Science, CMU
* Added Rich Goodwin's referenceStatusHnd
*
* 16-Oct-91 Christopher Fedor, School of Computer Science, CMU
* Moved the code to remove children from a killed node to happen
* before deleteTaskTree call for that node in both killTaskTree and
* killSubTaskTree.
*
* 21-Aug-91 Christopher Fedor, School of Computer Science, CMU
* Added changes from Reid includeing:
* Check that msgRef != NO_REF in referenceReleaseHnd.
* Tests on dispatch refCount and status for HandleKilAfterAttendingNodes
*
* 26-Jul-91, Reid Simmons, School of Computer Science, CMU
* Task tree killing had a bug -- did not correctly remove root node of kill
* tree from the children of its parent.
*
* 23-Jul-91, Reid Simmons, School of Computer Science, CMU
* Adding a new node to a goal node might nullify the "end of interval"
* completion constraints.  Need to retract those constraints before proceeding.
*
*  3-Jul-91, Reid Simmons, School of Computer Science, CMU
* Pre-remove *all* the rules from *all* the relations of the quantities 
* associated with the nodes to be killed.
* Also freed data used by the centrally registered handlers. 
*
* 13-Jun-91, Reid Simmons, School of Computer Science, CMU
* Added code to remove wiretap rules when task tree nodes are killed.
* Also, free dispatch when corresponding nodes are killed.
*
* 24-Oct-90 Christopher Fedor, School of Computer Science, CMU
* Spliced in the new resource routines were needed.
*
* 25-May-90 Christopher Fedor, School of Computer Science, CMU
* Revised to Software Standards.
* Removed addTreeNodeIfNecessary test because with the new flow and control,
* the message class is known, so only the existance of a tree node is needed
* and can be done easier elsewhere, i.e. in recvMessage.
*
* 30-Nov-89 Reid Simmons, School of Computer Science, CMU
* Display task tree now prints out status of node.
*
* 28-Nov-89 Long-Ji Lin, School of Computer Science, CMU
* Added addTreeNodeIfNecessary (dispatch). Added two slots (excHnds & ECNode)
* to tree node and updated "FreeTaskTreeNode()" and "taskTreeNodeCreate" 
* correspondingly.
*
* 27-Nov-89 Reid Simmons, School of Computer Science, CMU
* Added new routines for tracing task tree (find first and last child, 
* find child by name, get data of reference).
*
* 10-Nov-89 Reid Simmons, School of Computer Science, CMU
* First, primitive attempt at adding a mechanism to abort tasks that are 
* currently attending. Used now only for aborting a tcaDelay command that is
* to-be-killed (otherwise, the delay could linger for a long while).
*
* 10-Nov-89 Reid Simmons, School of Computer Science, CMU
* Bug fixes -- task tree tracing functions didn't work right for the 
* root node; displaying and re-killing a killed node didn't work; couldn't
* kill an inactive interval monitor.
*
* 18-Oct-89 Reid Simmons, School of Computer Science, CMU
* Added tracing code to track down "intermittent" task tree killing bug.
* (running unknown rule type).
*
*  2-Oct-89 Reid Simmons, School of Computer Science, CMU
* Moved "Kill" statuses to task tree nodes.
*
* 19-Sep-89 Reid Simmons, School of Computer Science, CMU
* Added a global task tree root node.
*
* 29-Aug-89 Long-Ji Lin, School of Computer Science, CMU
* Changed "DoList" in list.h and replaced "DoList_Star" with "DoList".
*
* 23-Aug-89 Long-Ji Lin, School of Computer Science, CMU
* Fixed a few bugs, including replacing DoList with DoList_Star.
* DoList may leave the "Cdr" pointer dangling, if the body of the 
* do loop involves destructive operations to the "Car" element.
*
* 23-Aug-89 Long-Ji Lin, School of Computer Science, CMU
* Added routines to get root and antecedent references, 
* and to display task trees.
*
* 14-Aug-89 Reid Simmons, School of Computer Science, CMU
* Added routines to kill task trees, get parent and
* children references, message handlers.
*
* 11-Aug-89 Reid Simmons, School of Computer Science, CMU
* Changed terms from "goal tree" to "task tree"
*
* 4-Aug-89 Reid Simmons, School of Computer Science, CMU
* Added warning if Query or tcaAddConstraint issue a
* goal, command, or monitor message.
*
* 2-Aug-89 Reid Simmons, School of Computer Science, CMU
* Added "lastChild" slot to nodes, to make sequentiality constraints easier
* to add. Also, changed so that CreateReference no longer makes the created 
* node the last child (this is done only when responding to an actual message).
*
* 28-Apr-89 Reid Simmons, School of Computer Science, CMU
* Created.
******************************************************************************/

#include "stdio.h"
#include "strings.h"

#include "tcaInternal.h"
#include "logging.h"
#include "idtable.h"
#include "hash.h"
#include "list.h"
#include "queue.h"
#include "key.h"

#include "formatters.h"
#include "datamsg.h"
#include "dispatch.h"
#include "res.h"

#include "tms.h"
#include "cells.h"
#include "rules.h"

#include "qlattice.h"
#include "tplConstr.h"
#include "taskTree.h"

extern QUANTITY_PTR Now;

extern ID_TABLE_PTR msgIdTable;

extern FORMAT_PTR createIntegerFormat();

extern MODULE_PTR tcaServerModGlobal;

extern queue_ptr tmsPriorityQueuesGlobal[]; /* defined in rules.c */
extern int numRuleQueuesGlobal; /* defined in rules.c */

extern void cancelIfAttendingIntervalMonitor();

extern void addMsgExceptions(); /* defined in exception.c */

static char *nodeStatusNames[DISPATCH_STATUS_TOTAL];

void resetLastChild(/*parentNode, childNode*/);

HASH_TABLE_PTR tcaAbortTableGlobal;
TASK_TREE_NODE_PTR tcaTaskTreeRootGlobal;


/******************************************************************************
*
* FUNCTION TASK_TREE_NODE_PTR taskTreeNodeFromRef(ref)
*
* DESCRIPTION: 
* Returns the task tree node associated with this dispatch ref.
* If NO_REF is passed, tcaTaskTreeRootGlobal is returned.
* If the dispatch reffered to is NULL, then NULL is returned.
*
* INPUTS: int ref
*
* OUTPUTS: TASK_TREE_NODE_PTR
*
******************************************************************************/

TASK_TREE_NODE_PTR taskTreeNodeFromRef(ref) 
int ref;
{ 
  DISPATCH_PTR dispatch;

  if (ref == NO_REF) {
    return tcaTaskTreeRootGlobal;
  } else {
    dispatch = DISPATCH_FROM_ID(ref);
    if (dispatch) {
      return dispatch->treeNode;
    } else {
      return NULL;
    }
  }
}


/******************************************************************************
*
* FUNCTION: DISPATCH_PTR findParentDispatch(child)
*
* DESCRIPTION:
* Given the child dispatch, return the dispatch of the 
* parent node of the child.
* Return NULL if no susch node exists.
*
* INPUTS: DISPATCH_PTR child;
*
* OUTPUTS: DISPATCH_PTR
*
******************************************************************************/

DISPATCH_PTR findParentDispatch(child)
DISPATCH_PTR child;
{ 
  TASK_TREE_NODE_PTR childNode, parentNode;

  if (child) {
    childNode = child->treeNode;
    if (childNode) {
      parentNode = childNode->parent;
      if (parentNode)
	return parentNode->dispatch;
    }
  }
  /* default case -- no parent */
  return NULL;
}


/******************************************************************************
*
* FUNCTION: TASK_TREE_NODE_PTR taskTreeNodeCreate(dispatch, parentRef)
*
* DESCRIPTION: Create a task tree node and attach it to its parent.
*
* INPUTS: 
* DISPATCH_PTR dispatch;
* int parentRef;
*
* OUTPUTS: TASK_TREE_NODE_PTR
*
******************************************************************************/

TASK_TREE_NODE_PTR taskTreeNodeCreate(dispatch, parentRef)
DISPATCH_PTR dispatch;
int parentRef;
{ 
  DISPATCH_PTR parentDispatch;
  TASK_TREE_NODE_PTR treeNode, parentNode;

  treeNode = NEW(TASK_TREE_NODE_TYPE);

  treeNode->dispatch = dispatch;
  treeNode->priority = NULL;
  treeNode->children = NULL;
  treeNode->lastChild = NULL;
  treeNode->handlingInterval = NULL;
  treeNode->achievementInterval = NULL;
  treeNode->planningInterval = NULL;
  treeNode->status = ActiveNode;

  treeNode->excHndList = NULL;

  treeNode->ECNode = NULL;
  treeNode->retries = 0;

  /* 12-May-91: fedor: important that root has a NULL parent for exceptions */

  /* No dispatch means no parent node */
  if (!dispatch) {
    treeNode->parent = NULL; 
  } else {
    parentNode = taskTreeNodeFromRef(parentRef);
    treeNode->parent = parentNode;
    dispatch->treeNode = treeNode;
    if (parentNode) {
      if (!parentNode->children)
	parentNode->children = listCreate();
      listInsertItem((char *)treeNode, parentNode->children);
    } else {
      parentDispatch = DISPATCH_FROM_ID(parentRef);

      if (!taskTreeClass(parentDispatch->msg->msgData->class)) {
	Log3("WARNING!! Issuing a %s class message (%s) from a %s class \n",
	     messageClassName(dispatch->msg->msgData->class),
	     dispatch->msg->msgData->name,
	     messageClassName(parentDispatch->msg->msgData->class));
	Log1("          message (%s) may yield unpredictable temporal results.\n",
	     parentDispatch->msg->msgData->name);
      } else {
	TCA_ERROR1("ERROR: Could not find a task tree node for message %s.\n",
		   parentDispatch->msg->msgData->name);
      }
    }

    /* Added 12/31/91: reids */
    if (dispatch->msg) {
      addMsgExceptions(dispatch->msg, treeNode);
    }
  }

  return treeNode;
}


/******************************************************************************
*
* FUNCTION: void FreeTaskTreeNode(TreeNode)
*
* DESCRIPTION:
*
* INPUTS: 
*
* OUTPUTS:
*
******************************************************************************/

void FreeTaskTreeNode(TreeNode)
TASK_TREE_NODE_PTR TreeNode;
{ 
  struct _interval_type *handlingInterval;

  /* 29-Sept-92 RTG : Remove reference from the parent to be safe */
  if ((TreeNode->parent != NULL) && listLength(TreeNode->parent->children) > 0) {
    /* RTG this is not quit right, so I'm commenting it out for now */
    /* resetLastChild(TreeNode->parent, TreeNode); */
    listDeleteItem((char *)TreeNode, TreeNode->parent->children);
  }

  listFree(TreeNode->children);
  TreeNode->children=NULL;
  if (TreeNode->priority) {
    FreeQuantity(TreeNode->priority);
    TreeNode->priority = NULL;
  }
  handlingInterval = TreeNode->handlingInterval;
  if (handlingInterval) {
    if (TreeNode->planningInterval != handlingInterval)
      FreeInterval(TreeNode->planningInterval);
    TreeNode->planningInterval = NULL;
    if (TreeNode->achievementInterval != handlingInterval) 
      FreeInterval(TreeNode->achievementInterval );
    TreeNode->achievementInterval = NULL;
    FreeInterval(handlingInterval);
    TreeNode->handlingInterval = NULL;
  }
  /* 29-Nov-89: ljl - free ECNode & excHnds here? */
  if (TreeNode->ECNode)  {
     tcaFree((char *)TreeNode->ECNode);
     TreeNode->ECNode = NULL;
  }

  if (TreeNode->excHndList) {
    listFree(TreeNode->excHndList);
    TreeNode->excHndList = NULL;
  }

  if (TreeNode->dispatch) {
    dispatchFree(TreeNode->dispatch);
    TreeNode->dispatch = NULL;
  }
  
  tcaFree((char *)TreeNode);
}

int freeTaskTreeNode1(dummy, node)
void *dummy;
TASK_TREE_NODE_PTR node;
{
  FreeTaskTreeNode(node);
  return TRUE;
}

/******************************************************************************
*
* FUNCTION: int taskTreeClass(msgClass)
*
* DESCRIPTION:
* Returns TRUE if the class of message is one that is added to the task 
* tree (currently, goals, commands, exceptions, point-monitors,
* interval-monitors).
*
* INPUTS: TCA_MSG_CLASS_TYPE msgClass;
*
* OUTPUTS: int
*
******************************************************************************/

int taskTreeClass(msgClass)
TCA_MSG_CLASS_TYPE msgClass;
{
  return ((msgClass == GoalClass) || 
	  (msgClass == CommandClass) ||
	  (msgClass == ExceptionClass) ||
	  (msgClass == PointMonitorClass) ||
	  (msgClass == PollingMonitorClass) ||
	  (msgClass == DemonMonitorClass));
}


/******************************************************************************
*
* FUNCTION: int Sequencing_Class(msgClass)
*
* DESCRIPTION:
* Returns TRUE if the class of message is one that is used for the
* sequential temporal constraints (SEQ_ACH, SEQ_PLANNING).
* These are the class of messages that can be a "lastChild"
* (currently, goals, commands, and point-monitors).
*
* INPUTS: TCA_MSG_CLASS_TYPE msgClass;
*
* OUTPUTS: int
*
******************************************************************************/

int Sequencing_Class(msgClass)
TCA_MSG_CLASS_TYPE msgClass;
{
  return ((msgClass == GoalClass) || 
	  (msgClass == CommandClass) ||
	  (msgClass == PointMonitorClass));
}


/******************************************************************************
*
* FUNCTION: int sameModule(module, modLastChild)
*
* DESCRIPTION:
*
* INPUTS: 
*
* OUTPUTS:
*
******************************************************************************/

int sameModule(module, modLastChild)
MODULE_PTR module;
MOD_LAST_CHILD_PTR modLastChild;
{
  return module == modLastChild->module;
}


/******************************************************************************
*
* FUNCTION: MOD_LAST_CHILD_PTR getModLastChild(rootNode, module)
*
* DESCRIPTION:
*
* INPUTS: 
*
* OUTPUTS:
*
******************************************************************************/

MOD_LAST_CHILD_PTR getModLastChild(rootNode, module)
TASK_TREE_NODE_PTR rootNode;
MODULE_PTR module;
{
  if (!IS_ROOT_NODE(rootNode)) {
    tcaError("Screwed up: getModLastChild");
  } 
    
  return (MOD_LAST_CHILD_PTR)listMemReturnItem(sameModule, (char *)module,
					       (LIST_PTR)(rootNode->lastChild));
}


/******************************************************************************
*
* FUNCTION: void removeModLastChild(rootNode, module)
*
* DESCRIPTION:
*
* INPUTS: 
*
* OUTPUTS:
*
******************************************************************************/

void removeModLastChild(rootNode, module)
TASK_TREE_NODE_PTR rootNode;
MODULE_PTR module;
{
  MOD_LAST_CHILD_PTR modLastChild;

  modLastChild = 
    (MOD_LAST_CHILD_PTR)listMemReturnItem(sameModule, (char *)module,
					  (LIST_PTR)rootNode->lastChild);
  if (modLastChild) {
    listDeleteItem((char *)modLastChild, (LIST_PTR)rootNode->lastChild);
    tcaFree((char *)modLastChild);
  }
}


/******************************************************************************
*
* FUNCTION: void setLastChild(parentNode, childNode, childModule)
*
* DESCRIPTION:
*
* INPUTS: 
*
* OUTPUTS:
*
******************************************************************************/

void setLastChild(parentNode, childNode, childModule)
TASK_TREE_NODE_PTR parentNode, childNode;
MODULE_PTR childModule;
{
  MOD_LAST_CHILD_PTR modLastChild;
 
  if (!IS_ROOT_NODE(parentNode)) {
    parentNode->lastChild = childNode;
  } else {
    /* root node has set of last children, one for each module */
    modLastChild = getModLastChild(parentNode, childModule);
    if (!modLastChild) {
      modLastChild = NEW(MOD_LAST_CHILD_TYPE);
      modLastChild->module = childModule;
      listInsertItem((char *)modLastChild, (LIST_PTR)parentNode->lastChild);
    }
    modLastChild->lastChild = childNode;
  }
}

int setLastChildIterate (parent, child)
TASK_TREE_NODE_PTR parent, child;
{
  setLastChild(parent, (TASK_TREE_NODE_PTR)NULL, child->dispatch->org);
  return TRUE;
}

/******************************************************************************
*
* FUNCTION: int Update_Last_Child(childNode)
*
* DESCRIPTION:
* If the childNode is goal, command, or point-monitor, it becomes
* the new "lastChild" of its parent.
* Returns TRUE if an update occurred.
*
* The root node (tcaTaskTreeRootGlobal), has multiple last
* children, one for each module.
*
* INPUTS: 
*
* OUTPUTS:
*
* NOTES: 25-Jul-91: Reid: Added the check for "DEAD_NODE" -- one process could
*        kill a node that another has created a reference to.
******************************************************************************/

int Update_Last_Child(childNode)
TASK_TREE_NODE_PTR childNode;
{ 
  TASK_TREE_NODE_PTR parentNode;

  parentNode = childNode->parent;
  if (parentNode && !DEAD_NODE(childNode)) {
    if (Sequencing_Class(TASK_TREE_NODE_CLASS(childNode))) {
      setLastChild(parentNode, childNode, childNode->dispatch->org);
      return TRUE;
    }
  }
  return FALSE;
}


/******************************************************************************
*
* FUNCTION: TASK_TREE_NODE_PTR getLastChild(treeNode, dispatch)
*
* DESCRIPTION:
* Returns the last goal, command, or point-monitor message added
* to the tree_node.
*
* The root node (tcaTaskTreeRootGlobal), has multiple last
* children, one for each module.  Returns the last child that
* matches the module of the dispatch's handler.
*
* INPUTS: 
*
* OUTPUTS:
*
******************************************************************************/

TASK_TREE_NODE_PTR getLastChild(treeNode, dispatch)
TASK_TREE_NODE_PTR treeNode;
DISPATCH_PTR dispatch;
{
  MOD_LAST_CHILD_PTR modLastChild;

  if (IS_ROOT_NODE(treeNode)) {
    modLastChild = getModLastChild(treeNode, dispatch->org);
    if (modLastChild) {
      return modLastChild->lastChild;
    } else {
      return NULL;
    }
  } else {
    return treeNode->lastChild;
  }
}


/******************************************************************************
*
* FUNCTION: TASK_TREE_NODE_PTR getFirstChildAfter(treeNode, childrenList)
*
* DESCRIPTION:
* Return the "first" child in the childrenList after the treeNode.
* This is actually fairly ambiguous, so we return one of the (possibly
* many) earliest nodes on the list whose achievement start time is
* greater than the end of the achievement time of the treeNode.
* If "treeNode" is NULL, the child with the earliest achievement
* time is returned.
*
* INPUTS: 
*
* OUTPUTS:
*
******************************************************************************/

TASK_TREE_NODE_PTR getFirstChildAfter(treeNode, childrenList)
TASK_TREE_NODE_PTR treeNode;
LIST_PTR childrenList;
{ 
  TASK_TREE_NODE_PTR child, nextChild = NULL;
  struct _QUANTITY_TYPE *treeNodeEnd, *nextChildStart=NULL, *childStart;

  if (treeNode) treeNodeEnd = treeNode->achievementInterval->end;
  child = (TASK_TREE_NODE_PTR)listFirst(childrenList);
  while (child) {
    if (child != treeNode && taskTreeClass(TASK_TREE_NODE_CLASS(child))) {
      childStart = child->achievementInterval->start;
      if ((!treeNode || QIs_True(childStart, ">=", treeNodeEnd)) &&
	  (!nextChild || QIs_True(childStart, "<=", nextChildStart))) {
	nextChild = child;
	nextChildStart = childStart;
      }
    }
    child = (TASK_TREE_NODE_PTR)listNext(childrenList);
  }
  return nextChild;
}

/******************************************************************************
*
* FUNCTION: TASK_TREE_NODE_PTR getLastChildBefore(treeNode, childrenList)
*
* DESCRIPTION:
* Return the "last" child in the childrenList before the treeNode.
* This is actually fairly ambiguous, so we return one of the (possibly
* many) latest nodes in the list whose achievement end time is less than
* the start of the achievement time of the treeNode.
* If "treeNode" is NULL, the node with the latest achievement time is
* returned.
*
* INPUTS: 
*
* OUTPUTS:
*
******************************************************************************/

TASK_TREE_NODE_PTR getLastChildBefore(treeNode, childrenList)
TASK_TREE_NODE_PTR treeNode;
LIST_PTR childrenList;
{ 
  TASK_TREE_NODE_PTR child, previousChild = NULL;
  struct _QUANTITY_TYPE *treeNodeStart, *previousChildEnd=NULL, *childEnd;

  if (treeNode) treeNodeStart = treeNode->achievementInterval->start;
  child = (TASK_TREE_NODE_PTR)listFirst(childrenList);
  while (child) {
    if (child != treeNode && taskTreeClass(TASK_TREE_NODE_CLASS(child))) {
      childEnd = child->achievementInterval->end;
      if ((!treeNode || QIs_True(childEnd, "<=", treeNodeStart)) &&
	  (!previousChild || QIs_True(childEnd, ">=", previousChildEnd))) {
	previousChild = child;
	previousChildEnd = childEnd;
      }
    }
    child = (TASK_TREE_NODE_PTR)listNext(childrenList);
  }
  return previousChild;
}

/******************************************************************************
*
* FUNCTION: Remove_TaskTree_Rules(TreeNode)
*
* DESCRIPTION:
*
* INPUTS: 
*
* OUTPUTS: Returns TRUE (needed by listIterate)
*
* NOTES: Changed (3-Jul-91) to free *all* rules associated with the
*        quantities to be killed in order to remove spurious firings.
*
******************************************************************************/

int removeTaskTreeRules1 (dummy, treeNode)
void* dummy;
TASK_TREE_NODE_PTR treeNode;
{ 
  DISPATCH_PTR dispatch;

  if (!DEAD_NODE(treeNode)) {

    dispatch = treeNode->dispatch;

    /* 16-Nov-90: fedor: there could still be problems with this! see mon.c */
    cancelIfAttendingIntervalMonitor(dispatch);

    removeQuantityRules(treeNode->handlingInterval->start);
    removeQuantityRules(treeNode->handlingInterval->end);
    if (treeNode->achievementInterval != treeNode->handlingInterval) {
      removeQuantityRules(treeNode->achievementInterval->start);
      removeQuantityRules(treeNode->achievementInterval->end);
    }
    if (treeNode->planningInterval != treeNode->handlingInterval) {
      removeQuantityRules(treeNode->planningInterval->start);
      removeQuantityRules(treeNode->planningInterval->end);
    }

    (void)listIterate(removeTaskTreeRules1, (char *)NULL, treeNode->children);
  }
  return TRUE;
}

Remove_TaskTree_Rules(treeNode)
TASK_TREE_NODE_PTR treeNode;
{
  (void)removeTaskTreeRules1((void *)NULL, treeNode);
}

/******************************************************************************
*
* FUNCTION: void resetLastChild(parentNode, currentChildNode)
*
* DESCRIPTION: If this is the "last" child node of the parent, need to
*    reset in order for future sequencing constraints to work correctly.
*    Some hair because the root node has multiple "last" children, one for
*    each module.
*
* INPUTS: 
*
* OUTPUTS:
*
******************************************************************************/

void resetLastChild(parentNode, childNode)
TASK_TREE_NODE_PTR parentNode, childNode;
{      
  TASK_TREE_NODE_PTR nextChild, lastChild;
  MODULE_PTR module;
  struct _QUANTITY_TYPE *lastChildEnd=NULL, *nextChildEnd;

  if (getLastChild(parentNode, childNode->dispatch) == childNode) {
    module = childNode->dispatch->org;
    if (IS_ROOT_NODE(parentNode)) {
      /* Get the last node before the "childNode",
	 but make sure it's from the same module */
      lastChild = NULL;
      nextChild = (TASK_TREE_NODE_PTR)listFirst(parentNode->children);
      while (nextChild) {
	if (nextChild != childNode && nextChild->dispatch->org == module &&
	    taskTreeClass(TASK_TREE_NODE_CLASS(nextChild))) {
	  nextChildEnd = nextChild->achievementInterval->end;
	  if (!lastChild || QIs_True(nextChildEnd, ">=", lastChildEnd)) {
	    lastChild = nextChild;
	    lastChildEnd = nextChildEnd;
	  }
	}
	nextChild = (TASK_TREE_NODE_PTR)listNext(parentNode->children);
      }
    } else {
      lastChild = getLastChildBefore(childNode, parentNode->children);
    }
    setLastChild(parentNode, lastChild, module);
  }
}

/******************************************************************************
*
* FUNCTION: void try_to_delete_node(TreeNode, firstCall)
*
* DESCRIPTION:
* Free the children nodes if they are all marked "KilledNode" and the TreeNode
* is marked "ToBeKilledNode".
* If children are freed, try to delete the parent of the TreeNode,
* and so on up the tree.
*
* INPUTS: 
*
* OUTPUTS:
*
******************************************************************************/

void try_to_delete_node(TreeNode, firstCall)
TASK_TREE_NODE_PTR TreeNode;
int firstCall;
{ 
  TASK_TREE_NODE_PTR child, parentNode;

  if (TreeNode->status == ToBeKilledNode) {
    child = (TASK_TREE_NODE_PTR)listFirst(TreeNode->children);
    while (child) {
      if (child->status != KilledNode) {
	if (firstCall)
	  Log_Status1("Will kill %s when submessages complete\n", 
		      TASK_TREE_NODE_MSG_NAME(TreeNode));
	return;
      }
      child = (TASK_TREE_NODE_PTR)listNext(TreeNode->children);
    }

    /* All children are marked "Killed" */
    if (TreeNode->dispatch->refCount <= 0) {
      (void)listIterate(freeTaskTreeNode1, (char *)NULL, TreeNode->children);
      TreeNode->status = KilledNode;
      Log_Status1("Killed %s\n", TASK_TREE_NODE_MSG_NAME(TreeNode));
      parentNode = TreeNode->parent;
      if (parentNode) {
	if (parentNode->status != ToBeKilledNode) {
	  /* Top of kill tree */
	  FreeTaskTreeNode(TreeNode);
	} else if (!parentNode->dispatch ||
		   parentNode->dispatch->status != AttendingDispatch) {
	  try_to_delete_node(parentNode, FALSE);
	}
      }
    } else if (firstCall) {
      Log_Status1("Will kill %s when all references to it are released\n", 
		  TASK_TREE_NODE_MSG_NAME(TreeNode));
    }
  }
}

/******************************************************************************
*
* FUNCTION: 
*
* DESCRIPTION:
* If the message has an abort handler registered, call the abort
* handler, and assume that the node is killed.
* 
* Abort handler must return TRUE (1) if the abort was successful.
*
* CURRENTLY ONLY WORKS FOR INTERNAL GOALS AND COMMANDS -- EXTERNAL
* ABORTS ARE HARDER BECAUSE OF TIMING PROBLEMS.
*
* INPUTS: 
*
* OUTPUTS:
*
******************************************************************************/

int tryToAbort(treeNode)
TASK_TREE_NODE_PTR treeNode;
{ 
  ABORT_MSG_PTR abortHnd;
  DISPATCH_PTR abortDispatch;

  abortDispatch = treeNode->dispatch;
  abortHnd = (ABORT_MSG_PTR)hashTableFind(abortDispatch->msg->msgData->name,
					  tcaAbortTableGlobal);
  if (abortHnd) {
    if ((abortHnd->hndProc)(abortDispatch)) {
      CompletionConstraints(abortDispatch);
      try_to_delete_node(treeNode, TRUE);
      return TRUE;
    } else {
      return FALSE;
    }
  } else {
    return FALSE;
  }
}


/******************************************************************************
*
* FUNCTION: ABORT_MSG_PTR addAbortHandler(name, hndProc)
*
* DESCRIPTION:
*
* INPUTS: 
*
* OUTPUTS:
*
******************************************************************************/

ABORT_MSG_PTR addAbortHandler(name, hndProc)
char *name;
int (*hndProc)();
{ 
  ABORT_MSG_PTR abortItem;

  abortItem = (ABORT_MSG_PTR)hashTableFind(name, tcaAbortTableGlobal);
  if (!abortItem) {
    abortItem = NEW(ABORT_MSG_TYPE);
    abortItem->name = name;
    abortItem->hndProc = hndProc;
    abortItem->hndOrg = tcaServerModGlobal;
    (void)hashTableInsert(name, strlen(name)+1, (char *)abortItem,
			  tcaAbortTableGlobal);
  } else {
    abortItem->hndProc = hndProc;
  }
}


/******************************************************************************
*
* FUNCTION: void deleteTaskTree(TreeNode)
*
* DESCRIPTION:
*
* INPUTS: 
*
* OUTPUTS:
*
******************************************************************************/

int deleteTaskTree1 (dummy, TreeNode)
void* dummy;
TASK_TREE_NODE_PTR TreeNode;
{ 
  DISPATCH_PTR dispatch;   
  DISPATCH_STATUS_TYPE CurrentStatus;
  
  if (!DEAD_NODE(TreeNode)) {

    dispatch = TreeNode->dispatch;   
    CurrentStatus = ((dispatch) ? dispatch->status : HandledDispatch);

    switch (CurrentStatus) {
    case AttendingDispatch:
    case ReplyDispatch:
    case HandledDispatch:
    case FailureDispatch:
      break;
    case InactiveDispatch:
      RetractHoldingConstraint(TreeNode);
      break;
    case PendingDispatch:
      RetractHoldingConstraint(TreeNode);
      resourceRemovePending(dispatch);
      break;
    case AllocatedDispatch:
      RetractHoldingConstraint(TreeNode);
      break;
    default:
      TCA_ERROR1("ERROR: Cannot handle killing node with status %d", 
		 CurrentStatus);
    }

    TreeNode->status = ToBeKilledNode;
    (void)listIterate(deleteTaskTree1, (char *)NULL, TreeNode->children);
    if (CurrentStatus != AttendingDispatch) {
      try_to_delete_node(TreeNode, TRUE);
    } else if (!tryToAbort(TreeNode)) {
      Log_Status1("Will kill %s when it completes\n",
		  TASK_TREE_NODE_MSG_NAME(TreeNode));
    }
  }
  return TRUE; /* Needed by listIterate */
}

void deleteTaskTree(TreeNode)
TASK_TREE_NODE_PTR TreeNode;
{
  (void)deleteTaskTree1((void *)NULL, TreeNode);
}

/******************************************************************************
*
* FUNCTION: void checkForQueuedRules(msg, causeError)
*
* DESCRIPTION:
* Try to trace down a mysterious "intermittent" bug in killing
* task trees.  Symptom is the bug "trying to run unknown rule type".
* Probably indicates that some rule is being run after it has been
* freed.  Perhaps this will show us where and why.
*
* INPUTS: 
*
* OUTPUTS:
*
******************************************************************************/
#if 0
void checkForQueuedRules(msg, causeError)
char *msg;
int causeError;
{
  RULE_PTR rule;
  int index, foundQueuedRule;

  foundQueuedRule = FALSE;
  for (index = 0; index < numRuleQueuesGlobal; index++)
    while (rule = (RULE_PTR)Dequeue(tmsPriorityQueuesGlobal[index])) {
      fprintf(stderr, "Rule Queued: %s at %s\n            ", rule->name, msg);
      switch (rule->class) {
      case InRule:
	sayNodeValue((TMS_NODE_PTR)listFirst(rule->innedNodes));
	fprintf(stderr, "\n");
	break;
      case OutRule:
	sayNodeValue((TMS_NODE_PTR)listFirst(rule->outtedNodes));
	fprintf(stderr, "\n");
      }

      foundQueuedRule = TRUE;
    }
  if (causeError && foundQueuedRule)
    tcaError("Rule(s) queued unexpectedly");
}
#endif

/******************************************************************************
*
* FUNCTION: void killTaskTreeHnd(dispatch, KillRefId)
*
* DESCRIPTION:
* Stop planning and execution of the task tree node and all its sub-nodes.
*
* The status "KilledNode" indicates that all the children of the node have
* been killed.
* The status "ToBeKilledNode" indicates that at least one child is not 
* yet "Killed".
*
* Algorithm has three phases:
*  1. Remove all activation/inactivation rules from the nodes to be killed.
*  2. Mark all the nodes as to be killed, remove from the pending/inactive
*     sets, and retract any necessary temporal assertions.
*  3. Try to delete all nodes whose children are already deleted (i.e.,
*     marked "Killed").
*
* INPUTS: 
*
* OUTPUTS:
*
******************************************************************************/

/* 3-July-91: Reid: also called by wire-tapping code */
void killTaskTree(killRefId)
int killRefId;
{
  TASK_TREE_NODE_PTR killNode, parentNode;

  killNode = taskTreeNodeFromRef(killRefId);
  if (killNode && !DEAD_NODE(killNode)) {
    if (IS_ROOT_NODE(killNode)) {
      tcaError("Cannot kill the root node");
    } else {
      (void)runRules(-1);
      /* checkForQueuedRules("Pre-removal", TRUE); */
      Remove_TaskTree_Rules(killNode);
      /* checkForQueuedRules("Post-removal", TRUE); */

      /* 26-Jul-91: Reid: Remove node from children list of parent */
      parentNode = killNode->parent;
      if (parentNode) {
	resetLastChild(parentNode, killNode);
	listDeleteItem((char *)killNode, parentNode->children);
      }

     /* Delay running rules until all retractions are done.  Efficiency hack */
      DELAY_RULES(deleteTaskTree(killNode));

      /* checkForQueuedRules("Post-kill", TRUE); */
    }
  }
}

void killTaskTreeHnd(dispatch, KillRefIdPtr)
DISPATCH_PTR dispatch;
int *KillRefIdPtr;
{ 
  killTaskTree(*KillRefIdPtr);

  tcaFree((char *)KillRefIdPtr); /* Use simple free: only 1 int */
}


/******************************************************************************
*
* FUNCTION: void killSubTaskTreeHnd(dispatch, KillRefId)
*
* DESCRIPTION:
* Like "KillTaskTreeHnd", except does not kill the task tree node itself.
*
* INPUTS: 
*
* OUTPUTS:
*
******************************************************************************/

/* 20-May-91: fedor: also called by exception handling code */
void killSubTaskTree(node)
TASK_TREE_NODE_PTR node;
{
  if (!DEAD_NODE(node)) {
    /* Delay running rules until all retractions are done.  Efficiency hack */
    (void)runRules(-1);
    /* checkForQueuedRules("Pre-removal", TRUE); */
    (void)listIterate(removeTaskTreeRules1, (char *)NULL, node->children);
    /* checkForQueuedRules("Post-removal", TRUE); */

    /* 26-Jul-91: Reid: Remove all children nodes */
    if (IS_ROOT_NODE(node)) {
      (void)listIterate(setLastChildIterate, (char *)node, node->children);
    } else {
      setLastChild(node, (TASK_TREE_NODE_PTR)NULL, (MODULE_PTR)NULL);
    }

    DELAY_RULES( (void)listIterate(deleteTaskTree1, (char *)NULL, node->children) );

    listFree(node->children);
    node->children = NULL;

    /* checkForQueuedRules("Post-kill", TRUE); */
  }
}

void killSubTaskTreeHnd(dispatch, KillRefId)
DISPATCH_PTR dispatch;
int *KillRefId;
{ 
  TASK_TREE_NODE_PTR Kill_TreeNode;

  Kill_TreeNode = taskTreeNodeFromRef(*KillRefId);
  if (Kill_TreeNode)
    killSubTaskTree(Kill_TreeNode);

  tcaFree((char *)KillRefId); /* Use simple free: only 1 int */
}


/******************************************************************************
*
* FUNCTION: int isToBeKilled(dispatch)
*
* DESCRIPTION:
* Return TRUE (1) if the status of the task tree node associated with
* the dispatch is "ToBeKilled".
*
* INPUTS: 
*
* OUTPUTS:
*
******************************************************************************/

int isToBeKilled(dispatch)
DISPATCH_PTR dispatch;
{
  return (dispatch->treeNode && dispatch->treeNode->status == ToBeKilledNode);
}


/******************************************************************************
*
* FUNCTION: void HandleKillAfterAttendingNodes(dispatch)
*
* DESCRIPTION:
* If the status of the task tree node associated with the dispatch is
* "ToBeKilled", then try to kill the task tree node and all its subnodes.
*
* INPUTS: 
*
* OUTPUTS:
*
******************************************************************************/

void HandleKillAfterAttendingNodes(dispatch)
DISPATCH_PTR dispatch;
{ 
  if (isToBeKilled(dispatch) &&
      (dispatch->refCount <= 0) && (dispatch->status != AttendingDispatch)) {
    try_to_delete_node(dispatch->treeNode, FALSE);
  }
}


/******************************************************************************
*
* FUNCTION: int NeedToKillMessage(dispatch)
*
* DESCRIPTION:
* Return TRUE (1) if the message is non-blocking and the parent is
* to be killed after it finishes attending.
*
* INPUTS: 
*
* OUTPUTS:
*
******************************************************************************/

int NeedToKillMessage(dispatch)
DISPATCH_PTR dispatch;
{ 
  DISPATCH_PTR parent;

  if (!ONE_WAY_MSG(dispatch->msg) && !TWO_WAY_MSG(dispatch->msg)) {
    /* non-blocking */
    parent = findParentDispatch(dispatch);

    return(parent && isToBeKilled(parent));
  } else {
    return FALSE;
  }
}


/******************************************************************************
*
* FUNCTION: void referenceDataHnd(dispatch, msgRef)
*
* DESCRIPTION:
* Send the data that is associated with the reference.
*
* INPUTS: 
*
* OUTPUTS:
*
******************************************************************************/

void referenceDataHnd(dispatch, msgRef)
DISPATCH_PTR dispatch;
int *msgRef;
{
  MSG_PTR msg;
  DISPATCH_PTR refDispatch;
 
  refDispatch = DISPATCH_FROM_ID(*msgRef);

  msg = dispatch->msg;

  msg->msgData->resForm = refDispatch->msg->msgData->resForm;

  dispatchSetResData(DISPATCH_MSG_DATA(refDispatch), dispatch);

  recvMessage(dispatch, ReplyClass, (char *)NULL);

  /* 9-Jul-91: Reid: Set the dispatch to ReplyClass so it will be freed */
  dispatch->class = ReplyClass;
  
  msg->msgData->resForm = NULL;

  tcaFree((char *)msgRef); /* Use simple free: only 1 int */
}

/******************************************************************************
*
* FUNCTION: void referenceReleaseHnd(dispatch, msgRef)
*
* DESCRIPTION:
* Release one reference to the associated dispatch.
*
* INPUTS: 
*
* OUTPUTS:
*
******************************************************************************/
void referenceReleaseHnd(dispatch, msgRef)
DISPATCH_PTR dispatch;
int *msgRef;
{
  DISPATCH_PTR refDispatch;
 
  if (*msgRef != NO_REF) {
    refDispatch = DISPATCH_FROM_ID(*msgRef);
    releaseDispatch(refDispatch);
  }
}

/******************************************************************************
*
* FUNCTION: DISPATCH_PTR CreateReferenceDispatch(msgRef, parentId, dispatchOrg)
*
* DESCRIPTION:
* Add a new reference to message msgRef to the parent node 
* ParentId, where the dispatch originated at DispatchOrg.
* Return the newly created dispatch.
*
* INPUTS: 
*
* OUTPUTS:
*
******************************************************************************/

DISPATCH_PTR CreateReferenceDispatch(msgRef, parentId, dispatchOrg)
int msgRef, parentId; 
MODULE_PTR dispatchOrg;
{ 
  DISPATCH_PTR refDispatch;

  refDispatch = dispatchAllocate();

  DISPATCH_SET_STATUS(AllocatedDispatch, refDispatch);

  refDispatch->msg = (MSG_PTR)idTableItem(msgRef, msgIdTable);
  /* "DispatchOrg" needed by CreateTaskNode, within Initialize_TaskTreeNode */
  refDispatch->org = dispatchOrg;

  switch (refDispatch->msg->msgData->class) {
  case GoalClass:
  case CommandClass:
  case PointMonitorClass:
  case PollingMonitorClass:
  case DemonMonitorClass:
    Initialize_TaskTreeNode(refDispatch, parentId);
  }
  return refDispatch;
}


/******************************************************************************
*
* FUNCTION: int retractEndOfConstraintsForNewChild
*
* DESCRIPTION: Adding a new child to a goal node might nullify the
*              "end of interval" completion constraints.  Need to retract
*              these constraints for *all* ancestor goal nodes.
*
* INPUTS: TASK_TREE_NODE_PTR childNode
*
* OUTPUTS: Returns TRUE if any of the relations were retracted.
*
******************************************************************************/
int retractEndOfConstraintsForNewChild (childNode)
TASK_TREE_NODE_PTR childNode;
{
  TASK_TREE_NODE_PTR parentNode;
  int retracted = FALSE;

  parentNode = childNode->parent;
  if (parentNode->dispatch && (TASK_TREE_NODE_CLASS(parentNode) == GoalClass)) {
    if (QIs_True(Now, ">", parentNode->achievementInterval->end)) {
      QRetract(Now, ">", parentNode->achievementInterval->end, "Completed Message");
      retracted = TRUE;
    }
    if ((TASK_TREE_NODE_CLASS(childNode) == GoalClass) &&
	QIs_True(Now, ">", parentNode->planningInterval->end)) {
      QRetract(Now, ">", parentNode->planningInterval->end, "Completed Message");
      retracted = TRUE;
    }

    if (retracted && !IS_ROOT_NODE(parentNode->parent)) {
      (void)retractEndOfConstraintsForNewChild(parentNode);
    }
  }

  return retracted;
}

/******************************************************************************
*
* FUNCTION: int retractEndOfConstraints
*
* DESCRIPTION: When certain task tree changes occur (such as retrying a
*              message), the assumptions that are needed to fire the
*              "end of interval" rules get violated.  Need to retract
*              these constraints for the node and all ancestor nodes.
*
* INPUTS: TASK_TREE_NODE_PTR node
*
* OUTPUTS: Returns TRUE if any of the relations were retracted.
*
******************************************************************************/
int retractEndOfConstraints (node)
     TASK_TREE_NODE_PTR node;
{
  TASK_TREE_NODE_PTR parentNode;
  int retracted = FALSE;
  
  if (QIs_True(Now, ">", node->achievementInterval->end)) {
    QRetract(Now, ">", node->achievementInterval->end, "Completed Message");
    retracted = TRUE;
  }
  if ((TASK_TREE_NODE_CLASS(node) == GoalClass) &&
      QIs_True(Now, ">", node->planningInterval->end)) {
    QRetract(Now, ">", node->planningInterval->end, "Completed Message");
    retracted = TRUE;
  }
  
  parentNode = node->parent;
  if (retracted && !IS_ROOT_NODE(parentNode)) {
    (void)retractEndOfConstraints(parentNode);
  }

  return retracted;
}

/******************************************************************************
*
* FUNCTION: void restoreEndOfRelations
*
* DESCRIPTION: After retracting the "end of" completion constraints, need to
*              re-establish (i.e., cache) the relationship between "Now" and
*              the end of the appropriate intervals, so that the "end of"
*              rules will fire when needed.
*
* INPUTS: TASK_TREE_NODE_PTR node
*
* OUTPUTS: none
*
******************************************************************************/
void restoreEndOfRelations (node)
TASK_TREE_NODE_PTR node;
{
  TASK_TREE_NODE_PTR parentNode;

  (void)QRelationship(Now, node->planningInterval->end);
  (void)QRelationship(Now, node->achievementInterval->end);

  parentNode = node->parent;
  if (!IS_ROOT_NODE(parentNode)) {
    restoreEndOfRelations(parentNode);
  }
}

/******************************************************************************
*
* FUNCTION: void AddChildRefHnd(dispatch, AddChild)
*
* DESCRIPTION:
*
* INPUTS: 
*
* OUTPUTS:
*
******************************************************************************/

void addChildRefHnd(dispatch, addChild)
DISPATCH_PTR dispatch;
ADD_CHILD_PTR addChild;
{ 
  DISPATCH_PTR refDispatch;
  int retracted;

  refDispatch = CreateReferenceDispatch(addChild->msgRef, 
					addChild->parentRefId, dispatch->org);
  reserveDispatch(refDispatch);

  /* 23-Jul-91: Reid: Nullify the "end of interval" completion constraints */
  retracted = retractEndOfConstraintsForNewChild(refDispatch->treeNode);

  PendingConstraints(refDispatch);

  /* 23-Jul-91: Reid: Now need to re-establish the "end of interval" relations */
  if (retracted) restoreEndOfRelations(refDispatch->treeNode->parent);

  tcaFree((char *)addChild); /* Use simple free: only 2 ints */

  centralReply(dispatch, (char *)&(refDispatch->locId));
}


/******************************************************************************
*
* FUNCTION: void createRefHnd(dispatch, msgRef)
*
* DESCRIPTION:
*
* INPUTS: 
*
* OUTPUTS:
*
******************************************************************************/

void createRefHnd(dispatch, msgRef)
DISPATCH_PTR dispatch;
int *msgRef;
{ 
  DISPATCH_PTR refDispatch;

  refDispatch = CreateReferenceDispatch(*msgRef, dispatch->pRef,dispatch->org);
  reserveDispatch(refDispatch);

  PendingConstraints(refDispatch);

  tcaFree((char *)msgRef);
  centralReply(dispatch, (char *)&(refDispatch->locId));
}


/******************************************************************************
*
* FUNCTION: void findParentRefHnd(dispatch, childRefId)
*
* DESCRIPTION:
*
* INPUTS: 
*
* OUTPUTS:
*
******************************************************************************/

void findParentRefHnd(dispatch, childRefId)
DISPATCH_PTR dispatch;
int *childRefId;
{ 
  TCA_REF_PTR ref = NULL;
  DISPATCH_PTR parentDispatch;
  TASK_TREE_NODE_PTR childNode;

  childNode = taskTreeNodeFromRef(*childRefId);
  if (childNode) {
    if (IS_ROOT_NODE(childNode->parent)) {
      ref = CREATE_NULL_REF();
    } else {
      parentDispatch = findParentDispatch(childNode->dispatch);
      if (parentDispatch) {
	ref = tcaRefCreate((MSG_PTR)NULL, parentDispatch->msg->msgData->name,
			   parentDispatch->locId);
	reserveDispatch(parentDispatch);
      }
    }
  }

  tcaFree((char *)childRefId); /* Use simple free: only 1 int */

  if (ref) {
    centralReply(dispatch, (char *)&ref);
    tcaRefFree(ref);
  } else {
    centralReply(dispatch, (char *)NULL);
  }
}

/******************************************************************************
*
* FUNCTION: void findChildReply(dispatch, child)
*
* DESCRIPTION: Generic reply routine for the various task-tree traversal
*              message handlers.
*
* INPUTS: 
*
* OUTPUTS:
*
******************************************************************************/
void findChildReply (dispatch, child)
DISPATCH_PTR dispatch;
TASK_TREE_NODE_PTR child;
{ 
  TCA_REF_PTR ref = NULL;

  if (child) {
    ref = tcaRefCreate((MSG_PTR)NULL, TASK_TREE_NODE_MSG_NAME(child),
		       child->dispatch->locId);
    reserveDispatch(child->dispatch);
  }

  if (ref) {
    centralReply(dispatch, (char *)&ref);
    tcaRefFree(ref);
  } else {
    centralReply(dispatch, (char *)NULL);
  }
}

/******************************************************************************
*
* FUNCTION: void findFirstChildHnd(dispatch, parentRefId)
*
* DESCRIPTION:
*
* INPUTS: 
*
* OUTPUTS:
*
******************************************************************************/

void findFirstChildHnd(dispatch, parentRefId)
DISPATCH_PTR dispatch;
int *parentRefId;
{ 
  TASK_TREE_NODE_PTR parentNode;

  parentNode = taskTreeNodeFromRef(*parentRefId);
  findChildReply(dispatch, (!parentNode ? (TASK_TREE_NODE_PTR)NULL : 
			    getFirstChildAfter((TASK_TREE_NODE_PTR)NULL,
					       parentNode->children)));

  tcaFree((char *)parentRefId); /* Use simple free: only 1 int */
}

/******************************************************************************
*
* FUNCTION: void findLastChildHnd(dispatch, parentRefId)
*
* DESCRIPTION:
*
* INPUTS: 
*
* OUTPUTS:
*
******************************************************************************/

void findLastChildHnd(dispatch, parentRefId)
DISPATCH_PTR dispatch;
int *parentRefId;
{ 
  TASK_TREE_NODE_PTR parentNode;

  parentNode = taskTreeNodeFromRef(*parentRefId);
  findChildReply(dispatch, (!parentNode ? (TASK_TREE_NODE_PTR)NULL :
			    getLastChildBefore((TASK_TREE_NODE_PTR)NULL, 
					       parentNode->children)));

  tcaFree((char *)parentRefId); /* Use simple free: only 1 int */
}

/******************************************************************************
*
* FUNCTION: void findNextChildHnd(dispatch, childRefId)
*
* DESCRIPTION:
*
* INPUTS: 
*
* OUTPUTS:
*
******************************************************************************/

void findNextChildHnd(dispatch, childRefId)
DISPATCH_PTR dispatch;
int *childRefId;
{ 
  TASK_TREE_NODE_PTR childNode;

  childNode = taskTreeNodeFromRef(*childRefId);
  findChildReply(dispatch, (!(childNode && childNode->parent) ? NULL :
			    getFirstChildAfter(childNode, 
					       childNode->parent->children)));

  tcaFree((char *)childRefId); /* Use simple free: only 1 int */
}

/******************************************************************************
*
* FUNCTION: void findPrevChildHnd(dispatch, childRefId)
*
* DESCRIPTION:
*
* INPUTS: 
*
* OUTPUTS:
*
******************************************************************************/

void findPrevChildHnd(dispatch, childRefId)
DISPATCH_PTR dispatch;
int *childRefId;
{ 
  TASK_TREE_NODE_PTR childNode;

  childNode = taskTreeNodeFromRef(*childRefId);
  findChildReply(dispatch, (!(childNode && childNode->parent) ? NULL :
			    getLastChildBefore(childNode, 
					       childNode->parent->children)));

  tcaFree((char *)childRefId); /* Use simple free: only 1 int */
}

/******************************************************************************
*
* FUNCTION: void findFailedRefHnd(dispatch, failedRefId)
*
* DESCRIPTION:
* Like "FindParentRefHnd", except goes up the tree to find the
* first node that is not of ExceptionClass.
*
* INPUTS: 
*
* OUTPUTS:
*
******************************************************************************/

void findFailedRefHnd(dispatch, failedRefId)
DISPATCH_PTR dispatch;
int *failedRefId;
{ 
  TCA_REF_PTR ref = NULL;
  TASK_TREE_NODE_PTR childNode, parentNode;

  childNode = taskTreeNodeFromRef(*failedRefId);
  
  if (childNode) 
    for(;;) {
      parentNode = childNode->parent;
      if (!parentNode)
	break;

      if (IS_ROOT_NODE(parentNode)) {
	ref = CREATE_NULL_REF();
	break;
      }

      if (TASK_TREE_NODE_CLASS(parentNode) != ExceptionClass) {
	ref = tcaRefCreate((MSG_PTR)NULL, TASK_TREE_NODE_MSG_NAME(parentNode),
			   parentNode->dispatch->locId);
	reserveDispatch(parentNode->dispatch);
	break;
      }

      childNode = parentNode;
    }

  tcaFree((char *)failedRefId); /* Use simple free: only 1 int */

  if (ref) {
    centralReply(dispatch, (char *)&ref);
    tcaRefFree(ref);
  } else {
    centralReply(dispatch, (char *)NULL);
  }
}

/******************************************************************************
*
* FUNCTION: void findTopLevelRefHnd(dispatch, childRefId)
*
* DESCRIPTION:
* Find the top level node (the node whose parent is tcaRootNodeGlobal
* or NULL) of the task tree in which the node of ChildRefId resides.
* In case tcaRootNodeGlobal is given as the child node, NULL
* is used as the answer.
*
* INPUTS: 
*
* OUTPUTS:
*
******************************************************************************/

void findTopLevelRefHnd(dispatch, childRefId)
DISPATCH_PTR dispatch;
int *childRefId;
{
  TASK_TREE_NODE_PTR childNode;
  DISPATCH_PTR parentDispatch, childDispatch = NULL;

  childNode = taskTreeNodeFromRef(*childRefId);
  if (childNode)
    childDispatch = childNode->dispatch;

  parentDispatch = childDispatch;
  while (parentDispatch && (!IS_ROOT_NODE(parentDispatch->treeNode))) {
      childDispatch = parentDispatch;
      parentDispatch = findParentDispatch(childDispatch);
    }
  
  findChildReply(dispatch, (childDispatch ? childDispatch->treeNode : NULL));

  tcaFree((char *)childRefId); /* Use simple free: only 1 int */
}

/******************************************************************************
*
* FUNCTION: void findAnteRefByNameHnd(dispatch, findAnteRef)
*
* DESCRIPTION:
* Find the child's antecedent node with a specified (message) name.
* If there are more than one antecedents with the name, get the one
* closest to the child.
*
* INPUTS: 
*
* OUTPUTS:
*
******************************************************************************/

void findAnteRefByNameHnd(dispatch, findAnteRef)
DISPATCH_PTR dispatch;
FIND_REF_PTR findAnteRef;
{
  char *anteName;
  DISPATCH_PTR childDispatch;
  TASK_TREE_NODE_PTR childNode;

  childNode = taskTreeNodeFromRef(findAnteRef->refId);
  if (childNode)
    childDispatch = childNode->dispatch;
  
  anteName = findAnteRef->name;
  while (childDispatch &&
	 !strKeyEqFunc(childDispatch->msg->msgData->name, anteName)) {
    childDispatch = findParentDispatch(childDispatch);
  }
  
  findChildReply(dispatch, (childDispatch ? childDispatch->treeNode : NULL));

  /* A bit more efficient than using tcaFreeData */
  freeDataStructure(dispatch->msg->msgData->msgForm, (char *)findAnteRef);
}

/******************************************************************************
*
* FUNCTION: void findChildByNameHnd(dispatch, findChildRef)
*
* DESCRIPTION:
* Find the child node of the parent reference which has the given message 
* name.  If more than one such child, returns the "first" one it comes
* across (where "first" is implementation dependent).
*
* INPUTS: 
*
* OUTPUTS:
*
******************************************************************************/

void findChildByNameHnd(dispatch, findChildRef)
DISPATCH_PTR dispatch;
FIND_REF_PTR findChildRef;
{  
  char *childName;
  TCA_REF_PTR ref = NULL;
  TASK_TREE_NODE_PTR parentNode, child;
  DISPATCH_PTR childDispatch;

  childName = findChildRef->name;

  parentNode = taskTreeNodeFromRef(findChildRef->refId);
  if (parentNode) {
    child = (TASK_TREE_NODE_PTR)listFirst(parentNode->children);
    while (child) {
      childDispatch = child->dispatch;
      if (childDispatch && 
	  strKeyEqFunc(childDispatch->msg->msgData->name, childName)) {
	ref = tcaRefCreate((MSG_PTR)NULL, childName, childDispatch->locId);
	reserveDispatch(childDispatch);
	centralReply(dispatch, (char *)&ref);
	freeDataStructure(dispatch->msg->msgData->msgForm, (char *)findChildRef);
	tcaRefFree(ref);
	return; 
      }
      child = (TASK_TREE_NODE_PTR)listNext(parentNode->children);
    }
  }
  /* Send a NULL reply on failure to find a message by that name */
  centralReply(dispatch, (char *)NULL);
  freeDataStructure(dispatch->msg->msgData->msgForm, (char *)findChildRef);
}


/******************************************************************************
*
* FUNCTION: int displayTaskTree(position, treeNode)
*
* DESCRIPTION:
*
* INPUTS: 
*
* OUTPUTS: Returns TRUE (needed by listIterate)
*
******************************************************************************/

int displayTaskTree(position, treeNode)
int position;
TASK_TREE_NODE_PTR treeNode;
{
  static int cursor = 0;
  char *nodeName, *status;

  if (!DEAD_NODE(treeNode)) {

    if (cursor > position) {
      Log("\n");
      cursor = 0;
    }
    pr_space(position - cursor);
    cursor = position;
    if (treeNode->dispatch || IS_ROOT_NODE(treeNode)) {
      if (IS_ROOT_NODE(treeNode)) {
	nodeName = "Root Node";
	status = "";
      } else {
	nodeName = TASK_TREE_NODE_MSG_NAME(treeNode);
	status = nodeStatusNames[(int)treeNode->dispatch->status];
      }

      Log2("(%s%s", status, nodeName);
      cursor += (1 + strlen(nodeName) + strlen(status));

      (void)listIterateFromFirst(displayTaskTree, (char *)(position + TAB),
				 treeNode->children);
      Log(")");
      ++cursor;
    } else {
      Log("(*FREED*)");
      cursor += 9;
    }
  }
  return TRUE;
}


/******************************************************************************
*
* FUNCTION: void displayTaskTreeHnd(dispatch, dispRefId)
*
* DESCRIPTION:
* Display the task tree rooted at DispRefId
*
* For each node, prints out its message name and its status:
*  "I"nactive 
*  "A"ttending
*  "C"ompleted (Handled)
*  "P"ending
*
* INPUTS: 
*
* OUTPUTS:
*
******************************************************************************/

void displayTaskTreeHnd(dispatch, dispRefId)
DISPATCH_PTR dispatch;
int *dispRefId;
{ 
  TASK_TREE_NODE_PTR dispNode;

  dispNode = taskTreeNodeFromRef(*dispRefId);
  if (dispNode) {
    (void)displayTaskTree (0, dispNode);
    Log("\n");
  } else {
    Log("DisplayTaskTreeHnd: Invalid Task Tree Node\n");
  }

  tcaFree((char *)dispRefId); /* Use simple free: only 1 int */
}


/******************************************************************************
*
* FUNCTION: void referenceStatusHnd(dispatch, msgRef)
*
* DESCRIPTION: Send the status of the message associated with the reference.
*              Possible status: UnknownRefStatus, InactiveRef, PendingRef,
*                               ActiveRef, HandledRef, PlannedRef, AchievedRef,
*                               KilledRef
* INPUTS: 
*
* OUTPUTS:
*
******************************************************************************/

void referenceStatusHnd(dispatch, msgRef)
DISPATCH_PTR dispatch;
int *msgRef;
{
  TASK_TREE_NODE_PTR node;
  TCA_REF_STATUS_TYPE status;

  status = UnknownRefStatus;
  node = taskTreeNodeFromRef(*msgRef);
  if (node) {
    switch (node->dispatch->status) {
    case AllocatedDispatch:
    case UnallocatedDispatch:
    case InactiveDispatch: status = InactiveRef; break;

    case AttendingDispatch: status = ActiveRef; break;

    case PendingDispatch: status = PendingRef; break;

    case KilledDispatch: status = KilledRef; break;

    case HandledDispatch: 
    case SuccessDispatch: 
    case FailureDispatch: 
      if (QIs_True(node->achievementInterval->end, "<=", Now))
	status = AchievedRef;
      else if (QIs_True(node->planningInterval->end, "<=", Now))
	status = PlannedRef;
      else
	status = HandledRef;
    }
  }
  centralReply(dispatch, (char *)&status);
}


/******************************************************************************
*
* FUNCTION: void taskTreeInitialize()
*
* DESCRIPTION:
* Initialize the messages and static variables for the task tree
* and temporal constraints.
*
* INPUTS: 
*
* OUTPUTS:
*
******************************************************************************/

void taskTreeInitialize()
{
  Tpl_Initialize();

  tcaAbortTableGlobal = hashTableCreate(10, strHashFunc, strKeyEqFunc);

  tcaTaskTreeRootGlobal = taskTreeNodeCreate((DISPATCH_PTR)NULL, NO_REF);
  tcaTaskTreeRootGlobal->lastChild = (TASK_TREE_NODE_PTR)listCreate();

  nodeStatusNames[(int)AllocatedDispatch] = "X:";
  nodeStatusNames[(int)AttendingDispatch] = "A:";
  nodeStatusNames[(int)PendingDispatch] = "P:";
  nodeStatusNames[(int)InactiveDispatch] = "I:";
  nodeStatusNames[(int)HandledDispatch] = "C:"; /* For "completed" */

  centralRegisterMessage("createRefMsg", QueryClass, "int", "int");
  centralRegisterHandler("createRefMsg", "createRefHnd", createRefHnd);

  centralRegisterMessage("addChildRef", QueryClass, "{int, int}", "int");
  centralRegisterHandler("addChildRef", "addChildRefHnd", addChildRefHnd);

  centralRegisterMessage("findParentRef", QueryClass, "int", "TCA_REF_PTR");
  centralRegisterHandler("findParentRef", "findParentRefHnd", 
			 findParentRefHnd);

  centralRegisterMessage("findFirstChild", QueryClass, "int", "TCA_REF_PTR");
  centralRegisterHandler("findFirstChild", "findFirstChildHnd", 
			 findFirstChildHnd);

  centralRegisterMessage("findLastChild", QueryClass, "int", "TCA_REF_PTR");
  centralRegisterHandler("findLastChild", "findLastChildHnd", 
			 findLastChildHnd);

  centralRegisterMessage("findNextChild", QueryClass, "int", "TCA_REF_PTR");
  centralRegisterHandler("findNextChild", "findNextChildHnd", 
			 findNextChildHnd);

  centralRegisterMessage("findPrevChild", QueryClass, "int", "TCA_REF_PTR");
  centralRegisterHandler("findPrevChild", "findPrevChildHnd", 
			 findPrevChildHnd);

  centralRegisterMessage("findFailedRef", QueryClass, "int", "TCA_REF_PTR");
  centralRegisterHandler("findFailedRef", "findFailedRefHnd",
			 findFailedRefHnd);

  centralRegisterMessage("findTopLevelRef", QueryClass, "int", "TCA_REF_PTR");
  centralRegisterHandler("findTopLevelRef", "findTopLevelRefHnd", 
		     findTopLevelRefHnd);

  centralRegisterMessage("findAnteRefByName", QueryClass, "{int, string}",
			 "TCA_REF_PTR");
  centralRegisterHandler("findAnteRefByName", "findAnteRefByNameHnd", 
			 findAnteRefByNameHnd);

  centralRegisterMessage("FindChildByName", QueryClass, "{int, string}",
			 "TCA_REF_PTR");
  centralRegisterHandler("findChildByName", "findChildByNameHnd", 
			 findChildByNameHnd);

  centralRegisterMessage("killTaskTree", InformClass, "int", (char *)NULL);
  centralRegisterHandler("killTaskTree", "killTaskTreeHnd", killTaskTreeHnd);

  centralRegisterMessage("killSubTaskTree", InformClass, "int", (char *)NULL);
  centralRegisterHandler("killSubTaskTree", "killSubTaskTreeHnd",
			 killSubTaskTreeHnd);
  
  centralRegisterMessage("displayTaskTree", InformClass, "int", (char *)NULL);
  centralRegisterHandler("displayTaskTree", "displayTaskTreeHnd", 
			 displayTaskTreeHnd);

  centralRegisterMessage("referenceData", QueryClass, "int", (char *)NULL);
  centralRegisterHandler("referenceData", "referenceDataHnd", 
			 referenceDataHnd);

  centralRegisterMessage("referenceRelease", InformClass, "int", (char *)NULL);
  centralRegisterHandler("referenceRelease", "referenceReleaseHnd", 
			 referenceReleaseHnd);
  Add_Message_To_Ignore("referenceRelease");

  centralRegisterMessage("referenceStatusMsg", QueryClass, "int", "int");
  centralRegisterHandler("referenceStatusMsg", "referenceStatusHnd", 
			 referenceStatusHnd);
}
