/******************************************************************************
*
* PROJECT: Carnegie Mellon Planetary Rover Project
*          Task Control Architecture 
* 
* (c) Copyright 1991 Christopher Fedor and Reid Simmons.  All rights reserved.
* 
* MODULE: taskTree
*
* FILE: taskTree.h
*
* ABSTRACT:
* 
* Creation and manipulation of hierarchical task trees
* (containing goal, command, and monitor message nodes).
*
* REVISION HISTORY
*
* 25-May-90 Christopher Fedor, School of Computer Science, CMU
* Revised to Software Standards.
*
* 28-Nov-89 Long-Ji Lin, School of Computer Science, CMU
* Added 'ECNode' & 'excHnds' slots to tree node
*
*  2-Oct-89 Reid Simmons, School of Computer Science, CMU
* Added status slot to task tree node.
*
* 19-Sep-89 Reid Simmons, School of Computer Science, CMU
* Added a global task tree root node.
*
* 14-Aug-89 Reid Simmons, School of Computer Science, CMU
* Added routines to kill task trees, get parent and children references.
*
* 11-Aug-89 Reid Simmons, School of Computer Science, CMU
*  Changed terms from "goal tree" to "task tree"
*
* 15-May-89 Christopher Fedor, School of Computer Science, CMU
* Changed interval type not to make use of quantity_ptr typedef to 
* minimize include files needed in modules that do not use qlattice.
*
* 28-Apr-89 Reid Simmons, School of Computer Science, CMU
* First attempt at goal structure. Added slots for handling temporal 
* constraints.
*
******************************************************************************/

/* For displaying task trees */
#define TAB 14
#define pr_space(n) { int i; i=(n); for(; i>0; --i) Log(" "); }

typedef enum {ActiveNode, AchievedNode, KilledNode, ToBeKilledNode, 
	      SuspendedNode, ToBeSuspendedNode} TASK_TREE_STATUS_TYPE;

#define DEAD_NODE(node) \
  (((node)->status == KilledNode) || ((node)->status == ToBeKilledNode))

typedef struct {
   struct _TREE_NODE *ENode, *CNode;
} ECNODE_TYPE, *ECNODE_PTR;

typedef struct _TREE_NODE {
  DISPATCH_PTR dispatch;
  struct _LIST *children;
  struct _TREE_NODE *parent;
  struct _TREE_NODE *lastChild; /* used for doing SEQ_ACH and SEQ_PLANNING */
  struct _QUANTITY_TYPE *priority;
  struct _interval_type *handlingInterval;
  struct _interval_type *achievementInterval;
  struct _interval_type *planningInterval;
  TASK_TREE_STATUS_TYPE status;

  int retries;
  ECNODE_PTR ECNode;
  struct _LIST *excHndList;

} TASK_TREE_NODE_TYPE, *TASK_TREE_NODE_PTR;

typedef struct {
  MODULE_PTR module;
  TASK_TREE_NODE_PTR lastChild;
} MOD_LAST_CHILD_TYPE, *MOD_LAST_CHILD_PTR;

typedef struct {
  int msgRef;
  int parentRefId;
} ADD_CHILD_TYPE, *ADD_CHILD_PTR;

typedef struct {
  int refId;
  char *name;
} FIND_REF_TYPE, *FIND_REF_PTR;

typedef struct { 
  char *name;
  int (*hndProc)();
  MODULE_PTR hndOrg;
} ABORT_MSG_TYPE, *ABORT_MSG_PTR;

#define TASK_TREE_NODE_CLASS(taskTreeNode) \
  (taskTreeNode)->dispatch->msg->msgData->class

#define TASK_TREE_NODE_MSG_NAME(taskTreeNode) \
  (taskTreeNode)->dispatch->msg->msgData->name

#define IS_ROOT_NODE(taskTreeNode) ((taskTreeNode) == tcaTaskTreeRootGlobal)

extern TASK_TREE_NODE_PTR tcaTaskTreeRootGlobal;

extern int taskTreeClass();
extern int Sequencing_Class();
extern int Update_Last_Child();
extern int isToBeKilled();
extern int NeedToKillMessage();

extern DISPATCH_PTR CreateReferenceDispatch();

extern TASK_TREE_NODE_PTR taskTreeNodeCreate();
extern TASK_TREE_NODE_PTR getLastChild();

extern void taskTreeInitialize();
extern void HandleKillAfterAttendingNodes();

extern ABORT_MSG_PTR addAbortHandler();

extern void removeModLastChild();
extern void restoreEndOfRelations();
