/******************************************************************************
*
* PROJECT: Carnegie Mellon Planetary Rover Project
*          Task Control Architecture 
* 
* (c) Copyright 1991 Christopher Fedor and Reid Simmons.  All rights reserved.
* 
* MODULE: taskTree
*
* FILE: taskTreeModule.c
*
* ABSTRACT:
* Modules interface with creation and manipulation of hierarchical trees.
* 
* REVISION HISTORY
*
* 22-Dec-90 Christopher Fedor, School of Computer Science, CMU
* Added tcaQuery NullReply checks to tcaFindParentReference, tcaFindFirstChild,
* tcaFindLastChild, tcaFindNextChild, tcaFindPreviousChild, 
* tcaFindTopLevelReference, tcaFindAnteReferenceByName, tcaFindChildByName
*
*
* 16-Nov-90 Christopher Fedor, School of Computer Science, CMU
* Moved Updated tcaCreateReference back here.
*
* 12-Sep-90 Christopher Fedor, School of Computer Science, CMU
* Updated for new TCA_REF_PTR.
* Revised to Software Standards.
*
* 26-Mar-90 Long-Ji Lin, School of Computer Science, CMU
* Modified code to allow interval monitors to be created by calling 
* tcaAddChildReference() or tcaCreateReference().
*
* 27-Nov-89 Reid Simmons, School of Computer Science, CMU
* Added new routines for tracing task tree (find first and last child, 
* find child by name, get name and data of reference).
*
* 14-Aug-89 Reid Simmons, School of Computer Science, CMU
* created.
*
******************************************************************************/

#ifdef VXWORKS
#include "stdioLib.h"
#else
#include "stdio.h"
#endif
#include "tcaInternal.h"

#include "formatters.h"
#include "hash.h"
#include "idtable.h"

#include "datamsg.h"
#include "dispatch.h"

#include "taskTree.h"

extern TCA_REF_PTR tcaRootNodeGlobal;


/******************************************************************************
*
* FUNCTION: int creationClass(msgClass)
*
* DESCRIPTION: 
* Returns TRUE if the class of message is one that can be created by calling
* tcaCreateReference() or tcaAddChildReference() (currently, goals, commands,
* point-monitors, interval-monitors).
*
* INPUTS: TCA_MSG_CLASS_TYPE msgClass;
*
* OUTPUTS: int
*
******************************************************************************/

int creationClass(msgClass)
TCA_MSG_CLASS_TYPE msgClass;
{
  return ((msgClass == GoalClass) || 
	  (msgClass == CommandClass) ||
	  (msgClass == PointMonitorClass) ||
	  (msgClass == PollingMonitorClass) ||
	  (msgClass == DemonMonitorClass));
}


/******************************************************************************
*
* FUNCTION: TCA_REF_PTR tcaCreateReference(name)
*
* DESCRIPTION: 
* Add a reference as a child of the currently handled message.
* Add to tcaRootNodeGlobal if no such message.
* Returns the created reference.
*
* INPUTS: char *name;
*
* OUTPUTS: TCA_REF_PTR
*
******************************************************************************/

TCA_REF_PTR tcaCreateReference(name)
char *name;
{
  int refId;
  MSG_PTR msg;
  TCA_MSG_CLASS_TYPE class;

  msg = msgFind(name);
  class = msg->msgData->class;

  if (!creationClass(class)) {
    tcaModError("ERROR: tcaCreateReference: Must be a goal, command or monitor.\n");
  }

  (void)tcaQuery("createRefMsg", (void *)&(msg->msgData->refId), (void *)&refId);

  return(tcaRefCreate(msg, name, refId));
}


/******************************************************************************
*
* FUNCTION: int tcaIsSameReference(ref1, ref2)
*
* DESCRIPTION: 
* Returns TRUE if ref1 and ref2 both represent the same task-tree node.
*
* INPUTS: TCA_REF_PTR ref1, ref2;
*
* OUTPUTS: int
*
******************************************************************************/

int tcaIsSameReference(ref1, ref2)
TCA_REF_PTR ref1, ref2;
{
  return ((ref1 == ref2) || 
	  (ref1->refId == ref2->refId && ref1->name == ref2->name) ||
	  (ref1->refId == ref2->refId && STREQ(ref1->name, ref2->name)));
}


/******************************************************************************
*
* FUNCTION: TCA_REF_PTR tcaAddChildReference(parentRef, name)
*
* DESCRIPTION: 
* Add a reference as a child of the parent reference.
* Returns the reference added.
*
* INPUTS:
* TCA_REF_PTR parentRef;
* char *name;
*
* OUTPUTS: TCA_REF_PTR
*
******************************************************************************/

TCA_REF_PTR tcaAddChildReference(parentRef, name)
TCA_REF_PTR parentRef;
char *name;
{
  int refId;
  MSG_PTR msg;
  TCA_MSG_CLASS_TYPE class;
  ADD_CHILD_TYPE addChild;

  msg = msgFind(name);
  class = msg->msgData->class;

  if (!creationClass(class)) {
    tcaModError("ERROR: tcaAddChildReference: Must be a goal, command or monitor.\n");
  }

  addChild.msgRef = msg->msgData->refId;
  addChild.parentRefId = parentRef->refId;

  (void)tcaQuery("addChildRef", (void *)&addChild, (void *)&refId);
    
  return(tcaRefCreate(msg, name, refId));
}


/******************************************************************************
*
* FUNCTION: TCA_REF_PTR tcaFindParentReference(childRef)
*
* DESCRIPTION: Return the parent reference of the child.
*
* INPUTS: TCA_REF_PTR childRef
*
* OUTPUTS: TCA_REF_PTR
*
* The returned value may be NULL.
*
******************************************************************************/

TCA_REF_PTR tcaFindParentReference(childRef)
TCA_REF_PTR childRef;
{ 
  TCA_REF_PTR ref;

  if (tcaQuery("FindParentRef", (void *)&(childRef->refId), (void *)&ref) 
      == NullReply)
    return NULL;
  else
    return ref;
}


/******************************************************************************
*
* FUNCTION: TCA_REF_PTR tcaFindFirstChild(parentRefPtr)
*
* DESCRIPTION: 
* Return the "first" child of the parent node.
* Since "first" is actually ambiguous, what is returned is one of the 
* children whose achievement time is no later than the achievement 
* times of the rest of the children (or NULL if there are no children).
*
* INPUTS: TCA_REF_PTR parentRefPtr;
*
* OUTPUTS: TCA_REF_PTR
*
******************************************************************************/

TCA_REF_PTR tcaFindFirstChild(parentRefPtr)
TCA_REF_PTR parentRefPtr;
{ 
  TCA_REF_PTR ref;

  if (tcaQuery("findFirstChild", (void *)&(parentRefPtr->refId), (void *)&ref)
      == NullReply)
    return NULL;
  else
    return ref;
}


/******************************************************************************
*
* FUNCTION: TCA_REF_PTR tcaFindLastChild(parentRefPtr)
*
* DESCRIPTION: 
* Return the "last" child of the parent node.
* Since "last" is actually ambiguous, what is returned is one of the 
* children whose achievement time is no earlier than the achievement 
* times of the rest of the children (or NULL if there are no children).
*
* INPUTS: TCA_REF_PTR parentRefPtr;
*
* OUTPUTS: TCA_REF_PTR
*
******************************************************************************/

TCA_REF_PTR tcaFindLastChild(parentRefPtr)
TCA_REF_PTR parentRefPtr;
{ 
  TCA_REF_PTR ref;

  if (tcaQuery("findLastChild", (void *)&(parentRefPtr->refId), (void *)&ref)
      == NullReply)
    return NULL;
  else
    return ref;
}


/******************************************************************************
*
* FUNCTION: TCA_REF_PTR tcaFindNextChild(childRefPtr)
*
* DESCRIPTION: 
* Return the "next" sibling of the child.
* "Next" is ambiguous, so what is returned is the earliest sibling
* whose achievement time is after the end of the child's achievement
* time (or NULL if there is no such sibling).
*
* INPUTS: TCA_REF_PTR childRefPtr;
*
* OUTPUTS: TCA_REF_PTR
*
******************************************************************************/

TCA_REF_PTR tcaFindNextChild(childRefPtr)
TCA_REF_PTR childRefPtr;
{ 
  TCA_REF_PTR ref;

  if (tcaQuery("findNextChild", (void *)&(childRefPtr->refId), (void *)&ref)
      == NullReply)
    return NULL;
  else
    return ref;
}


/******************************************************************************
*
* FUNCTION: TCA_REF_PTR tcaFindPreviousChild(childRefPtr)
*
* DESCRIPTION: 
* Return the "previous" sibling of the child.
* "Previous" is ambiguous, so what is returned is the latest sibling
* whose achievement time is before the start of the child's achievement
* time (or NULL if there is no such sibling).
*
* INPUTS: TCA_REF_PTR childRefPtr;
*
* OUTPUTS: TCA_REF_PTR
*
******************************************************************************/

TCA_REF_PTR tcaFindPreviousChild(childRefPtr)
TCA_REF_PTR childRefPtr;
{ 
  TCA_REF_PTR ref;

  if (tcaQuery("findPrevChild", (void *)&(childRefPtr->refId), (void *)&ref)
      == NullReply)
    return NULL;
  else
    return ref;
}


/******************************************************************************
*
* FUNCTION: TCA_REF_PTR tcaFindFailedReference(ERefPtr)
*
* DESCRIPTION: 
* Like "FindParentReference", except goes up the tree to find the
* first node that is not a failure handler (i.e., of ExceptionClass).
* ERefPtr must be a reference to an exception handler.
*
* INPUTS: TCA_REF_PTR ERefPtr;
*
* OUTPUTS: TCA_REF_PTR
*
******************************************************************************/

TCA_REF_PTR tcaFindFailedReference(ERefPtr)
TCA_REF_PTR ERefPtr;
{ 
  TCA_REF_PTR ref;

  if (ERefPtr) {
    if (!ERefPtr->msg)
      ERefPtr->msg = msgFind(ERefPtr->name);
  }
  else {
    tcaModError("ERROR: tcaFailedReference must be passed an exception class ref.");
  }

  if (ERefPtr->msg->msgData->class != ExceptionClass) {
    tcaModError("ERROR: tcaFailedReference must be passed an exception class ref.");
  }

  (void)tcaQuery("findFailedRef", (void *)&(ERefPtr->refId), (void *)&ref);
  return ref;
}


/******************************************************************************
*
* FUNCTION: TCA_REF_PTR tcaFindTopLevelReference(childRefPtr)
*
* DESCRIPTION: 
* Return the top level reference of the task tree.
* Return NULL if the given child is root node OR is not a valid task tree node.
*
* INPUTS: TCA_REF_PTR childRefPtr;
*
* OUTPUTS: TCA_REF_PTR
*
******************************************************************************/

TCA_REF_PTR tcaFindTopLevelReference(childRefPtr)
TCA_REF_PTR childRefPtr;
{
  TCA_REF_PTR ref;

  if (tcaQuery("findTopLevelRef", (void *)&(childRefPtr->refId), (void *)&ref)
      == NullReply)
    return NULL;
  else
    return ref;
}


/******************************************************************************
*
* FUNCTION: TCA_REF_PTR tcaFindAnteReferenceByName(childRefPtr, anteName)
*
* DESCRIPTION:
* Return the child's antecedent node with a specified (message) name.
*
* INPUTS:
* TCA_REF_PTR childRefPtr; 
* char *anteName;
*
* OUTPUTS: TCA_REF_PTR
*
******************************************************************************/

TCA_REF_PTR tcaFindAnteReferenceByName(childRefPtr, anteName)
TCA_REF_PTR childRefPtr; 
char *anteName;
{
  TCA_REF_PTR ref;
  FIND_REF_TYPE findAnteRef;

  findAnteRef.refId = childRefPtr->refId;
  findAnteRef.name = anteName;

  if (tcaQuery("findAnteRefByName", (void *)&findAnteRef, (void *)&ref)
      == NullReply)
    return NULL;
  else
    return ref;
}


/******************************************************************************
*
* FUNCTION: TCA_REF_PTR tcaFindChildByName(parentRefPtr, childName)
*
* DESCRIPTION: 
* Return the child node of the parent which has the specified (message) 
* name.  If more than one such child, returns the "first" one it comes
* across (where "first" is implementation dependent).
*
* INPUTS:
* TCA_REF_PTR parentRefPtr; 
* char *childName;
*
* OUTPUTS: TCA_REF_PTR
*
******************************************************************************/

TCA_REF_PTR tcaFindChildByName(parentRefPtr, childName)
TCA_REF_PTR parentRefPtr; 
char *childName;
{
  TCA_REF_PTR ref;
  FIND_REF_TYPE findChildRef;

  findChildRef.refId = parentRefPtr->refId;
  findChildRef.name = childName;

  if (tcaQuery("findChildByName", (void *)&findChildRef, (void *)&ref)
      == NullReply)
    return NULL;
  else
    return ref;
}


/******************************************************************************
*
* FUNCTION: void tcaKillTaskTree(ref)
*
* DESCRIPTION: Kill the task tree rooted at the node indicated by ref.
*
* INPUTS: TCA_REF_PTR ref;
*
* OUTPUTS: void
*
******************************************************************************/

void tcaKillTaskTree(ref)
TCA_REF_PTR ref;
{ 
  (void)tcaInform("killTaskTree", (void *)&(ref->refId));
}


/******************************************************************************
*
* FUNCTION: void tcaKillSubTaskTree(ref)
*
* DESCRIPTION: Kill the task trees rooted BELOW the node indicated by ref.
*
* INPUTS: TCA_REF_PTR ref;
*
* OUTPUTS: void
*
******************************************************************************/

void tcaKillSubTaskTree(ref)
TCA_REF_PTR ref;
{ 
  (void)tcaInform("killSubTaskTree", (void *)&(ref->refId));
}


/******************************************************************************
*
* FUNCTION: void tcaDisplayTaskTree(ref)
*
* DESCRIPTION: Display the task tree rooted at the node indicated by ref.
*
* INPUTS: TCA_REF_PTR ref;
*
* OUTPUTS: void
*
******************************************************************************/

void tcaDisplayTaskTree(ref)
TCA_REF_PTR ref;
{
  (void)tcaInform("displayTaskTree", (void *)&(ref->refId));
}
