/******************************************************************************
*
* PROJECT: Carnegie Mellon Planetary Rover Project
*          Task Control Architecture 
* 
* MODULE: tca inlcudes
*
* FILE: tcaInternal.h
*
* ABSTRACT:
* 
* Common include items and defines.
*
* REVISION HISTORY
*
* 31-Dec-91 Reid Simmons, School of Computer Science, CMU
* Added slot to MSG_TYPE to enable adding exception handlers to class 
*  of messages, rather than just a single task tree node.
*
* 27-May-91 Christopher Fedor, School of Computer Science, CMU
* Changed MSG_ASK_TYPE to allow for a different msg info cache scheme.
*
* 20-Aug-90 Christopher Fedor, School of Computer Science, CMU
* Revised to software standards.
*
* 18-Aug-90 Christopher Fedor, School of Computer Science, CMU
* Removed service define SERVER_NAME and explicitly specified the
* SERVER_PORT number to avoid calls to service lookup routines. 
* The number needed to be explicit for VxWorks and this simplified
* initiating a connection. 
*
* Lookup routines for a service should be added for an external release.
*
* 10-Oct-89 Christopher Fedor, School of Computer Science, CMU
* 4.2 fix added info to timing display try gerry rostons fix to socket opts.
*
* 19-Sep-89 Reid Simmons, School of Computer Science, CMU
* Deleted "RootNode" field from TCA_MOD_INFO, and made one
* global root node (tcaTaskTreeRootGlobal).
*
* 15-Sep-89 Christopher Fedor, School of Computer Science, CMU
* First release in a series to convert over to softwware standards
* and begin housecleaning.
*
*  1-May-89 Reid Simmons, School of Computer Science, CMU
* Added "RootNode" field to TC_ModInfo (for goal trees)
*
* 27-Apr-89 Christopher Fedor, School of Computer Science, CMU
* Copied Reids new tc3.h (find out what changed!)
* Added Resource field to TC_HndPtr
*
* 24-Feb-89 Christopher Fedor, School of Computer Science, CMU
* Added WAITFOREVER
*
*    Feb-89 Reid Simmons, School of Computer Science, CMU
* New formatters added.
*
*    Dec-89 Christopher Fedor, School of Computer Science, CMU
* Created.
*
******************************************************************************/

#include "basics.h"
#include "tca.h"

/* Integer Version Values for Handshaking */
#define TCA_VERSION_MAJOR      7
#define TCA_VERSION_MINOR      1
#define VERSION_DATE "*ALPHA RELEASE* Nov-92"

/* 1621 is debug. Correct reserved number is 1381 */
#define SERVER_PORT 1381
#define SERVER_NAME "task_control"
#define SERVER_MOD_NAME "TCA Server"

/* 
  For direct port connections start at LISTEN_PORT
  and if unavailable on current host increment by 1 until
  LISTEN_PORT_MAX and report error.
*/
#define LISTEN_PORT 8000
#define LISTEN_PORT_MAX 9001

#define CENTRAL_SERVER_ID  -2
#define QUERY_REPLY_INTENT -4

typedef struct {
  TCA_MSG_CLASS_TYPE className;
  struct _FORMAT_TYPE *format;
} CLASS_FORM_TYPE, *CLASS_FORM_PTR;

typedef struct {
  int tcaMajor;
  int tcaMinor;
} TCA_VERSION_TYPE, *TCA_VERSION_PTR;

typedef struct {
  char *modName;
  char *hostName;
} MOD_DATA_TYPE, *MOD_DATA_PTR;

typedef struct {
  int sd;
  int wait;
  int port;
  int alive;
  int repliesPending;
  MOD_DATA_PTR modData;
  struct _RESOURCE *impliedResource;
  struct _LIST *hndList, *resourceList;
} MODULE_TYPE, *MODULE_PTR;

typedef struct {
  char *msgName;
  char *hndName;
} MSG_ASK_TYPE, *MSG_ASK_PTR;

typedef struct {
  int refId;
  char *name;
  TCA_MSG_CLASS_TYPE class;
  struct _FORMAT_TYPE *msgForm, *resForm;
} MSG_DATA_TYPE, *MSG_DATA_PTR;

typedef struct _MSG {
  int direct;
  int intent, sd; /* 25-Jun-91: fedor: this should go away! */
  int parsedFormats;
  MSG_DATA_PTR msgData;
  struct _LIST *hndList;
  struct _LIST *tapList;
  char *msgFormatStr, *resFormatStr;
  struct _LIST *excepList;
} MSG_TYPE, *MSG_PTR;

typedef struct {
  int refId;
  char *msgName, *hndName;
} HND_DATA_TYPE, *HND_DATA_PTR;

typedef struct _HND {
  int sd;
  int localId;
  MSG_PTR msg;
  void (*hndProc)();
  MODULE_PTR hndOrg;
  HND_DATA_PTR hndData;
  struct _LIST *msgList;
  struct _RESOURCE *resource;
} HND_TYPE, *HND_PTR;

typedef struct {
  int sd;
  char *name;
} HND_KEY_TYPE, *HND_KEY_PTR;

typedef struct _TCA_REF {
  int refId;
  MSG_PTR msg;
  char *name;
} TCA_REF_TYPE;

typedef struct {
  char *name;
  char *format;
} REG_DATA_TYPE, *REG_DATA_PTR;

typedef struct {
  char *name;
  TCA_MSG_CLASS_TYPE class;
  char *msgForm;
  char *resForm;
} MSG_REG_DATA_TYPE, *MSG_REG_DATA_PTR;

typedef struct _BLOCK_COM_TYPE {
  int waitFlag;
  int tplConstr;
} BLOCK_COM_TYPE, *BLOCK_COM_PTR;

typedef struct {
  int port;
  char *name;
} DIRECT_TYPE, *DIRECT_PTR;

typedef struct {
  int sd;
  int port;
  int intent;
  char *host;
  char *module;
} DIRECT_INFO_TYPE, *DIRECT_INFO_PTR;

typedef struct {
  int sel;
  struct fd_set *readMask;
  TCA_REF_PTR ref;
  char *replyData;
  TCA_RETURN_VALUE_TYPE status;
} WAIT_FOR_REPLY_TYPE, *WAIT_FOR_REPLY_PTR;

typedef struct { int num; /* Number of handlers working on the multi-query */
		 int max; /* Maximum number of responses desired */
	       } MULTI_QUERY_CLASS_TYPE, *MULTI_QUERY_CLASS_PTR;

#define MULTI_QUERY_NUM(dispatch) \
  (((MULTI_QUERY_CLASS_PTR)(dispatch)->blockCom)->num)

#define MULTI_QUERY_MAX(dispatch) \
  (((MULTI_QUERY_CLASS_PTR)(dispatch)->blockCom)->max)

#define ONE_WAY_MSG(msg) \
  ((msg)->msgData->class == InformClass || (msg)->msgData->class == BroadcastClass)

#define RESPONSE_EXPECTED(msg) (!ONE_WAY_MSG(msg))

#define TWO_WAY_MSG(msg) \
  ((msg)->msgData->class == QueryClass || (msg)->msgData->class == MultiQueryClass)

extern TCA_REF_PTR tcaRefCreate(); /* Defined in tcaRef.c */
extern void tcaRefFree(); /* defined in tcaRef.c */

#define NO_REF -1
#define CREATE_NULL_REF() tcaRefCreate((MSG_PTR)NULL, (char *)NULL, NO_REF)

extern char *messageClassName(); /* defined in behaviors.c */

extern void recvMessage(); /* defined in comServer.c */
extern void centralRegisterMessage(); /* defined in comServer.c */
extern void centralRegisterHandler(); /* defined in comServer.c */
extern void centralRegisterNamedFormatter(); /* defined in comServer.c */
extern TCA_RETURN_VALUE_TYPE centralSendMessage(); /* defined in comServer.c */
extern void Add_Message_To_Ignore(); /* defined in comServer.c */

extern MSG_PTR msgFind(); /* defined in behaviors.c */


