/******************************************************************************
*
* PROJECT: Carnegie Mellon Planetary Rover Project
*          Task Control Architecture 
* 
* (c) Copyright 1991 Christopher Fedor and Reid Simmons.  All rights reserved.
* 
* MODULE: memory
*
* FILE: tcaMem.c
*
* ABSTRACT:
* 
* Interface to memory management routines.
*
* REVISION HISTORY
*
* 7-oct-92 Richard Goodwin : Changed free() to cfree().  Try to make it 
* easier to track down memory problems.  Also added an 
*
*  4-Jul-91 Christopher Fedor, School of Computer Science, CMU
* Added tcaRegisterMallocHnd to complete the routines so that modules
* can handle all memory allocation.
*
*  4-Jun-91 Christopher Fedor, School of Computer Science, CMU
* Added tcaRegisterFreeMemHnd so that modules can be called to free
* memory to satisfy a malloc request.
*
* 25-Oct-90 Christopher Fedor, School of Computer Science, CMU
* Created.
*
******************************************************************************/

#ifdef VXWORKS
#include "stdioLib.h"
#include "memLib.h"
#else
#include "stdio.h"
#include "malloc.h"
#endif


#include "basics.h"

static unsigned totalMemRequest = 0;

extern void dataMsgDisplayStats();

int freeMemRetryAmount = 0;
void (*tcaFreeMemoryHnd)() = NULL;

int mallocMemRetryAmount = 1;
void *(*tcaMallocMemHnd)() = (void *) malloc;


/******************************************************************************
*
* FUNCTION: void tcaRegisterFreeMemHnd(func, retry)
*
* DESCRIPTION: 
* Sets a function for freeing memory when malloc returns NULL.
* The function is passed an unsigned int amount of the memory requested.
* The number of times this routine is called is set by retry.
* The default is 1.
*
* INPUTS: 
* void (*func)();
* int retry;
*
* OUTPUTS: void.
*
******************************************************************************/

void tcaRegisterFreeMemHnd(func, retry)
void (*func)();
int retry;
{
  freeMemRetryAmount = retry;

  if (freeMemRetryAmount < 1)
    freeMemRetryAmount = 1;

  tcaFreeMemoryHnd = func;
}


/******************************************************************************
*
* FUNCTION: void tcaRegisterMallocHnd(func, retry)
*
* DESCRIPTION: 
* Registers a function to call in place of malloc.
* The routine will be passed an unsigned int of the amount of storage needed.
* The routine will be called a max of retry times if NULL is returned.
* The default retry amount is 1.
*
* INPUTS: 
* void *(*func)();
* int retry;
*
* OUTPUTS:
*
******************************************************************************/

void tcaRegisterMallocHnd(func, retry)
void *(*func)();
int retry;
{
  mallocMemRetryAmount = retry;

  if (mallocMemRetryAmount < 1)
    mallocMemRetryAmount = 1;

  if (func)
    tcaMallocMemHnd = func;
}


/******************************************************************************
*
* FUNCTION: void tcaFree(item)
*
* DESCRIPTION: 
* An interface to free - should use tcaFreeData or tcaFreeReply in most cases.
*
* INPUTS: char *item;
*
* OUTPUTS: void.
*
******************************************************************************/

void tcaFree(item)
char *item;
{
  cfree(item);
}


/******************************************************************************
*
* FUNCTION: char *tcaMalloc(amount)
*
* DESCRIPTION: Interface to malloc requests from tca.
*
* INPUTS: unsigned int amount;
*
* OUTPUTS: char * - generic pointr to the memory
*
* NOTES: 
* Stops everything if we have run out of memory. 
* Note there may not be enough memory to print that we have run out.
*
******************************************************************************/

char *tcaMalloc(amount)
unsigned int amount;
{
  int i, j;
  char *mem;

  mem = NULL;

  if (tcaMallocMemHnd) 
    for(i=0;!mem && i < mallocMemRetryAmount;i++) {
      mem = (*tcaMallocMemHnd)(amount);
    }

  if (mem) {
    totalMemRequest += amount;
    bzero(mem, (int)amount);  /* RTG: make sure the new memory is "clean" */
    return(mem);
  }

  if (tcaFreeMemoryHnd) 
    for(j=0;!mem && j < freeMemRetryAmount;j++) {
      (*tcaFreeMemoryHnd)(amount);

      if (tcaMallocMemHnd) 
	for(i=0;!mem && i < mallocMemRetryAmount;i++) {
	  mem = (*tcaMallocMemHnd)(amount);
	}
    }

  if (mem) {
    totalMemRequest += amount;
    bzero(mem, (int)amount);  /* RTG: make sure the new memory is "clean" */
    return(mem);
  }
  
  TCA_MOD_ERROR1("tcaMalloc: NULL returned from malloc for request: %d\n", 
		 amount);
  return NULL;
}


/******************************************************************************
*
* FUNCTION: void tcaStats()
*
* DESCRIPTION: Quick hack to display some memory stats.
*
* INPUTS: none.
*
* OUTPUTS: void.
*
******************************************************************************/

void tcaStats()
{
  printf("Total Memory Requests Filled: %d\n", totalMemRequest);
  printf("\n");
  dataMsgDisplayStats();
  printf("\n");
#ifdef LISP
  fflush(stdout);
#endif
}
