/*****************************************************************************
*
* PROJECT: Carnegie Mellon Planetary Rover Project
*          Task Control Architecture
*
* (c) Copyright 1991 Christopher Fedor and Reid Simmons.  All rights reserved.
*
* MODULE: temporal constraints
*
* FILE: tplConstr.c
*
* ABSTRACT:
* Partial temporal ordering constraints
*
* REVISION HISTORY
*
*  3-Jul-91, Reid Simmons, School of Computer Science, CMU
* Freed data used the TplConstrainHandler. 
*
* 24-Jun-91, Reid Simmons, School of Computer Science, CMU
* Changed the "name" of a quantity to be a TCA_TIME_POINT_PTR.  
* This makes it possible to refer back to the dispatch with which the
* quantities are associated.
*
* 20-Jun-91, Reid Simmons, School of Computer Science, CMU
* Added rule to explicitly assert that "Now" is greater than the end of
* achievement or planning intervals when their subtrees are
* achieved/planned.
*
* 30-Nov-89, Reid Simmons, School of Computer Science, CMU
* Needed to make the handling and achievement intervals of commands
* distinct in order for exception handling with commands to do the
* temporally right thing.
*
* 28-Nov-89, Long-Ji Lin, School of Computer Science, CMU
* The code (in AddTplConstraints()) that calls Initialize_TaskTreeNode()
* to create a new tree node was moved to dispatchServer().
*
* 14-Aug-89, Reid Simmons, School of Computer Science, CMU
* Added routines for killing task trees
* (Cancel_Activation_Message, FreeInterval, and KillAfterAttending check).
*
*  2-Aug-91, Reid Simmons, School of Computer Science, CMU
* Changed "printf"s to use logging facilities.
* Made changes so you can now issue commands and goals from a Query.
*
* 28-Apr-89, Reid Simmons, School of Computer Science, CMU
* Created (based on existing QLattice code)
*
*****************************************************************************/

#include "stdio.h"
#include "tcaInternal.h"
#include "idtable.h"
#include "datamsg.h"
#include "dispatch.h"
#include "logging.h"
#include "formatters.h"

#include "queue.h"
#include "list.h"
#include "tms.h"
#include "cells.h"
#include "rules.h"
#include "qlattice.h"

#include "tplConstr.h"
#include "taskTree.h"

#define Is_Temporal_Dispatch(dispatch) \
  ((dispatch)->treeNode && (dispatch)->treeNode->handlingInterval)

#define Attendable(tree_node) \
  !QIs_True(Now, "<", tree_node->handlingInterval->start)

QUANTITY_PTR Now;

extern CELL_PTR relationGetOrMake();

extern QUANTITY_PTR relation_other_arg(); /* Defined in qlattice.c */
extern QRELATION_TYPE true_relationship(); /* Defined in qlattice.c */
extern QRELATION_TYPE qrelationship1(); /* Defined in qlattice.c */

static RULE_PTR activationRule;

static RULE_PTR endOfConstraintRule;


/*****
 * The quantity "name" is now a reference back to the dispatch with which it
 *  is associated.  Coerce to a string (char *) to conform to the qlattice code.
 *****/
char *createQName(point, interval, dispatch)
     TCA_POINT_CLASS_TYPE point; 
     TCA_INTERVAL_CLASS_TYPE interval;
     DISPATCH_PTR dispatch;
{ 
  TCA_TIME_POINT_PTR timePoint;

  timePoint = NEW(TCA_TIME_POINT_TYPE);
  timePoint->class = point;
  timePoint->interval.class = interval;
  timePoint->interval.msgRef = dispatch->locId;

  return (char *)timePoint;
}

/******
 * This uses a 2D array of characters to avoid having to malloc up
 * space when names are printed.  Cycles through the array so that up
 * to five qName calls can be made before a buffer is overwritten
 ******/

char *qName(q)
     QUANTITY_PTR q;
{
  static char name[5][100];
  static int last = 0;
  TCA_TIME_POINT_PTR timePoint;
  DISPATCH_PTR dispatch;

  if (q == Now) {
    return "Now";
  } else {
    timePoint = (TCA_TIME_POINT_PTR)(q->name);
    dispatch = DISPATCH_FROM_ID(timePoint->interval.msgRef);
    last = (last+1)%5;
    (void)sprintf(name[last], "%s(%s(%s))", 
		  ((timePoint->class == StartPoint) ? "Start" : "End"),
		  ((timePoint->interval.class == HandlingInterval) ? "Handling" 
		   : ((timePoint->interval.class == AchievementInterval)
		      ? "Achievement" : "Planning")),
		  dispatch->msg->msgData->name);
    return name[last];
  }
}

interval_ptr createInterval (class, dispatch) 
     TCA_INTERVAL_CLASS_TYPE class;
     DISPATCH_PTR dispatch;
{ 
  interval_ptr interval;
  
  interval = NEW(interval_type);
  interval->start = createQuantity(createQName(StartPoint, class, dispatch));
  interval->end = createQuantity(createQName(EndPoint, class, dispatch));
  QAssert(interval->start, "<", interval->end, "Physics");

  return interval;
}

/***
 * The achievement or planning of a node is detected when "Now" cannot be shown
 *  to be less than end of the associated achievement or planning interval.
 *  However, "Now" cannot actually be inferred to be greater than the end point.
 *  This causes the temporal constraint mechanism to do a lot of extra work.
 * This rule rectifies the situation by adding the explicit constraint
 *  whenever the interval is achieved.
 *
 * MIGHT INTERACT WITH THE TASK-TREE KILLING CODE
 *
 ***/
void assertEndOfInterval (ruleName, relationshipNode)
     char *ruleName;
     TMS_NODE_PTR relationshipNode;
{
  if (Less_Than_Now_Relp(relationshipNode) &&
      Relp_Still_Out(relationshipNode)) {
    QAssert(Now, ">", relation_other_arg(relationshipNode->cell, Now),
	    "Completed Message");
  }
}

void addEndOfIntervalRule (interval)
     interval_ptr interval;
{
  if (QIs_True(Now, "<", interval->end)) {
    addOutRule(endOfConstraintRule, relationGetOrMake(Now, interval->end));
  }
}

void addEndOfConstraints (treeNode)
     TASK_TREE_NODE_PTR treeNode;
{
  if (treeNode->dispatch->msg->msgData->class == GoalClass) {
    addEndOfIntervalRule(treeNode->achievementInterval);
    addEndOfIntervalRule(treeNode->planningInterval);
  }
}

interval_ptr createNodeInterval(node, type)
     TASK_TREE_NODE_PTR node;
     TCA_INTERVAL_CLASS_TYPE type;
{
  return createInterval(type, node->dispatch);
}

void add_temporal_intervals_to_node (newTreeNode) 
TASK_TREE_NODE_PTR newTreeNode;
{
  TCA_MSG_CLASS_TYPE nodeClass;

  nodeClass = TASK_TREE_NODE_CLASS(newTreeNode);

  /* Set up Handling Interval */
  newTreeNode->handlingInterval = createNodeInterval(newTreeNode, 
						     HandlingInterval);
  /* Set up Achievement Interval */
  switch (nodeClass) {
  case GoalClass:
  case PollingMonitorClass:
  case DemonMonitorClass:
    newTreeNode->achievementInterval = createNodeInterval(newTreeNode, 
							  AchievementInterval);
    QAssert(newTreeNode->achievementInterval->start, ">=", 
	    newTreeNode->handlingInterval->start, "Task Control");
    QAssert(newTreeNode->achievementInterval->end, ">=", 
	    newTreeNode->handlingInterval->end, "Task Control");
    break;

    /******************
     * Need to make the handling and achievement intervals distinct, so that
     * adding an exception handler node under the command node prevents 
     * subsequent messages from firing until the exception is handled.
     ******************/
  case CommandClass:
    newTreeNode->achievementInterval = createNodeInterval(newTreeNode, 
							  AchievementInterval);
    QAssert(newTreeNode->achievementInterval->start, "=",
	    newTreeNode->handlingInterval->start, "Task Control");
    QAssert(newTreeNode->achievementInterval->end, ">=", 
	    newTreeNode->handlingInterval->end, "Task Control");
    break;

  case PointMonitorClass:
    newTreeNode->achievementInterval = createNodeInterval(newTreeNode, 
							  AchievementInterval);
    QAssert(newTreeNode->achievementInterval->start, "=", 
	    newTreeNode->handlingInterval->start, "Task Control");
    QAssert(newTreeNode->achievementInterval->end, ">=", 
	    newTreeNode->handlingInterval->end, "Task Control");
    break;

  case ExceptionClass:
    newTreeNode->achievementInterval = createNodeInterval(newTreeNode, 
							  AchievementInterval);
    QAssert(newTreeNode->achievementInterval->end, ">=", 
	    newTreeNode->handlingInterval->end, "Task Control");
    break;
  }

  /* Set up Planning Interval */
  switch (nodeClass) {
  case GoalClass:
  case PollingMonitorClass:
  case DemonMonitorClass:
    newTreeNode->planningInterval = createNodeInterval(newTreeNode, 
						       PlanningInterval);
    QAssert(newTreeNode->planningInterval->start, "=", 
	    newTreeNode->handlingInterval->start, "Task Control");
    QAssert(newTreeNode->planningInterval->end, ">=", 
	    newTreeNode->handlingInterval->end, "Task Control");
    break;

  case CommandClass:
    newTreeNode->planningInterval = newTreeNode->handlingInterval;
    break;

  case PointMonitorClass:
    newTreeNode->planningInterval = newTreeNode->handlingInterval;
    break;

  case ExceptionClass:
    newTreeNode->planningInterval = createNodeInterval(newTreeNode, 
						       PlanningInterval);
    QAssert(newTreeNode->planningInterval->end, ">=", 
	    newTreeNode->handlingInterval->end, "Task Control");
    break;
  }
}

/*****************************************************************
 * Set up additional temporal constraints imposed by the user.
 * Constraints can be:
 *   SEQ_ACH        -> End(Achievement(ParentNode->lastChild)) <= 
 *                        Start(Achievement(TreeNode))
 *   SEQ_PLANNING   -> End(Planning(ParentNode->lastChild)) <= 
 *                        Start(Planning(TreeNode))
 *   PLAN_FIRST     -> End(Planning(TreeNode)) <= Start(Achievement(TreeNode))
 *   DELAY_PLANNING -> Start(Planning(TreeNode)) = Start(Achievement(TreeNode))
 *
 * Multiple constraints can be specified by adding the constants together
 *   (e.g. SEQ_ACH+PLAN_FIRST).
 *****************************************************************/

/***********************************************************/

void addTplConstraints(dispatch, tplConstr)
DISPATCH_PTR dispatch;
int tplConstr;
{ 
  TASK_TREE_NODE_PTR TreeNode, parentNode, lastChild;

  TreeNode = dispatch->treeNode;
  parentNode = TreeNode->parent;

  if (SEQ_ACH & tplConstr) {
    if (parentNode) {
      lastChild = getLastChild(parentNode, TreeNode->dispatch);
      if (lastChild && !DEAD_NODE(lastChild)) {
	Log_Status2
	  ("Adding SEQUENTIAL ACHIEVEMENT constraint between %s and %s\n",
	   TASK_TREE_NODE_MSG_NAME(lastChild),
	   TASK_TREE_NODE_MSG_NAME(TreeNode));
	QAssert(lastChild->achievementInterval->end, "<=",
		TreeNode->achievementInterval->start, "Temporal Constraint");
      }
    }
  }

  if (SEQ_PLANNING & tplConstr) {
    if (parentNode) {
      lastChild = getLastChild(parentNode, TreeNode->dispatch);
      if (lastChild && !DEAD_NODE(lastChild)) {
	Log_Status2("Adding SEQUENTIAL PLANNING constraint between %s and %s\n",
		    TASK_TREE_NODE_MSG_NAME(lastChild),
		    TASK_TREE_NODE_MSG_NAME(TreeNode));
	QAssert(lastChild->planningInterval->end, "<=",
		TreeNode->planningInterval->start, "Temporal Constraint");
      }
    }
  }

  if (DELAY_PLANNING & tplConstr) {
    Log_Status1("Adding DELAY PLANNING constraint to %s\n",
		TASK_TREE_NODE_MSG_NAME(TreeNode));
    QAssert(TreeNode->planningInterval->start, "=", 
	    TreeNode->achievementInterval->start, "Temporal Constraint");
  }

  if (PLAN_FIRST & tplConstr) {
    Log_Status1("Adding PLAN FIRST constraint to %s\n", 
		TASK_TREE_NODE_MSG_NAME(TreeNode));
    QAssert(TreeNode->planningInterval->end, "<=", 
	    TreeNode->achievementInterval->start, "Temporal Constraint");
  }
}
/***********************************************************/

/* 
  Does the node represent the relationship "Now < X"?
*/
int Less_Than_Now_Relp(relationship_node)
TMS_NODE_PTR relationship_node;
{ 
  return (true_relationship(CV_RELP(relationship_node),
		    (CELL_PTR)relationship_node->cell, Now) == Less);
}

int Relp_Still_Out(relationship_node) 
TMS_NODE_PTR relationship_node;
{
  CELL_PTR relation;

  relation = (CELL_PTR)relationship_node->cell;

  (void)qrelationship1(REL_ARG1(relation), REL_ARG2(relation),
		       CV_RELP(relationship_node));

  return IS_OUT(relationship_node);
}

void Try_To_Activate(rule_name, relationship_node)
char *rule_name; 
TMS_NODE_PTR relationship_node; 
{ 
  DISPATCH_PTR dispatch;
  QUANTITY_PTR quantity;
  TCA_TIME_POINT_PTR timePoint;

  if (Less_Than_Now_Relp(relationship_node) && 
      Relp_Still_Out(relationship_node)) {
    quantity = relation_other_arg(relationship_node->cell, Now);
    timePoint = (TCA_TIME_POINT_PTR)(quantity->name);
    dispatch = DISPATCH_FROM_ID(timePoint->interval.msgRef);
    processActiveMessage(dispatch);
  }
}

void AssertHoldingConstraint(TreeNode)
TASK_TREE_NODE_PTR TreeNode;
{
  QAssert(Now, "<=", TreeNode->handlingInterval->start, "Scheduler");
}

void RetractHoldingConstraint(TreeNode)
TASK_TREE_NODE_PTR TreeNode;
{
  QRetract(Now, "<=", TreeNode->handlingInterval->start, "Scheduler");
}

void PendingConstraints(dispatch)
DISPATCH_PTR dispatch;
{
  if (Is_Temporal_Dispatch(dispatch))
    AssertHoldingConstraint(dispatch->treeNode);
}

void InactiveConstraints(dispatch) 
DISPATCH_PTR dispatch;
{ 
  TASK_TREE_NODE_PTR treeNode;

  treeNode = dispatch->treeNode;

  AssertHoldingConstraint(treeNode);
  addOutRule(activationRule, 
	     relationGetOrMake(Now, treeNode->handlingInterval->start));
}

/******************************************************************************
*
* FUNCTION: void Initialize_TaskTreeNode(dispatch, parentId)
*
* DESCRIPTION:
*  Creates a task tree node and sets up the temporal constraints between it and
*    its parent.
*
* INPUTS: 
* DISPATCH_PTR dispatch; 
* int parentId;
*
* OUTPUTS: TASK_TREE_NODE_PTR: the newly created task tree node
*
* NOTES: NEED TO CHANGE "removeParentChildRelations" IF THIS FUNCTION EVER CHANGES!!
*
******************************************************************************/
void Initialize_TaskTreeNode(dispatch, parentId)
DISPATCH_PTR dispatch; 
int parentId;
{ 
  TASK_TREE_NODE_PTR NewMsgNode, parentNode;

  NewMsgNode = taskTreeNodeCreate(dispatch, parentId);

  add_temporal_intervals_to_node(NewMsgNode);

  /* Set up temporal constraints with parent node */
  parentNode = NewMsgNode->parent;
  if (parentNode && !IS_ROOT_NODE(parentNode)) {
    if (parentNode->achievementInterval != parentNode->handlingInterval) {
      QAssert(parentNode->achievementInterval->start, "<=", 
              NewMsgNode->achievementInterval->start, "Task Control");
      QAssert(parentNode->achievementInterval->end, ">=", 
	      NewMsgNode->achievementInterval->end, "Task Control");
    }
    if ((parentNode->planningInterval != parentNode->handlingInterval) &&
	(NewMsgNode->planningInterval != NewMsgNode->handlingInterval)) {
      QAssert(parentNode->planningInterval->start, "<=", 
	      NewMsgNode->planningInterval->start, "Task Control");
      QAssert(parentNode->planningInterval->end, ">=", 
	      NewMsgNode->planningInterval->end, "Task Control");
    }
  }
}

/******************************************************************************
*
* FUNCTION: void removeParentChildRelations(parent, node)
*
* DESCRIPTION:
* Retracts the temporal constraints set up between a parent and a child node.
*
* INPUTS: 
* TASK_TREE_NODE_PTR parentNode, childNode;
*
* OUTPUTS: void. changes temporal constraints.
*
* NOTES:
* 29-May-91: fedor: should track down what retraction would do if the
* item was never asserted. should also see what effect this has on memory use.
*
******************************************************************************/
int removeParentChildRelations(parentNode, childNode)
TASK_TREE_NODE_PTR parentNode, childNode;
{
  if (parentNode && !IS_ROOT_NODE(parentNode)) {
    if (parentNode->achievementInterval != parentNode->handlingInterval) {
      QRetract(parentNode->achievementInterval->start, "<=", 
              childNode->achievementInterval->start, "Task Control");
      QRetract(parentNode->achievementInterval->end, ">=", 
	      childNode->achievementInterval->end, "Task Control");
    }
    if ((parentNode->planningInterval != parentNode->handlingInterval) &&
	(childNode->planningInterval != childNode->handlingInterval)) {
      QRetract(parentNode->planningInterval->start, "<=", 
	      childNode->planningInterval->start, "Task Control");
      QRetract(parentNode->planningInterval->end, ">=", 
	      childNode->planningInterval->end, "Task Control");
    }
  }
  return TRUE;
}

void AttendingConstraints(dispatch) 
DISPATCH_PTR dispatch;
{
  TASK_TREE_NODE_PTR treeNode;

  if (Is_Temporal_Dispatch(dispatch)) {
    treeNode = dispatch->treeNode;
    QAssert(Now, "<", treeNode->handlingInterval->end, "Scheduler");
    RetractHoldingConstraint(treeNode);
    
    addEndOfConstraints(treeNode);

    QAssert(Now, ">=", treeNode->handlingInterval->start, "Scheduler");
  }
}

void CompletionConstraints(dispatch)
DISPATCH_PTR dispatch;
{
  if (Is_Temporal_Dispatch(dispatch)) {
    QRetract(Now, "<", dispatch->treeNode->handlingInterval->end, "Scheduler");
    QAssert(Now, ">", dispatch->treeNode->handlingInterval->end,
	    "Completed Message");
    (void)runRules(-1);
  }
}

/*****************************************************************************
* Free the interval and its associated start and end points.
*****************************************************************************/
void FreeInterval(Interval)
interval_ptr Interval;
{
  FreeQuantity(Interval->start);
  FreeQuantity(Interval->end);
  tcaFree((char *)Interval);
}


/*
void print_reason(tms_node) tms_node_ptr tms_node;
{ 
  JUSTIFICATION_PTR just;
  cons_ptr supporters;

  just = tms_node->support_justification;
  supporters = just->supporters;
  if (supporters) {
    while (supporters) {
      print_reason((tms_node_ptr)Car(supporters));
      supporters = Cdr(supporters);
    }
  }
  else {
    say_node_value(tms_node);
    fprintf(stderr, " (%s)\n", 
	    Just_Informant(Node_Support_Justification(tms_node)));
  }
}

void Why_Rel(q1, q2) QUANTITY_PTR q1, q2;
{   
  Print_Relationship(q1,q2,TRUE);
  print_reason(Current_Value_Supporter(relationGetOrMake(q1,q2)));
}
*/

interval_ptr Interval_Of(Interval) 
TCA_INTERVAL_PTR Interval;
{ 
  DISPATCH_PTR dispatch;

  if (Interval->class == NoInterval) {
    return NULL;
  } else {
    dispatch = DISPATCH_FROM_ID(Interval->msgRef);
    if (!Is_Temporal_Dispatch(dispatch)) {
      Log("ERROR!!! Interval_Of: No interval associated with dispatch\n");
      return NULL;
    } else {
      switch (Interval->class) {
      case HandlingInterval: 
	return dispatch->treeNode->handlingInterval;
      case PlanningInterval: 
	return dispatch->treeNode->planningInterval;
      case AchievementInterval: 
	return dispatch->treeNode->achievementInterval;
      default: return NULL;
      }
    }
  }
}

QUANTITY_PTR Time_Of(timePoint) 
TCA_TIME_POINT_PTR timePoint;
{ 
  interval_ptr interval;

  if (timePoint->class == NoTime)
    return NULL;
  else {
    interval = Interval_Of(&(timePoint->interval));
    
    if (interval) 
      switch (timePoint->class) {
      case StartPoint: 
	return interval->start;
      case EndPoint: 
	return interval->end;
      default: return NULL;
      }
    else
      return NULL;
  }
}

void TplConstrainHandler(dispatch, TplConstraint)
DISPATCH_PTR dispatch; 
TCA_TPL_CONSTRAINT_PTR TplConstraint;
{
  QUANTITY_PTR q1, q2;

  q1 = Time_Of(&(TplConstraint->timePoint1));
  q2 = Time_Of(&(TplConstraint->timePoint2));

  if (q1 && q2) {
    Log3("   Asserting %s %s %s\n", 
	 qName(q1), TplConstraint->relationship, qName(q2));

    if (QIs(q1, TplConstraint->relationship, q2) == Is_False) {
      tcaError("TplConstrainHandler: Inconsistent Relationship Being Asserted\n");
    } else {
      QAssert(q1, TplConstraint->relationship, q2, "Temporal Constraint");
    }
    /* A bit more efficient than using tcaFreeData */
    freeDataStructure(dispatch->msg->msgData->msgForm, (char *)TplConstraint);
  }
}

void Tpl_Initialize ()
{
  qlatticeInit();

  Now = createQuantity("Now");

  activationRule = ruleCreateOut("Activate Message", Try_To_Activate, 
				 3, (char *)NULL);
  activationRule->used = 1000; /* Never free, Reid: 10-Jul-91 */

  endOfConstraintRule = ruleCreateOut("EndOf Rule", assertEndOfInterval,
				      3, (char *)NULL);
  endOfConstraintRule->used = 1000; /* Never free, Reid: 10-Jul-91 */

  /* "{Class, msgRef}" */
  centralRegisterNamedFormatter("Interval", "{int, int}"); 

  /* "{Class, Interval}" */
  centralRegisterNamedFormatter("TimePoint", "{int, Interval}"); 
  centralRegisterNamedFormatter("TimeList", "*{TimePoint, *!}"); 

  centralRegisterMessage("tplConstrain", InformClass, 
			 "{TimePoint, string, TimePoint}", (char *)NULL);
  centralRegisterHandler("tplConstrain", "TplConstrainHandler", 
		     TplConstrainHandler);
}
